#version 150 core

uniform sampler2D depthNormalMap;
uniform sampler2D noiseMap;
uniform mat4 _ProjectionMatrix;
uniform int uKernelSize;
const int MAX_KERNEL_SIZE = 128;
uniform vec4 uKernelOffsets[ MAX_KERNEL_SIZE ];
uniform float uRadius;
uniform float uPower;

in vec2 vTexCoord;
noperspective in vec3 vViewDir;

out vec4 fragColor;

float ssao( in mat3 kernelBasis, in vec3 originPos, in float radius )
{
    float occlusion = 0.0;
	  
    for (int i = 0; i < uKernelSize; ++i)
    {
        vec3 samplePos = kernelBasis * uKernelOffsets[i].xyz;
        samplePos = samplePos * radius + originPos;
		
        // project sample position:
        vec4 offset = _ProjectionMatrix * vec4( samplePos, 1.0 );
        offset.xy /= offset.w; // only need xy
        offset.xy = offset.xy * 0.5 + 0.5; // scale/bias to texcoords
		
        // get sample depth:
        float sampleDepth = texture( depthNormalMap, offset.xy ).x;
        
        float rangeCheck = smoothstep( 0.0, 1.0, radius / abs( originPos.z - sampleDepth ) );
        occlusion += rangeCheck * step( sampleDepth, samplePos.z );
    }
	
    // Normalizes and inverts occlusion so it can be used in lighting calc.
    occlusion = 1.0 - (occlusion / float( uKernelSize ));
    
    return pow( abs( occlusion ), uPower );
}

void main()
{
    vec2 noiseTexCoords = vec2( textureSize( depthNormalMap, 0 ) ) / vec2( textureSize( noiseMap, 0 ) );
    noiseTexCoords *= vTexCoord;
	
    // get view space origin:
    float originDepth = texture( depthNormalMap, vTexCoord ).x;

    vec3 originPos = vViewDir * originDepth;

    // get view space normal (no *2-1 needed because it's a float texture):
    vec3 normal = -normalize( texture( depthNormalMap, vTexCoord ).gba );
		
    // construct kernel basis matrix:
    vec3 rvec = normalize( texture( noiseMap, noiseTexCoords ).xyz ); // (no *2-1 needed because it's a float texture).
    vec3 tangent = normalize( rvec - normal * dot( rvec, normal ) );
    vec3 bitangent = cross( tangent, normal );
    mat3 kernelBasis = mat3( tangent, bitangent, normal );
	
    fragColor = vec4( ssao( kernelBasis, originPos, uRadius ) );
}

