#version 430 core

uniform sampler2D diffuseMap;
uniform sampler2D normalMap;
uniform sampler2D specularMap;
uniform samplerCube envMap;
uniform sampler2D _ShadowMap;

uniform vec3 materialSpecular;
uniform vec3 _LightColor;
uniform float roughness;
uniform float reflectivity;
uniform int _LightCastsShadow;
uniform int _LightType;

in vec2 vTexCoord;
in vec3 vViewVecTS;
in vec3 vLightVecTS;
#ifdef ENVIRONMENT_MAP
in vec3 vEnvTexCoord;
#endif
in vec3 vNormalWS;
in vec3 vTangentWS;
in vec3 vBitangentWS;

// Shadow stuff
in vec4 vPositionWS;
in vec4 vProjCoord;
uniform float uLinearDepthConstant;
uniform vec3 _LightPositionWorld;

// Forward+ stuff
uniform vec3 _CameraPosition;
uniform uint windowWidth;
uniform uint windowHeight;
uniform uint maxNumLightsPerTile;

out vec4 fragColor;

float linstep(float low, float high, float v)
{
    return clamp( (v - low) / (high - low), 0.0f, 1.0f );
}

float VSM( in float depth )
{
    vec2 moments = texture( _ShadowMap, vProjCoord.xy / vProjCoord.w ).rg;
    
    float variance = moments.y - moments.x * moments.x;
    
    float delta = depth - moments.x;
    float pMax = linstep( 0.2f, 1.0f, variance / (variance + delta * delta) );
    
    return (depth <= moments.x) ? 1.0f : pMax;
}

#define TILE_RES 16
#define LIGHT_INDEX_BUFFER_SENTINEL 0x7fffffff

layout (rgba32f, binding = 0) uniform imageBuffer pointLightBufferCenterAndRadius;
layout (r32ui, binding = 2) uniform uimageBuffer perTileLightIndexBuffer;
layout (rgba32f, binding = 3) uniform imageBuffer pointLightColors;
layout (rgba32f, binding = 4) uniform imageBuffer spotLightBufferCenterAndRadius;
layout (rgba32f, binding = 5) uniform imageBuffer spotLightColors;
layout (rgba32f, binding = 6) uniform imageBuffer spotLightParams;

uint GetNumLightsInThisTile( uint nTileIndex )
{
    uint nNumLightsInThisTile = 0;
    uint nIndex = maxNumLightsPerTile * nTileIndex;
    uint nNextLightIndex = imageLoad( perTileLightIndexBuffer, int( nIndex ) ).x;

    // count point lights
    while (nNextLightIndex != LIGHT_INDEX_BUFFER_SENTINEL)
    {
        nNumLightsInThisTile++;
        nIndex++;
        nNextLightIndex = imageLoad( perTileLightIndexBuffer, int( nIndex ) ).x;
    }

    return nNumLightsInThisTile;
}

// calculate the number of tiles in the horizontal direction
uint GetNumTilesX()
{
    return uint(((windowWidth + TILE_RES - 1) / float(TILE_RES)));
}

// calculate the number of tiles in the vertical direction
uint GetNumTilesY()
{
    return uint(((windowHeight + TILE_RES - 1) / float(TILE_RES)));
}

uint GetTileIndex( in vec2 ScreenPos )
{
    float fTileRes = float( TILE_RES );
    uint nNumCellsX = (windowWidth + TILE_RES - 1) / TILE_RES;
    uint nTileIdx = uint( floor( ScreenPos.x / fTileRes ) + floor( ScreenPos.y / fTileRes ) * nNumCellsX );
    return nTileIdx;
}

// Based on http://content.gpwiki.org/D3DBook:%28Lighting%29_Cook-Torrance
void main()
{
    vec3 albedoTex = texture( diffuseMap, vTexCoord ).rgb;
#ifdef SPECULAR_MAP
    vec3 specularTex = texture( specularMap, vTexCoord ).rgb;
#else
    vec3 specularTex = materialSpecular;
#endif
#ifdef ENVIRONMENT_MAP
    vec3 envTex = texture( envMap, vEnvTexCoord ).rgb;
    albedoTex = mix( albedoTex, envTex, reflectivity ) * specularTex; // Could use a reflection mask texture or channel.
#endif
    
    // input vectors
    vec3 N = normalize( texture( normalMap, vTexCoord ).rgb * 2.0f - 1.0f );

    // Directional light.
    if (_LightType == 0)
    {
        vec3 V = normalize( vViewVecTS );
        vec3 L = normalize( vLightVecTS );
        vec3 H = normalize( L + V );

        // dot products
        float dotNL = clamp( dot( N, L ), 0.0f, 1.0f );
        float dotNH = clamp( dot( N, H ), 0.0f, 1.0f );
        float dotNV = clamp( dot( N, V ), 0.0f, 1.0f );
        float dotVH = clamp( dot( V, H ), 0.0f, 1.0f );
        float r_sq = roughness * roughness;

        // Geometric term.
        float geoNumerator = 2.0f * dotNH;
        float geoB = (geoNumerator * dotNV) / dotVH;
        float geoC = (geoNumerator * dotNL) / dotVH;
        float geo = min( 1.0f, min( geoB, geoC ) );

        // Beckmann roughness.
        float roughness_a = 1.0f / (4.0f * r_sq * pow( dotNH, 4 ));
        float roughness_b = dotNH * dotNH - 1.0;
        float roughness_c = r_sq * dotNH * dotNH;
        float roughness = roughness_a * exp( roughness_b / roughness_c );

        // Schlick Fresnel.
        float ref_at_norm_incidence = 1.0f;
        float fresnel = pow( 1.0f - dotVH, 5.0f );
        fresnel *= (1.0f - ref_at_norm_incidence);
        fresnel += ref_at_norm_incidence;

        // Specular BRDF.
        float Rs_numerator = fresnel * geo * roughness;
        float Rs_denominator = dotNV * dotNL;
        float Rs = max( 0.0f, Rs_numerator / Rs_denominator ); // Added max() to prevent black areas.

        vec3 ambient = vec3( 0.1f, 0.1f, 0.1f );
        vec3 final = ambient * albedoTex + max( 0.0f, dotNL ) * (specularTex * Rs + albedoTex);

        if (_LightCastsShadow == 1)
        {
            float depth = vProjCoord.z / vProjCoord.w;
            float shadow = VSM( depth );
            final *= shadow;
        }

        fragColor = vec4( final, 1.0f );
        return;
    }

    // Transforms normal from tangent-space into world-space.
    N = (N.x * vTangentWS) + (N.y * vBitangentWS) + (N.z * vNormalWS);
    N = normalize( N );

    vec3 accumDiffuseAndSpecular = vec3( 0.0f, 0.0f, 0.0f );

    uint nTileIndex = GetTileIndex( gl_FragCoord.xy );
    uint nIndex = maxNumLightsPerTile * nTileIndex;
    uint nNextLightIndex = imageLoad( perTileLightIndexBuffer, int( nIndex ) ).x;

    vec3 viewDirWS = normalize( _CameraPosition - vPositionWS.xyz );

    // Loops over point lights.
    while (nNextLightIndex != LIGHT_INDEX_BUFFER_SENTINEL)
    {
        uint nLightIndex = nNextLightIndex;
        nIndex++;
        nNextLightIndex = imageLoad( perTileLightIndexBuffer, int( nIndex ) ).x;

        vec4 CenterAndRadius = imageLoad( pointLightBufferCenterAndRadius, int( nLightIndex ) );

        vec3 vecToLightWS = CenterAndRadius.xyz - vPositionWS.xyz;
        vec3 lightDirWS = normalize( vecToLightWS );

        float lightDistance = length( vecToLightWS );

        vec3 lightColorDiffuseAndSpecular = vec3( 0.0f, 0.0f, 0.0f );

        float radius = CenterAndRadius.w;

        if (lightDistance < radius)
        {
            float x = lightDistance / radius;
            // fake inverse squared falloff:
            // -(1/k)*(1-(k+1)/(1+k*x^2))
            // k=20: -(1/20)*(1 - 21/(1+20*x^2))
            float falloff = -0.05f + 1.05f / (1.0f + 20.0f * x*x);
        
            vec4 lightColor = imageLoad( pointLightColors, int( nLightIndex ) );

            vec3 V = viewDirWS;
            vec3 L = lightDirWS;
            vec3 H = normalize( L + V );

            // dot products
            float dotNL = clamp( dot( N, L ), 0.0f, 1.0f );
            float dotNH = clamp( dot( N, H ), 0.0f, 1.0f );
            float dotNV = clamp( dot( N, V ), 0.0f, 1.0f );
            float dotVH = clamp( dot( V, H ), 0.0f, 1.0f );
            float r_sq = roughness * roughness;

            // Geometric term.
            float geoNumerator = 2.0f * dotNH;
            float geoB = (geoNumerator * dotNV) / dotVH;
            float geoC = (geoNumerator * dotNL) / dotVH;
            float geo = min( 1.0f, min( geoB, geoC ) );

            // Beckmann roughness.
            float roughness_a = 1.0f / (4.0f * r_sq * pow( dotNH, 4 ));
            float roughness_b = dotNH * dotNH - 1.0;
            float roughness_c = r_sq * dotNH * dotNH;
            float roughness = roughness_a * exp( roughness_b / roughness_c );

            // Schlick Fresnel.
            float ref_at_norm_incidence = 1.0f;
            float fresnel = pow( 1.0f - dotVH, 5.0f );
            fresnel *= (1.0f - ref_at_norm_incidence);
            fresnel += ref_at_norm_incidence;

            // Specular BRDF.
            float Rs_numerator = fresnel * geo * roughness;
            float Rs_denominator = dotNV * dotNL;
            float Rs = max( 0.0f, Rs_numerator / Rs_denominator ); // Added max() to prevent black areas.

            lightColorDiffuseAndSpecular = max( 0.0f, dotNL ) * (specularTex * Rs + albedoTex) * lightColor.rgb * falloff;
        }

        accumDiffuseAndSpecular += lightColorDiffuseAndSpecular;
    }

    // Moves past the first sentinel to get to the spot lights.
    nIndex++;
    nNextLightIndex = imageLoad( perTileLightIndexBuffer, int( nIndex ) ).x;

    // Loops over spot lights.
    while (nNextLightIndex != LIGHT_INDEX_BUFFER_SENTINEL)
    {
        uint nLightIndex = nNextLightIndex;
        nIndex++;
        nNextLightIndex = imageLoad( perTileLightIndexBuffer, int( nIndex ) ).x;

        vec4 centerAndRadius = imageLoad( spotLightBufferCenterAndRadius, int( nLightIndex ) );
        vec4 spotParams = imageLoad( spotLightParams, int( nLightIndex ) );

        // reconstruct z component of the light dir from x and y
        vec3 spotLightDir;
        spotLightDir.xy = spotParams.xy;
        spotLightDir.z = sqrt( 1.0f - spotLightDir.x * spotLightDir.x - spotLightDir.y * spotLightDir.y );

        // the sign bit for cone angle is used to store the sign for the z component of the light dir
        // TODO: if GLSL has a sign() method, use it.
        spotLightDir.z = (spotParams.z > 0.0f) ? spotLightDir.z : -spotLightDir.z;

        vec3 vToLight = centerAndRadius.xyz - vPositionWS.xyz;
        vec3 vToLightNormalized = normalize( vToLight );
        float fLightDistance = length( vToLight );
        
        vec4 lightColor = imageLoad( spotLightColors, int( nLightIndex ) );

        vec3 vecToLight = normalize( centerAndRadius.xyz - vPositionWS.xyz );
        float spotAngle = dot( -spotLightDir.xyz, vecToLight );
        float cosineOfConeAngle = (spotParams.z > 0.0f) ? spotParams.z : -spotParams.z;

        // Falloff
        float dist = distance( vPositionWS.xyz, centerAndRadius.xyz );
        float a = dist / centerAndRadius.w + 1.0f;
        float att = 1.0f / (a * a);
        vec3 color = lightColor.rgb * att * specularTex;

        accumDiffuseAndSpecular += spotAngle > cosineOfConeAngle ? color : vec3( 0.0, 0.0, 0.0 );
    }

    accumDiffuseAndSpecular *= 2.0f;
    
    vec3 ambient = vec3( 0.1f, 0.1f, 0.1f );

    fragColor = vec4( albedoTex * ambient + accumDiffuseAndSpecular, 1.0 );

    //float col = coord.y / float( windowHeight);
    //fragColor = vec4( col, col, col, 1.0f);

    //float col = float( (nTileIndex % GetNumTilesX() )) / float( GetNumTilesX() );
    //fragColor = vec4( col, col, col, 1.0f);

    //float col = nTileIndex / float(GetNumTilesX() * GetNumTilesY());
    //fragColor = vec4( col, col, col, 1.0f );

    /*uint tileCount = GetNumLightsInThisTile( nTileIndex );
    if (tileCount == 0)
    {
        fragColor = vec4( 1.0f, 0.0f, 0.0f, 1.0f );
    }*/
}
