cbuffer buffer : register(b0)
{
    column_major matrix _ViewProjectionMatrix;
    column_major matrix modelMatrix;
    column_major matrix _ViewMatrix;
#ifdef SKIN
    column_major matrix boneMatrices[8];
#endif
    column_major matrix _TPMatrix;
    float4 diffuseMap_ST;
    float3 _LightDirection;
    float3 _LightColor;
    float3 _LightPositionWorld;
    float _LightConeAngleCos;
    int _LightCastsShadow;
    int _LightType;
    float3 materialSpecular;
    float3 _CameraPosition;
    float roughness;
    float reflectivity;
}

cbuffer cbPerFrame : register(b1)
{
    matrix              projection           : packoffset(c0);
    matrix              invPojection         : packoffset(c4);
    float3              cameraPos            : packoffset(c8);
    float               alphaTest            : packoffset(c8.w);
    uint                numLights            : packoffset(c9);
    uint                windowWidth          : packoffset(c9.y);
    uint                windowHeight         : packoffset(c9.z);
    uint                maxNumLightsPerTile  : packoffset(c9.w);
    float4              linearDepthConstant  : packoffset(c10); // Stored in .x.
    matrix              viewMatrix           : packoffset(c11);
};

struct VS_INPUT
{
    float3 Pos : POSITION;
    float2 Tex : TEXCOORD0;
    float3 Normal : NORMAL;
    float4 Tangent : TANGENT;
    float4 Color : COLOR;
#ifdef SKIN
    float4 BoneInd : BONE_INDEX;
    float4 Weights : BONE_WEIGHT;
#endif
};

struct PS_INPUT
{
    float4 Pos : SV_POSITION;
    float2 Tex : TEXCOORD0;
    float3 viewVec : VIEW_VEC;
    float3 lightVec : LIGHT_VEC;
#ifdef ENVIRONMENT_MAP
    float3 envTexCoord : ENV_TEXCOORD;
#endif
    float4 Position : POSITION;
    float4 ProjCoord : PROJ_COORD;
    float3 NormalWS : NORMAL;
    float3 TangentWS : TANGENT;
    float3 BitangentWS : BITANGENT;
};

PS_INPUT VS( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
    float4 position = float4( input.Pos, 1.0f );
#ifdef SKIN
    int4 boneIndex = int4( input.BoneInd );
    
    position  = mul( boneMatrices[ boneIndex.x ], float4(input.Pos, 1.0f) * input.Weights.x );
    position += mul( boneMatrices[ boneIndex.y ], float4(input.Pos, 1.0f) * input.Weights.y );
    position += mul( boneMatrices[ boneIndex.z ], float4(input.Pos, 1.0f) * input.Weights.z );
    position += mul( boneMatrices[ boneIndex.w ], float4(input.Pos, 1.0f) * input.Weights.w );
#endif

    output.Pos = mul( _ViewProjectionMatrix, mul( modelMatrix, position ) );
    output.Tex = input.Tex * diffuseMap_ST.xy + diffuseMap_ST.zw;
    
    matrix modelViewMatrix = mul( _ViewMatrix, modelMatrix );
    float4 positionVS = mul( modelViewMatrix, position );

    float3x3 normalMatrix = (float3x3)( modelViewMatrix );

    float3 t = normalize( mul( normalMatrix, input.Tangent.xyz ) );
    float3 b = normalize( mul( normalMatrix, (cross( input.Normal, input.Tangent.xyz ) * input.Tangent.w)) );
    float3 n = normalize( mul( normalMatrix, input.Normal ) );
    float3x3 tbn = float3x3( t, b, n );

    output.lightVec = (mul( _ViewMatrix, float4( -_LightDirection, 0.0f ) )).xyz;
    output.lightVec = mul( tbn, output.lightVec );
    output.viewVec = -positionVS.xyz;
    output.viewVec = mul( tbn, output.viewVec );

#ifdef ENVIRONMENT_MAP
    float4 worldPos = mul( modelMatrix, position );
    float4 worldNormal = normalize( mul( modelMatrix, float4(input.Normal, 0.0f) ) );
    float4 worldEye = normalize( worldPos - float4(_CameraPosition, 1.0f) );

    output.envTexCoord = reflect( worldEye.xyz, worldNormal.xyz );
#endif

    output.Position = mul( modelMatrix, position );
    output.ProjCoord = _LightCastsShadow == 1 ? mul( _TPMatrix, position ) : float4(0.0f, 0.0f, 0.0f, 0.0f);

    output.NormalWS  = normalize( mul( modelMatrix, float4( input.Normal,  0.0f ) ) ).xyz;
    output.TangentWS = normalize( mul( modelMatrix, float4( input.Tangent.xyz, 0.0f ) ) ).xyz;
    float3 bitangentOS = cross( input.Normal, input.Tangent.xyz ) * input.Tangent.w;
    output.BitangentWS = normalize( mul( modelMatrix, float4( bitangentOS, 0.0f ) ) ).xyz;

    return output;
}

Texture2D diffuseMap : register(t0);
Texture2D normalMap : register(t1);
TextureCube envMap : register(t2);
Texture2D specularMap : register(t3);
Texture2D _ShadowMap : register(t4);
Buffer<float4> pointLightBufferCenterAndRadius : register(t5);
Buffer<uint>   perTileLightIndexBuffer         : register(t6);
Buffer<float4> pointLightColors : register(t7);
Buffer<float4> spotLightBufferCenterAndRadius : register(t8);
Buffer<float4> spotLightColors : register(t9);
Buffer<float4> spotLightBufferSpotParams : register(t10);

SamplerState SampleTypeDiffuse : register(s0);
SamplerState SampleTypeNormal : register(s1);
#ifdef ENVIRONMENT_MAP
SamplerState SampleTypeEnv : register(s2);
#endif
#ifdef SPECULAR_MAP
SamplerState SampleTypeSpecular : register(s3);
#endif
SamplerState SampleTypeShadow : register(s4);

#define TILE_RES 16
#define LIGHT_INDEX_BUFFER_SENTINEL 0x7fffffff

uint GetNumLightsInThisTile( uint nTileIndex )
{
    uint nNumLightsInThisTile = 0;
    uint nIndex = maxNumLightsPerTile*nTileIndex;
    uint nNextLightIndex = perTileLightIndexBuffer[ nIndex ];

    // count point lights
    while (nNextLightIndex != LIGHT_INDEX_BUFFER_SENTINEL)
    {
        nNumLightsInThisTile++;
        nIndex++;
        nNextLightIndex = perTileLightIndexBuffer[ nIndex ];
    }

    return nNumLightsInThisTile;
}

// calculate the number of tiles in the horizontal direction
uint GetNumTilesX()
{
    return (uint)((windowWidth + TILE_RES - 1) / (float)TILE_RES);
}

// calculate the number of tiles in the vertical direction
uint GetNumTilesY()
{
    return (uint)((windowHeight + TILE_RES - 1) / (float)TILE_RES);
}

uint GetTileIndex( float2 ScreenPos )
{
    float fTileRes = (float)TILE_RES;
    uint nNumCellsX = (windowWidth + TILE_RES - 1) / TILE_RES;
    uint nTileIdx = floor( ScreenPos.x / fTileRes ) + floor( ScreenPos.y / fTileRes ) * nNumCellsX;
    return nTileIdx;
}

float linstep( float low, float high, float v )
{
    return saturate( (v - low) / (high - low) );
}

float VSM( float4 projCoord, float depth )
{
    /*float2 coord = projCoord.xy;
    coord.y = 1.0f - coord.y;

    float2 moments = _ShadowMap.Sample( SampleTypeShadow, coord ).rg;

    float variance = moments.y - moments.x * moments.x;

    float delta = depth - moments.x;
    float pMax = linstep( 0.2, 1.0, variance / (variance + delta * delta) );
    
    return (depth <= moments.x) ? 1.0f : pMax;
    */

    float2 coord = projCoord.xy;
    coord.y = 1.0f - coord.y;

    float2 moments = _ShadowMap.Sample(SampleTypeShadow, coord).rg;

    float variance = max( moments.y - moments.x * moments.x, -0.001 );

    float delta = depth - moments.x;
    float p = smoothstep(depth - 0.02, depth, moments.x);
    float pMax = linstep(0.2, 1.0, variance / (variance + delta * delta));

    return saturate( max( p, pMax ) );
}

float4 PS( PS_INPUT input ) : SV_Target
{
    float3 albedoTex = diffuseMap.Sample( SampleTypeDiffuse, input.Tex ).rgb;
#ifdef SPECULAR_MAP
    float3 specularTex = specularMap.Sample( SampleTypeSpecular, input.Tex ).rgb;
#else
    float3 specularTex = materialSpecular;
#endif
#ifdef ENVIRONMENT_MAP
    float3 envTex = envMap.Sample( SampleTypeEnv, input.envTexCoord ).rgb;
    albedoTex = lerp( albedoTex, envTex, reflectivity ) * specularTex; // Could use a reflection mask texture or channel.
#endif

    float3 N = normalize( normalMap.Sample( SampleTypeNormal, input.Tex ).rgb * 2.0f - 1.0f );
    float3 ambient = float3( 0.1f, 0.1f, 0.1f );

    if (_LightType == 0)
    {
        float3 V = normalize( input.viewVec );
        float3 L = normalize( input.lightVec );
        float3 H = normalize( L + V );

        // dot products
        float dotNL = saturate( dot( N, L ) );
        float dotNH = saturate( dot( N, H ) );
        float dotNV = saturate( dot( N, V ) );
        float dotVH = saturate( dot( V, H ) );
        float r_sq = roughness * roughness;

        // Geometric term.
        float geoNumerator = 2.0f * dotNH;
        float geoB = (geoNumerator * dotNV) / dotVH;
        float geoC = (geoNumerator * dotNL) / dotVH;
        float geo = min( 1.0f, min( geoB, geoC ) );

        // Beckmann roughness.
        float roughness_a = 1.0f / (4.0f * r_sq * pow( dotNH, 4 ));
        float roughness_b = dotNH * dotNH - 1.0f;
        float roughness_c = r_sq * dotNH * dotNH;
        float roughness = roughness_a * exp( roughness_b / roughness_c );

        // Schlick Fresnel.
        float ref_at_norm_incidence = 1.0f;
        float fresnel = pow( 1.0f - dotVH, 5.0f );
        fresnel *= (1.0f - ref_at_norm_incidence);
        fresnel += ref_at_norm_incidence;

        // Specular BRDF.
        float Rs_numerator = fresnel * geo * roughness;
        float Rs_denominator = dotNV * dotNL;
        float Rs = Rs_numerator / Rs_denominator;

        // FIXME: Should not need saturate() here, probably something is not normalized earlier.
        float3 final = ambient * albedoTex + saturate(max(0.0f, dotNL) * (specularTex * Rs + albedoTex));

        float depth = input.ProjCoord.z / input.ProjCoord.w;
        float shadow = (_LightCastsShadow == 1) ? VSM( input.ProjCoord / input.ProjCoord.w, depth ) : 1.0f;

        return float4(final * shadow, 1.0f);
    }

    // Transforms normal from tangent-space into world-space.
    N = (N.x * input.TangentWS) + (N.y * input.BitangentWS) + (N.z * input.NormalWS);
    N = normalize( N );

    float3 viewDirWS = normalize( _CameraPosition - input.Position.xyz );

    uint nTileIndex = GetTileIndex( input.Pos.xy );
    uint nIndex = maxNumLightsPerTile * nTileIndex;
    uint nNextLightIndex = perTileLightIndexBuffer[ nIndex ];
    float3 accumDiffuseAndSpecular = float3( 0, 0, 0 );
    
    [loop] // Point lights
    while (nNextLightIndex != LIGHT_INDEX_BUFFER_SENTINEL)
    {
        uint nLightIndex = nNextLightIndex;
        nIndex++;
        nNextLightIndex = perTileLightIndexBuffer[ nIndex ];

        float4 CenterAndRadius = pointLightBufferCenterAndRadius[ nLightIndex ];

        float3 vecToLightWS = CenterAndRadius.xyz - input.Position.xyz;
        float3 lightDirWS = normalize( vecToLightWS );
        
        float lightDistance = length( vecToLightWS );

        float3 lightColorDiffuseAndSpecular = float3(0.0f, 0.0f, 0.0f);

        float radius = CenterAndRadius.w;

        if (lightDistance < radius)
        {
            float x = lightDistance / radius;
            // fake inverse squared falloff:
            // -(1/k)*(1-(k+1)/(1+k*x^2))
            // k=20: -(1/20)*(1 - 21/(1+20*x^2))
            float falloff = -0.05 + 1.05 / (1 + 20 * x*x);

            float3 V = viewDirWS;
            float3 L = lightDirWS;
            float3 H = normalize( L + V );

            // dot products
            float dotNL = clamp( dot( N, L ), 0.0f, 1.0f );
            float dotNH = clamp( dot( N, H ), 0.0f, 1.0f );
            float dotNV = clamp( dot( N, V ), 0.0f, 1.0f );
            float dotVH = clamp( dot( V, H ), 0.0f, 1.0f );
            float r_sq = roughness * roughness;

            // Geometric term.
            float geoNumerator = 2.0f * dotNH;
            float geoB = (geoNumerator * dotNV) / dotVH;
            float geoC = (geoNumerator * dotNL) / dotVH;
            float geo = min( 1.0f, min( geoB, geoC ) );

            // Beckmann roughness.
            float roughness_a = 1.0f / (4.0f * r_sq * pow( dotNH, 4 ));
            float roughness_b = dotNH * dotNH - 1.0;
            float roughness_c = r_sq * dotNH * dotNH;
            float roughness = roughness_a * exp( roughness_b / roughness_c );

            // Schlick Fresnel.
            float ref_at_norm_incidence = 1.0f;
            float fresnel = pow( 1.0f - dotVH, 5.0f );
            fresnel *= (1.0f - ref_at_norm_incidence);
            fresnel += ref_at_norm_incidence;

            // Specular BRDF.
            float Rs_numerator = fresnel * geo * roughness;
            float Rs_denominator = dotNV * dotNL;
            float Rs = max( 0.0f, Rs_numerator / Rs_denominator ); // Added max() to prevent black areas.

            lightColorDiffuseAndSpecular = max( 0.0f, dotNL ) * (specularTex * Rs + albedoTex) * pointLightColors[ nLightIndex ].rgb * falloff;
        }

        accumDiffuseAndSpecular += lightColorDiffuseAndSpecular;
    }

    // move past the first sentinel to get to the spot lights
    nIndex++;
    nNextLightIndex = perTileLightIndexBuffer[ nIndex ];

    [loop] // Spot lights
    while (nNextLightIndex != LIGHT_INDEX_BUFFER_SENTINEL)
    {
        uint nLightIndex = nNextLightIndex;
        nIndex++;
        nNextLightIndex = perTileLightIndexBuffer[ nIndex ];

        float4 centerAndRadius = spotLightBufferCenterAndRadius[ nLightIndex ];
        float4 spotParams = spotLightBufferSpotParams[ nLightIndex ];
        
        // reconstruct z component of the light dir from x and y
        float3 spotLightDir;
        spotLightDir.xy = spotParams.xy;
        spotLightDir.z = sqrt( 1 - spotLightDir.x * spotLightDir.x - spotLightDir.y * spotLightDir.y );

        // the sign bit for cone angle is used to store the sign for the z component of the light dir
        spotLightDir.z = (spotParams.z > 0) ? spotLightDir.z : -spotLightDir.z;

        float3 vecToLight = normalize( centerAndRadius.xyz - input.Position.xyz );
        float spotAngle = dot( -spotLightDir.xyz, vecToLight );
        float cosineOfConeAngle = (spotParams.z > 0) ? spotParams.z : -spotParams.z;

        // Falloff
        float dist = distance( input.Position.xyz, centerAndRadius.xyz );
        float a = dist / centerAndRadius.w + 1;
        float att = 1.0f / (a * a);
        float3 color = spotLightColors[ nLightIndex ].rgb * att * specularTex;

        accumDiffuseAndSpecular += spotAngle > cosineOfConeAngle ? color : float3(0.0, 0.0, 0.0);

        //accumDiffuseAndSpecular += LightColorDiffuse + LightColorSpecular;
        //accumDiffuseAndSpecular = float3( 0, 1, 0 );
    }

    accumDiffuseAndSpecular *= 2.0f;

    return float4( albedoTex * ambient + accumDiffuseAndSpecular, 1.0f );

    //float col = screenPos.y / float( g_uWindowHeight );
    //return float4(col, col, col, 1.0f);

    //float col = nTileIndex / float( GetNumTilesX() * GetNumTilesY() );
    //return float4( col, col, col, 1.0f );

    //float col = (float)(nTileIndex % GetNumTilesX()) / (float)GetNumTilesX();
    //return float4( col, col, col, 1.0f);
    
    //uint numLightsInThisTile = GetNumLightsInThisTile( nTileIndex );
    //float percentOfMax = (float)numLightsInThisTile / 100.0f;// (float)g_uMaxNumLightsPerTile;
    //return float4( percentOfMax, percentOfMax, percentOfMax, 1.0f );
    
    //uint numLightsInThisTile = GetNumLightsInThisTile( nTileIndex );
    //float color = numLightsInThisTile > 0 ? 1 : 0;
    //return float4( 0, color, 0, 1);
}