#include "Aether3D.hpp"
#include <cassert>
#include <fstream>
#include <iostream>
#include <string>

#if !defined AETHER3D_WINDOWS && !defined AETHER3D_UNIX
#    error "Platform not defined! It must be either AETHER3D_WINDOWS or AETHER3D_UNIX."
#endif
#if defined AETHER3D_WINDOWS && !defined NDEBUG
#    define _CRTDBG_MAP_ALLOC
#    include <stdlib.h>
#    include <crtdbg.h>
#endif
#ifdef AETHER3D_WINDOWS
    #define API_FUNC GetProcAddress
    #define WIN32_LEAN_AND_MEAN
    #define VC_EXTRALEAN
    #include <windows.h>
#endif
#ifdef AETHER3D_UNIX
    #include <dlfcn.h>
    #define API_FUNC dlsym
#endif

typedef Audio* (*createAudio_t)();
static createAudio_t CreateAudio = nullptr;

typedef Camera* (*createCamera_t)();
static createCamera_t CreateCamera = nullptr;

typedef KeyboardInput* (*createKeyboardInput_t)();
static createKeyboardInput_t CreateKeyboardInput = nullptr;

typedef Light* (*createLight_t)( Light::Type lightType );
static createLight_t CreateLight = nullptr;

typedef MouseInput* (*createMouseInput_t)();
static createMouseInput_t CreateMouseInput = nullptr;

typedef Material* (*createMaterial_t)();
static createMaterial_t CreateMaterial = nullptr;

typedef Model* (*createModel_t)(const char*, Error::Enum*);
static createModel_t CreateModel = nullptr;

typedef Renderer* (*createRenderer_t)(const char*);
static createRenderer_t CreateRenderer = nullptr;

typedef Scene* (*createScene_t)();
static createScene_t CreateScene = nullptr;

typedef Texture* (*createTexture_t)(const char*, bool, bool, Texture::Tiling, Texture::Filter);
static createTexture_t CreateTexture = nullptr;

typedef Texture* (*createTextureCube_t)(const char*, const char*,
                                        const char*, const char*,
                                        const char*, const char*);
static createTextureCube_t CreateTextureCube = nullptr;

typedef Texture* (*createRenderTexture_t)(int, int);
static createRenderTexture_t CreateRenderTexture = nullptr;

typedef Texture* (*createRenderTextureCube_t)(int);
static createRenderTextureCube_t CreateRenderTextureCube = nullptr;

typedef Texture* (*createTextureFromAtlas_t)(const char*, const char*, const char*);
static createTextureFromAtlas_t CreateTextureFromAtlas = nullptr;

typedef unsigned (*rendererVersion_t)();
static rendererVersion_t RendererVersion = nullptr;

namespace Global
{
    std::string returnCodeToErrorString[ Error::count ];
}

static void InitErrorMap()
{
    Global::returnCodeToErrorString[ (int)Error::NO_ERRORS ] = "No errors.";
    Global::returnCodeToErrorString[ (int)Error::DLL_NOT_FOUND ] = "DLL not found or unreadable.";
    Global::returnCodeToErrorString[ (int)Error::UNSUPPORTED_RENDERER ] = "Unsupported renderer on this system.";
    Global::returnCodeToErrorString[ (int)Error::FACTORY_NOT_FOUND ] = "Class factory not found.";
    Global::returnCodeToErrorString[ (int)Error::DLL_VERSION_MISMATCH ] = "DLL version does not match header version.";
    Global::returnCodeToErrorString[ (int)Error::CONTRACT_VIOLATION ] = "Expected result does not match real result.";
    Global::returnCodeToErrorString[ (int)Error::SHADER_NOT_FOUND ] = "Shader source file not found.";
    Global::returnCodeToErrorString[ (int)Error::VERTEX_SHADER_COMPILE_FAILED ] = "Vertex shader compile failed.";
    Global::returnCodeToErrorString[ (int)Error::FRAGMENT_SHADER_COMPILE_FAILED ] = "Fragment shader compile failed.";
    Global::returnCodeToErrorString[ (int)Error::GEOMETRY_SHADER_COMPILE_FAILED ] = "Geometry shader compile failed.";
    Global::returnCodeToErrorString[ (int)Error::SHADER_LINK_FAILED ] = "Shader linking failed.";
    Global::returnCodeToErrorString[ (int)Error::FILE_NOT_FOUND ] = "File not found.";
    Global::returnCodeToErrorString[ (int)Error::CORRUPTED_DATA ] = "Corrupted data.";
    Global::returnCodeToErrorString[ (int)Error::OUT_OF_MEMORY ] = "Out of memory.";
    Global::returnCodeToErrorString[ (int)Error::COMPUTE_SHADER_COMPILE_FAILED ] = "Compute shader compile failed.";
    Global::returnCodeToErrorString[ (int)Error::GENERIC ] = "Generic error.";
}

struct Aether3D_API::Impl
{
#ifdef AETHER3D_UNIX
    void* api;
#endif
#ifdef AETHER3D_WINDOWS
    HMODULE api;
#endif
};

Aether3D_API::Aether3D_API() : m( new Impl )
{
    m->api = nullptr;
}

Aether3D_API::~Aether3D_API()
{
#ifdef AETHER3D_WINDOWS
    if (m->api)
    {
        FreeLibrary( m->api );
    }
#endif
}

Error::Enum Aether3D_API::LoadDLL( const std::string& path )
{
#if defined AETHER3D_WINDOWS && !defined NDEBUG && defined _MSC_VER
	_CrtSetDbgFlag(_CRTDBG_LEAK_CHECK_DF | _CRTDBG_ALLOC_MEM_DF);
	_CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_FILE);
	_CrtSetReportFile(_CRT_ASSERT, _CRTDBG_FILE_STDERR);
#endif

    std::string pathWithExtension;

    if (path.find( ".dll" ) == std::string::npos &&
        path.find( ".so" ) == std::string::npos &&
        path.find( ".dylib" ) == std::string::npos)
    {
#if AETHER3D_WINDOWS
        pathWithExtension = path + ".dll";
#endif
#if __linux__
        pathWithExtension = path + ".so";
#endif
#if __APPLE__
        pathWithExtension = path + ".dylib";
#endif
    }
    else
    {
        pathWithExtension = path;
    }

    std::ifstream apiFile( pathWithExtension.c_str() );

    if (!apiFile.is_open())
    {
        std::cerr << "DLL not found: " << pathWithExtension << std::endl;
#if AETHER3D_WINDOWS
        MessageBoxA( nullptr, "Could not find engine DLL!", "Error", MB_OK | MB_ICONERROR );
#endif
    	return Error::DLL_NOT_FOUND;
    }

#ifdef AETHER3D_UNIX
    void* api = dlopen( pathWithExtension.c_str(), RTLD_NOW );
#endif
#ifdef AETHER3D_WINDOWS
	HMODULE api = LoadLibraryA( pathWithExtension.c_str() );
#endif

    if (!api)
    {
#ifdef AETHER3D_UNIX
        char* error = dlerror();
        if (error)
        {
            std::cerr << std::string( error ) << std::endl;
        }
#endif
#if AETHER3D_WINDOWS
        std::cerr << "LoadLibrary() failed with error: " << GetLastError() << std::endl;
        LPSTR messageBuffer = nullptr;
        size_t size = FormatMessageA( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
            NULL, GetLastError(), MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ), (LPSTR)&messageBuffer, 0, NULL );

        std::string message( messageBuffer, size );
        std::cerr << message << std::endl;
        MessageBoxA( nullptr, "Engine DLL found but LoadLibraryA() failed!", "Error", MB_OK );
#endif
        return Error::FACTORY_NOT_FOUND;
    }

    CreateRenderer          = (createRenderer_t)API_FUNC( api, "CreateRenderer" );
    CreateAudio             = (createAudio_t)API_FUNC( api, "CreateAudio" );
    CreateCamera            = (createCamera_t)API_FUNC( api, "CreateCamera" );
    CreateKeyboardInput     = (createKeyboardInput_t)API_FUNC( api, "CreateKeyboardInput" );
    CreateLight             = (createLight_t)API_FUNC( api, "CreateLight" );
    CreateMaterial          = (createMaterial_t)API_FUNC( api, "CreateMaterial" );
    CreateModel             = (createModel_t)API_FUNC( api, "CreateModel" );
    CreateMouseInput        = (createMouseInput_t)API_FUNC( api, "CreateMouseInput" );
    CreateScene             = (createScene_t)API_FUNC( api, "CreateScene" );
    CreateTexture           = (createTexture_t)API_FUNC( api, "CreateTexture" );
    CreateTextureCube       = (createTextureCube_t)API_FUNC( api, "CreateTextureCube" );
    CreateRenderTexture     = (createRenderTexture_t)API_FUNC( api, "CreateRenderTexture" );
    CreateRenderTextureCube = (createRenderTextureCube_t)API_FUNC( api, "CreateRenderTextureCube" );
    CreateTextureFromAtlas  = (createTextureFromAtlas_t)API_FUNC( api, "CreateTextureFromAtlas" );
    RendererVersion         = (rendererVersion_t)API_FUNC( api, "RendererVersion" );

    if (!RendererVersion)
    {
        std::cerr << "Could not load method RendererVersion from DLL!" << std::endl;
#if AETHER3D_WINDOWS
        MessageBoxA( nullptr, "Could not load method RendererVersion from engine DLL!", "Error", MB_OK );
#endif
        return Error::FACTORY_NOT_FOUND;
    }

    if (RendererVersion() == Renderer::apiVersion)
    {
        return Error::NO_ERRORS;
    }
    else
    {
#if AETHER3D_WINDOWS
        MessageBoxA( nullptr, "Engine DLL version does not match Include/Aether3D.hpp version!", "Error", MB_OK );
#endif
        return Error::DLL_VERSION_MISMATCH;
    }
}

Audio* Aether3D_API::MakeAudio()
{
    assert( CreateAudio );
    return (CreateAudio)();
}

Camera* Aether3D_API::MakeCamera()
{
    assert( CreateCamera );
    return (CreateCamera)();
}

KeyboardInput* Aether3D_API::MakeKeyboardInput()
{
    assert( CreateKeyboardInput );
    return (CreateKeyboardInput)();
}

Light* Aether3D_API::MakeLight( Light::Type lightType )
{
    assert( CreateLight );
    return (CreateLight)( lightType );
}

Material* Aether3D_API::MakeMaterial()
{
    assert( CreateMaterial );
    return (CreateMaterial)();
}

MouseInput* Aether3D_API::MakeMouseInput()
{
    assert( CreateMouseInput );
    return (CreateMouseInput)();
}

Model* Aether3D_API::MakeModel( const std::string& path, Error::Enum& outError )
{
    assert( CreateModel );
    return (CreateModel)( path.c_str(), &outError );
}

Renderer* Aether3D_API::MakeRenderer( const std::string& builtinAssetPath )
{
    assert( CreateRenderer );
    std::string path = builtinAssetPath;
    return (CreateRenderer)(path.c_str());
}

Scene* Aether3D_API::MakeScene()
{
    assert( CreateScene );
    return (CreateScene)();
}

Texture* Aether3D_API::MakeTexture( const std::string& path, bool useMipmaps, bool sRGB, Texture::Tiling tiling, Texture::Filter filter )
{
    assert( CreateTexture );
    return (CreateTexture)( path.c_str(), useMipmaps, sRGB, tiling, filter );
}

Texture* Aether3D_API::MakeTexture( const std::string& atlasPath, const std::string& atlasMetadataPath, const std::string& fileName )
{
    assert( CreateTextureFromAtlas );
    return (CreateTextureFromAtlas)( atlasPath.c_str(), atlasMetadataPath.c_str(), fileName.c_str() );
}

Texture* Aether3D_API::MakeTexture( const char* left,  const char* right,
                                    const char* top,   const char* bottom,
                                    const char* front, const char* back )
{
    assert( CreateTextureCube );
    return (CreateTextureCube)( left, right, top, bottom, front, back );
}

Texture* Aether3D_API::MakeRenderTexture( int width, int height )
{
    assert( CreateRenderTexture );
    return (CreateRenderTexture)( width, height );
}

Texture* Aether3D_API::MakeRenderTextureCube( int size )
{
    assert( CreateRenderTextureCube );
    return (CreateRenderTextureCube)( size );
}

const std::string& Aether3D_API::ReturnCodeToErrorString( int returnCode )
{
    static bool sWasInit = false;

    if (!sWasInit)
    {
        sWasInit = true;
        InitErrorMap();
    }

    std::cout << "return code: " << (returnCode >> 8) << std::endl;
    if ((returnCode >> 8) >= (int)Error::count)
    {
        static const std::string sUnknown( "Unknown error." );
        return sUnknown;
    }

    return Global::returnCodeToErrorString[ returnCode >> 8 ];
}
