#ifndef CAMERA_H
#define CAMERA_H

#ifndef SCENEGRAPHNODE_H
#    include "SceneGraphNode.hpp"
#endif

struct Vec3;
class Texture;

/// Perspective or orthographic camera.
class Camera : public SceneGraphNode
{
  public:
    virtual ~Camera() {}

    /**
       Looks at a point.

       \param eye Camera's new position.
       \param center Point in space that's looked at.
       \param up Up-vector.
     */
    virtual void LookAt( const Vec3& eye, const Vec3& center, const Vec3& up ) = 0;

    /// \param amount Movement amount.
    virtual void MoveForward( float amount ) = 0;

    /// \param amount Movement amount.
    virtual void MoveRight( float amount ) = 0;

    /// \param amount Movement amount.
    virtual void MoveUp( float amount ) = 0;

    /**
       \param axis Axis, should be normalized.
       \param angleDegrees Angle in degrees.
    */
    virtual void OffsetRotation( const Vec3& axis, float angleDegrees ) = 0;

    /**
       Gets screen coordinate of a world point.
     
       \param worldPoint Point in scene.
       \return Vector containing screen position of worldPoint. z is unused. \todo More specific.
     */
    virtual Vec3 ScreenPoint( const Vec3& worldPoint ) const = 0;

    /**
       Sets perspective projection.
     
       \param fieldOfViewDegrees Field-of-View in degrees.
       \param aspect Aspect ratio.
       \param nearDepth Near plane distance.
       \param farDepth Far plane distance.
     */
    virtual void SetProjection( float fieldOfViewDegrees, float aspect, float nearDepth, float farDepth ) = 0;

    /**
      Sets the viewport. Defaults to the window dimension.
     
      \param left Left.
      \param right Right.
      \param top Top.
      \param bottom Bottom.
     */
    virtual void SetViewport( int left, int right, int top, int bottom ) = 0;
    
    /**
       Sets orthographic projection.

       \param left Left.
       \param right Right.
       \param bottom Bottom.
       \param top Top.
       \param nearDepth Near plane distance.
       \param farDepth Far plane distance.
     */
    virtual void SetProjection( float left, float right, float bottom, float top, float nearDepth, float farDepth ) = 0;

    /**
       Sets render target.  Defaults to screen backbuffer.
     
       \param target Target texture. Must have been created using dll.MakeRenderTexture().  Passing 0 renders to the back buffer.
     */
    virtual void SetRenderTarget( Texture* target ) = 0;

    /** \return View direction. */
    virtual Vec3 ViewDirection() const = 0;
    
    /** \return Far clip plane distance. */
    virtual float FarClipPlane() const = 0;
    
    /** \return Near clip plane distance. */
    virtual float NearClipPlane() const = 0;
};

#endif
