#ifndef MATERIAL_H
#define MATERIAL_H

#ifndef ERRORS_H
#    include "Errors.hpp"
#endif

#ifndef INCLUDE_STRING_H
#    include <string>
#    define INCLUDE_STRING_H
#endif

struct Matrix44;
class Texture;
struct Vec3;
struct Vec4;

/// Stores material properties for Mesh.  Their names/values are directly mapped into shader uniforms.
class Material
{
  public:
    virtual ~Material() {}

    /**
     Loads parameters and shader names from an ASCII file.
     File is not hot-reloaded.
     If parsing fails, parameters parsed before failure are kept.

     Example file contents:
     
     texture checker checker.png
     texture checkerNormal checker_n.png
     shaders cook_torrance cook_torrance
     param_float roughness 0.2
     param_vec3 materialSpecular 0.2 0.2 0.2
     param_texture diffuseMap checker
     param_texture normalMap checkerNormal
     
     \param path Path.
     */
    virtual void LoadFromFile( const char* path ) = 0;

    /// \return Name.
    virtual const char* Name() const = 0;
    
    /// \return opacity.
    virtual bool Opaque() const = 0;
    
    /**
       Sets opacity on/off.  Default is opaque.

       If it's off, the mesh using this material will have its textures's alpha channel applied.
     
       \param opaque Opacity.
     */
    virtual void SetOpaque( bool opaque ) = 0;

    /**
       Sets shaders.  Source language must match renderer API,
       eg. you can't use HLSL shaders if you're using OpenGL renderer.
       HLSL shaders are compiled in HLSL version 5.

       \param vertexShaderPath Vertex shader path or shader's file name without path or extension to use search paths.
       \param fragmentShaderPath Fragment (Pixel) shader path or shader's file name without path or extension to use search paths.
       \param geometryShaderPath Geometry shader path, can be an empty string if not used. Alternatively shader's file name without path or extension to use search paths. Ignored on unsupported renderers.
       \param defines Space separated shader source defines.
       \see Renderer.SetShaderSearchPaths.
       \return Error::NO_ERRORS on success, error otherwise.
     */
    virtual Error::Enum SetShaders( const std::string& vertexShaderPath,
                                    const std::string& fragmentShaderPath,
                                    const std::string& geometryShaderPath,
                                    const std::string& defines ) = 0;
    /**
      Sets a float.

      \param name Shader uniform name.
      \param value Value.
     */
    virtual void SetFloat( const std::string& name, float value ) = 0;

    /**
      Sets a Matrix.

      \param name Shader uniform name.
      \param matrix Matrix.
    */
    virtual void SetMatrix( const std::string& name, const Matrix44& matrix ) = 0;

    /// \param name Name
    virtual void SetName( const char* name ) = 0;
    
    /**
       Sets a texture.
       If the texture name is "textureMap", its UV scale and offset are in 4-dimensional vector "textureMap_ST".

      \param name Shader uniform name.
      \param texture Texture.
     */
    virtual void SetTexture( const std::string& name, const Texture* texture ) = 0;

    /**
     Sets a vector.

     \param name Shader uniform name.
     \param value Value.
     */
    virtual void SetVector( const std::string& name, const Vec3& value ) = 0;

    /**
     Sets a vector.

     \param name Shader uniform name.
     \param value Value.
     */
    virtual void SetVector( const std::string& name, const Vec4& value ) = 0;
};
#endif
