#include "MathUtil.hpp"
#include <cmath>
#include <random>
#include <ctime>

// Taken from AMD's ForwardPlus11 sample.
unsigned short MathUtil::ConvertF32ToF16( float fValueToConvert )
{
    // single precision floating point format:
    // |1|   8    |          23           |
    // |s|eeeeeeee|mmmmmmmmmmmmmmmmmmmmmmm|

    // half precision floating point format:
    // |1|  5  |    10    |
    // |s|eeeee|mmmmmmmmmm|

    unsigned int uFloatBits = (*(unsigned int *)&fValueToConvert);

    // note, this functions does not handle values that are too large (i.e. overflow), 
    // nor does it handle NaNs or infinity
    int  nExponent = (int)((uFloatBits & 0x7F800000u) >> 23) - 127 + 15;
    //assert( nExponent < 31 );

    // if the resulting value would be a denorm or underflow, then just return a (signed) zero
    if (nExponent <= 0)
    {
        return (unsigned short)((uFloatBits & 0x80000000u) >> 16);
    }

    // else, exponent is in the range [1,30], and so we can represent 
    // the value to convert as a normalized 16-bit float (with some loss of precision, of course)
    unsigned int uSignBit = (uFloatBits & 0x80000000u) >> 16;
    unsigned int uExponentBits = (unsigned int)nExponent << 10;
    unsigned int uMantissaBits = (uFloatBits & 0x007FFFFFu) >> 13;

    return (unsigned short)(uSignBit | uExponentBits | uMantissaBits);
}

float PlaneDistance( const Vec3& planePosition, const Vec3& planeNormal, const Vec3& point )
{
    return Vec3::Dot( point - planePosition, planeNormal );
}

bool SphereInsidePlane( const Vec3& sphereOrigin, float sphereRadius, const Vec3& planePosition, const Vec3& planeNormal )
{
    return -PlaneDistance( planePosition, planeNormal, sphereOrigin ) > sphereRadius;
}

bool SphereOutsidePlane( const Vec3& sphereOrigin, float sphereRadius, const Vec3& planePosition, const Vec3& planeNormal )
{
  return PlaneDistance( planePosition, planeNormal, sphereOrigin ) > sphereRadius;
}

bool SphereIntersectsPlane( const Vec3& sphereOrigin, float sphereRadius, const Vec3& planePosition, const Vec3& planeNormal )
{
    return std::fabs( PlaneDistance( planePosition, planeNormal, sphereOrigin ) ) <= sphereRadius;
}

// http://theorangeduck.com/page/correct-box-sphere-intersection
bool SphereInsideAABB( const Vec3& sphereOrigin, float sphereRadius, const Vec3& aabbMin, const Vec3& aabbMax )
{
    const Vec3 frontPos = aabbMin;
    const Vec3 frontDir( 0, 0, -1 );
    const Vec3 backPos = aabbMax;
    const Vec3 backDir( 0, 0, 1 );
    const Vec3 topPos = aabbMax;
    const Vec3 topDir( 0, 1, 0 );
    const Vec3 botPos = aabbMin;
    const Vec3 botDir( 0, -1, 0 );
    const Vec3 leftPos = aabbMin;
    const Vec3 leftDir( -1, 0, 0 );
    const Vec3 rightPos = aabbMax;
    const Vec3 rightDir( 1, 0, 0 );

    if (!SphereInsidePlane( sphereOrigin, sphereRadius, frontPos, frontDir )) { return false; }
    if (!SphereInsidePlane( sphereOrigin, sphereRadius, backPos,  backDir  )) { return false; }
    if (!SphereInsidePlane( sphereOrigin, sphereRadius, topPos,   topDir   )) { return false; }
    if (!SphereInsidePlane( sphereOrigin, sphereRadius, botPos,   botDir   )) { return false; }
    if (!SphereInsidePlane( sphereOrigin, sphereRadius, leftPos,  leftDir  )) { return false; }
    if (!SphereInsidePlane( sphereOrigin, sphereRadius, rightPos, rightDir )) { return false; }
  
    return true;
}

bool MathUtil::IntersectSphereAndAABB( const Vec3& sphereOrigin, float sphereRadius, const Vec3& aabbMin, const Vec3& aabbMax )
{
    const Vec3 frontPos = aabbMin;
    const Vec3 frontDir( 0, 0, -1 );
    const Vec3 backPos = aabbMax;
    const Vec3 backDir( 0, 0, 1 );
    const Vec3 topPos = aabbMax;
    const Vec3 topDir( 0, 1, 0 );
    const Vec3 botPos = aabbMin;
    const Vec3 botDir( 0, -1, 0 );
    const Vec3 leftPos = aabbMin;
    const Vec3 leftDir( -1, 0, 0 );
    const Vec3 rightPos = aabbMax;
    const Vec3 rightDir( 1, 0, 0 );

    const bool in_left   = !SphereOutsidePlane( sphereOrigin, sphereRadius, leftPos, leftDir );
    const bool in_right  = !SphereOutsidePlane( sphereOrigin, sphereRadius, rightPos, rightDir );
    const bool in_front  = !SphereOutsidePlane( sphereOrigin, sphereRadius, frontPos, frontDir );
    const bool in_back   = !SphereOutsidePlane( sphereOrigin, sphereRadius, backPos, backDir );
    const bool in_top    = !SphereOutsidePlane( sphereOrigin, sphereRadius, topPos, topDir );
    const bool in_bottom = !SphereOutsidePlane( sphereOrigin, sphereRadius, botPos, botDir );
  
    if (SphereIntersectsPlane(sphereOrigin, sphereRadius, topPos, topDir) && 
        in_left && in_right && in_front && in_back)
    {
        return true;
    }
    
    if (SphereIntersectsPlane(sphereOrigin, sphereRadius, botPos, botDir) && 
        in_left && in_right && in_front && in_back)
    {
        return true;
    }
  
    if (SphereIntersectsPlane(sphereOrigin, sphereRadius, leftPos, leftDir) &&
        in_top && in_bottom && in_front && in_back)
    {
        return true;
    }
  
    if (SphereIntersectsPlane(sphereOrigin, sphereRadius, rightPos, rightDir) &&
        in_top && in_bottom && in_front && in_back)
    {
        return true;
    }
  
    if (SphereIntersectsPlane(sphereOrigin, sphereRadius, frontPos, frontDir) &&
        in_top && in_bottom && in_left && in_right)
    {
        return true;
    }
  
    if (SphereIntersectsPlane(sphereOrigin, sphereRadius, backPos, backDir) &&
        in_top && in_bottom && in_left && in_right)
    {
        return true;
    }

    // Not in original algorithm, but added because we want to treat 'inside' as intersection.
    if (SphereInsideAABB( sphereOrigin, sphereRadius, aabbMin, aabbMax ))
    {
        return true;
    }

    return false;
}

bool MathUtil::IntersectAABBandAABB( const Vec3& aabb1Min, const Vec3& aabb1Max, const Vec3& aabb2Min, const Vec3& aabb2Max )
{
    return (aabb1Min.x < aabb2Max.x &&
            aabb1Max.x > aabb2Min.x &&
            aabb1Min.y < aabb2Max.y &&
            aabb1Max.y > aabb2Min.y &&
            aabb1Min.z < aabb2Max.z &&
            aabb1Max.z > aabb2Min.z);
}

bool MathUtil::AlmostEquals( float f1, float f2 )
{
    return std::fabs( f1 - f2 ) < tolerance;
}

float MathUtil::Ceil( float f )
{
    return std::ceil( f );
}

float MathUtil::Floor( float f )
{
    return std::floor( f );
}

float MathUtil::Round( float f )
{
    return std::floor( f + 0.5f );
}

void MathUtil::GetMinMax( const std::vector< Vec3 >& aPoints, Vec3& outMin, Vec3& outMax )
{
    if (!aPoints.empty())
    {
        outMin = aPoints[ 0 ];
        outMax = aPoints[ 0 ];
    }

    for (std::size_t i = 1; i < aPoints.size(); ++i)
    {
        if (aPoints[ i ].x < outMin.x)
        {
            outMin.x = aPoints[ i ].x;
        }
        
        if (aPoints[ i ].y < outMin.y)
        {
            outMin.y = aPoints[ i ].y;
        }
        
        if (aPoints[ i ].z < outMin.z)
        {
            outMin.z = aPoints[ i ].z;
        }
        
        if (aPoints[ i ].x > outMax.x)
        {
            outMax.x = aPoints[ i ].x;
        }
        
        if (aPoints[ i ].y > outMax.y)
        {
            outMax.y = aPoints[ i ].y;
        }
        
        if (aPoints[ i ].z > outMax.z)
        {
            outMax.z = aPoints[ i ].z;
        }
    }
}

void MathUtil::GetCorners( const Vec3& min, const Vec3& max, std::vector< Vec3 >& outCorners )
{
    outCorners =
    {
        Vec3( min.x, min.y, min.z ),
        Vec3( max.x, min.y, min.z ),
        Vec3( min.x, max.y, min.z ),
        Vec3( min.x, min.y, max.z ),
        Vec3( max.x, max.y, min.z ),
        Vec3( min.x, max.y, max.z ),
        Vec3( max.x, max.y, max.z ),
        Vec3( max.x, min.y, max.z )
    };
}

float MathUtil::IntersectRayAABB( const Vec3& origin, const Vec3& target, const Vec3& min, const Vec3& max )
{
    Vec3 dir = (target - origin).Normalized();
    
    Vec3 dirfrac;
    dirfrac.x = 1.0f / dir.x;
    dirfrac.y = 1.0f / dir.y;
    dirfrac.z = 1.0f / dir.z;
    
    const float t1 = (min.x - origin.x) * dirfrac.x;
    const float t2 = (max.x - origin.x) * dirfrac.x;
    const float t3 = (min.y - origin.y) * dirfrac.y;
    const float t4 = (max.y - origin.y) * dirfrac.y;
    const float t5 = (min.z - origin.z) * dirfrac.z;
    const float t6 = (max.z - origin.z) * dirfrac.z;
    
    const float tmin = Max2< float >( Max2< float >( Min2< float >(t1, t2), Min2< float >(t3, t4)), Min2< float >(t5, t6) );
    const float tmax = Min2< float >( Min2< float >( Max2< float >(t1, t2), Max2< float >(t3, t4)), Max2< float >(t5, t6) );
    
    // if tmax < 0, ray (line) is intersecting AABB, but whole AABB is behing us
    if (tmax < 0)
    {
        return -1.0f;
    }
    
    // if tmin > tmax, ray doesn't intersect AABB
    if (tmin > tmax)
    {
        return -1.0f;
    }
    
    return tmin;
}

float MathUtil::IntersectRayTriangles( const Vec3& origin, const Vec3& target, const std::vector< Vec3 >& vertices, const std::vector< Face >& faces )
{
    for (const auto& face : faces)
    {
        const Vec3& v0 = vertices[ face.a ];
        const Vec3& v1 = vertices[ face.b ];
        const Vec3& v2 = vertices[ face.c ];
        
        const Vec3 e1 = v1 - v0;
        const Vec3 e2 = v2 - v0;
        
        const Vec3 d = (target - origin).Normalized();
        
        const Vec3 h = Vec3::Cross( d, e2 );
        const float a = Vec3::Dot( e1, h );
        
        if (AlmostEquals( a, 0 ))
        {
            continue;
        }
        
        const float f = 1.0f / a;
        const Vec3 s = origin - v0;
        const float u = f * Vec3::Dot( s, h );
        
        if (u < 0 || u > 1)
        {
            continue;
        }
        
        const Vec3 q = Vec3::Cross( s, e1 );
        const float v = f * Vec3::Dot( d, q );
        
        if (v < 0 || u + v > 1)
        {
            continue;
        }

        const float t = f * Vec3::Dot( e2, q );
        
        if (t > 0.0001f)
        {
            return t;
        }
    }
    
    return -1;
}

float MathUtil::Lerp( float start, float end, float amount )
{
    return start + (end - start) * amount;
}

float MathUtil::Random( float aMin, float aMax )
{
    static std::mt19937 mt( static_cast<unsigned int>( std::time( nullptr ) ) );
    std::uniform_real_distribution<float> ud( aMin, aMax );
    return ud( mt );
}

float MathUtil::Remap( float value, float oldMin, float oldMax, float newMin, float newMax )
{
    const float oldRange = (oldMax - oldMin);
    const float newRange = (newMax - newMin);
    return (((value - oldMin) * newRange) / oldRange) + newMin;
}

bool MathUtil::IsNaN( float f )
{
    return f != f;
}

bool MathUtil::IsPowerOfTwo( unsigned i )
{
    return ((i & (i - 1)) == 0);
}
