#ifndef MODEL_H
#define MODEL_H

#ifndef INCLUDE_STRING_H
#    include <string>
#    define INCLUDE_STRING_H
#endif
#ifndef INCLUDE_VECTOR_H
#    include <vector>
#    define INCLUDE_VECTOR_H
#endif
#ifndef SCENEGRAPHNODE_H
#    include "SceneGraphNode.hpp"
#endif

class Material;
class Quaternion;
struct Vec3;

/// Contains meshes loaded from Aether3D's own .ae3d file format.
class Model : public SceneGraphNode
{
  public:
    virtual ~Model() {}

    /**
       \param outMin Minimum corner of AABB in world-space.
       \param outMax Maximum corner of AABB in world-space.
     */
    virtual void GetAABB( Vec3& outMin, Vec3& outMax ) const = 0;
    
    /// \return Mesh names.
    virtual std::vector< std::string > GetMeshNames() const = 0;

    /**
       Gets joint names for a Mesh.
     
       \param meshName Mesh name.
       \return Array of joint names for meshName.  If mesh or joints not found, returns an empty array.
     */
    virtual std::vector< std::string > GetMeshJointNames( const std::string& meshName ) const = 0;

    /// \return Path that was given to Aether3D_API::MakeModel().
    virtual const std::string& GetPath() const = 0;
    
    /**
       Applies skeletal or blendshape animation for all meshes.  Use parameters 0, 0 to apply the whole animation.
     
       \param firstKeyFrame First key frame of the animation that is applied.
       \param lastKeyFrame Last key frame of the animation that is applied.
     */
    virtual void ApplyAnimation( int firstKeyFrame, int lastKeyFrame ) = 0;

    /**
       Gets Mesh's material.

       \param meshName Mesh's name.
       \return Material or null if mesh or material is invalid.
    */
    virtual Material* GetMeshMaterial( const char* meshName ) = 0;

    /// \param enable True, if the model should be rendered as a wireframe.
    virtual void EnableWireframe( bool enable ) = 0;

    /**
       Sets a material for a mesh.  If mesh is not found, does nothing.
     
       \param material Material.
       \param meshName Mesh's name (case-sensitive).
     */
    virtual void SetMaterialForMesh( Material* material, const std::string& meshName ) = 0;

    /**
       Sets a material for meshes whose name contain meshNameSubstring.  If meshes are not found, does nothing.
     
       \param material Material.
       \param meshNameSubstring Part of mesh's name (case-sensitive).
     */
    virtual void SetMaterialForMeshSubstring( Material* material, const std::string& meshNameSubstring ) = 0;
};
#endif
