#ifndef SCENEGRAPHNODE_H
#define SCENEGRAPHNODE_H

#ifndef INCLUDE_LIST_H
#    include <list>
#    define INCLUDE_LIST_H
#endif
#ifndef INCLUDE_STRING_H
#    include <string>
#    define INCLUDE_STRING_H
#endif
#ifndef MATRIX_H
#    include "Matrix.hpp"
#endif
#ifndef VEC3_H
#    include "Vec3.hpp"
#endif
#ifndef QUATERNION_H
#    include "Quaternion.hpp"
#endif

class Scene;

/// Scene Graph Node. Models, Lights and Cameras are derived from this class.
#ifdef SIMD_SSE3
class BEGIN_ALIGNED( SceneGraphNode, 16 )
#else
class SceneGraphNode
#endif
{
  public:
    /** Graph node type. */
    enum class Type { None, Model, Camera, Light };

    /**
       Registers a Scene for listening node additions.
       Called automatically when creating scene from a DLL,
       but on iOS you must call this because it doesn't use a DLL.

       \param scene Scene.
     
       FIXME: This is a serious violation of good OO design: base classes should not know about their derived classes.
    */
    static void RegisterScene( Scene* scene );
    
    /**
      Unregisters a Scene from node addition listening.
     */
    static void UnregisterScene();
    
    /** Constructor */
    SceneGraphNode();
    
    /** Destructor */
    virtual ~SceneGraphNode();
    
    /**
       \param child Child. 
       \return True if succeeded, false if failed.
     */
    virtual bool AddChild( SceneGraphNode* child );

    /**
       \param node Node that's searched in this node and its children.
       \return true if this node or one of its children is node.
    */    
    virtual bool ContainsChild( const SceneGraphNode* node ) const;

    /**
        \param name Node's name that's searched in this node and its children.
        \param outNode First found node that matches name (case-sensitive) or unset if not found.
     */
    virtual void FindFirstChildWithName( const char* name, SceneGraphNode** outNode );
    
    /**
       Sets layer mask.  Used to select which cameras should render which models.
       Does not affect lights.

       \param mask layerMask.
    */
    virtual void SetLayerMask( unsigned mask ) { layerMask = mask; }
    
    /** newName Name. */
    virtual void SetName( const std::string& newName ) { name = newName; }
    
    /** \param aPosition Absolute position. */
    virtual void SetPosition( const Vec3& aPosition ) { position = aPosition; dirtyTransform = true; }
    
    /** \param aRotation Absolute rotation. */
    virtual void SetRotation( const Quaternion& aRotation ) { rotation = aRotation; dirtyTransform = true; }

    /** \return Local rotation. */
    virtual Quaternion GetRotation() const { return rotation; }
    
    /** \param aScale Absolute scale */
    virtual void SetScale( float aScale ) { scale = aScale; dirtyTransform = true; }

    /** \return Local scale. */
    virtual float GetScale() const { return scale; }
    
    /** Traverses the graph and creates transformation matrices. */
    virtual void SolveMatrices();
    
    /** \return parent */
    virtual SceneGraphNode* Parent() { return parent; }
    
    /** \return name */
    virtual std::string GetName() const { return name; }
    
    /** \return type */
    virtual Type GetType() const { return type; }
    
    /** \return layer mask. */
    virtual unsigned GetLayerMask() const { return layerMask; }
    
    /** \return children */
    virtual std::list< SceneGraphNode* >& Children() { return children; }
    
    /** \return Local position. */
    virtual Vec3 GetPosition() const { return position; }

    /** \return root */
    virtual SceneGraphNode* Root();

    /** Children */
    std::list< SceneGraphNode* > children;

    /** Cleans all dirty flags.  Used internally after transform matrices are solved. */
    void CleanDirties();

    /**
     Used internally during rendering to determine if the transform
     has changed since the beginning of Renderer::Draw( scene ).
     Used to recalculate childrens's AABBs.
     
     \return True, if the transform or some of its parent has changed since the beginning of Renderer::Draw( scene )
     */
    bool IsDirty() const
    {
        bool outDirty = dirtyTransform;
        
        if (outDirty)
        {
            return true;
        }
        
        SceneGraphNode* testParent = parent;
        
        while (testParent)
        {
            if (testParent->dirtyTransform)
            {
                return true;
            }
            else
            {
                testParent = testParent->parent;
            }
        }
        
        return false;
    }

    /** Created from position, rotation and parent hierarchy starting from Scene's root node. */
    Matrix44 modelMatrix;

protected:
    /** Type */
    Type type;
    /** Scale */
    float scale;
    /** Position */
    Vec3 position;
    /** Parent */
    SceneGraphNode* parent;
    /** Rotation */
    Quaternion rotation;
    /** Needed currently only in Editor. */
    std::string name;
    /** Needed to update children's transform. */
    bool dirtyTransform = true;
    /** Layer mask.  Affects Camera-Model interaction. */
    unsigned layerMask = 0xFFFFFFFF;

  private:
    /**
       Finds a node in this node or children.

       \param node Node.
       \param outFound If node is found, set to true, else not set.
    */
    void FindChild( const SceneGraphNode* node, bool& outFound ) const;

    /** Used to notify Scene when nodes are added. FIXME: Serious violation of good OO design:
        base classes should not know about their derived classes. */
    static Scene* sScene;
}
#ifdef SIMD_SSE3
END_ALIGNED( SceneGraphNode, 16 )
#endif
;
#endif
