#ifndef VEC3_H
#define VEC3_H

#ifndef INCLUDE_MATH_H
#    include <cmath>
#    define INCLUDE_MATH_H
#endif
#ifndef INCLUDE_ALGORITHM_H
#    include <algorithm>
#    define INCLUDE_ALGORITHM_H
#endif

/**
  \brief
   3-component vector.

   Aether3D's coordinate system is right-handed:
     up: 0, 1, 0
     forward: 0, 0, -1
     right: 1, 0, 0
 */
struct Vec3
{
    float x, y, z;

    /**
       \brief Cross product.

       \param v1 Vector.
       \param v2 Another vector.
       \return Cross product of this vector and v.
     */
    static Vec3 Cross( const Vec3& v1, const Vec3& v2 )
    {
        return Vec3( v1.y * v2.z - v1.z * v2.y,
                     v1.z * v2.x - v1.x * v2.z,
                     v1.x * v2.y - v1.y * v2.x );
    }

    /**
     Solves distance between two vectors.
     
     \param v1 Vector.
     \param v2 Other vector.
     \return Distance.
     \see DistanceSquared()
     */
    static float Distance( const Vec3& v1, const Vec3& v2 )
    {
        return (v2 - v1).Length();
    }

    /**
     Solves squared distance between two vectors.
     
     \param v1 Vector.
     \param v2 Other vector.
     \return Squared distance.
     \see Distance()
     */
    static float DistanceSquared( const Vec3& v1, const Vec3& v2 )
    {
        const Vec3 sub = v2 - v1;
        return sub.x * sub.x + sub.y * sub.y + sub.z * sub.z;
    }

    /**
     Dot product.
     
     \param v1 Vector.
     \param v2 Other vector.
     \return Dot product of vectors v1 and v2.
     */
    static float Dot( const Vec3& v1, const Vec3& v2 )
    {
        return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
    }

    /**
       Gets minimum components and creates a new vector from them.
       Example: \code v1( 1, 0, 2 ), v2( 2, 1, 1 ) -> result( 1, 0, 1 ) \endcode.

       \param v1 Vector.
       \param v2 Other vector.
       \return Vector that has v1's and v2's minimum components.
       \see Max2()
     */
    static Vec3 Min2( const Vec3& v1, const Vec3& v2 )
    {
        return Vec3( std::min( v1.x, v2.x ), std::min( v1.y, v2.y ), std::min( v1.z, v2.z ) );
    }

    /**
     Gets maximum components and creates a new vector from them.
     Example: v1( 1, 0, 2 ), v2( 2, 1, 1 ) -> result( 2, 1, 2 ).

     \param v1 Vector.
     \param v2 Other vector.
     \return Vector that has v1's and v2's maximum components.
     */
    static Vec3 Max2( const Vec3& v1, const Vec3& v2 )
    {
        return Vec3( std::max( v1.x, v2.x ), std::max( v1.y, v2.y ), std::max( v1.z, v2.z ) );
    }

    /**
     Reflects a vector with normal.
     
     \param vec Vector.
     \param normal Normal.
     \return Reflected vector.
     */
    static Vec3 Reflect( const Vec3& vec, const Vec3& normal ) { return vec - normal * Dot( normal, vec ) * 2; }

    /// Copy constructor.
    inline Vec3( const Vec3& other ) = default;

    /// Constructor.
    inline Vec3() : x( 0 ), y( 0 ), z( 0 ) {}

    /**
     Constructor.

     \param ax X-coordinate.
     \param ay Y-coordinate.
     \param az Z-coordinate.
     */
    inline Vec3( float ax, float ay, float az ) : x( ax ), y( ay ), z( az ) {}

    /**
       Almost equality.

       \param v another vector.
       \return true, if vectors are equal.
     */
    bool IsAlmost( const Vec3& v ) const
    {
        const float acceptableDelta = 0.0001f;
        if (std::fabs( x - v.x ) > acceptableDelta) { return false; }
        if (std::fabs( y - v.y ) > acceptableDelta) { return false; }
        if (std::fabs( z - v.z ) > acceptableDelta) { return false; }
        return true;
    }

    /**
       Divides by a scalar operator.

       \param f Scalar value.
       \return a vector that has its values divided by f.
     */
    inline Vec3 operator/( float f ) const
    {
        const float inv = 1.0f / f;
        return Vec3( x * inv, y * inv, z * inv );
    }

    /**
       Divides by a vector operator.
     
       \param v Vector value.
       \return a vector that has its values divided by v.
     */
    inline Vec3 operator/( const Vec3& v ) const
    {
        return Vec3( x / v.x, y / v.y, z / v.z );
    }

    /**
       Negation operator.

       \return a vector that has negative values.
     */
    inline Vec3 operator-() const
    {
        return Vec3( -x, -y, -z );
    }

    /**
       Subtraction operator.

       \param v a vector that is to be subtracted from this vector.
       \return result of the subtraction.
     */
    inline Vec3 operator-( const Vec3& v ) const
    {
        return Vec3( x - v.x, y - v.y, z - v.z );
    }

    /**
       Addition operator.

       \param v a vector that is to be added to this vector.
       \return A vector that is the sum of this vector and v.
     */
    inline Vec3 operator+( const Vec3& v ) const
    {
        return Vec3( x + v.x, y + v.y, z + v.z );
    }

    /**
       Addition operator.

       \param f a value that is to be added to this vector.
       \return A vector that is the sum of this vector and c.
     */
    inline Vec3 operator+( float f ) const
    {
        return Vec3( x + f, y + f, z + f );
    }

    /**
       Equality operator

       \param v Vector that is tested for equality.
       \return True, if component-wise values are equal, false otherwise.
     */
    inline bool operator==( const Vec3& v) const
    {
        return x == v.x && y == v.y && z == v.z;
    }
    
    /**
       Inequality operator

       \param v Vector that is tested for equality.
       \return True, if component-wise values are not equal, false otherwise.
     */
    inline bool operator!=( const Vec3& v) const
    {
        return x != v.x || y != v.y || z != v.z;
    }

    /**
       Operator subtract and assign.

       \param v Vector that is to be subtracted from this vector.
       \return Reference to this vector.
     */
    inline Vec3& operator-=( const Vec3& v )
    {
        x -= v.x;
        y -= v.y;
        z -= v.z;
        
        return *this;
    }

    /**
     Operator add and assign.

     \param v Vector that is to be added to this vector.
     \return Reference to this vector.
     */
    inline Vec3& operator+=( const Vec3& v )
    {
        x += v.x;
        y += v.y;
        z += v.z;

        return *this;
    }

    /**
       Operator multiply and assign.

       \param f A value that is to be multiplied with this vector.
       \return Reference to this vector.
     */
    inline Vec3& operator*=( float f )
    {
        x *= f;
        y *= f;
        z *= f;
        
        return *this;
    }

    /**
    Operator divide and assign.

    \param f A value that is to be divided with this vector.
    \return Reference to this vector.
    */
    inline Vec3& operator/=(float f)
    {
        const float inv = 1.0f / f;

        x *= inv;
        y *= inv;
        z *= inv;

        return *this;
    }

    /**
       Operator multiply and assign.
     
       \param v A value that is to be multiplied with this vector.
       \return Reference to this vector.
     */
    inline Vec3& operator*=( const Vec3& v )
    {
        x *= v.x;
        y *= v.y;
        z *= v.z;
        
        return *this;
    }

    /**
       Assignment operator.

       \param v a vector that is assigned to this vector.
       \return Reference to this vector.
     */
    inline Vec3& operator=( const Vec3& v ) { x = v.x; y = v.y; z = v.z; return *this; }

    /**
       Multiplication operator.

       \param v A vector that is multiplied with this vector
       \return A vector that is the product of this vector and v.
     */
    inline Vec3 operator*( const Vec3& v ) const
    {
        return Vec3( x * v.x, y * v.y, z * v.z );
    }

    /**
       Multiplication operator.

       \param f A value that is multiplied with this vector.
       \return This vector multiplied by f.
     */
    inline Vec3 operator*( float f ) const
    {
        return Vec3( x * f, y * f, z * f );
    }

    /**
       Returns a normalized vector
     */
    inline Vec3 Normalized() const
    {
        Vec3 out = *this;
        
        const float len = Length();

        if( std::abs( len ) < 0.0001f )
        {
            out.x = 1.0f;
            out.y = 0.0f;
            out.z = 0.0f;
            return out;
        }

        out *= (1.0f / len);
        return out;
    }

    /// \return length.
    inline float Length() const { return std::sqrt( x * x + y * y + z * z ); }
    
    /// Resets this vector's values to 0.
    inline void Zero() { x = y = z = 0; }
};

/// 4-component vector.
#if defined( _MSC_VER ) || defined( COMPILER_INTEL )
#       define BEGIN_ALIGNED( n, a )                            __declspec( align( a ) ) n
#       define END_ALIGNED( n, a )
#elif defined( __GNUC__ ) || defined( __clang__ )
#       define BEGIN_ALIGNED( n, a )                            n
#       define END_ALIGNED( n, a )                              __attribute__( (aligned( a ) ) )
#endif

#ifdef SIMD_SSE3
struct BEGIN_ALIGNED( Vec4, 16 )
#else
struct Vec4
#endif
{
    float x, y, z, w;

    /// Constructor.
    inline Vec4() : x( 0 ), y( 0 ), z( 0 ), w( 0 ) {}

    /**
     Constructor.

     \param ax X-coordinate.
     \param ay Y-coordinate.
     \param az Z-coordinate.
     \param aw W-coordinate.
     */
    inline Vec4( float ax, float ay, float az, float aw ) : x( ax ), y( ay ), z( az ), w( aw ) {}

    /// \param v Vec3
    inline explicit Vec4( const Vec3& v ) : x( v.x ), y( v.y ), z( v.z ), w( 1 ) {}

    /**
       Constructor.

       \param v Vec3
       \param aW .w component
    */
    inline Vec4( const Vec3& v, float aW ) : x( v.x ), y( v.y ), z( v.z ), w( aW ) {}

    /**
     Multiplication operator.

     \param f A value that is multiplied with this vector.
     \return This vector multiplied by f.
     */
    inline Vec4 operator*( float f ) const
    {
        return Vec4( x * f, y * f, z * f, w * f );
    }

    /**
     Operator addition and assign.

     \param v Vector that is to be added to this vector.
     */
    inline Vec4& operator+=( const Vec4& v )
    {
        x += v.x;
        y += v.y;
        z += v.z;

        return *this;
    }

    /**
     Operator subtract and assign.

     \param v Vector that is to be subtracted from this vector.
     */
    inline Vec4& operator-=( const Vec4& v )
    {
        x -= v.x;
        y -= v.y;
        z -= v.z;

        return *this;
    }

    /**
     Dot product.

     \param v another vector.
     \return Dot product of this vector and v.
     */
    inline float Dot( const Vec4& v ) const
    {
        return x * v.x + y * v.y + z * v.z + w * v.w;
    }

    /// \return length.
    inline float Length() const { return std::sqrt( x * x + y * y + z * z ); }

    /**
     Normalizes this vector.
     */
    inline void Normalize()
    {
        const float len = Length();

        if( std::abs( len ) < 0.0001f )
        {
            x = 1;
            y = 0;
            z = 0;
            return;
        }

        const float iLength = 1 / len;

        x *= iLength;
        y *= iLength;
        z *= iLength;
    }
}
#ifdef SIMD_SSE3
END_ALIGNED( Vec4, 16 )
#endif
;

#endif
