#define WIN32_LEAN_AND_MEAN
#define VC_EXTRALEAN
#include "GfxDevice.hpp"
#include <map>
#include <windows.h>
#include <d3d11.h>
#include "Log.hpp"
#include "RendererD3D11State.hpp"
#include "RendererImpl.hpp"
#include "Statistics.hpp"

#define AE3D_SAFE_RELEASE(x) if (x) { x->Release(); x = nullptr; }

namespace Global
{
    extern std::map< ID3D11ShaderResourceView*, RenderTextureData > textureToRT;
}

// Initialized by RendererD3D11
namespace GfxDeviceD3D11
{
    ID3D11Device* d3dDev = nullptr;
    ID3D11DeviceContext* context = nullptr;
}

struct GfxDevice::Impl
{
    Impl()
    {
        clearColor[ 0 ] = clearColor[ 1 ] = clearColor[ 2 ] = 0;
    }

    void CreateDepthStencilStates();
    void CreateBlendStates();
    void CreateRasterizerStates();

    ID3D11DepthStencilState* lessEqualWriteOn = nullptr;
    ID3D11DepthStencilState* lessEqualWriteOff = nullptr;
    ID3D11DepthStencilState* neverWriteOff = nullptr;
    ID3D11RasterizerState* uiRasterizerState = nullptr;
    ID3D11RasterizerState* meshRasterizerState = nullptr;
    ID3D11RasterizerState* transparentMeshRasterizerState = nullptr;
    ID3D11BlendState* alphaBlendState = nullptr;
    ID3D11BlendState* additiveBlendState = nullptr;
    float clearColor[ 4 ];
};

GfxDevice::GfxDevice() : m( new Impl() )
{
}

GfxDevice::~GfxDevice()
{
}

const TextureImpl* GfxDevice::CurrentRenderTarget()
{
    return rendererD3D11State.currentRenderTarget;
}

void GfxDevice::Init()
{
    Log::Assert( GfxDeviceD3D11::d3dDev != nullptr, "d3dDev is nullptr." );
    Log::Assert( GfxDeviceD3D11::context != nullptr, "context is nullptr." );

    m->CreateDepthStencilStates();
    m->CreateBlendStates();
    m->CreateRasterizerStates();
}

void GfxDevice::Deinit()
{
    GfxDeviceD3D11::d3dDev = nullptr;
    GfxDeviceD3D11::context = nullptr;

    AE3D_SAFE_RELEASE( m->lessEqualWriteOn )
    AE3D_SAFE_RELEASE( m->lessEqualWriteOff )
    AE3D_SAFE_RELEASE( m->neverWriteOff )
    AE3D_SAFE_RELEASE( m->alphaBlendState )
    AE3D_SAFE_RELEASE( m->additiveBlendState )
    AE3D_SAFE_RELEASE( m->transparentMeshRasterizerState )
    AE3D_SAFE_RELEASE( m->meshRasterizerState )
    AE3D_SAFE_RELEASE( m->uiRasterizerState )
}

void GfxDevice::SetBlendMode( BlendMode blendMode )
{
    Log::Assert( GfxDeviceD3D11::context != nullptr, "Context not initialized" );
    Log::Assert( m->alphaBlendState != nullptr, "Blend State not initialized" );
    Log::Assert( m->additiveBlendState != nullptr, "Blend State not initialized" );

    if (blendMode == BlendMode::Off)
    {
        GfxDeviceD3D11::context->OMSetBlendState( nullptr, nullptr, 0xffffffff );
    }
    else if (blendMode == BlendMode::AlphaBlend)
    {
        // null blend factor means that all values are 1.
        GfxDeviceD3D11::context->OMSetBlendState( m->alphaBlendState, nullptr, 0xffffffff );
    }
    else if (blendMode == BlendMode::Additive)
    {
        float blendFactor[ 4 ] = { 0 };

        GfxDeviceD3D11::context->OMSetBlendState( m->additiveBlendState, blendFactor, 0xffffffff );
    }
    else
    {
        Log::Assert( false, "Unhandled blend state." );
    }
}

GfxDevice& GfxDevice::Instance()
{
    static GfxDevice gfxDevice;
    return gfxDevice;
}

void GfxDevice::Impl::CreateBlendStates()
{
    Log::Assert( GfxDeviceD3D11::d3dDev != nullptr, "Device not initialized." );

    D3D11_BLEND_DESC alphaBlendDesc;
    ZeroMemory( &alphaBlendDesc, sizeof( D3D11_BLEND_DESC ) );
    alphaBlendDesc.RenderTarget[ 0 ].BlendEnable = TRUE;
    alphaBlendDesc.RenderTarget[ 0 ].SrcBlend = D3D11_BLEND_SRC_ALPHA;
    alphaBlendDesc.RenderTarget[ 0 ].DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
    alphaBlendDesc.RenderTarget[ 0 ].SrcBlendAlpha = D3D11_BLEND_ONE;
    alphaBlendDesc.RenderTarget[ 0 ].DestBlendAlpha = D3D11_BLEND_ZERO;
    alphaBlendDesc.RenderTarget[ 0 ].BlendOpAlpha = D3D11_BLEND_OP_ADD;
    alphaBlendDesc.RenderTarget[ 0 ].BlendOp = D3D11_BLEND_OP_ADD;
    alphaBlendDesc.RenderTarget[ 0 ].RenderTargetWriteMask = 0x0F;
    HRESULT hr = GfxDeviceD3D11::d3dDev->CreateBlendState( &alphaBlendDesc, &alphaBlendState );
    if (FAILED( hr ))
    {
        MessageBoxA( 0, "Failed to create alpha blend state!", "Error", MB_OK );
    }

    D3D11_BLEND_DESC additiveBlendDesc;
    ZeroMemory( &additiveBlendDesc, sizeof( D3D11_BLEND_DESC ) );
    additiveBlendDesc.RenderTarget[ 0 ].BlendEnable = TRUE;
    additiveBlendDesc.RenderTarget[ 0 ].SrcBlend = D3D11_BLEND_ONE;
    additiveBlendDesc.RenderTarget[ 0 ].DestBlend = D3D11_BLEND_ONE;
    additiveBlendDesc.RenderTarget[ 0 ].SrcBlendAlpha = D3D11_BLEND_ONE;
    additiveBlendDesc.RenderTarget[ 0 ].DestBlendAlpha = D3D11_BLEND_ONE; // FIXME: http://rastertek.com/tertut11.html has this as zero.
    additiveBlendDesc.RenderTarget[ 0 ].BlendOpAlpha = D3D11_BLEND_OP_ADD;
    additiveBlendDesc.RenderTarget[ 0 ].BlendOp = D3D11_BLEND_OP_ADD;
    additiveBlendDesc.RenderTarget[ 0 ].RenderTargetWriteMask = 0x0F;
    hr = GfxDeviceD3D11::d3dDev->CreateBlendState( &additiveBlendDesc, &additiveBlendState );
    if (FAILED( hr ))
    {
        MessageBoxA( 0, "Failed to create additive blend state!", "Error", MB_OK );
    }
}

void GfxDevice::SetViewport( int left, int right, int top, int bottom )
{

}

void GfxDevice::SetWireframe( bool enable )
{

}

void GfxDevice::Impl::CreateDepthStencilStates()
{
    Log::Assert( GfxDeviceD3D11::d3dDev != nullptr, "Device not initialized." );

    D3D11_DEPTH_STENCIL_DESC dsDesc;

    dsDesc.DepthEnable = TRUE;
    dsDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;
    dsDesc.DepthFunc = D3D11_COMPARISON_LESS_EQUAL;
    dsDesc.StencilEnable = FALSE;
    GfxDeviceD3D11::d3dDev->CreateDepthStencilState( &dsDesc, &lessEqualWriteOn );

    dsDesc.DepthEnable = TRUE;
    dsDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;
    dsDesc.DepthFunc = D3D11_COMPARISON_LESS_EQUAL;
    dsDesc.StencilEnable = FALSE;
    GfxDeviceD3D11::d3dDev->CreateDepthStencilState( &dsDesc, &lessEqualWriteOff );

    dsDesc.DepthEnable = FALSE;
    dsDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;
    dsDesc.DepthFunc = D3D11_COMPARISON_NEVER;
    dsDesc.StencilEnable = FALSE;
    GfxDeviceD3D11::d3dDev->CreateDepthStencilState( &dsDesc, &neverWriteOff );
}

void GfxDevice::Impl::CreateRasterizerStates()
{
    D3D11_RASTERIZER_DESC uiRasterizerDesc;
    ZeroMemory( &uiRasterizerDesc, sizeof( D3D11_RASTERIZER_DESC ) );
    uiRasterizerDesc.CullMode = D3D11_CULL_NONE;
    uiRasterizerDesc.DepthBias = 0;
    uiRasterizerDesc.DepthBiasClamp = 0;
    uiRasterizerDesc.DepthClipEnable = true;
    uiRasterizerDesc.FillMode = D3D11_FILL_SOLID;
    uiRasterizerDesc.FrontCounterClockwise = false;
    uiRasterizerDesc.MultisampleEnable = false;
    uiRasterizerDesc.ScissorEnable = false;
    uiRasterizerDesc.SlopeScaledDepthBias = 0;

    HRESULT hr = GfxDeviceD3D11::d3dDev->CreateRasterizerState( &uiRasterizerDesc, &uiRasterizerState );
    if (FAILED( hr ))
    {
        MessageBoxA( 0, "Failed to create UI rasterizer state!", "Error", MB_OK );
    }

    D3D11_RASTERIZER_DESC meshRasterizerDesc;
    ZeroMemory( &meshRasterizerDesc, sizeof( D3D11_RASTERIZER_DESC ) );
    meshRasterizerDesc.CullMode = D3D11_CULL_BACK;
    meshRasterizerDesc.DepthBias = 0;
    meshRasterizerDesc.DepthBiasClamp = 0;
    meshRasterizerDesc.DepthClipEnable = true;
    meshRasterizerDesc.FillMode = D3D11_FILL_SOLID;
    meshRasterizerDesc.FrontCounterClockwise = true;
    meshRasterizerDesc.MultisampleEnable = false;
    meshRasterizerDesc.ScissorEnable = false;
    meshRasterizerDesc.SlopeScaledDepthBias = 0;

    hr = GfxDeviceD3D11::d3dDev->CreateRasterizerState( &meshRasterizerDesc, &meshRasterizerState );
    if (FAILED( hr ))
    {
        MessageBoxA( 0, "Failed to create mesh rasterizer state!", "Error", MB_OK );
    }

    D3D11_RASTERIZER_DESC transparentMeshRasterizerDesc;
    ZeroMemory( &transparentMeshRasterizerDesc, sizeof( D3D11_RASTERIZER_DESC ) );
    transparentMeshRasterizerDesc.CullMode = D3D11_CULL_NONE;
    transparentMeshRasterizerDesc.DepthBias = 0;
    transparentMeshRasterizerDesc.DepthBiasClamp = 0;
    transparentMeshRasterizerDesc.DepthClipEnable = true;
    transparentMeshRasterizerDesc.FillMode = D3D11_FILL_SOLID;
    transparentMeshRasterizerDesc.FrontCounterClockwise = true;
    transparentMeshRasterizerDesc.MultisampleEnable = false;
    transparentMeshRasterizerDesc.ScissorEnable = false;
    transparentMeshRasterizerDesc.SlopeScaledDepthBias = 0;

    hr = GfxDeviceD3D11::d3dDev->CreateRasterizerState( &transparentMeshRasterizerDesc, &transparentMeshRasterizerState );
    if (FAILED( hr ))
    {
        MessageBoxA( 0, "Failed to create transparent mesh rasterizer state!", "Error", MB_OK );
    }
}

void GfxDevice::ClearScreen()
{
    if (GfxDeviceD3D11::context != nullptr)
    {
        GfxDeviceD3D11::context->ClearRenderTargetView( rendererD3D11State.currentRenderTargetView, m->clearColor );
        GfxDeviceD3D11::context->ClearDepthStencilView( rendererD3D11State.currentDepthStencilView, D3D11_CLEAR_DEPTH, 1, 0 );
    }
}

void GfxDevice::SetClearColor( float red, float green, float blue )
{
    m->clearColor[ 0 ] = red;
    m->clearColor[ 1 ] = green;
    m->clearColor[ 2 ] = blue;
    m->clearColor[ 3 ] = 1;
}

void GfxDevice::SetDepthFunc( DepthFunc func )
{
    if (func == DepthFunc::None)
    {
        GfxDeviceD3D11::context->OMSetDepthStencilState( m->neverWriteOff, 1 );
    }
    else if (func == DepthFunc::EqualWriteOn)
    {
        Log::Assert( false, "Unimplemented depth func." );

    }
    else if (func == DepthFunc::LessWriteOn)
    {
        Log::Assert( false, "Unimplemented depth func." );

    }
    else if (func == DepthFunc::LessOrEqualWriteOn)
    {
        GfxDeviceD3D11::context->OMSetDepthStencilState( m->lessEqualWriteOn, 1 );
    }
    else if (func == DepthFunc::EqualWriteOff)
    {
        Log::Assert( false, "Unimplemented depth func." );

    }
    else if (func == DepthFunc::LessWriteOff)
    {
        Log::Assert( false, "Unimplemented depth func." );

    }
    else if (func == DepthFunc::LessOrEqualWriteOff)
    {
        GfxDeviceD3D11::context->OMSetDepthStencilState( m->lessEqualWriteOff, 1 );
    }
    else
    {
        Log::Assert( false, "Unimplemented depth func." );
    }
}

void GfxDevice::SetRasterizerState( RasterizerState rasterizerState )
{
    if (rasterizerState == RasterizerState::CullBackCCW)
    {
        GfxDeviceD3D11::context->RSSetState( m->meshRasterizerState );
    }
    else if (rasterizerState == RasterizerState::CullNoneCCW)
    {
        GfxDeviceD3D11::context->RSSetState( m->transparentMeshRasterizerState );
    }
    else if (rasterizerState == RasterizerState::CullNoneCW)
    {
        GfxDeviceD3D11::context->RSSetState( m->uiRasterizerState );
    }
    else
    {
        Log::Assert( false, "Unimplemented rasterizer state." );
    }
}

void GfxDevice::SetRenderTarget( const TextureImpl* target, unsigned cubeMapFace )
{
    Log::Assert( cubeMapFace < 6, "Cube map face index too high." );

    D3D11_VIEWPORT viewport;
    viewport.MinDepth = 0;
    viewport.MaxDepth = 1;
    viewport.TopLeftX = 0;
    viewport.TopLeftY = 0;

    rendererD3D11State.currentRenderTarget = target;

    if (target == nullptr)
    {
        rendererD3D11State.immContext->OMSetRenderTargets( 1, &rendererD3D11State.backBufferRenderTargetView, rendererD3D11State.lequalDepthView );
        Statistics::IncRenderTargetSwitches();

        rendererD3D11State.currentRenderTargetView = rendererD3D11State.backBufferRenderTargetView;
        rendererD3D11State.currentDepthStencilView = rendererD3D11State.lequalDepthView;
        viewport.Width = static_cast<float>(RendererImpl::Instance().Width());
        viewport.Height = static_cast<float>(RendererImpl::Instance().Height());
        rendererD3D11State.immContext->RSSetViewports( 1, &viewport );
        return;
    }

    ID3D11ShaderResourceView* id = target->SRV();

    if (Global::textureToRT.find( id ) == Global::textureToRT.end())
    {
        Log::Print( "Tried to set a render texture that does not exist!" );
    }

    if (target->IsCubemap())
    {
        rendererD3D11State.immContext->OMSetRenderTargets( 1, &Global::textureToRT[ id ].cubeRTVs[ cubeMapFace ], Global::textureToRT[ id ].depthStencilView );
        Statistics::IncRenderTargetSwitches();

        rendererD3D11State.currentRenderTargetView = Global::textureToRT[ id ].cubeRTVs[ cubeMapFace ];
        rendererD3D11State.currentDepthStencilView = Global::textureToRT[ id ].depthStencilView;
    }
    else
    {
        rendererD3D11State.immContext->OMSetRenderTargets( 1, &Global::textureToRT[ id ].rtView, Global::textureToRT[ id ].depthStencilView );
        Statistics::IncRenderTargetSwitches();

        rendererD3D11State.currentRenderTargetView = Global::textureToRT[ id ].rtView;
        rendererD3D11State.currentDepthStencilView = Global::textureToRT[ id ].depthStencilView;
    }

    viewport.Width = static_cast<float>(target->Width());
    viewport.Height = static_cast<float>(target->Height());
    rendererD3D11State.immContext->RSSetViewports( 1, &viewport );
}

void GfxDevice::Set_sRGB_Writes( bool /*enable*/ )
{
// In D3D this is a property of render target object.
}
