#include "RenderQueue.hpp"
#include <cassert>
#include <d3d11.h>
#include "Matrix.hpp"
#include "RendererD3D11State.hpp"
//#include "RendererImpl.hpp"
#include "Shader.hpp"
#include "Statistics.hpp"
#include "TextureImpl.hpp"

RenderQueue::Vec3uv::Vec3uv() : x( 0 ), y( 0 ), z( 0 ), u( 0 ), v( 0 ) {}

RenderQueue::Vec3uv::Vec3uv( float aX, float aY, float aZ, float aU, float aV ) : x( aX ), y( aY ), z( aZ ), u( aU ), v( aV ) {}

struct RenderQueue::Impl
{
    Impl() : lastEntry( -1 ), shader( nullptr ) {}
    explicit Impl( Shader& aShader ) : lastEntry( -1 ), shader( &aShader ) {}
    void Draw( float width, float height, const Matrix44& viewProjection );
    void Sort();
    
    std::vector< ID3D11Buffer* > vertexBuffers;
    std::vector< RenderQueueEntry > entries;
    int lastEntry;
    Shader* shader;
};

RenderQueue::RenderQueue( Shader& shader ) : m( new Impl( shader ) ) {}

RenderQueue::~RenderQueue()
{
    m->shader = nullptr;
}

void RenderQueue::Add( PrimitiveType aPrimitive, const std::vector< Vec3uv >& aVertices,
                       const TextureImpl* aTexture,
                       int aSortingValue, const Vec4& color )
{
    // Finds an entry that has same state as the arguments and inserts vertices
    // there so we can render them using one draw call.
    for (auto& i : m->entries)
    {
        if (i.sortingValue == aSortingValue &&
            ((i.texture && aTexture) ? (i.texture->SRV() == aTexture->SRV()) : &i.texture == &aTexture) &&
            i.red == color.x && i.green == color.y && i.blue == color.z && i.opacity == color.w)
        {
            i.vertices.insert( std::end( i.vertices ), std::begin( aVertices ), std::end( aVertices ) );
            return;
        }
    }
    
    // Creates a new entry.
    ++m->lastEntry;
    auto& entries = m->entries;

    if ((std::size_t)m->lastEntry == entries.size())
    {
        const std::size_t newSize = entries.size() == 0 ? 20 : entries.size() * 2;
        
        entries.resize( newSize );
        m->vertexBuffers.resize( newSize );
    }
    
    RenderQueueEntry e( aSortingValue, aTexture, aVertices, color.x, color.y, color.z, color.w, aPrimitive );
    entries[ m->lastEntry ] = e;
}

void RenderQueue::Flush( float width, float height, const Matrix44& viewProjection )
{
    m->Sort();
    m->Draw( width, height, viewProjection );
    m->entries.clear();
    m->lastEntry = -1;
}

void RenderQueue::Deinit() const
{
}

void RenderQueue::Impl::Draw( float width, float height, const Matrix44& viewProjection )
{
    // TODO: Really need to create buffer objects beforehand.

    assert( shader );

	if (lastEntry == -1)
	{
		return;	
	}

    Matrix44 identity;
    Matrix44 projection;
    projection.MakeProjection( 0, (float)width, (float)height, 0, -1, 1 );

    shader->Use();
    // Scale and offset are baked into geometry to enable batching, so identity is set here.
    shader->SetVector( "textureMap_ST", Vec4( 1, 1, 0, 0 ) );
    shader->SetMatrix( "modelMatrix", identity );

    for (std::size_t e = 0; e <= (std::size_t)lastEntry; ++e)
    {
        assert( e < vertexBuffers.size() );

        auto& i = entries[ e ];
        assert( i.texture );

        const Matrix44& matrix = i.primitiveType == RenderQueue::PrimitiveType::Triangles ? projection : viewProjection;
        shader->SetMatrix( "_ViewProjectionMatrix", matrix );

        ID3D11SamplerState* const sampler = rendererD3D11State.samplers.Get( i.texture->GetSampler() );
        rendererD3D11State.immContext->PSSetSamplers( 0, 1, &sampler );

        D3D11_BUFFER_DESC bd;
        ZeroMemory( &bd, sizeof( bd ) );
        bd.Usage = D3D11_USAGE_DEFAULT;
        bd.ByteWidth = (unsigned)(sizeof( RenderQueue::Vec3uv ) * entries[ e ].vertices.size());
        bd.BindFlags = D3D11_BIND_VERTEX_BUFFER;
        bd.CPUAccessFlags = 0;
        bd.MiscFlags = 0;
        bd.StructureByteStride = 0;

        D3D11_SUBRESOURCE_DATA vertexDesc; 
        ZeroMemory( &vertexDesc, sizeof( vertexDesc ) );
        vertexDesc.pSysMem = (D3D11_SUBRESOURCE_DATA*)(&entries[ e ].vertices[0]);
        vertexDesc.SysMemPitch = 0;
	    vertexDesc.SysMemSlicePitch = 0;

        if (FAILED( rendererD3D11State.d3dDev->CreateBuffer( &bd, &vertexDesc, &vertexBuffers[ e ] ) ))
        {
            MessageBoxA( 0, "Failed to create vertex buffer!", "RenderQueue Error", MB_OK );
        }
        
        const Vec4 color( i.red, i.green, i.blue, i.opacity );
        shader->SetVector( "tint", color );

        // Binds the texture.
        ID3D11ShaderResourceView* texture = i.texture->SRV();
        assert( texture );
        rendererD3D11State.immContext->PSSetShaderResources( 0, 1, &texture );
        
        // Sets the vertex buffer.
        UINT stride = sizeof( Vec3uv );
        UINT offset = 0;
        rendererD3D11State.immContext->IASetVertexBuffers( 0, 1, &vertexBuffers[ e ], &stride, &offset );
        rendererD3D11State.immContext->IASetPrimitiveTopology( entries[ e ].primitiveType == PrimitiveType::Triangles ? D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST : D3D11_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP );
        shader->SetConstantBuffers();
#if DEBUG
        rendererD3D11State.debugDevice->ValidateContext( rendererD3D11State.immContext );
#endif
        rendererD3D11State.immContext->Draw( (unsigned)entries[ e ].vertices.size(), 0 );
        Statistics::IncDrawCalls();

        vertexBuffers[ e ]->Release();
        vertexBuffers[ e ] = nullptr;
    }
}

void RenderQueue::Impl::Sort()
{
    // \fixme This code assumes that sorting values are initialized
    // to a very large value in empty constructor.
    std::sort( std::begin( entries ), std::end( entries ),
               [](const RenderQueueEntry& e1, const RenderQueueEntry& e2) { return e1.sortingValue < e2.sortingValue; });
}
