#include <cassert>
#include <fstream>
#include <sstream>
#include <map>
#include <list>
#include <unordered_map>
#include <vector>
#include <d3d11.h>
#include <d3dcompiler.h>
#include "FileWatcher.hpp"
#include "Log.hpp"
#include "Shader.hpp"
#include "Matrix.hpp"
#include "TextureImpl.hpp"
#include "Platform.hpp"
#include "RendererD3D11State.hpp"
#include "Statistics.hpp"
#include "Vec3.hpp"

namespace Global
{
    extern std::map< ID3D11ShaderResourceView*, RenderTextureData > textureToRT;
}

// Used to keep track of created resources so that they can be released on exit.
namespace D3DShaderResource
{
    std::list< ID3D11VertexShader* > vertexShaders;
    std::list< ID3D11PixelShader* > pixelShaders;
    std::list< ID3D11GeometryShader* > geometryShaders;

    std::list< ID3D11ShaderReflection* > geometryReflectors;
    std::list< ID3D11ShaderReflection* > pixelReflectors;
    std::list< ID3D11ShaderReflection* > vertexReflectors;

    std::list< ID3D11Buffer* > constantBuffers;
    std::list< ID3D11InputLayout* > inputLayouts;
}

static HRESULT CreateInputLayoutDescFromVertexShaderSignature( ID3DBlob* pShaderBlob, ID3D11Device* pD3DDevice, ID3D11InputLayout** pInputLayout )
{
    // Reflect shader info
    ID3D11ShaderReflection* pVertexShaderReflection = nullptr; 
    if (FAILED( D3DReflect( pShaderBlob->GetBufferPointer(), pShaderBlob->GetBufferSize(), IID_ID3D11ShaderReflection, (void**) &pVertexShaderReflection ) )) 
    {
        return S_FALSE;
    }
     
    // Get shader info
    D3D11_SHADER_DESC shaderDesc;
    pVertexShaderReflection->GetDesc( &shaderDesc );
     
    // Read input layout description from shader info
    std::vector< D3D11_INPUT_ELEMENT_DESC > inputLayoutDesc;
    
    for (unsigned i = 0; i < shaderDesc.InputParameters; ++i)
    {
        D3D11_SIGNATURE_PARAMETER_DESC paramDesc;       
        pVertexShaderReflection->GetInputParameterDesc(i, &paramDesc );

        // fill out input element desc
        D3D11_INPUT_ELEMENT_DESC elementDesc;   
        elementDesc.SemanticName = paramDesc.SemanticName;
        elementDesc.SemanticIndex = paramDesc.SemanticIndex;
        elementDesc.InputSlot = 0;
        elementDesc.AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
        elementDesc.InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
        elementDesc.InstanceDataStepRate = 0;   
        //Log::PrintFormat( "semantic name: %s, offset: %d", paramDesc.SemanticName, byteOffset );

        // determine DXGI format
        if (paramDesc.Mask == 1)
        {
            if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_UINT32 ) elementDesc.Format = DXGI_FORMAT_R32_UINT;
            else if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_SINT32 ) elementDesc.Format = DXGI_FORMAT_R32_SINT;
            else if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_FLOAT32 ) elementDesc.Format = DXGI_FORMAT_R32_FLOAT;
            else
            {
                Log::Assert( false, "Unhandled shader input element format!" );
            }
        }
        else if (paramDesc.Mask <= 3)
        {
            if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_UINT32 ) elementDesc.Format = DXGI_FORMAT_R32G32_UINT;
            else if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_SINT32 ) elementDesc.Format = DXGI_FORMAT_R32G32_SINT;
            else if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_FLOAT32 ) elementDesc.Format = DXGI_FORMAT_R32G32_FLOAT;
            else
            {
                Log::Assert( false, "Unhandled shader input element format!" );
            }
        }
        else if (paramDesc.Mask <= 7)
        {
            if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_UINT32 ) elementDesc.Format = DXGI_FORMAT_R32G32B32_UINT;
            else if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_SINT32 ) elementDesc.Format = DXGI_FORMAT_R32G32B32_SINT;
            else if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_FLOAT32 ) elementDesc.Format = DXGI_FORMAT_R32G32B32_FLOAT;
            else
            {
                Log::Assert( false, "Unhandled shader input element format!" );
            }
        }
        else if (paramDesc.Mask <= 15)
        {
            if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_UINT32 ) elementDesc.Format = DXGI_FORMAT_R32G32B32A32_UINT;
            else if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_SINT32 ) elementDesc.Format = DXGI_FORMAT_R32G32B32A32_SINT;
            else if ( paramDesc.ComponentType == D3D_REGISTER_COMPONENT_FLOAT32 ) elementDesc.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
            else
            {
                Log::Assert( false, "Unhandled shader input element format!" );
            }
        }
             
        //save element desc
        inputLayoutDesc.push_back(elementDesc);
    }       
 
    // Try to create Input Layout
    HRESULT hr = pD3DDevice->CreateInputLayout( &inputLayoutDesc[0], (UINT)inputLayoutDesc.size(), pShaderBlob->GetBufferPointer(), pShaderBlob->GetBufferSize(), pInputLayout );

    if (FAILED( hr ))
    {
        MessageBoxA( 0, "Could not reflect input layout!", "Shader Error", MB_OK );
        return hr;
    }

    D3DShaderResource::inputLayouts.push_back( *pInputLayout );

    pVertexShaderReflection->Release();
    return hr;
}

struct ConstantBuffer
{
    ConstantBuffer() : buffer( nullptr ) {}

    HRESULT CreateBufferObject( unsigned byteWidth )
    {
        D3D11_BUFFER_DESC cbDesc;
        ZeroMemory( &cbDesc, sizeof( cbDesc ) );

        cbDesc.Usage = D3D11_USAGE_DYNAMIC;
	    cbDesc.ByteWidth = byteWidth;
	    cbDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
	    cbDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
	    cbDesc.MiscFlags = 0;
	    cbDesc.StructureByteStride = 0;
    
        HRESULT hr = rendererD3D11State.d3dDev->CreateBuffer( &cbDesc, nullptr, &buffer );
#if DEBUG
        static int i = 0;
        const std::string bufferName = "constant_buffer_" + std::to_string( i );
        ++i;
        buffer->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)bufferName.length(), bufferName.c_str() );
#endif
        D3DShaderResource::constantBuffers.push_back( buffer );

        return hr;
    }

    D3D11_SHADER_BUFFER_DESC desc;
    std::vector< D3D11_SHADER_VARIABLE_DESC > variables;
    ID3D11Buffer* buffer;
};

struct Shader::Impl
{
    /// Id is shader's source paths and defines appended together.
    static std::map< std::string, Shader > sIdToShaderCache;

    Impl();
    void GetUniformLocations( ID3DBlob* vertexShaderBuf, ID3DBlob* pixelShaderBuf, ID3DBlob* geometryShaderBuf );
    // Returns true if succeeded, false otherwise.
    bool GetConstantBufferIndexAndOffset( const std::string& name, unsigned& outIndex, int& outOffset );
    void Deinit();

    ID3D11VertexShader* vertexShader;
    ID3D11PixelShader* pixelShader;
    ID3D11GeometryShader* geometryShader;

    ID3D11ShaderReflection* geometryReflector;
    ID3D11ShaderReflection* pixelReflector;
    ID3D11ShaderReflection* vertexReflector;

    std::vector< ConstantBuffer > constantBuffers;

    // Used for caching and hot-loading.
    std::string vertexPath;
    std::string fragmentPath;
    std::string geometryPath;
    std::string defines;

    std::unordered_map< std::string, float > floats;
    std::unordered_map< std::string, int > ints;
    std::unordered_map< std::string, Matrix44 > matrices;
    std::unordered_map< std::string, std::vector< Matrix44 > > matrixArrays;
    std::unordered_map< std::string, Vec3 > vec3s;
    std::unordered_map< std::string, std::vector< Vec3 > > vec3Arrays;
    std::unordered_map< std::string, Vec4 > vec4s;
    std::unordered_map< std::string, std::vector< Vec4 > > vec4Arrays;
    std::unordered_map< std::string, const TextureImpl* > textures;

    ID3D11InputLayout* inputLayout;
    double compileTimeStamp;
};

Shader::Impl::Impl() 
    : vertexShader( nullptr )
    , pixelShader( nullptr )
    , geometryShader( nullptr )
    , geometryReflector( nullptr )
    , pixelReflector( nullptr )
    , vertexReflector( nullptr )
    , inputLayout( nullptr )
    , compileTimeStamp( 0 ) {}

std::map< std::string, Shader > Shader::Impl::sIdToShaderCache;

bool Shader::Impl::GetConstantBufferIndexAndOffset( const std::string& name, unsigned& outIndex, int& outOffset )
{
    unsigned cbIndex = 0;

    for (auto& cb : constantBuffers)
    {
        for (auto& var : cb.variables)
        {
            if (std::string( var.Name ) == name)
            {
                outIndex = cbIndex;
                outOffset = var.StartOffset;
                return true;
            }
        }

        ++cbIndex;
    }

    outIndex = 666;
    return false;
    //Log::PrintFormat( "None of the shader %s constant buffers contain variable %s.", vertexPath.c_str(), name.c_str() );
}

void Shader::SetConstantBuffers()
{
    for (std::size_t index = 0; index < m->constantBuffers.size(); ++index)
    {
        D3D11_MAPPED_SUBRESOURCE mappedResource;
        HRESULT result = rendererD3D11State.immContext->Map( m->constantBuffers[ index ].buffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource );
	    
        if (FAILED( result ) )
	    {
            MessageBoxA( 0, "Could not lock constant buffer!", "Shader Error", MB_OK );
            assert( !"Could not lock constant buffer!" );
		    return;
	    }

        int offset;
        unsigned varBufInd;

        for (const auto& integer : m->ints)
        {
            if (m->GetConstantBufferIndexAndOffset( integer.first, varBufInd, offset ) && varBufInd == index)
            {
                std::memcpy( (char*)mappedResource.pData + offset, (void*)&integer.second, sizeof( int ) );
            }
        }

        for (const auto& flo : m->floats)
        {
            if (m->GetConstantBufferIndexAndOffset( flo.first, varBufInd, offset ) && varBufInd == index)
            {
                std::memcpy( (char*)mappedResource.pData + offset, (void*)&flo.second, sizeof( float ) );
            }
        }

        for (const auto& vec3 : m->vec3s)
        {
            if (m->GetConstantBufferIndexAndOffset( vec3.first, varBufInd, offset ) && varBufInd == index)
            {
                std::memcpy( (char*)mappedResource.pData + offset, (void*)&vec3.second.x, sizeof( Vec3 ) );
            }
        }

        for (const auto& vec3Array : m->vec3Arrays)
        {
            if (m->GetConstantBufferIndexAndOffset( vec3Array.first, varBufInd, offset ) && varBufInd == index)
            {
                std::memcpy( (char*)mappedResource.pData + offset, (void*)&vec3Array.second[ 0 ].x, sizeof(Vec3) * vec3Array.second.size() );
            }
        }

        for (const auto& vec4 : m->vec4s)
        {
            if (m->GetConstantBufferIndexAndOffset( vec4.first, varBufInd, offset ) && varBufInd == index)
            {
                std::memcpy( (char*)mappedResource.pData + offset, (void*)&vec4.second.x, sizeof( Vec4 ) );
            }
        }

        for (const auto& vec4Array : m->vec4Arrays)
        {
            if (m->GetConstantBufferIndexAndOffset( vec4Array.first, varBufInd, offset ) && varBufInd == index)
            {
                std::memcpy( (char*)mappedResource.pData + offset, (void*)&vec4Array.second[ 0 ].x, sizeof(Vec4) * vec4Array.second.size() );
            }
        }

        for (const auto& matrix : m->matrices)
        {
            if (m->GetConstantBufferIndexAndOffset( matrix.first, varBufInd, offset ) && varBufInd == index)
            {
                std::memcpy( (char*)mappedResource.pData + offset, (void*)&matrix.second.m, sizeof( Matrix44 ) );
            }
        }

        for (const auto& matrixArray : m->matrixArrays)
        {
            if (m->GetConstantBufferIndexAndOffset( matrixArray.first, varBufInd, offset ) && varBufInd == index)
            {
                std::memcpy( (char*)mappedResource.pData + offset, (void*)&matrixArray.second[0].m, sizeof( Matrix44 ) * matrixArray.second.size() );
            }
        }

        rendererD3D11State.immContext->Unmap( m->constantBuffers[ index ].buffer, 0 );

        rendererD3D11State.immContext->VSSetConstantBuffers( (unsigned)index, 1, &m->constantBuffers[ index ].buffer );
        rendererD3D11State.immContext->PSSetConstantBuffers( (unsigned)index, 1, &m->constantBuffers[ index ].buffer );
        
        if (m->geometryShader)
        {
            rendererD3D11State.immContext->GSSetConstantBuffers( (unsigned)index, 1, &m->constantBuffers[ index ].buffer );
        }
    }
}

void Shader::Impl::Deinit()
{
    if (vertexShader)
    {
        vertexShader->Release();
        vertexShader = nullptr;
    }

    if (pixelShader)
    {
        pixelShader->Release();
        pixelShader = nullptr;
    }

    if (geometryShader)
    {
        geometryShader->Release();
        geometryShader = nullptr;
    }

    if (vertexReflector)
    {
        vertexReflector->Release();
        vertexReflector = nullptr;    
    }

    if (pixelReflector)
    {
        pixelReflector->Release();
        pixelReflector = nullptr;
    }

    if (geometryReflector)
    {
        geometryReflector->Release();
        geometryReflector = nullptr;    
    }
}

void Shader::Impl::GetUniformLocations( ID3DBlob* vertexShaderBuf, ID3DBlob* pixelShaderBuf, ID3DBlob* geometryShaderBuf )
{
    Log::Assert( vertexShader && pixelShader, "Vertex or pixel shader missing." );

    D3DReflect( vertexShaderBuf->GetBufferPointer(), vertexShaderBuf->GetBufferSize(), IID_ID3D11ShaderReflection, (void**) &vertexReflector );
    Log::Assert( vertexReflector != nullptr, "vertexReflector missing" );
    D3DShaderResource::vertexReflectors.push_back( vertexReflector );

    D3DReflect( pixelShaderBuf->GetBufferPointer(), pixelShaderBuf->GetBufferSize(), IID_ID3D11ShaderReflection, (void**) &pixelReflector );
    Log::Assert( pixelReflector != nullptr, "pixelReflector missing" );
    D3DShaderResource::pixelReflectors.push_back( pixelReflector );

    if (geometryShaderBuf)
    {
        D3DReflect( geometryShaderBuf->GetBufferPointer(), geometryShaderBuf->GetBufferSize(), IID_ID3D11ShaderReflection, (void**) &geometryReflector );
        assert( geometryReflector );
        D3DShaderResource::geometryReflectors.push_back( geometryReflector );
    }

    D3D11_SHADER_DESC vertexDesc;
    vertexReflector->GetDesc( &vertexDesc );

    for (unsigned i = 0; i < vertexDesc.ConstantBuffers; ++i)
    {
        ConstantBuffer constantBuffer;
        ID3D11ShaderReflectionConstantBuffer* pConstBuffer = vertexReflector->GetConstantBufferByIndex( i );
        pConstBuffer->GetDesc( &constantBuffer.desc );

        for (unsigned j = 0; j < constantBuffer.desc.Variables; ++j)
        {
            ID3D11ShaderReflectionVariable* pVariable = pConstBuffer->GetVariableByIndex( j );
            D3D11_SHADER_VARIABLE_DESC var_desc;
            pVariable->GetDesc( &var_desc );
            constantBuffer.variables.push_back( var_desc );
        }

        constantBuffer.CreateBufferObject( constantBuffer.desc.Size );
        constantBuffers.push_back( constantBuffer );
    }

    if (constantBuffers.empty())
    {
        D3D11_SHADER_DESC pixelDesc;
        pixelReflector->GetDesc( &pixelDesc );

        for (unsigned i = 0; i < pixelDesc.ConstantBuffers; ++i)
        {
            ConstantBuffer constantBuffer;
            ID3D11ShaderReflectionConstantBuffer* pConstBuffer = pixelReflector->GetConstantBufferByIndex( i );
            pConstBuffer->GetDesc( &constantBuffer.desc );

            for (unsigned j = 0; j < constantBuffer.desc.Variables; ++j)
            {
                ID3D11ShaderReflectionVariable* pVariable = pConstBuffer->GetVariableByIndex( j );
                D3D11_SHADER_VARIABLE_DESC var_desc;
                pVariable->GetDesc( &var_desc );
                constantBuffer.variables.push_back( var_desc );
            }

            constantBuffer.CreateBufferObject( constantBuffer.desc.Size );
            constantBuffers.push_back( constantBuffer );
        }
    }
}

Shader::Shader() : m( new Impl )
{
}

Shader::Shader( const Shader& other ) : m( new Impl )
{
    *this = other;
}

/// Cannot release shader objects here because destructor is 
/// called after Renderer::DeInit().  Solution: Shader::DeInit() is called there.
Shader::~Shader()
{
}

Shader& Shader::operator=( const Shader& other )
{
    m->vertexPath = other.m->vertexPath;
    m->fragmentPath = other.m->fragmentPath;
    m->geometryPath = other.m->geometryPath;
    
    m->vertexShader = other.m->vertexShader;
    m->pixelShader = other.m->pixelShader;
    m->geometryShader = other.m->geometryShader;
    
    m->vertexReflector = other.m->vertexReflector;
    m->pixelReflector = other.m->pixelReflector;
    m->geometryReflector = other.m->geometryReflector;

    m->defines = other.m->defines;
    m->inputLayout = other.m->inputLayout;
    m->constantBuffers = other.m->constantBuffers;
    m->compileTimeStamp = other.m->compileTimeStamp;

    return *this;
}

void Shader::Deinit()
{
    for (auto i : D3DShaderResource::vertexShaders)
    {
        i->Release();
    }

    for (auto i : D3DShaderResource::pixelShaders)
    {
        i->Release();
    }

    for (auto i : D3DShaderResource::geometryShaders)
    {
        i->Release();
    }

    for (auto i : D3DShaderResource::vertexReflectors)
    {
        i->Release();
    }

    for (auto i : D3DShaderResource::pixelReflectors)
    {
        i->Release();
    }

    for (auto i : D3DShaderResource::geometryReflectors)
    {
        i->Release();
    }

    for (auto i : D3DShaderResource::constantBuffers)
    {
        i->Release();
    }

    for (auto i : D3DShaderResource::inputLayouts)
    {
        i->Release();
    }
}

const std::string& Shader::FragmentPath() const
{
    return m->fragmentPath;
}

std::string SourceWithDefines( const std::string& source, const std::string& defines )
{
    if (defines.empty())
    {
        return source;
    }

    std::string sourceWithDefines;

    std::stringstream defineStm( defines );

    while (!defineStm.eof())
    {
        std::string define;
        defineStm >> define;
        define = std::string( "#define " ) + define;
        sourceWithDefines += define + "\n";
    }

    sourceWithDefines += "\n" + source;
    return sourceWithDefines;
}

bool Shader::Equals( const Shader& /*other*/ ) const
{
    // FIXME: If the commented line is uncommented, 2 or more cloned models using different materials cause visual corruption on D3D renderer.
    return false;
    //return m->vertexPath == other.m->vertexPath && m->fragmentPath == other.m->fragmentPath && m->geometryPath == other.m->geometryPath && m->defines == other.m->defines;
}

void Shader::Load( const std::string& vertexShaderPath,
                   const std::string& fragmentShaderPath,
                   const std::string& geometryShaderPath,
                   const std::string& defines, Error::Enum& outError )
{
    const bool firstTimeLoad = m->vertexShader == nullptr && m->pixelShader == nullptr && m->geometryShader == nullptr;

    if (firstTimeLoad && Shader::Impl::sIdToShaderCache.find( vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ) != Shader::Impl::sIdToShaderCache.end())
    {
        *this = Shader::Impl::sIdToShaderCache[ vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ];
        return;
    }

    m->vertexPath = vertexShaderPath;
    m->fragmentPath = fragmentShaderPath;
    m->geometryPath = geometryShaderPath;
    m->defines = defines;

    DWORD flags = D3DCOMPILE_ENABLE_STRICTNESS | D3DCOMPILE_WARNINGS_ARE_ERRORS;
#if defined( DEBUG ) || defined( _DEBUG )
    flags |= D3DCOMPILE_DEBUG;
#endif

    ID3DBlob* pErrorBlob = nullptr;
    ID3DBlob* pVSBlob    = nullptr;
    ID3DBlob* pPSBlob    = nullptr;
    ID3DBlob* pGSBlob    = nullptr;
    
    // Vertex shader.
    std::ifstream vertexStream( vertexShaderPath.c_str() );
    if (!vertexStream.is_open())
    {
        Log::Print( "Could not find vertex shader %s", vertexShaderPath.c_str() );
        outError = Error::SHADER_NOT_FOUND;
        return;
    }

    std::string vertexSrc = std::string( std::istreambuf_iterator< char >( vertexStream ),
                                         std::istreambuf_iterator< char >() );
    vertexSrc = SourceWithDefines( vertexSrc, defines );

    HRESULT hr = D3DCompile( vertexSrc.c_str(), vertexSrc.size(), nullptr, nullptr, nullptr, "VS", "vs_5_0", flags, 0, &pVSBlob, &pErrorBlob );
    if (FAILED( hr ))
    {
        Log::Print( "Could not compile vertex shader %s\nError: %s", vertexShaderPath.c_str(), pErrorBlob->GetBufferPointer() );
        outError = Error::VERTEX_SHADER_COMPILE_FAILED;
        return;
    }

    CreateInputLayoutDescFromVertexShaderSignature( pVSBlob, rendererD3D11State.d3dDev, &m->inputLayout );

    hr = rendererD3D11State.d3dDev->CreateVertexShader( pVSBlob->GetBufferPointer(), pVSBlob->GetBufferSize(), nullptr, &m->vertexShader );
    D3DShaderResource::vertexShaders.push_back( m->vertexShader );

    if (FAILED( hr ))
    {    
        pVSBlob->Release();
        Log::Print( "Could not create vertex shader %s", vertexShaderPath.c_str() );
        outError = Error::VERTEX_SHADER_COMPILE_FAILED;
        return;
    }

#if DEBUG
    m->vertexShader->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)vertexShaderPath.length(), vertexShaderPath.c_str() );
#endif

    // Pixel shader.
    std::ifstream fragmentStream( fragmentShaderPath.c_str() );
    if (!fragmentStream.is_open())
    {
        Log::Print( "Could not find fragment shader %s", fragmentShaderPath.c_str() );
        outError = Error::SHADER_NOT_FOUND;
        return;
    }

    std::string pixelSrc = std::string( std::istreambuf_iterator< char >( fragmentStream ),
                                        std::istreambuf_iterator< char >() );
    pixelSrc = SourceWithDefines( pixelSrc, defines );

    hr = D3DCompile( pixelSrc.c_str(), pixelSrc.size(), nullptr, nullptr, nullptr, "PS", "ps_5_0", flags, 0, &pPSBlob, &pErrorBlob );
    if (FAILED( hr ))
    {
        Log::Print( "Could not compile pixel shader %s\nError: %s", fragmentShaderPath.c_str(), pErrorBlob->GetBufferPointer() );
        outError = Error::FRAGMENT_SHADER_COMPILE_FAILED;
        return;
    }

    hr = rendererD3D11State.d3dDev->CreatePixelShader( pPSBlob->GetBufferPointer(), pPSBlob->GetBufferSize(), nullptr, &m->pixelShader );
    D3DShaderResource::pixelShaders.push_back( m->pixelShader );

    if (FAILED( hr ))
    {
        pPSBlob->Release();
        Log::Print( "Could not create pixel shader %s", fragmentShaderPath.c_str() );
        outError = Error::FRAGMENT_SHADER_COMPILE_FAILED;
        return;
    }

#if DEBUG
    m->pixelShader->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)fragmentShaderPath.length(), fragmentShaderPath.c_str() );
#endif

    if (!geometryShaderPath.empty())
    {
        // Geometry shader.
        std::ifstream geometryStream( geometryShaderPath.c_str() );
        if (!geometryStream.is_open())
        {
            Log::Print( "Could not find geometry shader %s", geometryShaderPath.c_str() );
            outError = Error::SHADER_NOT_FOUND;
            return;
        }
        
        std::string geometrySrc = std::string( std::istreambuf_iterator< char >( geometryStream ),
                                               std::istreambuf_iterator< char >() );
        geometrySrc = SourceWithDefines( geometrySrc, defines );

        hr = D3DCompile( geometrySrc.c_str(), geometrySrc.size(), nullptr, nullptr, nullptr, "GS", "gs_5_0", flags, 0, &pGSBlob, &pErrorBlob );
        if (FAILED( hr ))
        {
            MessageBoxA( 0, (LPCSTR)pErrorBlob->GetBufferPointer(), "Error", MB_OK );
            outError = Error::GEOMETRY_SHADER_COMPILE_FAILED;
            return;
        }

        hr = rendererD3D11State.d3dDev->CreateGeometryShader( pGSBlob->GetBufferPointer(), pGSBlob->GetBufferSize(), nullptr, &m->geometryShader );
        D3DShaderResource::geometryShaders.push_back( m->geometryShader );

        if (FAILED( hr ))
        {    
            pGSBlob->Release();
            MessageBoxA( 0, "Could not create a geometry shader from a file!", "Error", MB_OK );
            outError = Error::GEOMETRY_SHADER_COMPILE_FAILED;
            return;
        }

#if DEBUG
        m->geometryShader->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)geometryShaderPath.length(), geometryShaderPath.c_str() );
#endif

    }

    m->GetUniformLocations( pVSBlob, pPSBlob, pGSBlob );
    m->compileTimeStamp = Time::GetTime();

    if (firstTimeLoad)
    {
        FileWatcher::Instance().AddShader( vertexShaderPath.c_str(), fragmentShaderPath.c_str(), geometryShaderPath.c_str(), defines.c_str() );
        Impl::sIdToShaderCache[ vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ] = *this;
    }

    return;
}

void Shader::SetInt( const std::string& name, int value )
{
    m->ints[ name ] = value;
}

void Shader::SetFloat( const std::string& name, float value )
{
    m->floats[ name ] = value;
}

void Shader::SetMatrix( const std::string& name, const Matrix44& matrix )
{
    m->matrices[ name ] = matrix;
}

void Shader::SetMatrices( const std::string& name, const Matrix44* matrices, int count )
{
    std::vector< Matrix44 > matrixArray( count );

    for (int i = 0; i < count; ++i)
    {
        matrixArray[ i ] = matrices[ i ];    
    }

    m->matrixArrays[ name ] = matrixArray;
}

void Shader::SetTexture( const std::string& name, const TextureImpl* texture, int unit )
{
    assert( m->pixelReflector != nullptr );

    D3D11_SHADER_INPUT_BIND_DESC bindDesc;
    const HRESULT hr = m->pixelReflector->GetResourceBindingDescByName( name.c_str(), &bindDesc );
    if (hr != S_OK)
    {
        return;
    }

    if (!texture)
    {
        // Set a dummy sampler state to prevent log flood about unused samplers.
        TextureImpl::Sampler sampler;
        sampler.tiling = TextureImpl::Tiling::Clamp;
        sampler.filter = TextureImpl::Filter::Linear;
        
        const auto samplerState = rendererD3D11State.samplers.Get( sampler );
        rendererD3D11State.immContext->PSSetSamplers( bindDesc.BindPoint, 1, &samplerState );
        return;
    }

    assert( texture != nullptr );
    
    // Prevents feedback.
    if (!texture->IsCubemap() && unit != 666 && rendererD3D11State.currentRenderTargetView == texture->RTV())
    {
        //Log::Print( "Shader %s using method SetTexture: %s is a bound render target.  Unbind it before trying to read it.", m->fragmentPath.c_str(), name.c_str() );
        return;
    }
    
    if (texture->IsCubemap() && unit != 666)
    {
        ID3D11ShaderResourceView* id = texture->SRV();

        for (int i = 0; i < 6; ++i)
        {
            if (rendererD3D11State.currentRenderTargetView == Global::textureToRT[ id ].cubeRTVs[ i ])
            {
                return;
            }
        }
    }

    if (unit != 666)
    {
        m->textures[ name ] = texture;
    }

    SetVector( name + "_ST", texture->ScaleOffset() );

    ID3D11ShaderResourceView* textureView = texture->SRV();
    if (textureView != nullptr)
    {
        ID3D11ShaderResourceView* nullView = nullptr;
        rendererD3D11State.immContext->PSSetShaderResources( bindDesc.BindPoint, 1, unit == 666 ? &nullView : &textureView );
    }
    else
    {
        Log::Print( "SetTexture: %s has invalid ID3D11ShaderResourceView.", name.c_str() );
    }

    ID3D11SamplerState* sampler = rendererD3D11State.samplers.Get( texture->GetSampler() );
    rendererD3D11State.immContext->PSSetSamplers( bindDesc.BindPoint, 1, &sampler );
}

void Shader::SetVector( const std::string& name, const Vec3& value )
{
    m->vec3s[ name ] = value;
}

void Shader::SetVectors( const std::string& name, const Vec3* values, int count )
{
    std::vector< Vec3 > vec3Array( count );

    for (int i = 0; i < count; ++i)
    {
        vec3Array[ i ] = values[ i ];
    }

    m->vec3Arrays[ name ] = vec3Array;
}

void Shader::SetVector( const std::string& name, const Vec4& value )
{
    m->vec4s[ name ] = value;
}

void Shader::SetVectors( const std::string& name, const Vec4* values, int count )
{
    std::vector< Vec4 > vec4Array( count );

    for (int i = 0; i < count; ++i)
    {
        vec4Array[ i ] = values[ i ];
    }

    m->vec4Arrays[ name ] = vec4Array;
}

Error::Enum Shader::Reload( const std::string& vertexShaderPath,
                            const std::string& fragmentShaderPath,
                            const std::string& geometryShaderPath,
                            const std::string& defines )
{
    bool found = false;
    Shader* original = nullptr;

    if (Shader::Impl::sIdToShaderCache.find( vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ) != Shader::Impl::sIdToShaderCache.end())
    {
        original = &(Shader::Impl::sIdToShaderCache[ vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ]);
        found = true;
    }

    if (!found)
    {
        Log::Print( "Warning: Trying to reload a shader that is not loaded." );
        return Error::NO_ERRORS;
    }

    Shader reloaded = *original;
    Error::Enum error = Error::NO_ERRORS;
    reloaded.Load( vertexShaderPath, fragmentShaderPath, geometryShaderPath, defines, error );
    
    if (error != Error::NO_ERRORS)
    {
        Log::Print( "Warning: Tried to reload shader %s or %s but compilation failed.", vertexShaderPath.c_str(), fragmentShaderPath.c_str() );
        return Error::VERTEX_SHADER_COMPILE_FAILED;
    }

    *original = reloaded;
    
    return Error::NO_ERRORS;
}

void Shader::Use()
{
    Log::Assert( m->vertexShader != nullptr && m->pixelShader != nullptr && m->inputLayout != nullptr, "Uninitialized shader." );

    if (Shader::Impl::sIdToShaderCache[ m->vertexPath + m->fragmentPath + m->geometryPath + m->defines ].m->compileTimeStamp > m->compileTimeStamp)
    {
        Log::Print( "Shader::Use(): Updating shader from cache." );
        *this = Shader::Impl::sIdToShaderCache[ m->vertexPath + m->fragmentPath + m->geometryPath + m->defines ];
    }

    rendererD3D11State.immContext->VSSetShader( m->vertexShader, nullptr, 0 );
    rendererD3D11State.immContext->PSSetShader( m->pixelShader, nullptr, 0 );

    if (m->geometryShader)
    {
        rendererD3D11State.immContext->GSSetShader( m->geometryShader, nullptr, 0 );
    }

    Statistics::IncShaderSwitches();
    rendererD3D11State.immContext->IASetInputLayout( m->inputLayout );
}

void Shader::UnbindTextures()
{
    for (auto tex : m->textures)
    {
        SetTexture( tex.first, tex.second, 666 );
    }
}
