#include "TextureImpl.hpp"
#include <map>
#include <list>
#include <locale>
#include <vector>
#include <d3d11.h>
#include "Log.hpp"
#include "MathUtil.hpp"
#include "RendererImpl.hpp"
#include "RendererD3D11State.hpp"

#define AE3D_SAFE_RELEASE(x) if (x) { x->Release(); x = nullptr; }

unsigned char* LoadTextureData( const std::string& path, int& outWidth, int& outHeight, int& outComponents );
void FreeTextureData( unsigned char* data );
void SetDebugLabel( ID3D11DeviceChild* d3dObj, const std::string& name );

std::wstring stringToWstring( const std::string& s )
{
    std::string curLocale = setlocale( LC_ALL, "" );
    const char* _Source = s.c_str( );
    size_t _Dsize = mbstowcs( NULL, _Source, 0 ) + 1;
    wchar_t *_Dest = new wchar_t[ _Dsize ];
    wmemset( _Dest, 0, _Dsize );
    mbstowcs( _Dest, _Source, _Dsize );
    std::wstring result = _Dest;
    delete[ ]_Dest;
    setlocale( LC_ALL, curLocale.c_str( ) );
    return result;
}

namespace Global
{
    extern std::map< std::string, TextureImpl > pathToCachedTexture;
    std::map< std::string, ID3D11Texture2D* > pathToCachedTexture2D;
    /// Keeps info about which texture instances are created from a given file, used in hot-loading.
    std::map< std::string, std::list< TextureImpl* > > pathToTextureInstances;
    std::map< ID3D11ShaderResourceView*, RenderTextureData > textureToRT; 
}

static DXGI_FORMAT SolveD3DFormat( TextureImpl::Format format )
{
    if (format == TextureImpl::Format::Float)
    {
        return DXGI_FORMAT_R32G32B32A32_FLOAT;
    }
    else if (format == TextureImpl::Format::HalfFloat)
    {
        return DXGI_FORMAT_R16G16B16A16_FLOAT;
    }
    else if (format == TextureImpl::Format::HalfFloatRG)
    {
        return DXGI_FORMAT_R16G16_FLOAT;
    }
    else if (format == TextureImpl::Format::UnsignedByte)
    {
        return DXGI_FORMAT_R8G8B8A8_UNORM;
    }
    else
    {
        Log::Assert( false, "Unhandled texture format." );
    }

    return DXGI_FORMAT_R8G8B8A8_UNORM;
}

TextureImpl::~TextureImpl()
{
}

void TextureImpl::Bind( int ) const
{
    // Shader::SetTexture handles this for D3D11 renderer.
}

void TextureImpl::CreateNoise( int size )
{
    sampler.filter = Filter::Point;
    sampler.tiling = Tiling::Repeat;
    sRGB = false;
    width = height = size;

    std::vector< Vec4 > noise( size * size );

    for (int i = 0; i < size * size; ++i)
    {
        noise[ i ] = Vec4(
            MathUtil::Random( -1.0f, 1.0f ),
            MathUtil::Random( -1.0f, 1.0f ),
            0,
            0
            );
        noise[ i ].Normalize();
    }

    opaque = true;

    D3D11_TEXTURE2D_DESC desc;
    memset( &desc, 0, sizeof(D3D11_TEXTURE2D_DESC) );
    desc.Width = width;
    desc.Height = height;
    desc.MipLevels = 1;
    desc.ArraySize = 1;
    desc.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
    desc.SampleDesc.Count = 1;
    desc.Usage = D3D11_USAGE_DEFAULT;
    desc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
    desc.CPUAccessFlags = 0;
    desc.MiscFlags = 0;

    D3D11_SUBRESOURCE_DATA initData;
    initData.pSysMem = &noise[ 0 ].x;
    initData.SysMemPitch = width * 4 * sizeof( float );
    initData.SysMemSlicePitch = 0;// width * height * 4;

    ID3D11Texture2D *pTexture = nullptr;
    HRESULT result = rendererD3D11State.d3dDev->CreateTexture2D( &desc, &initData, &pTexture );
    if (result != S_OK)
    {
        Log::Assert( false, "Could not create noise texture object!" );
        return;
    }
    
    const std::string textureName = "noise";
    SetDebugLabel( pTexture, textureName );

    D3D11_SHADER_RESOURCE_VIEW_DESC SRVDesc;
    memset( &SRVDesc, 0, sizeof(SRVDesc) );
    SRVDesc.Format = desc.Format;
    SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
    SRVDesc.Texture2D.MipLevels = 1;

    ID3D11ShaderResourceView* textureView = nullptr;
    HRESULT hr = rendererD3D11State.d3dDev->CreateShaderResourceView( pTexture, &SRVDesc, &textureView );
    if (FAILED( hr ))
    {
        pTexture->Release();
        Log::Assert( false, "Failed to create shader resource view for noise texture." );
        return;
    }

    Log::Assert( textureView != nullptr, "Texture view is missing." );

    srv = textureView;

    Global::pathToCachedTexture2D[ textureName ] = pTexture;
    Global::pathToCachedTexture[ textureName ] = *this;
    Global::pathToTextureInstances[ textureName ].push_back( this );
}

void TextureImpl::CreateRenderTexture2D( int aWidth, int aHeight, Format format, Filter filter, Tiling tiling, const std::string& debugName )
{
    width = aWidth;
    height = aHeight;
    sampler.filter = filter;
    sampler.tiling = tiling;

    D3D11_RENDER_TARGET_VIEW_DESC renderTargetViewDesc;
	D3D11_SHADER_RESOURCE_VIEW_DESC shaderResourceViewDesc;
	D3D11_TEXTURE2D_DESC textureDesc;
	ZeroMemory( &textureDesc, sizeof( textureDesc ) );

	textureDesc.Width = width;
	textureDesc.Height = height;
	textureDesc.MipLevels = 1;
	textureDesc.ArraySize = 1;
    textureDesc.Format = SolveD3DFormat( format );
	textureDesc.SampleDesc.Count = 1;
	textureDesc.Usage = D3D11_USAGE_DEFAULT;
	textureDesc.BindFlags = D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE;
	textureDesc.CPUAccessFlags = 0;
    textureDesc.MiscFlags = 0;

    ID3D11Texture2D *pTexture = nullptr;
    HRESULT result = rendererD3D11State.d3dDev->CreateTexture2D( &textureDesc, nullptr, &pTexture );

	if (FAILED( result ))
	{
        Log::Print( "Failed to create render texture: CreateTexture2D" );
		return;
	}
    
    static int counter = 0;
    std::string textureName( "render texture " + std::to_string( counter ) );
    ++counter;

    SetDebugLabel( pTexture, debugName );

    // Cached so resources are released on exit.
    Global::pathToCachedTexture2D[ textureName ] = pTexture;

    renderTargetViewDesc.Format = textureDesc.Format;
	renderTargetViewDesc.ViewDimension = D3D11_RTV_DIMENSION_TEXTURE2D;
	renderTargetViewDesc.Texture2D.MipSlice = 0;
    
    ID3D11RenderTargetView* rtView = nullptr;

	// Create the render target view.
    result = rendererD3D11State.d3dDev->CreateRenderTargetView( pTexture, &renderTargetViewDesc, &rtView );
	if (FAILED( result ))
	{
        Log::Print( "Failed to create render texture: render target view." );
		return;
	}
   
    SetDebugLabel( rtView, debugName );

	// Setup the description of the shader resource view.
	shaderResourceViewDesc.Format = textureDesc.Format;
	shaderResourceViewDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
	shaderResourceViewDesc.Texture2D.MostDetailedMip = 0;
	shaderResourceViewDesc.Texture2D.MipLevels = 1;

    ID3D11ShaderResourceView* shaderResourceView = nullptr;

    result = rendererD3D11State.d3dDev->CreateShaderResourceView( pTexture, &shaderResourceViewDesc, &shaderResourceView );
	
    if (FAILED( result ))
	{
        Log::Print( "Failed to create render texture: shader resource view." );
		return;
	}

    SetDebugLabel( shaderResourceView, debugName );

    srv = shaderResourceView;
    rtv = rtView;
    
    D3D11_TEXTURE2D_DESC depthStencilDesc;
    ZeroMemory( &depthStencilDesc, sizeof( depthStencilDesc ) );
	depthStencilDesc.Width     = width;
	depthStencilDesc.Height    = height;
	depthStencilDesc.MipLevels = 1;
	depthStencilDesc.ArraySize = 1;
    depthStencilDesc.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;
	depthStencilDesc.SampleDesc.Count   = 1;
	depthStencilDesc.SampleDesc.Quality = 0;
	depthStencilDesc.Usage          = D3D11_USAGE_DEFAULT;
    depthStencilDesc.BindFlags = D3D11_BIND_DEPTH_STENCIL;
	depthStencilDesc.CPUAccessFlags = 0; 
	depthStencilDesc.MiscFlags      = 0;
    
    RenderTextureData rtData;
    rtData.rtView = rtView;
    rtData.srView = shaderResourceView;

    rendererD3D11State.d3dDev->CreateTexture2D( &depthStencilDesc, nullptr, &rtData.depthStencilBuffer );
    SetDebugLabel( rtData.depthStencilBuffer, textureName );

    rendererD3D11State.d3dDev->CreateDepthStencilView( rtData.depthStencilBuffer, nullptr, &rtData.depthStencilView );
    SetDebugLabel( rtData.depthStencilView, textureName );

    float white[] = { 1, 1, 1, 1 };
    rendererD3D11State.immContext->ClearRenderTargetView( rtData.rtView, white );
    rendererD3D11State.immContext->ClearDepthStencilView( rtData.depthStencilView, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 1, 0 );

    Global::textureToRT[ shaderResourceView ] = rtData;
    Global::pathToCachedTexture[ textureName ] = *this;
}

void TextureImpl::CreateRenderTextureCube( int size )
{
    width = height = size;
    cubeMap = true;
    sampler.filter = Filter::Linear;
    sampler.tiling = Tiling::Clamp;

	D3D11_TEXTURE2D_DESC textureDesc;
	ZeroMemory( &textureDesc, sizeof( textureDesc ) );

	textureDesc.Width = size;
	textureDesc.Height = size;
	textureDesc.MipLevels = 1;
	textureDesc.ArraySize = 6;
	textureDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
	textureDesc.SampleDesc.Count = 1;
	textureDesc.Usage = D3D11_USAGE_DEFAULT;
	textureDesc.BindFlags = D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE;
	textureDesc.CPUAccessFlags = 0;
	textureDesc.MiscFlags = D3D11_RESOURCE_MISC_TEXTURECUBE;

    ID3D11Texture2D *pTexture = nullptr;
    HRESULT result = rendererD3D11State.d3dDev->CreateTexture2D( &textureDesc, nullptr, &pTexture );

	if (FAILED( result ))
	{
        Log::Print( "Failed to create render texture: CreateTexture2D" );
		return;
	}

    static int counter = 0;
    std::string textureName( "render texture cube " + std::to_string( counter ) );
    ++counter;
    SetDebugLabel( pTexture, textureName );

    Global::pathToCachedTexture2D[ textureName ] = pTexture;

    RenderTextureData rtData;
    
	// Create the render target views.
    for (int i = 0; i < 6; ++i)
    {
        D3D11_RENDER_TARGET_VIEW_DESC renderTargetViewDesc;
        renderTargetViewDesc.Format = textureDesc.Format;
        renderTargetViewDesc.ViewDimension = D3D11_RTV_DIMENSION_TEXTURE2DARRAY;
        renderTargetViewDesc.Texture2DArray.ArraySize = 1;
        renderTargetViewDesc.Texture2DArray.MipSlice = 0;
        renderTargetViewDesc.Texture2DArray.FirstArraySlice = i;

        result = rendererD3D11State.d3dDev->CreateRenderTargetView( pTexture, &renderTargetViewDesc, &rtData.cubeRTVs[ i ] );
        
        if (FAILED( result ))
        {
            Log::Print( "Failed to create a cube render texture: render target view." );
            return;
        }
    }

	D3D11_SHADER_RESOURCE_VIEW_DESC shaderResourceViewDesc;
   	shaderResourceViewDesc.Format = textureDesc.Format;
	shaderResourceViewDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURECUBE;
	shaderResourceViewDesc.Texture2D.MostDetailedMip = 0;
	shaderResourceViewDesc.Texture2D.MipLevels = 1;

    ID3D11ShaderResourceView* shaderResourceView = nullptr;
    result = rendererD3D11State.d3dDev->CreateShaderResourceView( pTexture, &shaderResourceViewDesc, &shaderResourceView );
	
    if (FAILED( result ))
	{
        Log::Print( "Failed to create render texture: shader resource view." );
		return;
	}

    srv = shaderResourceView;
    rtv = rtData.cubeRTVs[ 0 ];

    D3D11_TEXTURE2D_DESC depthStencilDesc;
    ZeroMemory( &depthStencilDesc, sizeof( depthStencilDesc ) );
	depthStencilDesc.Width     = width;
	depthStencilDesc.Height    = height;
	depthStencilDesc.MipLevels = 1;
	depthStencilDesc.ArraySize = 1;
	depthStencilDesc.Format    = DXGI_FORMAT_D24_UNORM_S8_UINT;
	depthStencilDesc.SampleDesc.Count   = 1;
	depthStencilDesc.SampleDesc.Quality = 0;
	depthStencilDesc.Usage          = D3D11_USAGE_DEFAULT;
	depthStencilDesc.BindFlags      = D3D11_BIND_DEPTH_STENCIL;
	depthStencilDesc.CPUAccessFlags = 0; 
	depthStencilDesc.MiscFlags      = 0;
    
    rtData.rtView = rtData.cubeRTVs[ 0 ];
    rtData.srView = shaderResourceView;

    rendererD3D11State.d3dDev->CreateTexture2D( &depthStencilDesc, nullptr, &rtData.depthStencilBuffer );
    std::string textureName2( "render texture " + std::to_string( counter ) + " depthStencilBuffer" );
    SetDebugLabel( rtData.depthStencilBuffer, textureName2 );

    rendererD3D11State.d3dDev->CreateDepthStencilView( rtData.depthStencilBuffer, nullptr, &rtData.depthStencilView );

    Global::textureToRT[ shaderResourceView ] = rtData;
    Global::pathToCachedTexture[ textureName ] = *this;
    //Global::pathToTextureInstances[ textureName ].push_back( this );
}

bool TextureImpl::Load( const std::string& path, bool useMipmaps, bool usesRGB )
{
    sRGB = usesRGB;

    const bool firstTimeLoad = Global::pathToCachedTexture2D.find( path ) == Global::pathToCachedTexture2D.end();

    if (path.find( ".dds" ) != std::string::npos || path.find( ".DDS" ) != std::string::npos)
    {
        Log::Print( "MakeTexture: DDS loading not supported on D3D11, tried to load %s", path.c_str() );

        return false;
    }

    // Checks for uncompressed formats in texture's file name.
    const std::string extensions[] =
    {
        ".png", ".PNG", ".jpg", ".JPG", ".tga", ".TGA",
        ".bmp", ".BMP", ".gif", ".GIF"
    };

    bool loadedSuccessfully = true;
    bool foundExtension = false;

    for (auto& e : extensions)
    {
        if (path.find( e ) != std::string::npos)
        {
            foundExtension = true;

            int comp;
            unsigned char* data = LoadTextureData( path, width, height, comp );
            
            if (!data)
            {
                if (path == RendererImpl::Instance().BuiltinAssetPath() + "textures/checker.png")
                {
                    Log::Assert( false, "Trying to load default texture unsuccessfully." );
                    return false;
                }

                Log::Print( "MakeTexture: Could not load %s", path.c_str() );
                loadedSuccessfully = false;

                bool ok = Load( RendererImpl::Instance().BuiltinAssetPath() + "textures/checker.png", true, true );

                if (!ok)
                {
                    Log::Assert( false, "Could not load default texture." );
                }

                return false;
            }
            
            // The texture wasn't loaded, so create it.
            opaque = comp == 3;

            D3D11_TEXTURE2D_DESC desc;
            memset( &desc, 0, sizeof( D3D11_TEXTURE2D_DESC ) );
            desc.Width = width;
            desc.Height = height;
            desc.MipLevels = useMipmaps ? 0 : 1;
            desc.ArraySize = 1;
            desc.Format = sRGB ? DXGI_FORMAT_R8G8B8A8_UNORM_SRGB : DXGI_FORMAT_R8G8B8A8_UNORM;
            desc.SampleDesc.Count = 1;
            desc.Usage = useMipmaps ? D3D11_USAGE_DEFAULT : D3D11_USAGE_IMMUTABLE;
            desc.BindFlags = D3D11_BIND_SHADER_RESOURCE | (useMipmaps ? D3D11_BIND_RENDER_TARGET : 0);
            desc.CPUAccessFlags = 0;
            desc.MiscFlags = useMipmaps ? D3D11_RESOURCE_MISC_GENERATE_MIPS : 0;

            D3D11_SUBRESOURCE_DATA initData;
            initData.pSysMem = data;
            initData.SysMemPitch = width * 4;
            initData.SysMemSlicePitch = width * height * 4;

            ID3D11Texture2D *pTexture = nullptr;
            HRESULT result = rendererD3D11State.d3dDev->CreateTexture2D( &desc, useMipmaps ? nullptr : &initData, &pTexture );
            if (result != S_OK)
            {
                Log::Assert( false, "Could not create texture object!" );
                return false;
            }
#if DEBUG
            pTexture->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)path.length(), path.c_str() );
#endif
            D3D11_SHADER_RESOURCE_VIEW_DESC SRVDesc;
            memset( &SRVDesc, 0, sizeof( SRVDesc ) );
            SRVDesc.Format = desc.Format;
            SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
            SRVDesc.Texture2D.MipLevels = useMipmaps ? -1 : 1;

            ID3D11ShaderResourceView* textureView = nullptr;
            HRESULT hr = rendererD3D11State.d3dDev->CreateShaderResourceView( pTexture, &SRVDesc, &textureView );
            if (FAILED( hr ))
            {
                pTexture->Release();
                Log::Assert( false, "Failed to create shader resource view for texture." );
                return false;
            }

            Log::Assert( textureView != nullptr, "Texture view is missing." );
#if DEBUG
            textureView->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)path.length(), path.c_str() );
#endif
            Global::pathToCachedTexture2D[ path ] = pTexture;

            if (useMipmaps)
            {
                rendererD3D11State.immContext->UpdateSubresource( pTexture, 0, nullptr, data, static_cast<UINT>(width * 4), static_cast<UINT>(width * height * 4) );
                rendererD3D11State.immContext->GenerateMips( textureView );
            }

            srv = textureView;

            FreeTextureData( data );
            data = nullptr;
            break; // Done iterating extensions.
        }
    }

    if (!loadedSuccessfully || !foundExtension)
    {
        Log::Print( "MakeTexture: Could not load %s", path.c_str() );

    }

    if (firstTimeLoad)
    {
        Global::pathToTextureInstances[ path ].push_back( this );
    }
    else // Hot-reloading.
    {
        for (auto instance : Global::pathToTextureInstances[ path ])
        {
            *instance = Global::pathToCachedTexture[ path ];
        }
    }

    return loadedSuccessfully;
}

bool TextureImpl::Load( const std::string& left,  const std::string& right,
                        const std::string& top,   const std::string& bottom,
                        const std::string& front, const std::string& back,
                        bool useMipmaps, bool usesRGB )
{
    sRGB = usesRGB;
    sampler.filter = Filter::Linear;
    sampler.tiling = Tiling::Clamp;

    const std::string fileNames[] = { right, left, bottom, top, front, back };

    // Checks for uncompressed formats in texture's file name.
    const std::string extensions[] =
    {
        ".png", ".PNG", ".jpg", ".JPG", ".tga", ".TGA",
        ".bmp", ".BMP", ".gif", ".GIF"
    };
    
    // \todo Get mipmaps working for cube maps.
    useMipmaps = false;

    D3D11_SUBRESOURCE_DATA cubeData[ 6 ];
    
    D3D11_TEXTURE2D_DESC desc;
    memset( &desc, 0, sizeof( D3D11_TEXTURE2D_DESC ) );
    desc.Width = width;
    desc.Height = height;
    desc.MipLevels = useMipmaps ? 0 : 1;
    desc.ArraySize = 6;
    desc.Format = sRGB ? DXGI_FORMAT_R8G8B8A8_UNORM_SRGB : DXGI_FORMAT_R8G8B8A8_UNORM;
    desc.SampleDesc.Count = 1;
    desc.Usage = D3D11_USAGE_DEFAULT;
    desc.BindFlags = D3D11_BIND_SHADER_RESOURCE | (useMipmaps ? D3D11_BIND_RENDER_TARGET : 0);
    desc.CPUAccessFlags = 0;
    desc.MiscFlags = useMipmaps ? (D3D11_RESOURCE_MISC_GENERATE_MIPS | D3D11_RESOURCE_MISC_TEXTURECUBE) : D3D11_RESOURCE_MISC_TEXTURECUBE;

    unsigned char* faceData[ 6 ] = { 0 };

    for (int cubeMapFace = 0; cubeMapFace < 6; ++cubeMapFace)
    {
        for (auto& e : extensions)
        {
            if (fileNames[ cubeMapFace ].find( e ) == std::string::npos)
            {
                continue;
            }

            int comp;
            faceData[ cubeMapFace ] = LoadTextureData( fileNames[ cubeMapFace ], width, height, comp );
            
            if (!faceData[ cubeMapFace ])
            {
                Log::Print( "Could not load cubemap texture %s", fileNames[ cubeMapFace ].c_str() );

                comp = 3;
                width = 64;
                height = 64;
                faceData[ cubeMapFace ] = new unsigned char[ width * height * 4 ];
                std::fill_n( faceData[ cubeMapFace ], width * height * 4, 0 );
            }
            
            // The texture wasn't loaded, so create it.
            opaque = comp == 3;

            Log::Assert( &faceData[ cubeMapFace ][0] != nullptr, "Cube map face is uninitialized." );

            cubeData[ cubeMapFace ].pSysMem = faceData[ cubeMapFace ];
            cubeData[ cubeMapFace ].SysMemPitch = width * 4;
            cubeData[ cubeMapFace ].SysMemSlicePitch = 0;
        }
    }

    desc.Width = width;
    desc.Height = height;

    ID3D11Texture2D *pTexture = nullptr;
    HRESULT result = rendererD3D11State.d3dDev->CreateTexture2D( &desc, cubeData, &pTexture );
    if (result != S_OK)
    {
        Log::Assert( false, "Could not create texture object!" );
        return false;
    }

    static int counter = 0;
    std::string textureName( "cubemap texture " + std::to_string( counter ) );
    ++counter;
#if DEBUG
    pTexture->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)textureName.length(), textureName.c_str() );
#endif

    Global::pathToCachedTexture2D[ textureName ] = pTexture;

    D3D11_SHADER_RESOURCE_VIEW_DESC SRVDesc;
    memset( &SRVDesc, 0, sizeof( SRVDesc ) );
    SRVDesc.Format = sRGB ? DXGI_FORMAT_R8G8B8A8_UNORM_SRGB : DXGI_FORMAT_R8G8B8A8_UNORM;
    SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURECUBE;
    SRVDesc.TextureCube.MipLevels = useMipmaps ? -1 : 1;

    ID3D11ShaderResourceView* textureView = nullptr;
    HRESULT hr = rendererD3D11State.d3dDev->CreateShaderResourceView( pTexture, &SRVDesc, &textureView );
    if (FAILED( hr ))
    {
        pTexture->Release();
        Log::Assert( false, "Failed to create shader resource view for cube map." );
        return false;
    }
            
    for (int i = 0; i < 6; ++i)
    {
        FreeTextureData( faceData[ i ] );
        faceData[ i ] = nullptr;
    }

    if (useMipmaps)
    {
        rendererD3D11State.immContext->GenerateMips( textureView );
    }

    srv = textureView;

    Log::Assert( textureView != nullptr, "Texture view is missing." );

    return true;
}

ID3D11ShaderResourceView* TextureImpl::SRV() const
{
    return srv;
}

ID3D11RenderTargetView* TextureImpl::RTV() const
{
    return rtv;
}

ID3D11UnorderedAccessView* TextureImpl::UAV() const
{
    return uav;
}
