#include "ComputeShader.hpp"
#include <fstream>
#include <string>
#include <unordered_map>
#if AETHER3D_GLES3
#    include <GLES3/gl3.h>
#else
#    include <GL/glxw.h>
#endif
#include "Log.hpp"
#include "Matrix.hpp"
#include "Statistics.hpp"
#include "TextureImpl.hpp"

const char* GetFullPath( const char* path );
GLuint Compile( const char* path, GLenum shaderType, const std::string& defines );
void PrintInfoLog( GLuint shader, bool isProgram );
void ErrorCheck( const std::string& info );

namespace ae3d
{
    extern GLuint sCachedProgram;
}

struct ComputeShader::Impl
{
    int UniformLocation( const std::string& name );

    GLuint program = 0;
    std::unordered_map< std::string, int > uniformNameToLocation;
};

int ComputeShader::Impl::UniformLocation( const std::string& name )
{
    if (uniformNameToLocation.find( name ) == uniformNameToLocation.end())
    {
        uniformNameToLocation[ name ] = glGetUniformLocation( program, name.c_str() );
    }

    return uniformNameToLocation[ name ];
}

ComputeShader::ComputeShader() : m( new Impl ) {}

ComputeShader::~ComputeShader()
{
}

void ComputeShader::Load( const char* path, Error::Enum& outError )
{
    ErrorCheck( "ComputeShader Load begin" );

    m->program = glCreateProgram();
    const GLuint shader = Compile( path, GL_COMPUTE_SHADER, "" );
    glAttachShader( m->program, shader );
    glLinkProgram( m->program );
    GLint wasLinked;
    glGetProgramiv( m->program, GL_LINK_STATUS, &wasLinked );

    if (!wasLinked)
    {
        PrintInfoLog( m->program, true );
        outError = Error::COMPUTE_SHADER_COMPILE_FAILED;
        return;
    }

    ErrorCheck( "ComputeShader Load end" );
}

void ComputeShader::Use()
{
    if (m->program != ae3d::sCachedProgram)
    {
        ae3d::sCachedProgram = m->program;
        glUseProgram( ae3d::sCachedProgram );
        Statistics::IncShaderSwitches();
    }
}

void ComputeShader::Dispatch( int numGroupsX, int numGroupsY, int numGroupsZ )
{
    glDispatchCompute( numGroupsX, numGroupsY, numGroupsZ );
}

void ComputeShader::SetFloat( const char* name, float value )
{
    glUniform1f( m->UniformLocation( name ), value );
}

void ComputeShader::SetInt( const char* name, int value )
{
    glUniform1i( m->UniformLocation( name ), value );
}

void ComputeShader::SetUInt( const char* name, unsigned value )
{
    glUniform1ui( m->UniformLocation( name ), value );
}

void ComputeShader::SetVector( const char* name, const Vec3& value )
{
    glUniform3fv( m->UniformLocation( name ), 1, &value.x );
}

void ComputeShader::SetVector( const char* name, const Vec4& value )
{
    glUniform4fv( m->UniformLocation( name ), 1, &value.x );
}

void ComputeShader::SetMatrix( const char* name, const Matrix44& matrix )
{
    glUniformMatrix4fv( m->UniformLocation( name ), 1, false, matrix.m );
}
