#include "AudioImpl.hpp"
#include <cstdlib>
#include <fstream>
#include <map>
#include <string>
#if defined __APPLE__
#    include <OpenAL/al.h>
#    include <OpenAL/alc.h>
#else
#    include "AL/al.h"
#    include "AL/alc.h"
#endif
#define STB_VORBIS_HEADER_ONLY
#include "stb_vorbis.c"
#include "FileWatcher.hpp"
#include "Log.hpp"

#ifdef AETHER3D_WINDOWS
#define AETHER3D_API __declspec(dllexport)
#else
#define AETHER3D_API 
#endif

static_assert( sizeof( unsigned int   ) == 4, "" );
static_assert( sizeof( unsigned short ) == 2, "" );

struct SoundInfo
{
    int fileSize = 0;
    int frequency = 0;
    int channels = 0;
    int bitsPerChannel = 0;
};

struct Impl
{
    /**
     Loads a .wav file. Called by AddSound().
     
     \param fileName File name of .wav file.
     */
    void LoadWav( const std::string& fileName );
    
    /**
     Loads a .ogg file. Called by AddSound().

     \param fileName File name of .ogg file.
     */
    void LoadOgg( const std::string& fileName );

    /**
      \param soundName Sound name. Must have been already loaded with LoadWav.
      \return Clip length in milliseconds.
    */
    int ClipLength( const std::string& soundName );

    std::map< std::string, std::string > soundNameToFileName;
    std::map< std::string, unsigned > fileNameToSourceId;
    std::map< std::string, unsigned > fileNameToBufferId;
    std::map< std::string, SoundInfo > fileNameToSoundInfo;
    ALCdevice* device; // Audio device.
    ALCcontext* context; // Audio context.
    bool muted;
};

int Impl::ClipLength( const std::string& soundName )
{
    const auto& sound = fileNameToSoundInfo[ soundNameToFileName[ soundName ] ];
    int duration = sound.fileSize / (sound.frequency * sound.channels * sound.bitsPerChannel / 8);
    return duration;
}

namespace Global
{
    Impl impl;
}

/**
 * WAVE structure which is used by .wav files.
 * Follows the specification at the following URI:
 * <http://ccrma.stanford.edu/courses/422/projects/WaveFormat/>
 *
 * Assumes that unsigned int is 4 bytes and unsigned short is 2 bytes.
 */
struct WAVE
{
    unsigned char chunkID[ 4 ];
    unsigned chunkSize;
    unsigned char format[4];

    // Format sub chunk
    unsigned char subchunk1ID[ 4 ];
    unsigned subchunk1Size;
    unsigned short audioFormat;
    unsigned short numChannels;
    unsigned sampleRate;
    unsigned byteRate;
    unsigned short blockAlign;
    unsigned short bitsPerSample;

    // Data sub chunk
    unsigned char subchunk2ID[ 4 ]; // Data sub chunk ID.
    unsigned subchunk2Size; // Data sub chunk size.
    unsigned char* data; // Chunk data.
};

extern "C" AETHER3D_API Audio* CreateAudio()
{
    return new AudioImpl();
}

const char * GetOpenALErrorString(int errID)
{
    if (errID == AL_NO_ERROR) return "";
    if (errID == AL_INVALID_NAME) return "Invalid name";
    if (errID == AL_INVALID_ENUM) return " Invalid enum ";
    if (errID == AL_INVALID_VALUE) return " Invalid value ";
    if (errID == AL_INVALID_OPERATION) return " Invalid operation ";
    if (errID == AL_OUT_OF_MEMORY) return " Out of memory! ";

    return " Don't know ";
}

void CheckOpenALError( const char* stmt )
{
    const ALenum err = alGetError();

    if (err != AL_NO_ERROR)
    {
        Log::Print( "OpenAL error (%s) for %s", GetOpenALErrorString( err ), stmt );
    }
};

void AudioImpl::Reload( const char* path )
{
    const std::string pathStr( path );
    const std::string extension = pathStr.substr( pathStr.length() - 3, pathStr.length() );

    if (extension == "wav" || extension == "WAV")
    {
        Global::impl.LoadWav( pathStr );
    }
    else if (extension == "ogg" || extension == "OGG")
    {
        Global::impl.LoadOgg( pathStr );
    }
}

AudioImpl::AudioImpl()
{
    Global::impl.device = alcOpenDevice( nullptr );
    Global::impl.muted = false;

    if (!Global::impl.device)
    {
        Log::Print( "Error opening audio device!" );
    }

    Global::impl.context = alcCreateContext( Global::impl.device, nullptr );
    alcMakeContextCurrent( Global::impl.context );

    if (!Global::impl.context)
    {
        Log::Print( "Error creating audio context!" );
    }

    alListener3f( AL_POSITION, 0.0f, 0.0f, 1.0f );
    CheckOpenALError( "AudioImpl constructor" );
}

AudioImpl::~AudioImpl()
{
    for (const auto& it : Global::impl.fileNameToSourceId)
    {
        alSourceStopv( 1, &it.second );
        alDeleteSources( 1, &it.second );
    }

    for (const auto& it : Global::impl.fileNameToBufferId)
    {
        alDeleteBuffers( 1, &it.second );
    }

    alcMakeContextCurrent( nullptr );

    if (Global::impl.context != nullptr)
    {
        alcDestroyContext( Global::impl.context );
    }

    if (Global::impl.device != nullptr)
    {
        alcCloseDevice( Global::impl.device );
    }
}

void AudioImpl::AddSound( const char* soundName, const char* soundFile )
{
    const std::string soundFileStr( soundFile );

    if (Global::impl.fileNameToBufferId.find( soundFileStr ) != Global::impl.fileNameToBufferId.end())
    {
        Global::impl.soundNameToFileName[ soundName ] = soundFile;
        return;
    }

    const std::string extension = soundFileStr.substr( soundFileStr.length() - 3, soundFileStr.length() );

    if (extension == "wav" || extension == "WAV")
    {
        Global::impl.LoadWav( soundFileStr );
    }
    else if (extension == "ogg" || extension == "OGG")
    {
        Global::impl.LoadOgg( soundFileStr );
    }
    else
    {
        Log::Print( "Unknown audio file extension in file %s", soundFile );
        return;
    }

    Global::impl.soundNameToFileName[ soundName ] = soundFileStr;
    FileWatcher::Instance().AddAudioClip( soundFile );
}

void Impl::LoadOgg( const std::string& aFileName )
{
    const std::string systemFileName = aFileName;

    std::ifstream ifs( systemFileName.c_str(), std::ios::binary );
    
    if (!ifs)
    {
        Log::Print( "AddSound: %s not found!", systemFileName.c_str() );
        return;
    }

    ifs.close();

    short* decoded = nullptr;
    int channels = 0;
    int samplerate = 0;
    const int len = stb_vorbis_decode_filename( const_cast<char*>(systemFileName.c_str()), &channels, &samplerate, &decoded );
    
    if (len == 0)
    {
        Log::Print( "AddSound: Could not open %s", systemFileName.c_str() );
        return;
    }

    ALuint bufID;
    ALuint srcID;
    
    alGenBuffers( 1, &bufID );
    fileNameToBufferId[ aFileName ] = bufID;
    alGenSources( 1, &srcID );
    fileNameToSourceId[ aFileName ] = srcID;
    alListener3f( AL_POSITION, 0.0f, 0.0f, 0.0f );
    alSource3f( srcID, AL_POSITION, 0.0f, 0.0f, 0.0f );
    
    int error;
    stb_vorbis* vorbis = stb_vorbis_open_filename( (char*)systemFileName.c_str(), &error, nullptr );
    const stb_vorbis_info info = stb_vorbis_get_info( vorbis );
    
    const ALenum format = info.channels == 2 ? AL_FORMAT_STEREO16 : AL_FORMAT_MONO16;
    
    alBufferData( bufID, format, &decoded[0], len, info.sample_rate );
    alSourcei( srcID, AL_BUFFER, bufID );

    fileNameToSoundInfo[systemFileName].fileSize = len;
    fileNameToSoundInfo[systemFileName].channels = info.channels;
    fileNameToSoundInfo[systemFileName].frequency = info.sample_rate;
    fileNameToSoundInfo[systemFileName].bitsPerChannel = format == AL_FORMAT_STEREO16 || format == AL_FORMAT_MONO16 ? 16 : 8;

    CheckOpenALError("Loading ogg");
}

void Impl::LoadWav( const std::string& fileName )
{
    const std::string systemFileName = fileName;
    std::ifstream ifs( systemFileName.c_str(), std::ios::binary );

    if (!ifs)
    {
        Log::Print( "AddSound: %s not found!", systemFileName.c_str() );
        return;
    }

    WAVE wav;

    ifs.read( (char*)&wav.chunkID, 4 );
    if (wav.chunkID[0] != 'R' && wav.chunkID[1] != 'I' &&
        wav.chunkID[2] != 'F' && wav.chunkID[3] != 'F')
    {
        Log::Print( "AddSound: %s is not a valid .wav file!", systemFileName.c_str() );
        return;
    }

    ifs.read( (char*)&wav.chunkSize, 4 );

    ifs.read( (char*)&wav.format, 4 );
    if (wav.format[0] != 'W' && wav.format[1] != 'A' &&
        wav.format[2] != 'V' && wav.format[3] != 'E')
    {
        Log::Print( "%s is not a valid .wav file!", systemFileName.c_str() );
        return;
    }
    
    ifs.read( (char*)&wav.subchunk1ID, 4 );
    if (wav.subchunk1ID[0] != 'f' && wav.subchunk1ID[1] != 'm' &&
        wav.subchunk1ID[2] != 't' && wav.subchunk1ID[3] != ' ')
    {
        Log::Print( "%s is not a valid .wav file!", systemFileName.c_str() );
        return;
    }
    
    ifs.read( (char*)&wav.subchunk1Size, 4 );
    if (wav.subchunk1Size != 16)
    {
        Log::Print( "%s does not contain PCM audio data!", systemFileName.c_str() );
        return;
    }
    
    ifs.read( (char*)&wav.audioFormat, 2 );
    if (wav.audioFormat != 1)
    {
        Log::Print( "%s does not contain uncompressed PCM audio data!", systemFileName.c_str() );
        return;
    }
    
    ifs.read( (char*)&wav.numChannels,   2 );
    ifs.read( (char*)&wav.sampleRate,    4 );
    ifs.read( (char*)&wav.byteRate,      4 );
    ifs.read( (char*)&wav.blockAlign,    2 );
    ifs.read( (char*)&wav.bitsPerSample, 2 );
    
    ifs.read( (char*)&wav.subchunk2ID, 4 );

    if (wav.subchunk2ID[0] != 'd' && wav.subchunk2ID[1] != 'a' &&
        wav.subchunk2ID[2] != 't' && wav.subchunk2ID[3] != 'a')
    {
        Log::Print( "%s does not contain data chunk!", systemFileName.c_str() );
        return;
    }

    ifs.read( (char*)&wav.subchunk2Size, 4 );

    // Data size is file size minus header size.
    // We get it by saving the current position, moving
    // to the EOF, checking EOF's position and moving back to
    // the saved location for data reading.
    std::streampos dataPos = ifs.tellg();
    ifs.seekg( 0, std::ios::end );
    std::streampos fileSize = ifs.tellg();
    const ALsizei dataSize = (ALsizei)fileSize - 44; // Header is 44 bytes.

    ifs.seekg( dataPos );

    wav.data = new unsigned char[ dataSize ];

    ifs.read( (char*)wav.data, dataSize );

    ALenum format = AL_FORMAT_MONO8;

    if (wav.numChannels == 1 && wav.bitsPerSample == 8)
    {
        format = AL_FORMAT_MONO8;
    }
    else if (wav.numChannels == 2 && wav.bitsPerSample == 8)
    {
        format = AL_FORMAT_STEREO8;
    }
    else if (wav.numChannels == 1 && wav.bitsPerSample == 16)
    {
        format = AL_FORMAT_MONO16;
    }
    else if (wav.numChannels == 2 && wav.bitsPerSample == 16)
    {
        format = AL_FORMAT_STEREO16;
    }
    else
    {
        Log::Print( "Audio: Unknown format in file %s", systemFileName.c_str() );
    }
    
    ALuint bufID;
    alGenBuffers( 1, &bufID );

    ALuint srcID;
    alGenSources( 1, &srcID );

    alBufferData( bufID, format, &wav.data[0], dataSize, wav.sampleRate );
    alSourcei( srcID, AL_BUFFER, bufID );
    alSourcef( srcID, AL_GAIN, 1.0f );
    
    if ((alGetError()) != AL_NO_ERROR)
    {
        Log::Print( "OpenAL error loading %s!", fileName.c_str() );
        delete[] wav.data;
        return;
    }

    delete[] wav.data;

    fileNameToSourceId[ fileName ] = srcID;
    fileNameToBufferId[ fileName ] = bufID;

    fileNameToSoundInfo[fileName].fileSize = static_cast< int >( fileSize );
    fileNameToSoundInfo[fileName].channels = wav.numChannels;
    fileNameToSoundInfo[fileName].frequency = wav.sampleRate;
    fileNameToSoundInfo[fileName].bitsPerChannel = wav.bitsPerSample;

    CheckOpenALError( "Loading wav" );
}

void AudioImpl::Play( const char* soundName )
{
    if (Global::impl.muted)
    {
        return;    
    }

    if (Global::impl.fileNameToSourceId.find( Global::impl.soundNameToFileName[ soundName ] ) == Global::impl.fileNameToSourceId.end())
    {
        Log::Print( "Audio::Play: Tried to play non-existing sound %s", soundName );
        return;
    }

    ALint state;
    alGetSourcei( Global::impl.fileNameToSourceId[ Global::impl.soundNameToFileName[ soundName ] ], AL_SOURCE_STATE, &state );

    if (state != AL_PLAYING)
    {
        alSourcePlay( Global::impl.fileNameToSourceId[ Global::impl.soundNameToFileName[ soundName ] ] );
    }

    if ((alGetError()) != AL_NO_ERROR)
    {
        Log::Print( "OpenAL error playing sound %s", soundName );
    }

    CheckOpenALError( "Play" );
}

void AudioImpl::SetDopplerFactor( float factor )
{
    alDopplerFactor( factor );
}

void AudioImpl::SetDopplerVelocity( float velocity )
{
    alDopplerVelocity( velocity );
}

void AudioImpl::SetGain( float gain )
{
    alListenerf( AL_GAIN, gain );
}

void AudioImpl::SetMute( bool aMute )
{
    Global::impl.muted = aMute;
}

bool AudioImpl::IsMuted() const
{
    return Global::impl.muted;
}
