#include "CameraImpl.hpp"
#include "Frustum.hpp"
#include "Matrix.hpp"
#include "MathUtil.hpp"
#include "Platform.hpp"
#include "Vec3.hpp"
#include "Quaternion.hpp"

#ifdef AETHER3D_WINDOWS
#    define AETHER3D_API __declspec(dllexport)
#else
#    define AETHER3D_API
#endif

extern "C" AETHER3D_API Camera* CreateCamera()
{
    return new CameraImpl();
}

#ifdef SIMD_SSE3
struct BEGIN_ALIGNED( CameraImpl::Impl, 16 )
#else
struct CameraImpl::Impl
#endif
{
    Impl()
    {
        viewport[ 0 ] = 0;
        viewport[ 1 ] = Screen::Width();
        viewport[ 2 ] = 0;
        viewport[ 3 ] = Screen::Height();
    }
    
    ~Impl() {}

#if defined SIMD_SSE3 && !defined AETHER3D_UNIX
    void* operator new( std::size_t size )
    {
        void* p = _aligned_malloc( size, 16 );
        return p;
    }

    void operator delete( void* p )
    {
        Impl* ptr = static_cast< Impl* >( p );
        _aligned_free( ptr );
    }
#endif // SIMD_SSE3

    void UpdateViewMatrix( const Vec3& position, const Quaternion& rotation );

    Matrix44 projection;
    Matrix44 view;
    Frustum frustum;
    float fovDegrees = 0;
    float aspect = 0;
    Texture* target = nullptr;
    int viewport[ 4 ]; // left, right, top, bottom
    bool dirtyView = true;
}
#ifdef SIMD_SSE3
END_ALIGNED( CameraImpl::Impl, 16 )
#endif
;

CameraImpl::CameraImpl() : m( new Impl() )
{
    type = Type::Camera;
}

CameraImpl::~CameraImpl() {}

float CameraImpl::Aspect() const
{
    return m->aspect;
}

float CameraImpl::FOV() const
{
    return m->fovDegrees;
}

float CameraImpl::FarClipPlane() const
{
    return m->frustum.FarClipPlane();
}

float CameraImpl::NearClipPlane() const
{
    return m->frustum.NearClipPlane();
}

void CameraImpl::LookAt( const Vec3& eye, const Vec3& center, const Vec3& up )
{
    Matrix44 lookAt;
    lookAt.MakeLookAt( eye, center, up );
    rotation.FromMatrix( lookAt );
    position = eye;
    m->dirtyView = true;
}

void CameraImpl::MoveForward( float amount )
{
    if (!MathUtil::AlmostEquals( amount, 0 ))
    {
        position += rotation * Vec3( 0, 0, amount );
        m->dirtyView = true;
    }
}

void CameraImpl::MoveRight( float amount )
{
    if (!MathUtil::AlmostEquals( amount, 0 ))
    {
        position += rotation * Vec3( amount, 0, 0 );
        m->dirtyView = true;
    }
}

void CameraImpl::MoveUp( float amount )
{
    position.y += amount;
    m->dirtyView = true;
}

void CameraImpl::OffsetRotation( const Vec3& axis, float angleDeg )
{
    Quaternion rot;
    rot.FromAxisAngle( axis, angleDeg );

    Quaternion newRotation;

    if (MathUtil::AlmostEquals( axis.y, 0 ))
    {
        newRotation = rotation * rot;
    }
    else
    {
        newRotation = rot * rotation;
    }

    newRotation.Normalize();

    if ((axis.x == 1.0f || axis.x == -1.0f) && axis.y == 0.0f && axis.z == 0.0f &&
        newRotation.FindTwist( Vec3( 1.0f, 0.0f, 0.0f ) ) > 0.9999f)
    {
        return;
    }

    rotation = newRotation;
    m->dirtyView = true;
}

Vec3 CameraImpl::ScreenPoint( const Vec3& worldPoint ) const
{
    Matrix44 viewProjection;
    Matrix44::Multiply( m->view, m->projection, viewProjection );
    Vec4 pos( worldPoint );
    Matrix44::TransformPoint( pos, viewProjection, &pos );
    pos.x /= pos.w;
    pos.y /= pos.w;

    const float halfWidth = Screen::Width() * 0.5f;
    const float halfHeight = Screen::Height() * 0.5f;

    return Vec3( pos.x * halfWidth + halfWidth, -pos.y * halfHeight + halfHeight, pos.z );
}

void CameraImpl::SetProjection( float fieldOfViewDegrees, float aspect, float nearDepth, float farDepth )
{
    m->fovDegrees = fieldOfViewDegrees;
    m->projection.MakeProjection( fieldOfViewDegrees, aspect, nearDepth, farDepth );
    m->frustum.SetProjection( fieldOfViewDegrees, aspect, nearDepth, farDepth );
    m->aspect = aspect;
}

void CameraImpl::SetProjection( float left, float right, float bottom, float top, float nearDepth, float farDepth )
{
    m->projection.MakeProjection( left, right, bottom, top, nearDepth, farDepth );
    m->frustum.SetProjection( left, right, bottom, top, nearDepth, farDepth );
}

void CameraImpl::SetViewport( int left, int right, int top, int bottom )
{
    m->viewport[ 0 ] = left;
    m->viewport[ 1 ] = right;
    m->viewport[ 2 ] = top;
    m->viewport[ 3 ] = bottom;
}

void CameraImpl::SetRenderTarget( Texture* target )
{
    m->target = target;
}

Texture* CameraImpl::GetRenderTarget() const
{
    return m->target;
}

const Frustum& CameraImpl::GetFrustum() const
{
    if (m->dirtyView)
    {
        m->UpdateViewMatrix( position, rotation );
    }

    return m->frustum;
}

void CameraImpl::GetViewport( int& outLeft, int& outRight, int& outTop, int& outBottom ) const
{
    outLeft = m->viewport[ 0 ];
    outRight = m->viewport[ 1 ];
    outTop = m->viewport[ 2 ];
    outBottom = m->viewport[ 3 ];
}

const Matrix44& CameraImpl::ProjectionMatrix() const
{
    if (m->dirtyView)
    {
        m->UpdateViewMatrix( position, rotation );
    }

    return m->projection;
}

Vec3 CameraImpl::ViewDirection() const
{
    if (m->dirtyView)
    {
        m->UpdateViewMatrix( position, rotation );    
    }
    
    Vec3 viewDir = Vec3( ViewMatrix().m[2], ViewMatrix().m[6], ViewMatrix().m[10] ).Normalized();
    return viewDir;
}

const Matrix44& CameraImpl::ViewMatrix() const
{
    if (m->dirtyView)
    {
        m->UpdateViewMatrix( position, rotation );
    }

    return m->view;
}

void CameraImpl::Impl::UpdateViewMatrix( const Vec3& aPosition, const Quaternion& aRotation )
{
    aRotation.GetMatrix( view );
    Matrix44 translation;
    translation.Translate( -aPosition );
    Matrix44::Multiply( translation, view, view );

    Vec3 viewDir = Vec3( view.m[2], view.m[6], view.m[10] ).Normalized();

    frustum.Update( aPosition, viewDir );
    dirtyView = false;
}
