#include "FileWatcher.hpp"
#include <ctime>
#include <sys/stat.h>
#include <locale.h>
#include <vector>
#include "AudioImpl.hpp"
#include "Log.hpp"
#include "ModelImpl.hpp"
#include "Shader.hpp"
#include "TextureImpl.hpp"

// \todo DRY

struct PathModTime
{
    explicit PathModTime( const std::string& aPath ) : path( aPath ), hour( 0 ), minute( 0 ), second( 0 ) {}
    PathModTime() : hour( 0 ), minute( 0 ), second( 0 ) {}

    std::string path;
    int hour;
    int minute;
    int second;
};

struct FileWatcherEntry
{
    enum class Type { Uninitialized, Model, Texture, Shader, AudioClip };

    FileWatcherEntry() : type( Type::Uninitialized ) {}
    FileWatcherEntry( Type aType, const std::string& path ) : type( aType )
    {
        if (aType == Type::Model)
        {
            model = PathModTime( path );
        }
        else if (aType == Type::Texture)
        {
            texture = PathModTime( path );
        }
        else if (aType == Type::AudioClip)
        {
            audioClip = PathModTime( path );
        }
        else
        {
            Log::Assert( false, "Type must be Model, AudioClip or Texture when calling this constructor." );
        }
    }

    FileWatcherEntry( const std::string& aVertexPath, const std::string& aFragmentPath, const std::string& aGeometryPath )
        : type( Type::Shader )
        , vertexShader( aVertexPath )
        , fragmentShader( aFragmentPath )
        , geometryShader( aGeometryPath ) {}

    Type type;
    PathModTime model;
    PathModTime texture;
    PathModTime vertexShader;
    PathModTime fragmentShader;
    PathModTime geometryShader;
    PathModTime audioClip;
    std::string shaderDefines;
};

struct FileWatcher::Impl
{
    void PollAudioClips( struct stat& inode );
    void PollTextures( struct stat& inode );
    void PollModels( struct stat& inode );
    void PollShaders( struct stat& inode );

    std::vector< FileWatcherEntry > models;
    std::vector< FileWatcherEntry > textures;
    std::vector< FileWatcherEntry > shaders;
    std::vector< FileWatcherEntry > audioClips;
};

void FileWatcher::Impl::PollAudioClips( struct stat& inode )
{
    for (auto& audioClip : audioClips)
    {
        if (stat( audioClip.audioClip.path.c_str(), &inode ) != -1)
        {
            tm* tm = localtime( &inode.st_mtime );
            if (tm->tm_hour != audioClip.audioClip.hour || tm->tm_min != audioClip.audioClip.minute || tm->tm_sec != audioClip.audioClip.second)
            {
                AudioImpl::Reload( audioClip.audioClip.path.c_str() );
                audioClip.audioClip.hour = tm->tm_hour;
                audioClip.audioClip.minute = tm->tm_min;
                audioClip.audioClip.second = tm->tm_sec;
            }
        }
    }
}

void FileWatcher::Impl::PollModels( struct stat& inode )
{
    for (auto& model : models)
    {
        if (stat( model.model.path.c_str(), &inode ) != -1)
        {
            tm* tm = localtime( &inode.st_mtime );
            if (tm->tm_hour != model.model.hour || tm->tm_min != model.model.minute || tm->tm_sec != model.model.second)
            {
                ModelImpl::Reload( model.model.path );
                model.model.hour = tm->tm_hour;
                model.model.minute = tm->tm_min;
                model.model.second = tm->tm_sec;
            }
        }
    }
}

void FileWatcher::Impl::PollTextures( struct stat& inode )
{
    for (auto& texture : textures)
    {
        if (stat( texture.texture.path.c_str(), &inode ) != -1)
        {
            tm* tm = localtime( &inode.st_mtime );
            if (tm->tm_hour != texture.texture.hour || tm->tm_min != texture.texture.minute || tm->tm_sec != texture.texture.second)
            {
                TextureImpl::Reload( texture.texture.path );
                texture.texture.hour = tm->tm_hour;
                texture.texture.minute = tm->tm_min;
                texture.texture.second = tm->tm_sec;
            }
        }
    }
}

void FileWatcher::Impl::PollShaders( struct stat& inode )
{
    for (auto& shader : shaders)
    {
        bool needsReloading = false;
        
        if (stat( shader.vertexShader.path.c_str(), &inode ) != -1)
        {
            tm* tm = localtime( &inode.st_mtime );
            if (tm->tm_hour != shader.vertexShader.hour || tm->tm_min != shader.vertexShader.minute || tm->tm_sec != shader.vertexShader.second)
            {
                needsReloading = true;
                shader.vertexShader.hour = tm->tm_hour;
                shader.vertexShader.minute = tm->tm_min;
                shader.vertexShader.second = tm->tm_sec;
            }
        }
        
        if (stat( shader.fragmentShader.path.c_str(), &inode ) != -1)
        {
            tm* tm = localtime( &inode.st_mtime );
            if (tm->tm_hour != shader.fragmentShader.hour || tm->tm_min != shader.fragmentShader.minute || tm->tm_sec != shader.fragmentShader.second)
            {
                needsReloading = true;
                shader.fragmentShader.hour = tm->tm_hour;
                shader.fragmentShader.minute = tm->tm_min;
                shader.fragmentShader.second = tm->tm_sec;
            }
        }
        
        if (stat( shader.geometryShader.path.c_str(), &inode ) != -1)
        {
            tm* tm = localtime( &inode.st_mtime );
            if (tm->tm_hour != shader.geometryShader.hour || tm->tm_min != shader.geometryShader.minute || tm->tm_sec != shader.geometryShader.second)
            {
                needsReloading = true;
                shader.geometryShader.hour = tm->tm_hour;
                shader.geometryShader.minute = tm->tm_min;
                shader.geometryShader.second = tm->tm_sec;
            }
        }
        
        if (needsReloading)
        {
            Shader::Reload( shader.vertexShader.path, shader.fragmentShader.path, shader.geometryShader.path, shader.shaderDefines );
        }
    }
}

FileWatcher& FileWatcher::Instance()
{
    static FileWatcher watcher;

    return watcher;
}

FileWatcher::FileWatcher() : m( new Impl )
{
    m->textures.reserve( 10 );
    m->shaders.reserve( 10 );
    m->models.reserve( 10 );
    m->audioClips.reserve( 10);
}

FileWatcher::~FileWatcher()
{
}

void FileWatcher::AddAudioClip( const char* path )
{
    for (auto& audioClip : m->audioClips)
    {
        if (audioClip.audioClip.path == path)
        {
            return;
        }
    }

    m->audioClips.emplace_back( FileWatcherEntry( FileWatcherEntry::Type::AudioClip, path ) );

    struct stat inode;

    if (stat( path, &inode ) != -1)
    {
        tm* tm = localtime( &inode.st_mtime );
        auto& lastClip = m->audioClips.back();
        lastClip.audioClip.hour   = tm->tm_hour;
        lastClip.audioClip.minute = tm->tm_min;
        lastClip.audioClip.second = tm->tm_sec;
    }
}

void FileWatcher::AddModel( const char* path )
{
    if (std::any_of( std::begin( m->models ), std::end( m->models ),
                     [&](const FileWatcherEntry& entry) { return entry.model.path == path; } ))
    {
        return;
    }

    m->models.emplace_back( FileWatcherEntry( FileWatcherEntry::Type::Model, path ) );

    struct stat inode;
    if (stat( path, &inode ) != -1)
    {
        tm* tm = localtime( &inode.st_mtime );
        auto& lastModel = m->models.back();
        lastModel.model.hour   = tm->tm_hour;
        lastModel.model.minute = tm->tm_min;
        lastModel.model.second = tm->tm_sec;
    }
}

void FileWatcher::AddTexture( const char* path )
{
    for (auto& texture : m->textures)
    {
        if (texture.texture.path == path)
        {
            return;
        }
    }

    m->textures.emplace_back( FileWatcherEntry( FileWatcherEntry::Type::Texture, path ) );

    struct stat inode;
    if (stat( path, &inode ) != -1)
    {
        tm* tm = localtime( &inode.st_mtime );
        auto& lastTexture = m->textures.back();
        lastTexture.texture.hour   = tm->tm_hour;
        lastTexture.texture.minute = tm->tm_min;
        lastTexture.texture.second = tm->tm_sec;
    }
}

void FileWatcher::AddShader( const char* vertexPath, const char* fragmentPath, const char* geometryPath, const char* defines )
{
    for (auto& shader : m->shaders)
    {
        if (shader.vertexShader.path   == vertexPath &&
            shader.fragmentShader.path == fragmentPath &&
            shader.geometryShader.path == geometryPath &&
            shader.shaderDefines == defines)
        {
            return;
        }
    }

    m->shaders.emplace_back( FileWatcherEntry( vertexPath, fragmentPath, geometryPath ) );
    m->shaders.back().shaderDefines = defines;

    struct stat inode;
    if (stat( vertexPath, &inode ) != -1)
    {
        tm* tm = localtime( &inode.st_mtime );
        m->shaders.back().vertexShader.hour   = tm->tm_hour;
        m->shaders.back().vertexShader.minute = tm->tm_min;
        m->shaders.back().vertexShader.second = tm->tm_sec;
    }

    if (stat( fragmentPath, &inode ) != -1)
    {
        tm* tm = localtime( &inode.st_mtime );
        m->shaders.back().fragmentShader.hour   = tm->tm_hour;
        m->shaders.back().fragmentShader.minute = tm->tm_min;
        m->shaders.back().fragmentShader.second = tm->tm_sec;
    }

    if (stat( geometryPath, &inode ) != -1)
    {
        tm* tm = localtime( &inode.st_mtime );
        m->shaders.back().geometryShader.hour   = tm->tm_hour;
        m->shaders.back().geometryShader.minute = tm->tm_min;
        m->shaders.back().geometryShader.second = tm->tm_sec;
    }
}

void FileWatcher::Poll()
{
    struct stat inode;

    m->PollAudioClips( inode );
    m->PollTextures( inode );
    m->PollModels( inode );
    m->PollShaders( inode );
}
