#include "Fonter.hpp"
#include <cstdlib>
#include <cstring>
#include <string>
#include <fstream>
#include <sstream>
#include "Log.hpp"
#include "TextureImpl.hpp"

static_assert( sizeof( unsigned short ) == 2, "" );
static_assert( sizeof( unsigned ) == 4, "" );

const char* GetFullPath( const char* fileName );

namespace
{
namespace BlockType
{
    enum Enum
    {
        Info = 1,
        Common,
        Pages,
        Chars,
        Kerning
    };
}

struct CharacterBlock
{
    unsigned id;
    unsigned short x;
    unsigned short y;
    unsigned short width;
    unsigned short height;
    short xOffset;
    short yOffset;
    short xAdvance;
    unsigned char page;
    unsigned char channel;
};

static_assert( sizeof( CharacterBlock ) == 20, "" );

struct CommonBlock
{
    unsigned short lineHeight;
    unsigned short base;
    unsigned short scaleW;
    unsigned short scaleH;
    unsigned short pages;
    unsigned char  bitField;
    unsigned char  alpha;
    unsigned char  red;
    unsigned char  green;
    unsigned char  blue;
};

struct Character
{
    Character() :
          x( 0 )
        , y( 0 )
        , width( 0 )
        , height( 0 )
        , xOffset( 0 )
        , yOffset( 0 )
        , xAdvance( 0 ) {}

    float x, y;
    float width, height;
    float xOffset, yOffset;
    float xAdvance;
};

}

struct Fonter::Impl
{
    Impl() :
          fontTexture( nullptr )
        , lineHeight( 0 )
        , base( 0 ) 
    {
        spacing[ 0 ] = 0;
        spacing[ 1 ] = 0;

        padding[ 0 ] = 0;
        padding[ 1 ] = 0;
        padding[ 2 ] = 0;
        padding[ 3 ] = 0;
    }

    virtual ~Impl()
    {
        delete fontTexture;
    }

    /**
       Loads a BMFont text meta data.
       Not tested with variable width fonts.
       Limitation: Only one page is supported.
       \todo Support padding and spacing > 9.
      
       \param ifstream Input stream at the position where "info" token was just read.
     */
    void LoadBMFontMetaText( std::ifstream& ifstream );

    /**
       Loads a BMFont binary meta data.
       Not tested with variable width fonts.
       Limitation: Only one page is supported.

       \param path Path.
     */
    void LoadBMFontMetaBinary( const std::string& path );

    /// Font's texture.
    const TextureImpl* fontTexture;
    /// The spacing for each character (horizontal, vertical).
    int spacing[ 2 ];
    /// The padding for each character (up, right, down, left).
    int padding[ 4 ];
    /// This is the distance in pixels between each line of text.
    int lineHeight;
    /// The number of pixels from the absolute top of the line to the base of the characters.
    int base;

    Character chars[ 256 ];
};

void Fonter::Impl::LoadBMFontMetaText( std::ifstream& metaStream )
{
    std::string line;
    std::getline( metaStream, line );

    std::string token;

    // First line (info)
    std::stringstream infoStream( line );
    while (!infoStream.eof())
    {
        infoStream >> token;

        if (token.find( "padding" ) != std::string::npos)
        {
            // \todo Support padding over 9.
            const char a[ 2 ] = { token[  8 ], 0 };
            const char b[ 2 ] = { token[ 10 ], 0 };
            const char c[ 2 ] = { token[ 12 ], 0 };
            const char d[ 2 ] = { token[ 14 ], 0 };

            padding[ 0 ] = std::atoi( a );
            padding[ 1 ] = std::atoi( b );
            padding[ 2 ] = std::atoi( c );
            padding[ 3 ] = std::atoi( d );
        }
        else if (token.find( "spacing" ) != std::string::npos)
        {
            // \todo Support spacing over 9.
            const char a[ 2 ] = { token[  8 ], 0 };
            const char b[ 2 ] = { token[ 10 ], 0 };

            spacing[ 0 ] = std::atoi( a );
            spacing[ 1 ] = std::atoi( b );
        }
    }

    std::getline( metaStream, line );

    // Second line (common)
    std::stringstream commonStream( line );
    while (!commonStream.eof())
    {
        commonStream >> token;

        if (token.find( "lineHeight" ) != std::string::npos)
        {
            const char a[ 3 ] = { token[ 11 ], token[ 12 ], 0 };

            lineHeight = std::atoi( a );
        }
        else if (token.find( "base" ) != std::string::npos)
        {
            const char a[ 3 ] = { token[ 5 ], token[ 6 ], 0 };

            base = std::atoi( a );
        }
    }

    // Third line (page)
    std::getline( metaStream, line );
    // Fourth line (chars)
    std::getline( metaStream, line );

    // Character tags.
    while (!metaStream.eof())
    {
        std::getline( metaStream, line );
        std::stringstream charStream( line );

        charStream.seekg( line.find( "id=" ) + 3 );
        int id;
        charStream >> id;
        Log::Assert( id < 256, "Too high glyph id." );

        charStream.seekg( line.find( "x=" ) + 2 );
        charStream >> chars[ id ].x;

        charStream.seekg( line.find( "y=" ) + 2 );
        charStream >> chars[ id ].y;

        charStream.seekg( line.find( "width=" ) + 6 );
        charStream >> chars[ id ].width;

        charStream.seekg( line.find( "height=" ) + 7 );
        charStream >> chars[ id ].height;

        charStream.seekg( line.find( "xoffset=" ) + 8 );
        charStream >> chars[ id ].xOffset;

        charStream.seekg( line.find( "yoffset=" ) + 8 );
        charStream >> chars[ id ].yOffset;

        charStream.seekg( line.find( "xadvance=" ) + 9 );
        charStream >> chars[ id ].xAdvance;
    }
}

void Fonter::Impl::LoadBMFontMetaBinary( const std::string& path )
{
    std::ifstream ifs( path.c_str(), std::ios::binary );
    unsigned char header[ 4 ];
    ifs.read( (char*)&header, 4 );
    const bool validHeader = header[ 0 ] == 66 && header[ 1 ] == 77 && header[ 2 ] == 70 && header[ 3 ] == 3;
    if (!validHeader)
    {
        Log::Print( "%s does not contain a valid BMFont header!", path.c_str() );
        return;
    }

    while (!ifs.eof())
    {
        unsigned char blockId;
        ifs.read( (char*)&blockId, 1 );
        int blockType = (BlockType::Enum)blockId;
        
        int blockSize;
        ifs.read( (char*)&blockSize, 4 );

        std::vector< unsigned char > blockData( blockSize );

        if (blockType == BlockType::Info)
        {
            ifs.read( (char*)&blockData[ 0 ], blockSize );

            padding[ 0 ] = blockData[  7 ];
            padding[ 1 ] = blockData[  8 ];
            padding[ 2 ] = blockData[  9 ];
            padding[ 3 ] = blockData[ 10 ];

            spacing[ 0 ] = blockData[ 11 ];
            spacing[ 1 ] = blockData[ 12 ];
        }
        else if (blockType == BlockType::Common)
        {
            CommonBlock block;
            ifs.read( (char*)&block, blockSize );

            lineHeight = block.lineHeight;
            base = block.base;
        }
        else if (blockType == BlockType::Pages)
        {
            ifs.read( (char*)&blockData[ 0 ], blockSize );
        }
        else if (blockType == BlockType::Chars)
        {
            std::vector< CharacterBlock > blocks( blockSize / sizeof( CharacterBlock ) );
            ifs.read( (char*)&blocks[ 0 ].id, blockSize );
            Log::Assert( blocks.size() <= 256, "Too many glyphs." );

            for (std::size_t c = 0; c < blocks.size(); ++c)
            {
                chars[ blocks[ c ].id ].x = blocks[ c ].x;
                chars[ blocks[ c ].id ].y = blocks[ c ].y;
                chars[ blocks[ c ].id ].width = blocks[ c ].width;
                chars[ blocks[ c ].id ].height = blocks[ c ].height;
                chars[ blocks[ c ].id ].xOffset = blocks[ c ].xOffset;
                chars[ blocks[ c ].id ].yOffset = blocks[ c ].yOffset;
                chars[ blocks[ c ].id ].xAdvance = blocks[ c ].xAdvance;
            }

            return;
        }
        else if (blockType == BlockType::Kerning)
        {
            ifs.read( (char*)&blockData[ 0 ], blockSize );
        }
        else
        {
            Log::Assert( false, "Fonter: Corrupted binary meta data file or a bug in reader!" );
        }
    }
}

Fonter::Fonter() : m( new Impl )
{
}

Fonter::Fonter( const Fonter& fonter ) : m( new Impl )
{
    *this = fonter;
}

Fonter& Fonter::operator=( const Fonter& fonter )
{
    m->fontTexture = fonter.m->fontTexture;
    
    m->spacing[ 0 ] = fonter.m->spacing[ 0 ];
    m->spacing[ 1 ] = fonter.m->spacing[ 1 ];
    
    m->padding[ 0 ] = fonter.m->padding[ 0 ];
    m->padding[ 1 ] = fonter.m->padding[ 1 ];
    m->padding[ 2 ] = fonter.m->padding[ 2 ];
    m->padding[ 3 ] = fonter.m->padding[ 3 ];
    
    m->lineHeight = fonter.m->lineHeight;
    m->base = fonter.m->base;
    
    std::memcpy( &m->chars, fonter.m->chars, sizeof( m->chars ) );

    return *this;
}

Fonter::~Fonter()
{
}

std::vector< RenderQueue::Vec3uv > Fonter::CreateGeometry( const char* text, float x, float y, float scale ) const
{
    const std::string textStr( text );
    
    std::vector< RenderQueue::Vec3uv > quad( textStr.size() * 6 );

    float accumX = 0;

    for (std::size_t c = 0; c < textStr.size(); ++c)
    {
        if ((int)text[ c ] < 0)
        {
            continue;
        }

        const Character& ch = m->chars[ (int)text[ c ] ];

        float offx = x + ch.xOffset * scale + accumX * scale;
        float offy = y + ch.yOffset * scale;

        accumX += ch.xAdvance;
        
        auto tex = m->fontTexture;
        
        float u0 = ch.x / tex->Width();
        float u1 = (ch.x + ch.width) / tex->Width();

        float v0 = (ch.y + ch.height) / tex->Height();
        float v1 = (ch.y) / tex->Height();
        
        const float z = 0;

        // Upper triangle.
        quad[ c * 6 + 0 ] = RenderQueue::Vec3uv( offx, offy, z, u0, v1 );
        quad[ c * 6 + 1 ] = RenderQueue::Vec3uv( offx + ch.width * scale, offy, z, u1, v1 );
        quad[ c * 6 + 2 ] = RenderQueue::Vec3uv( offx, offy + ch.height * scale, z, u0, v0 );
        // Lower triangle.
        quad[ c * 6 + 3 ] = RenderQueue::Vec3uv( offx + ch.width * scale, offy, z, u1, v1 );
        quad[ c * 6 + 4 ] = RenderQueue::Vec3uv( offx + ch.width * scale, offy + ch.height * scale, z, u1, v0 );
        quad[ c * 6 + 5 ] = RenderQueue::Vec3uv( offx, offy + ch.height * scale, z, u0, v0 );
    }
    
    return quad;
}

Error::Enum Fonter::Load( const char* fontBitmapPath, const char* fontMetadataPath )
{
    std::ifstream stream( GetFullPath( fontBitmapPath ) );
    if (!stream.is_open())
    {
        Log::Print( "Fonter: Could not find %s", GetFullPath( fontBitmapPath ) );
    }

    m->fontTexture = new TextureImpl( fontBitmapPath, false, true, Texture::Tiling::Clamp, Texture::Filter::Point );

    std::ifstream metaStream( GetFullPath( fontMetadataPath ) );
    if (!metaStream.is_open())
    {
        Log::Print( "Fonter: Could not find %s", GetFullPath( fontMetadataPath ) );
        return Error::FILE_NOT_FOUND;
    }

    std::string token;
    // Determines the encoding (text or binary).
    metaStream >> token;
    if (token == "info")
    {
        m->LoadBMFontMetaText( metaStream );
    }
    else
    {
        m->LoadBMFontMetaBinary( fontMetadataPath );
    }

    return Error::NO_ERRORS;
}

int Fonter::TextWidth( const char* text ) const
{
    return (int)(std::string( text ).length() * m->chars[ (int)'a' ].width); // \fixme Does padding/spacing need to be added?
}

const TextureImpl* Fonter::Texture() const
{
    return m->fontTexture;
}
