#include "Frustum.hpp"
#include "Vec3.hpp"
#include <cmath>
#ifndef M_PI
#    define M_PI 3.14159265358979f
#endif

struct Frustum::Impl
{
    Impl() :
      zNear( 0 )
    , zFar( 0 )
    , nearWidth( 0 )
    , nearHeight( 0 )
    , farWidth( 0 )
    , farHeight( 0 )
    {}

    void UpdateCornersAndCenters( const Vec3& cameraPosition, const Vec3& zAxis );

    // Near clipping plane coordinates
    Vec3 nearTopLeft;
    Vec3 nearTopRight;
    Vec3 nearBottomLeft;
    Vec3 nearBottomRight;

    // Far clipping plane coordinates.
    Vec3 farTopLeft;
    Vec3 farTopRight;
    Vec3 farBottomLeft;
    Vec3 farBottomRight;

    Vec3 nearCenter; // Near clipping plane center coordinate.
    Vec3 farCenter;  // Far clipping plane center coordinate.
    float zNear, zFar; // The same as in gluPerspective()
    float nearWidth, nearHeight;
    float farWidth, farHeight;

    /**
     3 points define a plane.  The frustum has 6 planes.
     */
    struct Plane
    {
        /**
         \param point Point to be tested.
         \return point's distance to the plane.
         */
        float Distance( const Vec3& point ) const
        {
            return Vec3::Dot( normal, point ) + d;
        }

        /**
         Sets plane's normal and point.

         \param norm Normal.
         \param point Point.
         */
        void SetNormalAndPoint( const Vec3& normal, const Vec3& point );

        /// Calculates plane's normal.
        void CalculateNormal();

        Vec3 a; // Point a on the plane.
        Vec3 b; // Point b on the plane.
        Vec3 c; // Point c on the plane.
        Vec3 normal; // Plane's normal pointing inside the frustum.
        float d = 0;
    } planes[ 6 ]; // Clipping planes.
};

void Frustum::Impl::UpdateCornersAndCenters( const Vec3& cameraPosition, const Vec3& zAxis )
{
    const Vec3 up( 0, 1, 0 );

    Vec3 xAxis = Vec3::Cross( up, zAxis ).Normalized();

    const Vec3 yAxis = Vec3::Cross( zAxis, xAxis );

    // Computes the centers of near and far planes.
    nearCenter = cameraPosition - zAxis * zNear;
    farCenter  = cameraPosition - zAxis * zFar;
    
    // Computes the near plane corners.
    nearTopLeft     = nearCenter + yAxis * nearHeight - xAxis * nearWidth;
    nearTopRight    = nearCenter + yAxis * nearHeight + xAxis * nearWidth;
    nearBottomLeft  = nearCenter - yAxis * nearHeight - xAxis * nearWidth;
    nearBottomRight = nearCenter - yAxis * nearHeight + xAxis * nearWidth;

    // Computes the far plane corners.
    farTopLeft     = farCenter + yAxis * farHeight - xAxis * farWidth;
    farTopRight    = farCenter + yAxis * farHeight + xAxis * farWidth;
    farBottomLeft  = farCenter - yAxis * farHeight - xAxis * farWidth;
    farBottomRight = farCenter - yAxis * farHeight + xAxis * farWidth;
}

Frustum::Frustum() : m( new Impl )
{
}

Frustum::~Frustum()
{
}

float Frustum::NearClipPlane() const
{
    return m->zNear;
}

float Frustum::FarClipPlane() const
{
    return m->zFar;
}

Vec3 Frustum::Centroid() const
{
    return (m->nearCenter + m->farCenter) * 0.5f;
}

void Frustum::Impl::Plane::CalculateNormal()
{
    const Vec3 v1 = a - b;
    const Vec3 v2 = c - b;
    normal = Vec3::Cross( v2, v1 ).Normalized();
    d = -( Vec3::Dot( normal, b ) );
}

void Frustum::Impl::Plane::SetNormalAndPoint( const Vec3& aNormal, const Vec3& aPoint )
{
    normal = aNormal.Normalized();
    d = -( Vec3::Dot( normal, aPoint ) );
}

void Frustum::SetProjection( float fieldOfView, float aAspect, float aNear, float aFar )
{
    m->zNear = aNear;
    m->zFar  = aFar;

    // Computes width and height of the near and far plane sections.
    const float deg2rad = (float)M_PI / 180.0f;
    const float tang = std::tan( deg2rad * fieldOfView * 0.5f );
    m->nearHeight = m->zNear * tang;
    m->nearWidth  = m->nearHeight * aAspect;
    m->farHeight  = m->zFar * tang;
    m->farWidth   = m->farHeight * aAspect;
}

void Frustum::SetProjection( float aLeft, float aRight, float aBottom, float aTop, float aNear, float aFar )
{
    m->zNear = aNear;
    m->zFar  = aFar;
    
    m->nearHeight = aTop - aBottom;
    m->nearWidth  = aRight - aLeft;
    m->farHeight  = m->nearHeight;
    m->farWidth   = m->nearWidth;
}

void Frustum::Update( const Vec3& cameraPosition, const Vec3& cameraDirection )
{
    const Vec3 zAxis = cameraDirection;
    m->UpdateCornersAndCenters( cameraPosition, zAxis );

    enum FrustumPlane
    {
        FARP = 0,
        NEARP,
        BOTTOM,
        TOP,
        LEFT,
        RIGHT
    };

    m->planes[ FrustumPlane::TOP ].a = m->nearTopRight;
    m->planes[ FrustumPlane::TOP ].b = m->nearTopLeft;
    m->planes[ FrustumPlane::TOP ].c = m->farTopLeft;
    m->planes[ FrustumPlane::TOP ].CalculateNormal();

    m->planes[ FrustumPlane::BOTTOM ].a = m->nearBottomLeft;
    m->planes[ FrustumPlane::BOTTOM ].b = m->nearBottomRight;
    m->planes[ FrustumPlane::BOTTOM ].c = m->farBottomRight;
    m->planes[ FrustumPlane::BOTTOM ].CalculateNormal();

    m->planes[ FrustumPlane::LEFT ].a = m->nearTopLeft;
    m->planes[ FrustumPlane::LEFT ].b = m->nearBottomLeft;
    m->planes[ FrustumPlane::LEFT ].c = m->farBottomLeft;
    m->planes[ FrustumPlane::LEFT ].CalculateNormal();

    m->planes[ FrustumPlane::RIGHT ].a = m->nearBottomRight;
    m->planes[ FrustumPlane::RIGHT ].b = m->nearTopRight;
    m->planes[ FrustumPlane::RIGHT ].c = m->farBottomRight;
    m->planes[ FrustumPlane::RIGHT ].CalculateNormal();

    m->planes[ FrustumPlane::NEARP ].SetNormalAndPoint( -zAxis, m->nearCenter );
    m->planes[ FrustumPlane::FARP  ].SetNormalAndPoint(  zAxis, m->farCenter  );
}

// \todo refactor to also show if the box is completely inside frustum to disable culling of meshes if model is inside.
#if 1
bool Frustum::BoxInFrustum( const Vec3& min, const Vec3& max ) const
{
    bool result = true;

    Vec3 pos = min;
    //Vec3 neg = max;

    // Determines positive and negative vertex in relation to the normal.
    for (unsigned p = 0; p < 6; ++p)
    {
        pos = min;
        //neg = max;

        if (m->planes[ p ].normal.x >= 0)
        {
            pos.x = max.x;
        }
        if (m->planes[ p ].normal.y >= 0)
        {
            pos.y = max.y;
        }
        if (m->planes[ p ].normal.z >= 0)
        {
            pos.z = max.z;
        }

        /*if (m->planes[ p ].normal.x >= 0)
        {
            neg.x = min.x;
        }
        if (m->planes[ p ].normal.y >= 0)
        {
            neg.y = min.y;
        }
        if (m->planes[ p ].normal.z >= 0)
        {
            neg.z = min.z;
        }*/

        if (m->planes[ p ].Distance( pos ) < 0)
        {
            return false;
        }
    }

    return result;
}
#endif
#if 0
bool Frustum::BoxInFrustum( const Vec3& min, const Vec3& max ) const
{
    int in = 0, out = 0;

    const Vec3 corners[] =
    {
        Vec3( min.x, min.y, min.z ),
        Vec3( max.x, min.y, min.z ),
        Vec3( min.x, max.y, min.z ),
        Vec3( min.x, min.y, max.z ),
        Vec3( max.x, max.y, min.z ),
        Vec3( min.x, max.y, max.z ),
        Vec3( max.x, max.y, max.z ),
        Vec3( max.x, min.y, max.z )
    };

    // For each plane.
    for (unsigned p = 0; p < 6; ++p)
    {
        in = 0;
        out = 0;

        // For each corner of the box.
        for (int k = 0; k < 8 && (in == 0 || out == 0); ++k)
        {
            if (m->planes[ p ].Distance( corners[k] ) < 0.0f)
            {
                ++out;
            }
            else
            {
                ++in;
            }
        }

        if (in == 0)
        {
            return false;
        }
        else if (out != 0)
        {
            return true;
        }
    }
    return true;
}
#endif

const Vec3& Frustum::NearTopLeft() const { return m->nearTopLeft; }
const Vec3& Frustum::NearTopRight() const { return m->nearTopRight; }
const Vec3& Frustum::NearBottomLeft() const { return m->nearBottomLeft; }
const Vec3& Frustum::NearBottomRight() const { return m->nearBottomRight; }

const Vec3& Frustum::FarTopLeft() const { return m->farTopLeft; }
const Vec3& Frustum::FarTopRight() const { return m->farTopRight; }
const Vec3& Frustum::FarBottomLeft() const { return m->farBottomLeft; }
const Vec3& Frustum::FarBottomRight() const { return m->farBottomRight; }
