#ifndef FRUSTUM_H
#define FRUSTUM_H

#ifndef INCLUDE_MEMORY_H
#    include <memory>
#    define INCLUDE_MEMORY_H
#endif

struct Vec3;

/**
   View Frustum.

   Contains the view frustum methods used for frustum culling.
   All coordinates/tests are done in world space.
   Frustum plane normals point towards the inside of the frustum.
 */
class Frustum
{
  public:
    Frustum();
    Frustum( const Frustum& ) = delete;
    Frustum& operator=( const Frustum& ) = delete;
    virtual ~Frustum();

    const Vec3& NearTopLeft() const;
    const Vec3& NearTopRight() const;
    const Vec3& NearBottomLeft() const;
    const Vec3& NearBottomRight() const;

    const Vec3& FarTopLeft() const;
    const Vec3& FarTopRight() const;
    const Vec3& FarBottomLeft() const;
    const Vec3& FarBottomRight() const;

    /**
       Tests AABB against the frustum. 

       \param min AABB's minimum corner.
       \param max AABB's maximum corner.
       \return True, if part of the box is in the frustum.
       \return False, if the box is not in the frustum.
     */
    bool BoxInFrustum( const Vec3& min, const Vec3& max ) const;

    /**
       Sets values from which the frustum is calculated.
       Should be called when the perspective is changed.
       The values should be the same as in gluPerspective()

       \param fieldOfView Vertical field of view in degrees
       \param aaspect Aspect ratio (width / height)
       \param anear Near clipping plane z-coordinate
       \param afar Far clipping plane z-coordinate
     */
    void SetProjection( float fieldOfView, float aspect, float anear, float afar );

    /**
     Sets orthographic projection.

     \param left Left.
     \param right Right.
     \param bottom Bottom.
     \param top Top.
     \param nearDepth Near plane distance.
     \param farDepth Far plane distance.
     */
    void SetProjection( float left, float right, float bottom, float top, float nearp, float farp );

    /**
       Calculates the frustum planes. Should be called when the camera's origin
       or direction changes.

       \param cameraPosition Camera's position
       \param cameraDirection Camera's direction
     */
    void Update( const Vec3& cameraPosition, const Vec3& cameraDirection );

    /// \return Near clip plane.
    float NearClipPlane() const;

    /// \return Far clip plane.
    float FarClipPlane() const;

    /// \return Centroid.
    Vec3 Centroid() const;

  private:
    struct Impl;
    std::unique_ptr< Impl > m;
};
#endif

