#include "MaterialImpl.hpp"
#include <forward_list>
#include <fstream>
#include <map>
#include <sstream>
#include <unordered_map>
#include "Matrix.hpp"
#include "Log.hpp"
#include "Shader.hpp"
#include "TextureImpl.hpp"
#include "Vec3.hpp"

#ifdef AETHER3D_WINDOWS
#    define AETHER3D_API __declspec(dllexport)
#else
#    define AETHER3D_API
#endif

const char* GetFullPath( const char* fileName );
void ErrorCheck( const std::string& info );

extern "C" AETHER3D_API Material* CreateMaterial()
{
    return new MaterialImpl();
}

bool FileExists( const std::string& path )
{
    std::ifstream infile( GetFullPath( path.c_str() ) );
    return infile.good();
}

struct MaterialImpl::Impl
{
    Impl() : opaque( true ) {}
    
    Shader shader;
    std::unordered_map< std::string, float > floats;
    std::unordered_map< std::string, Matrix44 > matrices;
    std::unordered_map< std::string, const TextureImpl* > textures;
    std::unordered_map< std::string, Vec3 > vec3s;
    std::unordered_map< std::string, Vec4 > vec4s;
    bool opaque;
    std::string name = "material";
    std::string path;
};

std::forward_list< std::string > sShaderSearchPaths;

std::string ResolveVSPath( const std::string& vsPathOrName )
{
    // Finds vertex shader path.
    std::forward_list< std::string > vsExtensions;
#if AETHER3D_UNIX || AETHER3D_IOS || AETHER3D_GL32 || AETHER3D_GL43 || AETHER3D_GLES3
    vsExtensions.emplace_front( ".vsh" );
    vsExtensions.emplace_front( ".vert" );
    vsExtensions.emplace_front( ".glsl" );
#elif AETHER3D_D3D11
    vsExtensions.emplace_front( ".hlsl" );
#endif
    std::string vsPath;
    
    for (const auto& path : sShaderSearchPaths)
    {
        for (const auto& extension : vsExtensions)
        {
            const std::string testedPath = path + vsPathOrName + extension;
            
            if (vsPath.empty() && FileExists( testedPath ))
            {
                vsPath = testedPath;
            }
        }
        
        if (!vsPath.empty())
        {
            break;
        }
    }
    
    if (vsPath.empty())
    {
        vsPath = vsPathOrName;
    }
    
    return vsPath;
}

std::string ResolveFSPath( const std::string& fsPathOrName )
{
    // Finds fragment/pixel shader path.
    std::forward_list< std::string > fsExtensions;
#if AETHER3D_UNIX || AETHER3D_IOS || AETHER3D_GL32 || AETHER3D_GL43 || AETHER3D_GLES3
    fsExtensions.emplace_front( ".fsh" );
    fsExtensions.emplace_front( ".frag" );
    fsExtensions.emplace_front( ".glsl" );
#elif AETHER3D_D3D11
    fsExtensions.emplace_front( ".hlsl" );
#endif
    std::string fsPath;
    
    for (const auto& path : sShaderSearchPaths)
    {
        for (const auto& extension : fsExtensions)
        {
            const std::string testedPath = path + fsPathOrName + extension;
            
            if (fsPath.empty() && FileExists( testedPath ))
            {
                fsPath = testedPath;
            }
        }
        
        if (!fsPath.empty())
        {
            break;
        }
    }
    
    if (fsPath.empty())
    {
        fsPath = fsPathOrName;
    }
    
    return fsPath;
}

std::string ResolveGSPath( const std::string& gsPathOrName )
{
    // Finds geometry shader path.
    std::forward_list< std::string > gsExtensions;
#if AETHER3D_UNIX || AETHER3D_GL32 || AETHER3D_GL43
    gsExtensions.emplace_front( ".gsh" );
    gsExtensions.emplace_front( ".glsl" );
#elif AETHER3D_D3D11
    gsExtensions.emplace_front( ".hlsl" );
#endif
    std::string gsPath;
    
    for (const auto& path : sShaderSearchPaths)
    {
        for (const auto& extension : gsExtensions)
        {
            const std::string testedPath = path + gsPathOrName + extension;
            
            if (gsPath.empty() && FileExists( testedPath ))
            {
                gsPath = testedPath;
                break;
            }
        }
        
        if (!gsPath.empty())
        {
            break;
        }
    }
    
    if (gsPath.empty())
    {
        gsPath = gsPathOrName;
    }

    return gsPath;
}

MaterialImpl defaultMaterial;

MaterialImpl& MaterialImpl::Default()
{
    return defaultMaterial;
}

MaterialImpl::MaterialImpl() : m( new Impl )
{
}

MaterialImpl::~MaterialImpl() {}

void MaterialImpl::AddShaderSearchPath( const std::string& path )
{
    std::string fixedPath = path;
    
    if (path[ path.length() - 1 ] != '/' && path[ path.length() - 1 ] != '\\')
    {
        fixedPath += '/';
    }
    
    sShaderSearchPaths.push_front( fixedPath );
}

const char* MaterialImpl::Path() const
{
    return m->path.c_str();
}

void MaterialImpl::LoadFromFile( const char* path )
{
    std::ifstream ifs( path );
    if (!ifs)
    {
        Log::Print( "Material::LoadFromFile: Could not open %s", path );
        return;
    }
    
    m->path = path;

    std::stringstream stm;
    std::locale c_locale( "C" );
    stm.imbue( c_locale );
    
    std::map< std::string, TextureImpl* > textures;

    while (!ifs.eof())
    {
        std::string line;
        getline( ifs, line );
        stm.clear();
        stm.str( line );
        
        std::string token;
        stm >> token;

        if (token == "name")
        {
            stm >> m->name;
        }
        
        if (token == "texture")
        {
            std::string name, texPath;
            stm >> name;
            stm >> texPath;
            textures[ name ] = new TextureImpl( texPath.c_str(), true, true, TextureImpl::Tiling::Repeat, TextureImpl::Filter::Linear );
        }
        
        if (token == "transparent")
        {
            SetOpaque( false );
        }
        
        if (token == "param_texture")
        {
            std::string paramName, textureName;
            stm >> paramName >> textureName;
            
            if (textures.find( textureName ) == textures.end())
            {
                Log::Print( "Material::LoadFromFile: %s tries to use texture that is not loaded: %s", path, textureName.c_str() );
            }
            
            SetTexture( paramName, textures[ textureName ] );
        }
        
        if (token == "param_vec3")
        {
            std::string paramName;
            Vec3 vec;
            stm >> paramName >> vec.x >> vec.y >> vec.z;
            SetVector( paramName, vec );
        }
        
        if (token == "param_vec4")
        {
            std::string paramName;
            Vec4 vec;
            stm >> paramName >> vec.x >> vec.y >> vec.z >> vec.w;
            SetVector( paramName, vec );
        }
        
        if (token == "param_float")
        {
            std::string paramName;
            float f;
            stm >> paramName >> f;
            SetFloat( paramName, f );
        }

        if (token == "shaders")
        {
            std::string vertex, fragment;
            stm >> vertex >> fragment;
            
            std::string defines;
            
            while (!stm.eof())
            {
                std::string define;
                stm >> define;
                
                if (defines.empty())
                {
                    defines = define;
                }
                else
                {
                    defines += " " + define;
                }
            }
            
            SetShaders( vertex, fragment, "", defines );
        }
    }
}

const char* MaterialImpl::Name() const
{
    return m->name.c_str();
}

bool MaterialImpl::Opaque() const
{
    return m->opaque;
}

void MaterialImpl::SetOpaque( bool opaque )
{
    m->opaque = opaque;
}

Error::Enum MaterialImpl::SetShaders( const std::string& vertexShaderPathOrName,
                                      const std::string& fragmentShaderPathOrName,
                                      const std::string& geometryShaderPathOrName,
                                      const std::string& defines )
{
    const std::string vsPath = ResolveVSPath( vertexShaderPathOrName );
    const std::string fsPath = ResolveFSPath( fragmentShaderPathOrName );
    const std::string gsPath = ResolveGSPath( geometryShaderPathOrName );

    Error::Enum result = Error::NO_ERRORS;
    m->shader.Load( vsPath, fsPath, gsPath, defines, result );

    if (result != Error::NO_ERRORS)
    {
        m->shader = defaultMaterial.m->shader;
    }

    return result;
}

void MaterialImpl::SetFloat( const std::string& name, float value)
{
#if AETHER3D_GL32
    // TODO Query uniform type and ensure it's float
#endif
    m->floats[ name ] = value;
}

void MaterialImpl::SetMatrix( const std::string& name, const Matrix44& matrix )
{
#if AETHER3D_GL32
    // TODO Query uniform type and ensure it's matrix
#endif
    m->matrices[ name ] = matrix;
}

void MaterialImpl::SetName( const char* name )
{
    m->name = name;
}

void MaterialImpl::SetTexture( const std::string& name, const Texture* texture )
{
    if (m->textures.size() == TextureImpl::sMaxTextureUnit - 1)
    {
        Log::Print( "Material %s has too many textures", name.c_str() );
        return;
    }

    m->textures[ name ] = static_cast<const TextureImpl* >( texture );
}

void MaterialImpl::SetVector( const std::string& name, const Vec3& vector )
{
#if AETHER3D_GL32
    // TODO Query uniform type and ensure it's vec3
#endif

    m->vec3s[ name ] = vector;
}

void MaterialImpl::SetVector( const std::string& name, const Vec4& vector )
{
#if AETHER3D_GL32
    // TODO Query uniform type and ensure it's vec4
#endif

    m->vec4s[ name ] = vector;
}

void MaterialImpl::Apply()
{
    ErrorCheck( "MaterialImpl Apply begin" );

    auto& shader = m->shader;

    if (shader.FragmentPath() == "")
    {
        Log::Print( "Material doesn't contain shader, loading unlit." );
        SetShaders( "unlit", "unlit", "", "" );
    }

    shader.Use();

    int textureUnit = 0;
    for (const auto& texture : m->textures)
    {
        shader.SetTexture( texture.first, texture.second, textureUnit );
        ++textureUnit;
    }

    for (const auto& floatVar : m->floats)
    {
        shader.SetFloat( floatVar.first, floatVar.second );
    }

    for (const auto& matrix : m->matrices)
    {
        shader.SetMatrix( matrix.first, matrix.second );
    }

    for (const auto& vec : m->vec3s)
    {
        shader.SetVector( vec.first, vec.second );
    }

    for (const auto& vec : m->vec4s)
    {
        shader.SetVector( vec.first, vec.second );
    }

    ErrorCheck( "MaterialImpl Apply end" );
}

Shader& MaterialImpl::GetShader()
{
    if (m->shader.FragmentPath() == "")
    {
        Log::Print( "Material doesn't contain shader, loading unlit." );
        SetShaders( "unlit", "unlit", "", "" );
    }

    return m->shader;
}
