#include "Mesh.hpp"
#include <map>
#include <sstream>
#include <vector>
#include "CameraImpl.hpp"
#include "LightImpl.hpp"
#include "Log.hpp"
#include "MaterialImpl.hpp"
#include "MathUtil.hpp"
#include "Matrix.hpp"
#include "Platform.hpp"
#include "Shader.hpp"
#include "RendererImpl.hpp"
#include "Vec3.hpp"
#include "VertexBuffer.hpp"

void UploadModelMatrices( Matrix44* modelMatrices, int matrixCount, Shader& depthNormalShader );
void ErrorCheck( const std::string& info );

// Skinned or morph animation.
struct Animation
{
    enum Interpolation
    {
        Linear = 0,
        Bezier // Not implemented yet.
    };

    Animation() : inputIndex( 0 ), jointIndex( 0 ), lastKeyFrame( 0 ), stepStartTime( 0 ) {}

    std::string name;
    std::vector< float > input; // Time in seconds.
    std::vector< float > output; // Angle or position.
    int inputIndex;
    int jointIndex; // Index to Mesh::Impl.joints.
    int lastKeyFrame; // 0 means that the whole animation is played.
    std::string operation; // rotateX.ANGLE, translate, transform etc.
    std::vector< Interpolation > interpolation;
    std::vector< float > inTangents;
    std::vector< float > outTangents;

    double stepStartTime;
};

#ifdef SIMD_SSE3
struct BEGIN_ALIGNED( Joint, 16 )
#else
struct Joint
#endif
{
    std::vector< short > children; // Indices to Mesh.jointArray
    std::string name;
    Matrix44 originalTransform; // Original transform,
    Matrix44 worldTransform; // Transform multiplied by parent's world transform.
    Matrix44 transform;
    Matrix44 invBindMatrix;
}
#ifdef SIMD_SSE3
END_ALIGNED( Joint, 16 )
#endif
;

#ifdef SIMD_SSE3
struct BEGIN_ALIGNED( Mesh::Impl, 16 )
#else
struct Mesh::Impl
#endif
{
    Impl()
    : material( &MaterialImpl::Default() )
    , morphTarget( false )
    , animScale( 1, 1, 1 )
    {}

    ~Impl() {}

#if defined SIMD_SSE3 && !defined AETHER3D_UNIX
    void* operator new( std::size_t size )
    {
        void* p = _aligned_malloc( size, 16 );
        return p;
    }

    void operator delete( void* p )
    {
        Impl* ptr = static_cast< Impl* >( p );
        _aligned_free( ptr );
    }
#endif // SIMD_SSE3

    void ApplySkin( int jointIndex );
    void PrintGeometry() const;
    void SetLightingShaderParams( Shader* shader, LightImpl* light );
    
    const Matrix44* modelMatrix;
    Vec3 aabbMin;
    Vec3 aabbMax;
    Vec3 aabbMinWorld;
    Vec3 aabbMaxWorld;

    MaterialImpl* material;
    ModelImpl* model = nullptr;
    std::string name;
    VertexBuffer vertexBuffer;
    /** Lights used in current frame for lighting this mesh. */
    std::vector< LightImpl* > frameLights;

    // Skin
    Joint skeletonRoot;
    std::map< std::string, Animation > animations;

    std::vector< Influence > influences;

    // Used to determine which joint influences what.
    std::vector< std::string > jointNames;

    // Joints from skeleton organized so that influence indices point right.
    std::vector< Joint > joints;
    std::vector< std::string > morphTargets;
    std::vector< float > morphWeights;

    bool morphTarget;
    std::vector< Vec3 > morphPositions1;
    std::vector< Vec3 > morphPositions2;

    // \todo Move into UpdateAnimations() if possible.
    Matrix44 animTransformMatrix;
    Vec3 animRotation;
    Vec3 animPosition;
    Vec3 animScale;
    Vec3 animMorph; // Only .x and .y used because 2 targets are supported.    

    std::vector< Vec3 > faceNormals;
    std::vector< Face > indices;
    std::vector< Vertex > vertices;
}
#ifdef SIMD_SSE3
END_ALIGNED( Mesh::Impl, 16 )
#endif
;

void Mesh::Impl::SetLightingShaderParams( Shader* shader, LightImpl* light )
{
    shader->SetVector( "_LightColor", light->Color() );
    
    if (light->GetLightType() == Light::Type::Directional)
    {
        shader->SetVector( "_LightDirection", light->Direction() );
        shader->SetInt( "_LightType", 0 );
    }
    else if (light->GetLightType() == Light::Type::Point)
    {
        shader->SetVector( "_LightPosition", light->worldPosition );
        shader->SetFloat( "_LightRadius", light->Radius() );
        shader->SetInt( "_LightType", 1 );
    }
    else if (light->GetLightType() == Light::Type::Spot)
    {
        shader->SetVector( "_LightPosition", light->worldPosition );
        shader->SetFloat( "_LightRadius", light->Radius() );
        shader->SetVector( "_LightDirection", light->worldDirection );
        shader->SetFloat( "_LightConeAngleCos", light->GetConeAngleCos() );
        shader->SetInt( "_LightType", 2 );
    }
    
    if (light->CastsShadow())
    {
        shader->SetTexture( "_ShadowMap", &light->ShadowMap(), 4 );
        
        Matrix44 shadowTexProjMatrix = *modelMatrix;
        auto& shadowCamera = light->ShadowCamera();

        Matrix44::Multiply( shadowTexProjMatrix, shadowCamera.ViewMatrix(), shadowTexProjMatrix );
        Matrix44::Multiply( shadowTexProjMatrix, shadowCamera.ProjectionMatrix(), shadowTexProjMatrix );
        Matrix44::Multiply( shadowTexProjMatrix, Matrix44::bias, shadowTexProjMatrix );
        
        shader->SetMatrix( "_TPMatrix", shadowTexProjMatrix );
        shader->SetVector( "_LightPositionWorld", shadowCamera.GetPosition() );
    }
}

Mesh::Mesh() : m( new Impl ) {}

Mesh::Mesh( const Mesh& other ) : m( new Impl )
{
    *this = other;
}

Mesh::~Mesh() {}

Mesh& Mesh::operator=( const Mesh& other )
{
    *m = *other.m;
    return *this;
}

void Mesh::AddLight( LightImpl* light )
{
    m->frameLights.push_back( light );
}

const Vec3& Mesh::AABBMax() const
{
    return m->aabbMax;
}

const Vec3& Mesh::AABBMin() const
{
    return m->aabbMin;
}

const Vec3& Mesh::AABBMaxWorld() const
{
    return m->aabbMaxWorld;
}

const Vec3& Mesh::AABBMinWorld() const
{
    return m->aabbMinWorld;
}

const std::vector< Face >& Mesh::Indices() const
{
    return m->indices;
}

const std::vector< Vertex >& Mesh::Vertices() const
{
    return m->vertices;
}

void Mesh::ClearLights()
{
    m->frameLights.clear();
}

void Mesh::Deinit()
{
}

std::vector< Vec3 > Mesh::GetVertexPositions() const
{
    std::vector< Vec3 > outPositions;
    outPositions.reserve( m->vertices.size() );

    for (const auto& vertex : m->vertices)
    {
        outPositions.push_back( vertex.position );
    }

    return outPositions;
}

bool Mesh::HasSkin() const
{
    return !m->joints.empty();
}

bool Mesh::IsMorphBaseMesh() const
{
    return !m->morphTargets.empty();
}

bool Mesh::IsMorphTarget() const
{
    return m->morphTarget;
}

bool Mesh::IsOpaque() const
{
    return m->material != nullptr && m->material->Opaque();
}

std::vector< LightImpl* >& Mesh::Lights()
{
    return m->frameLights;
}

const ModelImpl* Mesh::Model() const
{
    return m->model;
}

const std::vector< std::string >& Mesh::JointNames() const
{
    return m->jointNames;
}

const std::vector< std::string >& Mesh::MorphTargetNames() const
{
    return m->morphTargets;
}

const std::string& Mesh::Name() const
{
    return m->name;
}

MaterialImpl* Mesh::Material()
{
    Log::Assert( m->material != nullptr, "Mesh doesn't have material." );
    
    return m->material;
}

const Matrix44& Mesh::ModelMatrix() const
{
    return *m->modelMatrix;
}

void Mesh::StartAnimations( int firstKeyFrame, int lastKeyFrame )
{
    for (auto& a : m->animations)
    {
        Animation& animation = a.second;

        animation.lastKeyFrame = lastKeyFrame;
        animation.inputIndex = firstKeyFrame < (int)animation.input.size() ? firstKeyFrame : (int)animation.input.size() - 1;
        animation.stepStartTime = firstKeyFrame < (int)animation.input.size() ? Time::GetTime() : 0;
    }
}

void Mesh::UpdateAnimations()
{
    for (auto& a : m->animations)
    {
        Animation& animation = a.second;

        // Skips this animation if it's not running.
        if (MathUtil::AlmostEquals( (float)animation.stepStartTime, 0 ))
        {
            continue;
        }

        // If true, proceed to the next time step.
        if ((Time::GetTime() - animation.stepStartTime) >= animation.input[ animation.inputIndex ])
        {
            ++animation.inputIndex;
            animation.stepStartTime = Time::GetTime();
        }

        // If true, the animation is finished.
        if (animation.inputIndex == (int)animation.input.size() ||
            (animation.lastKeyFrame != 0 && animation.inputIndex >= animation.lastKeyFrame))
        {
            animation.stepStartTime = 0;

            // If the animation was the last active animation, reset
            // joint transforms to their original state.
            for (std::size_t j = 0; j < m->joints.size(); ++j)
            {
                m->joints[ j ].transform = Matrix44::identity;
            }

            return; // \fixme Could this be "continue"?
        }

        float t = (float)(Time::GetTime( ) - animation.stepStartTime);
        const int& ind = animation.inputIndex;
        t /= animation.input[ ind ];

        bool matrixTransform = false;

        // Output values are changing location.
        if (animation.operation == "location")
        {
            const float xv = (animation.output[ ind - 1 == -1 ? 0 : (ind - 1)*3+0 ]);
            m->animPosition.x = MathUtil::Lerp( xv, animation.output[ ind * 3 + 0 ], t );

            const float yv = (animation.output[ ind - 1 == -1 ? 1 : (ind - 1)*3+1 ]);
            m->animPosition.y = MathUtil::Lerp( yv, animation.output[ ind * 3 + 1 ], t );

            const float zv = (animation.output[ ind - 1 == -1 ? 2 : (ind - 1)*3+2 ]);
            m->animPosition.z = MathUtil::Lerp( zv, animation.output[ ind * 3 + 2 ], t );
        }
        // Output values are changing X angle.
        else if (animation.operation.find( "X.ANGLE" ) != std::string::npos)
        {
            const float val1 = animation.output[ ind - 1 == -1 ? 0 : ind - 1 ];
            const float val2 = animation.output[ ind ];

            m->animRotation.x = MathUtil::Lerp( val1, val2, t );
        }
        // Output values are changing Y angle.
        else if (animation.operation.find( "Y.ANGLE" ) != std::string::npos)
        {
            float val1 = animation.output[ ind - 1 == -1 ? 0 : ind  -1 ];
            float val2 = animation.output[ ind ];

            m->animRotation.y = MathUtil::Lerp( val1, val2, t );
        }
        // Output values are changing Z angle.
        else if (animation.operation.find( "Z.ANGLE" ) != std::string::npos)
        {
            const float prev = animation.output[ ind - 1 == -1 ? 0 : ind - 1 ];
            const float next = animation.output[ ind ];

            m->animRotation.z = MathUtil::Lerp( prev, next, t );
        }
        // Output values are changing the whole transform.
        else if (animation.operation == "transform")
        {
            matrixTransform = true;

            // Generally matrices can not be interpolated, but this seems to work.
            for (int i = 0; i < 16; ++i)
            {
                const float prev = (animation.output[ ind - 1 == -1 ? i : (ind - 1) * 16 + i ]);
                const float next = (animation.output[ ind * 16 + i ]);

                m->animTransformMatrix.m[ i ] = MathUtil::Lerp( prev, next, t );
            }
        }
        // Output values are changing morph weights. \todo Make more sensible.
        else if (animation.operation.find( "morph" ) != std::string::npos)
        {
            const float prev = animation.output[ ind - 1 == -1 ? 0 : ind - 1 ];
            const float next = animation.output[ ind ];

           const float lerp = MathUtil::Lerp( prev, next, t );
           m->animMorph.x = lerp;
           m->animMorph.y = lerp;
        }
        else
        {
            Log::Assert( false, "Unimplemented animation operation!" );
        }

        if (matrixTransform)
        {
            m->joints[ animation.jointIndex ].transform = m->animTransformMatrix;
            m->joints[ animation.jointIndex ].transform.Transpose( m->joints[ animation.jointIndex ].transform );
        }
        else if (!m->joints.empty())
        {
            auto& matrix = m->animTransformMatrix;
            matrix.MakeRotationXYZ( -m->animRotation.x, -m->animRotation.y, -m->animRotation.z );
            matrix.m[ 12 ] = m->animPosition.x * m->animScale.x;
            matrix.m[ 13 ] = m->animPosition.y * m->animScale.y;
            matrix.m[ 14 ] = m->animPosition.z * m->animScale.z;

            m->joints[ animation.jointIndex ].transform = matrix;
        }

        if (!m->joints.empty())
        {
            // Applies each joint's inverse bind matrix.
            Matrix44::Multiply( m->joints[ animation.jointIndex ].invBindMatrix,
                                m->joints[ animation.jointIndex ].transform,
                                m->joints[ animation.jointIndex ].transform );
        }
    }

    if (!m->joints.empty())
    {
        m->ApplySkin( 0 );
    }
}

void Mesh::Impl::ApplySkin( int jointIndex )
{
    if (jointIndex >= (int)joints.size())
    {
        Log::Assert( false, "Invalid joint index!" );
        return;
    }

    Joint& joint = joints[ jointIndex ];

    for (std::size_t c = 0; c < joint.children.size(); ++c)
    {
        Matrix44::Multiply( joints[ joint.children[ c ] ].transform,
                            joint.worldTransform,
                            joints[ joint.children[ c ] ].worldTransform );

        ApplySkin( joint.children[ c ] );
    }
}

void Mesh::GenerateVertexBuffer()
{
    if (m->morphTargets.empty())
    {
        m->vertexBuffer.Generate( m->indices, m->vertices, m->influences );
#if AETHER3D_GL43 || AETHER3D_GL32
        m->vertexBuffer.SetName( m->name.c_str() );
#endif
    }
    else
    {
        m->vertexBuffer.Generate( m->indices, m->vertices, m->morphPositions1, m->morphPositions2 );
    }
}

Error::Enum Mesh::ReadAnimations( std::istringstream& is )
{
    // Reads # of animations.
    short nAnimations;
    is.read( (char*) &nAnimations, 2 );

    // Reads animations.
    for (short a = 0; a < nAnimations; ++a)
    {
        // Reads animation name length.
        short len;
        is.read( (char*) &len, 2 );

        // Reads animation name.
        char* animationName = new char[ len + 1 ];
        is.read( (char*) animationName, len );
        animationName[ len ] = '\0';
        std::string nameStr = std::string( animationName );

        m->animations[ nameStr ].name = nameStr;
        delete[] animationName;

        // Reads # of inputs.
        short nInputs;
        is.read( (char*) &nInputs, 2 );
        m->animations[ nameStr ].input.resize( nInputs );

        // Reads inputs.
        for (short i = 0; i < nInputs; ++i)
        {
            float input;
            is.read( (char*) &input, 4 );
            m->animations[ nameStr ].input[ i ] = input;
        }

        // Reads # of outputs.
        short nOutputs;
        is.read( (char*) &nOutputs, 2 );
        m->animations[ nameStr ].output.resize( nOutputs );

        // Reads outputs.
        for (short i = 0; i < nOutputs; ++i)
        {
            float output;
            is.read( (char*) &output, 4 );
            m->animations[ nameStr ].output[ i ] = output;
        }

        // Reads # of interpolations.
        short nInterpolations;
        is.read( (char*) &nInterpolations, 2 );

        m->animations[ nameStr ].interpolation.resize( nInterpolations );

        // Reads interpolations.
        for (short i = 0; i < nInterpolations; ++i)
        {
            unsigned interp = 0;
            is.read( (char*) &interp, 1 );
            m->animations[ nameStr ].interpolation[ i ] = (Animation::Interpolation)interp;
        }

        // Reads # of in tangents.
        short nInTangents;
        is.read( (char*) &nInTangents, 2 );

        m->animations[ nameStr ].inTangents.resize( nInTangents );

        // Reads in tangents.
        for (short i = 0; i < nInTangents; ++i)
        {
            float t = 0;
            is.read( (char*) &t, 4 );
            m->animations[ nameStr ].inTangents[ i ] = t;
        }

        // Reads # of out tangents.
        short nOutTangents;
        is.read( (char*) &nOutTangents, 2 );

        m->animations[ nameStr ].outTangents.resize( nOutTangents );

        // Reads out tangents.
        for (short i = 0; i < nOutTangents; ++i)
        {
            float t = 0;
            is.read( (char*) &t, 4 );
            m->animations[ nameStr ].outTangents[ i ] = t;
        }

        // Reads joint index.
        short j;
        is.read( (char*) &j, 2 );
        m->animations[ nameStr ].jointIndex = j;

        // Reads operation length.
        short opLen;
        is.read( (char*) &opLen, 2 );

        // Reads operation string.
        char* op = new char[ opLen + 1 ];
        is.read( (char*)op, opLen );
        op[ opLen ] = '\0';

        m->animations[ nameStr ].operation = std::string( op );
        delete[] op;
    }

    return Error::NO_ERRORS;
}

Error::Enum Mesh::ReadGeometry( std::istringstream& is )
{
    is.read( (char*)&m->aabbMin, sizeof( m->aabbMin ) );
    is.read( (char*)&m->aabbMax, sizeof( m->aabbMax ) );

    uint16_t nameLength;
    is.read( (char*)&nameLength, sizeof( nameLength ) );

    std::vector< char > meshName( nameLength + 1 );
    is.read( (char*)&meshName[ 0 ], nameLength );
    m->name = std::string( &meshName[ 0 ], meshName.size() - 1 );

    uint16_t vertexCount;
    is.read( (char*)&vertexCount, sizeof( vertexCount ) );

    try { m->vertices.resize( vertexCount ); }
    catch (std::bad_alloc&)
    { 
        Log::Print( "Ran out of memory." );
        return Error::OUT_OF_MEMORY;
    }

    is.read( (char*)&m->vertices[ 0 ].position.x, vertexCount * sizeof( Vertex ) );

    uint16_t faceCount;
    is.read( (char*)&faceCount, sizeof( faceCount ) );

    try { m->indices.resize( faceCount ); }
    catch (std::bad_alloc&)
    {
        Log::Print( "Ran out of memory." );
        return Error::OUT_OF_MEMORY;
    }

    try { m->faceNormals.resize( faceCount ); }
    catch (std::bad_alloc&)
    {
        Log::Print( "Ran out of memory." );
        return Error::OUT_OF_MEMORY;
    }

    is.read( (char*)&m->indices[ 0 ],     faceCount * sizeof( Face ) );
    is.read( (char*)&m->faceNormals[ 0 ], faceCount * sizeof( Vec3 ) );

    unsigned char isMorphTarget;
    is.read( (char*)&isMorphTarget, 1 );
    m->morphTarget = isMorphTarget == 1;

    unsigned char morphTargetNameCount = 0;
    is.read( (char*)&morphTargetNameCount, 1 );

    for (int i = 0; i < morphTargetNameCount; ++i)
    {
        unsigned char morphTargetNameLength = 0;
        is.read( (char*)&morphTargetNameLength, 1 );

        std::vector< char > targethName( morphTargetNameLength + 1 );
        is.read( (char*)&targethName[ 0 ], morphTargetNameLength );
        m->morphTargets.emplace_back( std::string( &targethName[ 0 ], targethName.size() - 1 ) );
    }

    return Error::NO_ERRORS;
}

Error::Enum Mesh::ReadSkin( std::istringstream& is )
{
    // Reads # of joint names.
    short nJointNames;
    is.read( (char*) &nJointNames, 2 );

    // Reads joint name lengths and names.
    for (short j = 0; j < nJointNames; ++j)
    {
        // Reads name length.
        uint16_t nameLength;
        is.read( (char*)&nameLength, 2 );
        char* jointName = new char[ nameLength + 1 ];
        // Reads name.
        is.read( (char*)jointName, nameLength );
        jointName[ nameLength ] = 0;
        m->jointNames.emplace_back( std::string( jointName ) );
        delete[] jointName;
    }

    m->influences.resize( m->vertices.size() );

    for (std::size_t i = 0; i < m->influences.size(); ++i)
    {
        is.read( (char*) &m->influences[ i ].jointIndices, 4 * sizeof( float ) );
        is.read( (char*) &m->influences[ i ].weights, 4 * sizeof( float ) );
    }

    m->joints.resize( m->jointNames.size() );

    // Reads joint hierarchy.
    for (std::size_t j = 0; j < m->jointNames.size(); ++j)
    {
        Joint joint;
        joint.name = m->jointNames[ j ];

        // Reads inverse bind matrix for this joint.
        is.read( (char*) &joint.invBindMatrix.m, 4 * 16 );

        // Reads transform matrix for this joint.
        is.read( (char*) &joint.transform.m, 4 * 16 );

        joint.originalTransform = joint.transform;

        // Reads # of children.
        short nChildren;
        is.read( (char*)&nChildren, 2 );
        joint.children.resize( nChildren );

        // Reads indices to children joints.
        for (short c = 0; c < nChildren; ++c)
        {
            short childIndex;
            is.read( (char*) &childIndex, 2 );
            joint.children[ c ] = childIndex;
        }

        m->joints[ j ] = joint;
    }

    // Root joint's world transform must be set to its transform, so applySkin() would work.
    m->joints[ 0 ].worldTransform = m->joints[ 0 ].transform;

    for (std::size_t j = 0; j < m->joints.size(); ++j)
    {
        m->joints[ j ].transform = Matrix44::identity;
    }

    return Error::NO_ERRORS;
}

void Mesh::Render( Shader* overrideShader, LightImpl* light )
{
    ErrorCheck( "Mesh Render begin." );
    
    Shader* shader = overrideShader ? overrideShader : &m->material->GetShader();

    if (!overrideShader)
    {
        m->material->Apply();
    }

    const auto& joints = m->joints;

    if (!joints.empty())
    {
        std::vector< Matrix44 > jointMatrices( joints.size() );

        for (std::size_t j = 0; j < joints.size(); ++j)
        {
            jointMatrices[ j ] = joints[ j ].worldTransform;
        }

        shader->SetMatrices( "boneMatrices", &jointMatrices[ 0 ], (int)joints.size() );
    }
    else if (!m->morphTargets.empty())
    {
        shader->SetVector( "morphWeights", m->animMorph );
    }

    shader->SetMatrix( "modelMatrix", *m->modelMatrix );
    shader->SetInt( "_LightCastsShadow", (light != nullptr && light->CastsShadow()) ? 1 : 0 );

    // Classic forward lighting.
    if (light)
    {
        m->SetLightingShaderParams( shader, light );
    }
    else
    {
        shader->SetTexture( "_ShadowMap", RendererImpl::Instance().DefaultTexture(), 4 );
        shader->SetInt( "_LightType", -1 );
    }

    if (overrideShader && !HasSkin())
    {
        std::vector< Matrix44 > modelMatrices;
        modelMatrices.push_back( *m->modelMatrix );
        UploadModelMatrices( modelMatrices.data(), 1, *overrideShader );
        shader->SetConstantBuffers();

        m->vertexBuffer.Bind();
        m->vertexBuffer.DrawInstanced( 1 );
    }
    else
    {
        shader->SetConstantBuffers();
        m->vertexBuffer.Bind();
        m->vertexBuffer.Draw();
    }
    
    shader->UnbindTextures();
    ErrorCheck( "Mesh Render end." );
}

void Mesh::SetMaterial( MaterialImpl* material )
{
    if (material != nullptr)
    {
        m->material = material;
    }
    else
    {
        Log::Print( "Tried to set a null material for mesh %s", m->name.c_str() );
    }
}

void Mesh::SetModel( ModelImpl* model )
{
    Log::Assert( model != nullptr, "Model is null" );
    m->model = model;
}

void Mesh::SetModelMatrix( const Matrix44& modelMatrix )
{
    m->modelMatrix = &modelMatrix;
}

void Mesh::SetMorphPositions1( const std::vector< Vec3 >& positions )
{
    m->morphPositions1 = positions;
}

void Mesh::SetMorphPositions2( const std::vector< Vec3 >& positions )
{
    m->morphPositions2 = positions;
}

void Mesh::SetWorldAABB( const Vec3& aMin, const Vec3& aMax )
{
    m->aabbMinWorld = aMin;
    m->aabbMaxWorld = aMax;
}

void Mesh::Impl::PrintGeometry() const
{
    Log::Print( "Indices: ", indices.size() );
    Log::Print( "Vertices: ", vertices.size() );

    for (const auto& vertex : vertices)
    {
        Log::Print( "vertex: %f, %f, %f", vertex.position.x, vertex.position.y, vertex.position.z );
    }
}

