#ifndef MESH_H
#define MESH_H

#include <iosfwd>
#ifndef INCLUDE_MEMORY_H
#    include <memory>
#    define INCLUDE_MEMORY_H
#endif
#ifndef INCLUDE_STRING_H
#    include <string>
#    define INCLUDE_STRING_H
#endif
#ifndef INCLUDE_VECTOR_H
#    include <vector>
#    define INCLUDE_VECTOR_H
#endif
#ifndef ERRORS_H
#    include "Errors.hpp"
#endif
#ifndef VEC3_H
#    include "Vec3.hpp"
#endif
#include "MathUtil.hpp"

class LightImpl;
class MaterialImpl;
class ModelImpl;
struct Matrix44;
class Shader;
struct Vertex;

/**
 Used to store all the data for a single vertex.
 Tangent's last element is the handedness of the vector.
 */
struct Vertex
{
    Vertex() : u( 0 ), v( 0 ) {}
    Vertex( const Vec3& pos, float aU, float aV, const Vec3& aNormal, const Vec4& aTangent )
    : position( pos )
    , u( aU )
    , v( aV )
    , normal( aNormal )
    , tangent( aTangent )
    {}
    
    Vec3 position;
    float u, v;
    Vec3 normal;
    /// w is the handedness.
    Vec4 tangent;
    Vec4 color;
};

/**
   Stores geometry, skinning and animation data.
 */
class Mesh
{
  public:
    Mesh();
    Mesh( const Mesh& other );
    virtual ~Mesh();
    Mesh& operator=( const Mesh& other );

    /**
       Adds a light to be used in current frame.
       SceneImpl sets visible lights for meshes in Cull().
     
       \param light Light.
     */
    void AddLight( LightImpl* light );
    
    /* \return AABB max in model space. */
    const Vec3& AABBMax() const;

    /* \return AABB min in model space. */
    const Vec3& AABBMin() const;

    /* \return AABB max in world space. */
    const Vec3& AABBMaxWorld() const;
    
    /* \return AABB min in world space. */
    const Vec3& AABBMinWorld() const;

    /** Clears lights used in current frame.  SceneImpl calls this in Cull(). */
    void ClearLights();
    
    /* Releases vertex buffer. */
    void Deinit();

    // \return Vertex positions.
    std::vector< Vec3 > GetVertexPositions() const;

    // Generates vertex buffer.
    void GenerateVertexBuffer();

    /** \return True, if this mesh has a skin. */
    bool HasSkin() const;

    /** \return Indices. */
    const std::vector< Face >& Indices() const;

    /** \return Vertices. */
    const std::vector< Vertex >& Vertices() const;

    // \return True, if this mesh is a morph base mesh.
    bool IsMorphBaseMesh() const;

    // \return True if this mesh is a morph target.  They should not be drawn like others.
    bool IsMorphTarget() const;

    bool IsOpaque() const;
    
    /** \return Lights that affect this mesh in this frame. Valid after SceneImpl::Cull() */
    std::vector< LightImpl* >& Lights();
    
    /** \return Model */
    const ModelImpl* Model() const;
    
    // \return Name.
    const std::string& Name() const;

    // \return Joint names.
    const std::vector< std::string >& JointNames() const;

    // \return Morph target names.
    const std::vector< std::string >& MorphTargetNames() const;

    // \return Model-to-World matrix.
    const Matrix44& ModelMatrix() const;

    // \return Material.
    MaterialImpl* Material();

    void StartAnimations( int firstKayFrame, int lastKeyFrame );

    /**
     Reads animations from an input stream that contains mesh data as in .ae3d file format.
     The stream must be positioned at animation data's begin point.
     */
    Error::Enum ReadAnimations( std::istringstream& is );

    /**
       Reads geometry from an input stream that contains mesh data as in .ae3d file format.
       The stream must be positioned at geometry data's begin point.
     */
    Error::Enum ReadGeometry( std::istringstream& is );

    /**
       Reads skin from an input stream that contains mesh data as in .ae3d file format.
       The stream must be positioned at skin data's begin point.
     */
    Error::Enum ReadSkin( std::istringstream& is );

    /**
       Renders the mesh.  If overrideShader is not used, lighting will be performed.
     
       \param overrideShader Shader to use for overriding material's shader.
       \param light LightImpl.
     */
    void Render( Shader* overrideShader, LightImpl* light );

    /**
       Sets material.
     
       \param material Material.
     */
    void SetMaterial( MaterialImpl* material );

    /** \param model Model that this Mesh belongs to. */
    void SetModel( ModelImpl* model );
    
    /// \param model Model matrix of the Object that owns this Mesh.
    void SetModelMatrix( const Matrix44& model );

    /// \param First morph positions.
    void SetMorphPositions1( const std::vector< Vec3 >& positions );

    /// \param Second morph positions.
    void SetMorphPositions2( const std::vector< Vec3 >& positions );

    void SetWorldAABB( const Vec3& aMin, const Vec3& aMax );

    void UpdateAnimations();
    
    struct Impl;

private:
    std::unique_ptr< Impl > m;
};

#endif
