#include "RenderQueue.hpp"
#include <algorithm>
#if defined AETHER3D_GLES3 && !defined AETHER3D_IOS
#    include <GLES3/gl3.h>
#elif defined AETHER3D_GLES3 && defined AETHER3D_IOS
#    include <OpenGLES/ES3/glext.h>
#else
#    include "GL/glxw.h"
#endif
#include "GfxDevice.hpp"
#include "Matrix.hpp"
#include "Shader.hpp"
#include "Statistics.hpp"
#include "TextureImpl.hpp"
#include "Vec3.hpp"
#include "Log.hpp"

void PrintInfoLog( GLuint shader, bool isProgram );
void ErrorCheck( const std::string& info );

// TODO: Abstract GL/D3D away.

RenderQueue::Vec3uv::Vec3uv() : x( 0 ), y( 0 ), z( 0 ), u( 0 ), v( 0 ) {}

RenderQueue::Vec3uv::Vec3uv( float aX, float aY, float aZ, float aU, float aV ) : x( aX ), y( aY ), z( aZ ), u( aU ), v( aV ) {}

RenderQueue::Vec3uv RenderQueue::Vec3uv::operator+( const Vec3& vec ) const
{
    return RenderQueue::Vec3uv( x + vec.x, y + vec.y, z + vec.z, u, v );
}

struct RenderQueue::Impl
{
    Impl() : shader( nullptr ) {}
    explicit Impl( Shader& aShader ) : shader( &aShader ) {}
    void Draw( float width, float height, const Matrix44& viewProjection );
    void Sort();

    std::vector< RenderQueueEntry > entries;
    std::vector< unsigned > VBOids;
    std::vector< unsigned > VAOids;
    int lastEntry = -1;
    Shader* shader;
};

RenderQueue::RenderQueue( Shader& shader ) : m( new Impl( shader ) ) {}

RenderQueue::~RenderQueue() {}

void RenderQueue::Add( PrimitiveType aPrimitive, const std::vector< Vec3uv >& aVertices,
                       const TextureImpl* aTexture,
                       int aSortingValue, const Vec4& color )
{
    // Finds an entry that has same state as the arguments and inserts vertices
    // there so we can render them using one draw call.
    for (int k = 0; k <= m->lastEntry; ++k)
    {
        auto& i = m->entries[ k ];
        
        if (i.primitiveType == aPrimitive && i.sortingValue == aSortingValue &&
            ((i.texture && aTexture) ? (i.texture->Id() == aTexture->Id()) : &i.texture == &aTexture) &&
            i.red == color.x && i.green == color.y && i.blue == color.z && i.opacity == color.w)
        {
            i.vertices.insert( std::end( i.vertices ), std::begin( aVertices ), std::end( aVertices ) );
            return;
        }
    }
    
    // Creates a new entry.
    ++m->lastEntry;
    auto& entries = m->entries;

    if ((std::size_t)m->lastEntry == entries.size())
    {
        if (!m->VBOids.empty() && m->lastEntry > 0)
        {
            glDeleteBuffers( m->lastEntry, &m->VBOids[ 0 ] );
            glDeleteBuffers( m->lastEntry, &m->VAOids[ 0 ] );
        }
        
        const std::size_t newSize = entries.size() == 0 ? 20 : entries.size() * 2;
        
        entries.resize( newSize );
        m->VBOids.resize( newSize );
        m->VAOids.resize( newSize );
        
        glGenBuffers( static_cast< int >( newSize ), m->VBOids.data() );
        glGenVertexArrays( static_cast< int >( newSize ), m->VAOids.data() );
    }
    
    RenderQueueEntry e( aSortingValue, aTexture, aVertices, color.x, color.y, color.z, color.w, aPrimitive );
    entries[ m->lastEntry ] = e;
}

void RenderQueue::Flush( float width, float height, const Matrix44& viewProjection )
{
    m->Sort();
    m->Draw( width, height, viewProjection );
    
    for (auto& e : m->entries)
    {
        e = RenderQueueEntry();
    }
    
    m->lastEntry = -1;
}

void RenderQueue::Deinit() const
{
    if (m->lastEntry != -1 && !m->VBOids.empty())
    {
        glDeleteBuffers( m->lastEntry, m->VBOids.data() );
        glDeleteVertexArrays( m->lastEntry, m->VAOids.data() );
    }
}

void RenderQueue::Impl::Draw( float width, float height, const Matrix44& viewProjection )
{
    ErrorCheck( "RenderQueue Draw begin" );

    if (lastEntry == -1)
    {
        return;	
    }
    
    glDisable( GL_CULL_FACE  );
    GfxDevice().Instance().SetDepthFunc( DepthFunc::None );

    shader->Use();

    Matrix44 projection;
    projection.MakeProjection( 0, width, height, 0, -1, 1 );

    // Scale and offset are baked into geometry to enable batching, so identity is set here.
    shader->SetVector( "textureMap_ST", Vec4( 1, 1, 0, 0 ) );
    shader->SetMatrix( "modelMatrix", Matrix44::identity );

    Vec4 tint;

    for (std::size_t e = 0; e <= (std::size_t)lastEntry; ++e)
    {
        auto& i = entries[ e ];
        const Matrix44& matrix = i.primitiveType == RenderQueue::PrimitiveType::Triangles ? projection : viewProjection;
        shader->SetMatrix( "_ViewProjectionMatrix", matrix );

        if (i.texture != nullptr)
        {
            i.texture->Bind( 0 );
            shader->SetInt( "textureMap", 0 );
        }

        tint = Vec4( i.red, i.green, i.blue, i.opacity );
        shader->SetVector( "tint", tint );

        glBindVertexArray( VAOids[ e ] );
        Statistics::IncBufferSwitches();

        glBindBuffer( GL_ARRAY_BUFFER, VBOids[ e ] );
        glBufferData( GL_ARRAY_BUFFER, sizeof( RenderQueue::Vec3uv ) * (GLsizeiptr)i.vertices.size(), &i.vertices[ 0 ], GL_STREAM_DRAW );

        glVertexAttribPointer( Shader::Channel::POSITION, 3, GL_FLOAT, GL_FALSE, sizeof( RenderQueue::Vec3uv ), (const GLvoid*)0 );
        glEnableVertexAttribArray( Shader::Channel::POSITION );

        glVertexAttribPointer( Shader::Channel::TEXCOORD, 2, GL_FLOAT, GL_FALSE, sizeof( RenderQueue::Vec3uv ), (const GLvoid*)(3 * 4) );
        glEnableVertexAttribArray( Shader::Channel::TEXCOORD );
        
#if DEBUG
        Shader::Validate();
#endif
        glDrawArrays( i.primitiveType == PrimitiveType::Triangles ? GL_TRIANGLES : GL_LINES, 0, (int)i.vertices.size() );
        Statistics::IncDrawCalls();
        
        glDisableVertexAttribArray( Shader::Channel::POSITION );
        glDisableVertexAttribArray( Shader::Channel::TEXCOORD );

        glBindVertexArray( 0 );
        Statistics::IncBufferSwitches();
        
        glBindBuffer( GL_ARRAY_BUFFER, 0 );
    }

    glEnable( GL_CULL_FACE );
    GfxDevice().Instance().SetDepthFunc( DepthFunc::LessWriteOn );

    ErrorCheck( "RenderQueue Draw end" );
}

void RenderQueue::Impl::Sort()
{
    // \fixme This code assumes that sorting values are initialized
    // to a very large value in empty constructor.
    auto sortFunction = [](const RenderQueueEntry& e1, const RenderQueueEntry& e2) { return e1.sortingValue < e2.sortingValue; };
    std::sort( std::begin( entries ), std::end( entries ), sortFunction );
}
