#include "RendererImpl.hpp"
#include <forward_list>
#include <string>
#include <vector>
#include "CameraImpl.hpp"
#include "GfxDevice.hpp"
#include "LightImpl.hpp"
#include "Log.hpp"
#include "MaterialImpl.hpp"
#include "MathUtil.hpp"
#include "Mesh.hpp"
#include "ModelImpl.hpp"
#include "TextureImpl.hpp"
#include "Platform.hpp"
#include "RenderQueue.hpp"
#include "SceneImpl.hpp"
#include "Statistics.hpp"
#include "VertexBuffer.hpp"
#include "Vec3.hpp"

#ifdef AETHER3D_WINDOWS
#    define AETHER3D_API __declspec(dllexport)
#else
#    define AETHER3D_API
#endif

void SetDebugEventBegin( const std::string& name );
void SetDebugEventEnd();
void ErrorCheck(const std::string& info);
void UploadModelMatrices(Matrix44* modelMatrices, int matrixCount, Shader& depthNormalShader);

namespace Screen
{
    int Width()  { return RendererImpl::Instance().Width(); }
    int Height() { return RendererImpl::Instance().Height(); }
    bool HasFocus() { return RendererImpl::Instance().HasFocus(); }
    void SetFocus(bool enable) { RendererImpl::Instance().SetFocus(enable); }
}

namespace Global
{
    // TODO: Get from driver.
#if AETHER3D_GLES3
    int maxInstanceCount = 256;
#else
    int maxInstanceCount = 1024;
#endif
}

extern "C" AETHER3D_API Renderer* CreateRenderer( const char* builtinAssetPath )
{
    std::string path( builtinAssetPath );
    RendererImpl::Instance().SetBuiltinAssetPath( path );
    return &(RendererImpl::Instance());
}

extern "C" AETHER3D_API unsigned RendererVersion()
{
    return Renderer::apiVersion;
}

// If inOutShadersUsedInFrame doesn't contain mesh's shader, it's added to the list.  List is used to apply global uniforms.
static void AddMeshShader( Mesh* mesh, std::forward_list< Shader* >& inOutShadersUsedInFrame )
{
    if (std::none_of( std::begin( inOutShadersUsedInFrame ), std::end( inOutShadersUsedInFrame ),
                     [&](Shader* sh) { return sh->Equals( mesh->Material()->GetShader() ); } ))
    {
        inOutShadersUsedInFrame.push_front( &mesh->Material()->GetShader() );
    }
}

std::vector< Vec4 > CreateSSAOKernel( int size )
{
    std::vector< Vec4 > kernel( size );

    for (int i = 0; i < size; ++i)
    {
        Vec3 v = Vec3(
            MathUtil::Random( -1.0f, 1.0f ),
            MathUtil::Random( -1.0f, 1.0f ),
            MathUtil::Random( 0.0f, 1.0f )
            ).Normalized();

        // Distributes the samples so that there are more near the origin.
        float scale = (float)i / (float)size;
        v *= MathUtil::Lerp( 0.1f, 1.0f, scale * scale );

        kernel[ i ].x = v.x;
        kernel[ i ].y = v.y;
        kernel[ i ].z = v.z;
        kernel[ i ].w = 0;
    }

    return kernel;
}

void CreateSkybox( VertexBuffer& outSkybox )
{
    const float s = 50;

    const std::vector< Vertex > vertices =
    {
        Vertex( Vec3( -s, -s, s ), 0, 0, Vec3(), Vec4() ),
        Vertex( Vec3( s, -s, s ), 0, 0, Vec3(), Vec4() ),
        Vertex( Vec3( s, -s, -s ), 0, 0, Vec3(), Vec4() ),
        Vertex( Vec3( -s, -s, -s ), 0, 0, Vec3(), Vec4() ),
        Vertex( Vec3( -s, s, s ), 0, 0, Vec3(), Vec4() ),
        Vertex( Vec3( s, s, s ), 0, 0, Vec3(), Vec4() ),
        Vertex( Vec3( s, s, -s ), 0, 0, Vec3(), Vec4() ),
        Vertex( Vec3( -s, s, -s ), 0, 0, Vec3(), Vec4() )
    };

    const std::vector< Face > indices =
    {
        Face( 0, 4, 1 ),
        Face( 4, 5, 1 ),
        Face( 1, 5, 2 ),
        Face( 2, 5, 6 ),
        Face( 2, 6, 3 ),
        Face( 3, 6, 7 ),
        Face( 3, 7, 0 ),
        Face( 0, 7, 4 ),
        Face( 4, 7, 5 ),
        Face( 5, 7, 6 ),
        Face( 3, 0, 2 ),
        Face( 2, 0, 1 )
    };

    std::vector< Influence > influencesEmpty;

    outSkybox.Generate( indices, vertices, influencesEmpty );
}

void CreateQuadBuffer( VertexBuffer& outQuadBuffer )
{
    std::vector< Vertex > quadVertices;
    quadVertices.emplace_back( Vertex( Vec3( -1, -1, 0 ), 0, 0, Vec3(), Vec4() ) );
    quadVertices.emplace_back( Vertex( Vec3(  1, -1, 0 ), 1, 0, Vec3(), Vec4() ) );
    quadVertices.emplace_back( Vertex( Vec3(  1,  1, 0 ), 1, 1, Vec3(), Vec4() ) );
    quadVertices.emplace_back( Vertex( Vec3( -1,  1, 0 ), 0, 1, Vec3(), Vec4() ) );

    std::vector< Face > indices;
    indices.emplace_back( Face( 0, 1, 2 ) );
    indices.emplace_back( Face( 2, 3, 0 ) );

    std::vector< Influence > influencesEmpty;

    outQuadBuffer.Generate( indices, quadVertices, influencesEmpty );
}

void DrawModelInstances( std::vector< const ModelImpl* >& models, Shader& depthNormalShader )
{
    std::unordered_map< std::string, std::list< const ModelImpl* > > modelInstances;

    for (auto model : models)
    {
        if (model->CombinedOpaqueMeshes().IsGenerated())
        {
            modelInstances[ model->GetPath() ].push_back( model );
        }
    }

    for (auto modelInstance : modelInstances)
    {
        std::vector< Matrix44 > modelMatrices;

        for (auto i : modelInstance.second)
        {
            modelMatrices.push_back( i->ModelMatrix() );
        }

        const int batches = (int)MathUtil::Ceil( (float)modelMatrices.size() / (float)Global::maxInstanceCount );
        const int lastBatchSize = (int)(modelMatrices.size() % Global::maxInstanceCount);

        for (int i = 0; i < batches; ++i)
        {
            const int drawCount = i != batches - 1 ? Global::maxInstanceCount : lastBatchSize;

            UploadModelMatrices( &modelMatrices[ i * Global::maxInstanceCount ], drawCount, depthNormalShader );
            depthNormalShader.SetConstantBuffers();

            modelInstance.second.front()->CombinedOpaqueMeshes().Bind();
            modelInstance.second.front()->CombinedOpaqueMeshes().DrawInstanced( (int)modelMatrices.size() );
        }
    }
}

RendererImpl& RendererImpl::Instance()
{
    static RendererImpl renderer;
    return renderer;
}

RendererImpl::RendererImpl() :
  transparentRenderQueue( builtinShaders.unlitShader )
, opaqueRenderQueue( builtinShaders.unlitShader )
{
}

RendererImpl::~RendererImpl() {}

void RendererImpl::AddShaderSearchPath( const char* path ) const
{
    std::string tmp( path );
    MaterialImpl::AddShaderSearchPath( tmp.c_str() );
}

void RendererImpl::Blur( TextureImpl& source, TextureImpl& target, bool horizontal )
{
    ErrorCheck( "Blur begin" );

    GfxDevice::Instance().SetRenderTarget( &target, 0 );

    const float texSize = (float)target.Width();

    // Texel offset for blur filter kernel
    float texelOffset = 1.0f / texSize;

    // Altered weights for the faster filter kernel
    const float w1 = 0.0555555f;
    const float w2 = 0.2777777f;
    const float intraTexelOffset = (w1 / (w1 + w2)) * texelOffset;
    texelOffset += intraTexelOffset;
    
    builtinShaders.blurShader.Use();
    builtinShaders.blurShader.SetVector( "texelOffset", horizontal ? Vec3( texelOffset, 0, 0 ) : Vec3( 0, texelOffset, 0 ) );
    builtinShaders.blurShader.SetTexture( "textureMap", &source, 0 );
    builtinShaders.blurShader.SetMatrix( "_ViewProjectionMatrix", Matrix44::identity );
    builtinShaders.blurShader.SetMatrix( "modelMatrix", Matrix44::identity );
    builtinShaders.blurShader.SetConstantBuffers();

    quadBuffer.Bind();
    quadBuffer.Draw();
    
    builtinShaders.blurShader.SetTexture( "textureMap", &source, 666 );

    ErrorCheck( "Blur end" );
}

void RendererImpl::ClearScreen() const
{
    GfxDevice::Instance().ClearScreen();
}

void RendererImpl::Draw( const Texture* texture, int x, int y, const Vec4& color )
{
    if (texture != nullptr)
    {
        Draw( texture, x, y, texture->Width(), texture->Height(), color );
    }
    else
    {
        Log::Print( "Draw: Texture is null, not drawing." );
    }
}

void RendererImpl::Draw( const Texture* texture, int x, int y, int width, int height, const Vec4& color )
{
    if (!texture)
    {
        Log::Print( "Draw: Texture is null, not drawing." );
        return;
    }
    
    const TextureImpl* textureImpl = static_cast< const TextureImpl* >( texture );
    
    const float w = (float)width;
    const float h = (float)height;
    
    const float fx = static_cast<float>( x );
    const float fy = static_cast<float>( y );
    
    const Vec4& so = textureImpl->ScaleOffset();
    
    const float u0 = 0.0f * so.x + so.z;
    const float u1 = 1.0f * so.x + so.z;
    
    const float v0 = 0.0f * so.y + so.w;
    const float v1 = 1.0f * so.y + so.w;
    
    std::vector< RenderQueue::Vec3uv > quad( 6 );
    quad[ 0 ] = RenderQueue::Vec3uv( fx    , fy    , 0, u0, v0 );
    quad[ 1 ] = RenderQueue::Vec3uv( fx + w, fy    , 0, u1, v0 );
    quad[ 2 ] = RenderQueue::Vec3uv( fx + w, fy + h, 0, u1, v1 );
    quad[ 3 ] = RenderQueue::Vec3uv( fx    , fy    , 0, u0, v0 );
    quad[ 4 ] = RenderQueue::Vec3uv( fx    , fy + h, 0, u0, v1 );
    quad[ 5 ] = RenderQueue::Vec3uv( fx + w, fy + h, 0, u1, v1 );
    
    ((!textureImpl->Opaque() || color.w != 1.0f) ? transparentRenderQueue : opaqueRenderQueue).Add( RenderQueue::PrimitiveType::Triangles, quad, textureImpl, 0,
                                                                                                   color );
}

void RendererImpl::Draw( const char* text, const char* fontName, float scale, int x, int y, const Vec4& color )
{
    if (fonts.find( fontName ) == fonts.end())
    {
        Log::Print( "Draw: Didn't find font %s.", fontName );
    }
    
    const Fonter& fonter = fonts.find( fontName ) != fonts.end() ? fonts[ fontName ] : fonts[ "aether3d_font" ];
    transparentRenderQueue.Add( RenderQueue::PrimitiveType::Triangles, fonter.CreateGeometry( text, (float)x, (float)y, scale ), fonter.Texture(), 1, color );
}

double RendererImpl::GetTime() const
{
    return Time::GetTime();
}

const TextureImpl* RendererImpl::DefaultTexture()
{
    return fonts[ "aether3d_font" ].Texture();
}

void RendererImpl::Draw( Scene* scene )
{
    if (scene == nullptr)
    {
        Log::Print( "Draw: Tried to draw a scene without a valid scene pointer." );
        return;
    }

    SceneImpl* sceneImpl = static_cast< SceneImpl* >( scene );

    sceneImpl->Root()->SolveMatrices();
    sceneImpl->TransformLights();
    sceneImpl->UpdateAABBs();
    sceneImpl->Root()->CleanDirties();

    DrawRenderTextureCameras( sceneImpl );
    DrawNormalCameras( sceneImpl );

    GfxDevice().Instance().SetDepthFunc( DepthFunc::None );

    if (bloomEnabled)
    {
        Blur( builtinTargets.hdrTarget, builtinTargets.bloomBlurHTarget, true );
        Blur( builtinTargets.bloomBlurHTarget, builtinTargets.bloomBlurVTarget, false );

        for (int i = 0; i < bloomBlurIterations - 1; ++i)
        {
            Blur( builtinTargets.bloomBlurVTarget, builtinTargets.bloomBlurHTarget, true );
            Blur( builtinTargets.bloomBlurHTarget, builtinTargets.bloomBlurVTarget, false );
        }
    }

    if (ssaoEnabled)
    {
        ApplySSAO( sceneImpl->ActiveCamera()->FOV() );
    }

    ApplyToneMap();

    ErrorCheck( "Draw Scene end" );
}

void RendererImpl::DrawRenderTextureCameras( SceneImpl* sceneImpl )
{
    std::list< CameraImpl* > rtCameras;
    sceneImpl->GetCameras( sceneImpl->Root(), true, rtCameras );

    for (auto cam : rtCameras)
    {
        sceneImpl->SetActiveCamera( cam );
        TextureImpl* renderTarget = static_cast< TextureImpl* >( sceneImpl->ActiveCamera()->GetRenderTarget() );
        skyboxTexture = sceneImpl->Skybox();

        if (renderTarget->IsCubemap())
        {
            CameraImpl* camera = sceneImpl->ActiveCamera();

            for (int cubeMapFace = 0; cubeMapFace < 6; ++cubeMapFace)
            {
                GfxDevice::Instance().SetRenderTarget( renderTarget, cubeMapFace );
                GfxDevice::Instance().ClearScreen();

                const float scale = 2000;

                static const Vec3 directions[ 6 ] =
                {
                    Vec3(  1, 0, 0 ) * scale, // posX
                    Vec3( -1, 0, 0 ) * scale, // negX
                    Vec3( 0, 1, 0 ) * scale, // posY
                    Vec3( 0, -1, 0 ) * scale, // negY
                    Vec3( 0, 0, -1 ) * scale, // posZ
                    Vec3( 0, 0, 1 ) * scale  // negZ
                };

                static const Vec3 ups[ 6 ] =
                {
                    Vec3( 0, -1, 0 ),
                    Vec3( 0, -1, 0 ),
                    Vec3( 0, 1, 0 ),
                    Vec3( 0, 1, 0 ),
                    Vec3( 0, -1, 0 ),
                    Vec3( 0, -1, 0 )
                };

                camera->LookAt( camera->GetPosition(), camera->GetPosition() + directions[ cubeMapFace ], ups[ cubeMapFace ] );

                state.view = camera->ViewMatrix();
                state.projection = camera->ProjectionMatrix();
                state.cameraPosition = camera->GetPosition();

                sceneImpl->Cull( opaqueMeshQueue, transparentMeshQueue );

                if (skyboxTexture)
                {
                    RenderSkybox();
                }

                FinishFrame();
            }
        }
        else // Not cube map.
        {
            GfxDevice::Instance().SetRenderTarget( renderTarget );
            GfxDevice::Instance().ClearScreen();

            state.view = sceneImpl->ActiveCamera()->ViewMatrix();
            state.projection = sceneImpl->ActiveCamera()->ProjectionMatrix();
            state.cameraPosition = sceneImpl->ActiveCamera()->GetPosition();

            sceneImpl->Cull( opaqueMeshQueue, transparentMeshQueue );

            if (skyboxTexture)
            {
                RenderSkybox();
            }

            FinishFrame();
        }
    }
}

void RendererImpl::DrawLines( const Vec3* lineArray, int arraySize, const Vec4& color )
{
    if (arraySize <= 0 || lineArray == nullptr)
    {
        Log::Print( "DrawLines: Line count must be > 0 and lines must not be null." );
        return;
    }
    
    std::vector< RenderQueue::Vec3uv > vertices( arraySize );
    
    for (int i = 0; i < arraySize; ++ i)
    {
        vertices[ i ].x = lineArray[ i ].x;
        vertices[ i ].y = lineArray[ i ].y;
        vertices[ i ].z = lineArray[ i ].z;
    }
    
    opaqueRenderQueue.Add( RenderQueue::PrimitiveType::Lines, vertices, &whiteTexture, 5, color );
}

void RendererImpl::DrawStatistics()
{
    static double lastTime = Time::GetTime();
    static int frame = 0;
    static int frameDisplay = 1000;
    double currentTime = Time::GetTime();
    
    ++frame;
    
    // Updates frame time every second.
    if (1 < currentTime - lastTime)
    {
        lastTime += 1.0;
        frameDisplay = frame;
        frame = 0;
    }
    
    const std::string str = std::string( "draw calls: " ) + std::to_string( Statistics::DrawCalls() );
    Draw( str.c_str(), "aether3d_font", 1, 5, 0, Vec4( 1, 1, 1, 1 ) );

    const double frameTime = 1000.0 / double( frameDisplay );
    const std::string cpuTime = std::string( "CPU: " ) + std::to_string( frameTime ) + " ms";
    Draw( cpuTime.c_str(), "aether3d_font", 1, 5, 25, Vec4( 1, 1, 1, 1 ) );
    
    const std::string rtStr = std::string( "rt switches: " ) + std::to_string( Statistics::RenderTargetSwitches() );
    Draw( rtStr.c_str(), "aether3d_font", 1, 5, 50, Vec4( 1, 1, 1, 1 ) );
    
    const std::string ssStr = std::string( "shader switches: " ) + std::to_string( Statistics::RenderTargetSwitches() );
    Draw( ssStr.c_str(), "aether3d_font", 1, 5, 75, Vec4( 1, 1, 1, 1 ) );

    const std::string bsStr = std::string( "VAO switches: " ) + std::to_string( Statistics::BufferSwitches() );
    Draw( bsStr.c_str(), "aether3d_font", 1, 5, 100, Vec4( 1, 1, 1, 1 ) );

    const std::string tsStr = std::string( "Texture switches: " ) + std::to_string( Statistics::TextureSwitches() );
    Draw( tsStr.c_str(), "aether3d_font", 1, 5, 125, Vec4( 1, 1, 1, 1 ) );
}

void RendererImpl::FinishFrame()
{
    ErrorCheck( "FinishFrame begin" );
    
    ApplySharedShaderParams();

    if (!opaqueMeshQueue.empty())
    {
#if AETHER3D_GL43 || AETHER3D_D3D11
        RenderDepthAndNormals();
        forwardPlus.CullLights( state, builtinShaders.cullLightShader, builtinTargets.depthNormalTarget );
#else
        if (ssaoEnabled)
        {
            RenderDepthAndNormals();
        }
#endif

        ErrorCheck( "Opaque pass begin." );
        
        auto meshSortFunctionNearToFar = [&]( Mesh* mesh1, Mesh* mesh2 )
        {
            return Vec3::DistanceSquared( mesh1->Model()->GetPosition(), state.cameraPosition ) <
            Vec3::DistanceSquared( mesh2->Model()->GetPosition(), state.cameraPosition );
        };
        
        opaqueMeshQueue.sort( meshSortFunctionNearToFar );
        
        GfxDevice::Instance().SetDepthFunc( DepthFunc::LessOrEqualWriteOn );
        GfxDevice::Instance().SetRasterizerState( RasterizerState::CullBackCCW );
        
        for (auto& mesh : opaqueMeshQueue)
        {
            RenderOpaqueMeshQueueItem( mesh );
        }
        
        ErrorCheck( "Opaque pass end." );
        
        opaqueMeshQueue.clear();
    }

    GfxDevice::Instance().SetBlendMode( BlendMode::AlphaBlend );
    GfxDevice::Instance().SetDepthFunc( DepthFunc::LessOrEqualWriteOff );
    GfxDevice::Instance().SetRasterizerState( RasterizerState::CullNoneCCW );
    
    auto meshSortFunctionFarToNear = [&]( Mesh* mesh1, Mesh* mesh2 )
    {
        return Vec3::DistanceSquared( mesh1->Model()->GetPosition(), state.cameraPosition ) >
        Vec3::DistanceSquared( mesh2->Model()->GetPosition(), state.cameraPosition );
    };
    
    transparentMeshQueue.sort( meshSortFunctionFarToNear );
    
    for (auto& mesh : transparentMeshQueue)
    {
        if (mesh->Lights().empty())
        {
            mesh->Render( nullptr, nullptr );
        }
        else
        {
            for (auto light : mesh->Lights())
            {
                mesh->Render( nullptr, light );
            }
        }
    }
    
    transparentMeshQueue.clear();
    GfxDevice::Instance().SetBlendMode( BlendMode::Off );
    GfxDevice::Instance().SetDepthFunc( DepthFunc::LessOrEqualWriteOn );
    GfxDevice::Instance().SetRasterizerState( RasterizerState::CullNoneCCW );
    opaqueRenderQueue.Flush( static_cast< float >( Width() ), static_cast< float >( Height() ), state.viewProjection );

    GfxDevice::Instance().SetBlendMode( BlendMode::AlphaBlend );
    transparentRenderQueue.Flush( static_cast< float >( Width() ), static_cast< float >( Height() ), state.viewProjection );
    GfxDevice::Instance().SetBlendMode( BlendMode::Off );
    
    ErrorCheck( "FinishFrame end" );
}

int RendererImpl::Height() const
{
    return screenSize[ 1 ];
}

Error::Enum RendererImpl::LoadFont( const char* fontName, const char* fontBitmapPath, const char* fontMetadataPath )
{
    if (fonts.find( fontName ) != fonts.end())
    {
        Log::Print( "Warning: RendererImpl::LoadFont(): %s already exists, overwriting.", fontName );
    }
    
    return fonts[ fontName ].Load( fontBitmapPath, fontMetadataPath );
}

void RendererImpl::SetBloom( bool enable, float factor, int blurIterations )
{
    if (blurIterations > 0)
    {
        bloomBlurIterations = blurIterations;
    }
    else
    {
        Log::Print( "SetBloom: iterations < 0, setting to 1." );
        bloomBlurIterations = 1;
    }
    
    bloomEnabled = enable;
    builtinShaders.toneMapShader.Use();
    builtinShaders.toneMapShader.SetFloat( "bloomFactor", factor );
}

void RendererImpl::SetClearColor( float red, float green, float blue )
{
    GfxDevice::Instance().SetClearColor( red, green, blue );
}

void RendererImpl::SetSSAO( bool enable, float radius, float power )
{
    ssaoEnabled = enable;
    builtinShaders.ssaoShader.Use();
    builtinShaders.ssaoShader.SetFloat( "uRadius", radius );
    builtinShaders.ssaoShader.SetFloat( "uPower", power );
}

void RendererImpl::SetBuiltinAssetPath( const std::string& path )
{
    builtinAssetPath = path;
    
    if (path[ path.size() - 1 ] != '\\' &&
        path[ path.size() - 1 ] != '/')
    {
        builtinAssetPath += '/';
    }
}

int RendererImpl::Width() const
{
    return screenSize[ 0 ];
}

void RendererImpl::DrawNormalCameras( SceneImpl* sceneImpl )
{
    Log::Assert( sceneImpl != nullptr, "sceneImpl is null." );

    std::list< CameraImpl* > cameras;
    sceneImpl->GetCameras( sceneImpl->Root(), false /* render texture camera */, cameras );

    for (auto cam : cameras)
    {
        DrawShadows( sceneImpl, cam );
        GfxDevice::Instance().SetRenderTarget( &builtinTargets.hdrTarget );

        sceneImpl->SetActiveCamera( cam );
        
        int viewport[ 4 ];
        cam->GetViewport( viewport[ 0 ], viewport[ 1 ], viewport[ 2 ], viewport[ 3 ] );
        GfxDevice::Instance().SetViewport( viewport[ 0 ], viewport[ 1 ], viewport[ 2 ], viewport[ 3 ] );
        GfxDevice::Instance().ClearScreen();

        state.view = cam->ViewMatrix();
        state.projection = cam->ProjectionMatrix();
        state.cameraPosition = cam->GetPosition();

        skyboxTexture = sceneImpl->Skybox();

        sceneImpl->Cull( opaqueMeshQueue, transparentMeshQueue );

        std::for_each( std::begin( opaqueMeshQueue ), std::end( opaqueMeshQueue ), []( Mesh* mesh ) { mesh->UpdateAnimations(); } );
        std::for_each( std::begin( transparentMeshQueue ), std::end( transparentMeshQueue ), []( Mesh* mesh ) { mesh->UpdateAnimations(); } );

        if (skyboxTexture)
        {
            RenderSkybox();
        }
        
        FinishFrame();
    }
}

void RendererImpl::DoShadowCasterPass( std::list< Mesh* >& meshQueue )
{
    // Depth-pass for all opaque combined geometry (= Model's meshes combined into one).
    std::vector< const ModelImpl* > models;
    for (auto& mesh : meshQueue)
    {
        bool foundModel = false;
        
        for (auto& model : models)
        {
            if (model == mesh->Model())
            {
                foundModel = true;
                break;
            }
        }
        
        if (!foundModel)
        {
            models.push_back( mesh->Model() );
        }
    }
    
    BuiltinShader().momentsShader.Use();
    
    DrawModelInstances( models, BuiltinShader().momentsShader );
    
    for (auto& mesh : meshQueue)
    {
        if (!mesh->HasSkin() && !mesh->Model()->CombinedOpaqueMeshes().IsGenerated())
        {
            BuiltinShader().momentsShader.Use();
            mesh->Render( &BuiltinShader().momentsShader, nullptr );
        }
        else if (mesh->HasSkin())
        {
            BuiltinShader().momentsSkinShader.Use();
            mesh->Render( &BuiltinShader().momentsSkinShader, nullptr );
        }
    }

    meshQueue.clear();
}

void RendererImpl::DrawShadows( SceneImpl* sceneImpl, CameraImpl* eyeCamera )
{
    Log::Assert( sceneImpl != nullptr, "SceneImpl is null." );
    Log::Assert( eyeCamera != nullptr, "eyeCamera is null." );

    SetDebugEventBegin("Render shadows");
    GfxDevice::Instance().SetRasterizerState(RasterizerState::CullBackCCW);

    for (auto& light : sceneImpl->GetLights())
    {
        if (light->CastsShadow())
        {
            auto& shadowCamera = light->ShadowCamera();

            sceneImpl->SetActiveCamera( &shadowCamera );

            // TODO: Set only once
            shadowCamera.SetRenderTarget( &light->ShadowMap() );

            GfxDevice::Instance().SetRenderTarget( &light->ShadowMap() );
            GfxDevice::Instance().Set_sRGB_Writes( false );
            GfxDevice::Instance().ClearScreen();

            sceneImpl->SolveShadowFrustum( light, eyeCamera, shadowCamera );
            sceneImpl->Cull( opaqueMeshQueue, transparentMeshQueue );

            state.view = shadowCamera.ViewMatrix();
            state.projection = shadowCamera.ProjectionMatrix();
            state.cameraPosition = shadowCamera.GetPosition();
            
            ApplySharedShaderParams();
            DoShadowCasterPass( opaqueMeshQueue );
            transparentMeshQueue.clear();
            
            GfxDevice().Instance().SetDepthFunc( DepthFunc::None );
            Blur( light->ShadowMap(), builtinTargets.shadowBlurHTarget, true );
            Blur( builtinTargets.shadowBlurHTarget, light->ShadowMap(), false );
            GfxDevice().Instance().SetDepthFunc( DepthFunc::LessOrEqualWriteOn );

            //Draw( &light->ShadowMap(), 0, 512, 512, -512, Vec4( 1, 1, 1, 1 ) );
            GfxDevice::Instance().Set_sRGB_Writes( true );
        }
    }

    SetDebugEventEnd();
}

void RendererImpl::RenderDepthAndNormals()
{
    ErrorCheck( "DepthNormal pass begin." );
    SetDebugEventBegin( "RenderDepthAndNormals" );

    GfxDevice::Instance().SetBlendMode( BlendMode::Off );
    GfxDevice::Instance().SetDepthFunc( DepthFunc::LessOrEqualWriteOn );
    GfxDevice::Instance().SetRasterizerState( RasterizerState::CullBackCCW );

    const TextureImpl* oldTarget = GfxDevice::Instance().CurrentRenderTarget();
    GfxDevice::Instance().SetRenderTarget( &builtinTargets.depthNormalTarget, 0 );
    GfxDevice::Instance().ClearScreen();

    GfxDevice::Instance().Set_sRGB_Writes( false );

    // Depth-pass for all combined geometry (= Model's meshes combined into one).
    std::vector< const ModelImpl* > models;
    for (auto& mesh : opaqueMeshQueue)
    {
        bool foundModel = false;

        for (auto& model : models)
        {
            if (model == mesh->Model())
            {
                foundModel = true;
                break;
            }
        }

        if (!foundModel)
        {
            models.push_back( mesh->Model() );
        }
    }

    builtinShaders.depthNormalShader.Use();

    DrawModelInstances( models, builtinShaders.depthNormalShader );

    // Depth-pass for all meshes that cannot be drawn using model's combined mesh buffer.
    for (auto& mesh : opaqueMeshQueue)
    {
        if (!mesh->HasSkin() && !mesh->Model()->CombinedOpaqueMeshes().IsGenerated())
        {
            builtinShaders.depthNormalShader.Use();

            mesh->Render( &builtinShaders.depthNormalShader, nullptr );
        }
        else if (mesh->HasSkin())
        {
            builtinShaders.depthNormalSkinShader.Use();

            mesh->Render( &builtinShaders.depthNormalSkinShader, nullptr );
        }
    }

    builtinShaders.depthNormalShader.SetTexture( "depthNormalMap", &builtinTargets.depthNormalTarget, 666 );
    builtinShaders.depthNormalSkinShader.SetTexture( "depthNormalMap", &builtinTargets.depthNormalTarget, 666 );

    GfxDevice::Instance().Set_sRGB_Writes( true );
    GfxDevice::Instance().SetRenderTarget( oldTarget, 0 );

    SetDebugEventEnd();
    ErrorCheck( "DepthNormal pass end." );
}

void RendererImpl::RenderSkybox()
{
    ErrorCheck( "RenderSkybox begin" );
    SetDebugEventBegin( "Render skybox" );

    Matrix44 model;
    Matrix44 modelView;
    Matrix44 modelViewProjection;

    model.Translate( state.cameraPosition );
    Matrix44::Multiply( model, state.view, modelView );
    Matrix44::Multiply( modelView, state.projection, modelViewProjection );

    builtinShaders.skyboxShader.Use();
    builtinShaders.skyboxShader.SetMatrix( "modelViewProjection", modelViewProjection );
    builtinShaders.skyboxShader.SetTexture( "textureMap", skyboxTexture, 0 );
    builtinShaders.skyboxShader.SetConstantBuffers();

    GfxDevice().Instance().SetDepthFunc( DepthFunc::LessOrEqualWriteOff );
    skyboxBuffer.Bind();
    skyboxBuffer.Draw();
    GfxDevice().Instance().SetDepthFunc( DepthFunc::LessOrEqualWriteOn );
    
    builtinShaders.skyboxShader.SetTexture( "textureMap", skyboxTexture, 666 );

    SetDebugEventEnd();
    ErrorCheck( "RenderSkybox end" );
}

void RendererImpl::RenderOpaqueMeshQueueItem( Mesh* mesh )
{
#if AETHER3D_GL43 || AETHER3D_D3D11
    // Renders with point lights, if any.
    mesh->Render( nullptr, nullptr );
    
    if (!mesh->Lights().empty())
    {
        GfxDevice::Instance().SetBlendMode( BlendMode::Additive );
        
        // Renders only directional lights because point lights are rendered earlier in one pass.
        for (auto light : mesh->Lights())
        {
            if (light->GetLightType() == Light::Type::Directional)
            {
                mesh->Render( nullptr, light );
            }
        }
    }
    
    GfxDevice::Instance().SetBlendMode( BlendMode::Off );
#else
    GfxDevice::Instance().SetWireframe( mesh->Model()->IsWireframe() );

    // Ambient pass.
    mesh->Render( nullptr, nullptr );
    
    // Lighting passes.
    GfxDevice::Instance().SetBlendMode( BlendMode::Additive );
    
    for (auto light : mesh->Lights())
    {
        mesh->Render( nullptr, light );
    }
    
    GfxDevice::Instance().SetBlendMode( BlendMode::Off );
    GfxDevice::Instance().SetWireframe( false );
#endif
}

void RendererImpl::ApplySSAO( float fovDegrees )
{
    GfxDevice::Instance().SetRenderTarget( &builtinTargets.ssaoTarget );
    GfxDevice::Instance().ClearScreen();

    builtinShaders.ssaoShader.Use();
    builtinShaders.ssaoShader.SetTexture( "depthNormalMap", &builtinTargets.depthNormalTarget, 0 );
    builtinShaders.ssaoShader.SetTexture( "noiseMap", &noiseTexture, 1 );
    builtinShaders.ssaoShader.SetMatrix( "_ProjectionMatrix", state.projection );
    
    const float tanHalfFov = std::tan( fovDegrees * 0.5f * MathUtil::deg2rad );
    builtinShaders.ssaoShader.SetFloat( "uTanHalfFov", tanHalfFov );
    builtinShaders.ssaoShader.SetFloat( "uAspectRatio", Width() / (float)Height() );
    builtinShaders.ssaoShader.SetConstantBuffers();

    GfxDevice::Instance().Set_sRGB_Writes( false );

    quadBuffer.Bind();
    quadBuffer.Draw();

    builtinShaders.ssaoShader.SetTexture( "depthNormalMap", &builtinTargets.depthNormalTarget, 666 );
    BlurSSAO( builtinTargets.ssaoTarget, builtinTargets.ssaoBlurTarget );
    GfxDevice::Instance().Set_sRGB_Writes( true );
}

void RendererImpl::ApplyToneMap()
{
    SetDebugEventBegin( "Tone map" );

    GfxDevice::Instance().SetRenderTarget( nullptr );
    
    builtinShaders.toneMapShader.Use();
    builtinShaders.toneMapShader.SetTexture( "textureMap", &builtinTargets.hdrTarget, 0 );
    builtinShaders.toneMapShader.SetTexture( "bloomMap", &builtinTargets.bloomBlurVTarget, 1 );
    builtinShaders.toneMapShader.SetTexture( "ssaoMap", &builtinTargets.ssaoBlurTarget, 2 );
    builtinShaders.toneMapShader.SetMatrix( "_ViewProjectionMatrix", Matrix44::identity );
    builtinShaders.toneMapShader.SetMatrix( "modelMatrix", Matrix44::identity );
    builtinShaders.toneMapShader.SetConstantBuffers();

    quadBuffer.Bind();
    quadBuffer.Draw();
    
    builtinShaders.toneMapShader.SetTexture( "textureMap", &builtinTargets.hdrTarget, 666 );
    builtinShaders.toneMapShader.SetTexture( "bloomMap", &builtinTargets.bloomBlurVTarget, 666 );
    builtinShaders.toneMapShader.SetTexture( "ssaoMap", &builtinTargets.ssaoBlurTarget, 666 );
    
    SetDebugEventEnd();
    
    GfxDevice().Instance().SetDepthFunc( DepthFunc::LessOrEqualWriteOn );
}

void RendererImpl::BlurSSAO( TextureImpl& source, TextureImpl& target )
{
    ErrorCheck( "BlurSSAO begin" );

    GfxDevice::Instance().SetRenderTarget( &target, 0 );

    builtinShaders.blurSSAOShader.Use();
    builtinShaders.blurSSAOShader.SetTexture( "textureMap", &source, 0 );
    builtinShaders.blurSSAOShader.SetMatrix( "_ViewProjectionMatrix", Matrix44::identity );
    builtinShaders.blurSSAOShader.SetMatrix( "modelMatrix", Matrix44::identity );
    builtinShaders.blurSSAOShader.SetConstantBuffers();

    quadBuffer.Bind();
    quadBuffer.Draw();

    ErrorCheck( "BlurSSAO end" );
}

void RendererImpl::ApplySharedShaderParams()
{
    // TODO: Make a uniform buffer of these params to avoid setting them to each shader.

    ErrorCheck( "ApplySharedShaderParams begin" );
    
    std::forward_list< Shader* > shadersUsedInFrame;
    
    shadersUsedInFrame.push_front( &builtinShaders.ssaoShader );
    shadersUsedInFrame.push_front( &builtinShaders.depthNormalShader );
    shadersUsedInFrame.push_front( &builtinShaders.depthNormalSkinShader );
    shadersUsedInFrame.push_front( &builtinShaders.toneMapShader );
    shadersUsedInFrame.push_front( &builtinShaders.momentsShader );
    shadersUsedInFrame.push_front( &builtinShaders.momentsSkinShader );
    
    // Finds all the different shaders used in this pass to render meshes.
    for (auto& mesh : opaqueMeshQueue)
    {
        AddMeshShader( mesh, shadersUsedInFrame );
    }
    
    for (auto& mesh : transparentMeshQueue)
    {
        AddMeshShader( mesh, shadersUsedInFrame );
    }
    
    Matrix44::Multiply( state.view, state.projection, state.viewProjection );
    
    // Shimmer reduction.
    /*const float halfSize = 512.0f;
    Vec3 origin;
    Matrix44::TransformPoint( Vec3( 0, 0, 0 ), state.viewProjection, &origin );
    origin *= halfSize;
    
    Vec3 rounding;
    
    const float xRound = MathUtil::Round( origin.x );
    rounding.x = xRound - origin.x;
    
    const float yRound = MathUtil::Round( origin.y );
    rounding.y = yRound - origin.y;

    rounding /= halfSize;
    
    Matrix44 roundMat;
    roundMat.Translate(rounding);
    
    Matrix44::Multiply(state.viewProjection, roundMat, state.viewProjection);
    //Matrix44::Multiply(roundMat, state.viewProjection, state.viewProjection);
// Shimmer reduction end.
    */
    for (auto& shader : shadersUsedInFrame)
    {
        shader->Use();
        shader->SetMatrix( "_ViewMatrix", state.view );
        shader->SetMatrix( "_ViewProjectionMatrix", state.viewProjection );
        shader->SetVector( "_CameraPosition", state.cameraPosition );
#if AETHER3D_GL43
        forwardPlus.SetUniformsForShading( *shader );
#endif
    }
    
    ErrorCheck( "ApplySharedShaderParams end" );
}
