#ifndef RENDERER_IMPL_H
#define RENDERER_IMPL_H

#ifndef INCLUDE_UNORDERED_MAP
#    include <unordered_map>
#    define INCLUDE_UNORDERED_MAP
#endif
#ifndef INCLUDE_LIST
#    include <list>
#    define INCLUDE_LIST
#endif
#ifndef INCLUDE_STRING
#    include <string>
#    define INCLUDE_STRING
#endif
#ifndef FONTER_H
#    include "Fonter.hpp"
#endif
#ifndef RENDERER_H
#    include "Renderer.hpp"
#endif
#ifndef RENDERQUEUE_H
#    include "RenderQueue.hpp"
#endif
#ifndef RENDERER_TYPES_H
#    include "RendererTypes.hpp"
#endif
#ifndef VERTEX_BUFFER_H
#    include "VertexBuffer.hpp"
#endif
#if AETHER3D_GL43 || AETHER3D_D3D11
#    include "ForwardPlus.hpp"
#endif

class CameraImpl;
class Mesh;
class SceneImpl;

/**
  Implements Renderer interface and provides high-level rendering algorithms.
*/
class RendererImpl : public Renderer
{
public:
    static RendererImpl& Instance();
    void AddShaderSearchPath( const char* path ) const override;
    void ClearScreen() const override;
    void Deinit() override;
    void Draw( const Texture* texture, int x, int y, const Vec4& color ) override;
    void Draw( const Texture* texture, int x, int y, int width, int height, const Vec4& color ) override;
    void Draw( const char* text, const char* fontName, float scale, int x, int y, const Vec4& color ) override;
    void Draw( Scene* scene ) override;
    void DrawLines( const Vec3* lines, int lineCount, const Vec4& color ) override;
    void DrawStatistics() override;
    void FinishFrame() override;
    // \return time in seconds.
    double GetTime() const override;
    int Height() const override;
    Error::Enum LoadFont( const char* fontName, const char* fontBitmapPath, const char* fontMetadataPath ) override;
    Error::Enum OpenWindow( int width, int height, bool fullScreen, int antialiasSamples, bool hasWindowAndContext ) override;
    void Resize( int width, int height ) override;
    void SetBloom( bool enable, float factor, int blurIterations ) override;
    void SetClearColor( float red, float green, float blue ) override;
    void SetSSAO( bool enable, float radius, float power ) override;
    void SetWindowTitle( const char* title ) const override;
    void SwapBuffers() override;
    int Width() const override;
    void BeginQuery() override;
    double EndQuery() override;
    bool HasFocus() const override { return hasFocus; }
    int MSAASamples() const { return antialiasSamples; }
    const TextureImpl* DefaultTexture();

    /** \return Asset path without the trailing '/' or '\' */
    std::string BuiltinAssetPath() const { return builtinAssetPath;  }
    void SetBuiltinAssetPath( const std::string& path );
#if AETHER3D_GL43 || AETHER3D_D3D11
    ForwardPlus& GetForwardPlus() { return forwardPlus; }
#endif
    BuiltinShaders& BuiltinShader() { return builtinShaders; }
    BuiltinTargets& BuiltinTarget() { return builtinTargets; }
    void SetFocus(bool enable) { hasFocus = enable; }

private:
    RendererImpl();
    ~RendererImpl();
    void ApplySSAO( float cameraFOV );
    void ApplyToneMap();
    void ApplySharedShaderParams();
    void Blur( TextureImpl& source, TextureImpl& target, bool horizontal );
    void BlurSSAO( TextureImpl& source, TextureImpl& target );
    void CreateTimerQueries();
    void CreateRenderTargets();
    void DrawRenderTextureCameras( SceneImpl* sceneImpl );
    void DrawNormalCameras( SceneImpl* sceneImpl );
    void DrawShadows( SceneImpl* sceneImpl, CameraImpl* eyeCamera );
    Error::Enum LoadBuiltinAssets();
    void RenderDepthAndNormals();
    void RenderSkybox();
    void RenderOpaqueMeshQueueItem( Mesh* mesh );
    void ResolveMSAA();
    void SolveLightListsForTiles();
    void DoShadowCasterPass( std::list< Mesh* >& meshQueue );
    
    VertexBuffer skyboxBuffer;
    VertexBuffer quadBuffer; // Screen-space quad for image effects and HDR.
    BuiltinShaders builtinShaders;
    BuiltinTargets builtinTargets;
    RenderQueue transparentRenderQueue;
    RenderQueue opaqueRenderQueue;
    std::list< Mesh* > transparentMeshQueue;
    std::list< Mesh* > opaqueMeshQueue;
    RendererState state;
    std::string builtinAssetPath;
    std::unordered_map< std::string, Fonter > fonts;
    const TextureImpl* skyboxTexture;
    TextureImpl noiseTexture;
    TextureImpl whiteTexture;
    int screenSize[ 2 ];
    int antialiasSamples = 1;
    int bloomBlurIterations = 1;
    bool bloomEnabled = false;
    bool ssaoEnabled = false;
    bool hasWindowAndContext = true;
#if AETHER3D_GL43 || AETHER3D_D3D11
    ForwardPlus forwardPlus;
#endif
    bool hasFocus = true;
};
#endif
