#include "SceneImpl.hpp"
#include <algorithm>
#include <map>
#include <sstream>
#include <fstream>
#include <vector>
#include "CameraImpl.hpp"
#include "Frustum.hpp"
#include "Log.hpp"
#include "LightImpl.hpp"
#include "MaterialImpl.hpp"
#include "MathUtil.hpp"
#include "Mesh.hpp"
#include "ModelImpl.hpp"
#include "Platform.hpp"
#include "SceneGraphNode.hpp"
#include "TextureImpl.hpp"

#ifdef AETHER3D_WINDOWS
#    define AETHER3D_API __declspec(dllexport)
#else
#    define AETHER3D_API
#endif

extern "C" AETHER3D_API Scene* CreateScene()
{
    SceneImpl* outSceneImpl = new SceneImpl();
    SceneGraphNode::RegisterScene( outSceneImpl );
    return outSceneImpl;
}

std::list< std::pair< Model*, std::list< std::string > > > SortIntersectedModelsAndMeshes( std::list< std::pair< Model*, std::list< Mesh* > > >& intersections, const Vec3& cameraPosition )
{
    auto meshSortFunctionNearToFar = [&]( Mesh* mesh1, Mesh* mesh2 )
    {
        return Vec3::DistanceSquared( mesh1->Model()->GetPosition(), cameraPosition ) <
               Vec3::DistanceSquared( mesh2->Model()->GetPosition(), cameraPosition );
    };

    // Sorts mesh lists.
    for (auto& pair : intersections)
    {
        pair.second.sort( meshSortFunctionNearToFar );
    }

    auto pairSortFunctionNearToFar = [&]( std::pair< Model*, std::list< Mesh* > >& list1, std::pair< Model*, std::list< Mesh* > >& list2 )
    {
        if (list1.second.empty() || list2.second.empty())
        {
            return true;
        }
        // TODO: Replace second... with first
        return Vec3::DistanceSquared( list1.second.front()->Model()->GetPosition(), cameraPosition ) <
               Vec3::DistanceSquared( list2.second.front()->Model()->GetPosition(), cameraPosition );
    };

    // Sorts model list based on mesh list sorting.
    intersections.sort( pairSortFunctionNearToFar );

    std::list< std::pair< Model*, std::list< std::string > > > aabbIntersectionNames;

    // Stores mesh names.
    for (const auto& pair : intersections)
    {
        Model* model = pair.first;
        std::list< std::string > meshNames;

        for (const auto& mesh : pair.second)
        {
            meshNames.push_back( mesh->Name() );
        }

        aabbIntersectionNames.push_front( std::make_pair( model, meshNames ) );
    }
        
    return aabbIntersectionNames;
}

#ifdef SIMD_SSE3
struct BEGIN_ALIGNED( SceneImpl::Impl, 16 )
#else
struct SceneImpl::Impl
#endif
{
    Impl() : activeCamera( nullptr ), skybox( nullptr ) {}
    void AddVisibleMeshes( SceneGraphNode* node, std::list< Mesh* >& outOpaqueMeshes, std::list< Mesh* >& outTransparentMeshes );
    void AddVisibleLights( std::list< Mesh* >& outOpaqueMeshes, std::list< Mesh* >& outTransparentMeshes );
    void TryAddNodeToList( SceneGraphNode* root, SceneGraphNode* node );
    void ScreenPointToRay( int screenX, int screenY, Vec3& outRayOrigin, Vec3& outRayTarget ) const;
    
    SceneGraphNode root;
    CameraImpl* activeCamera;
    const TextureImpl* skybox;
    std::vector< CameraImpl* > cameras;
    std::vector< ModelImpl* > models;
    std::vector< LightImpl* > lights;
    std::vector< SceneGraphNode* > emptyNodes;
    std::map< std::string, MaterialImpl* > materials;
    Vec3 aabbMin, aabbMax;

    std::list< ModelImpl* > modelsLoadedFromFile;

}
#ifdef SIMD_SSE3
END_ALIGNED( SceneImpl::Impl, 16 )
#endif
;

SceneImpl::SceneImpl() : m( new Impl )
{
    m->lights.reserve( 30 );
}

SceneImpl::~SceneImpl()
{
    for (auto& a : m->modelsLoadedFromFile)
    {
        delete a;
    }
}

CameraImpl* SceneImpl::ActiveCamera() const
{
    if (m->activeCamera == nullptr)
    {
        Log::Print( "No active camera" );
        return nullptr;
    }

    return m->activeCamera;
}

void SceneImpl::Add( Camera* camera )
{
    m->root.AddChild( camera );
}

void SceneImpl::Add( Model* model )
{
    m->root.AddChild( model );
}

void SceneImpl::Add( Light* light )
{
    if (light && light->GetLightType() == Light::Type::Point)
    {
        LightImpl* li = static_cast< LightImpl* >( light );
        li->AddPointLightToTileCuller();
    }
    else if (light && light->GetLightType() == Light::Type::Spot)
    {
        LightImpl* li = static_cast< LightImpl* >(light);
        li->AddSpotLightToTileCuller();
    }

    m->root.AddChild( light );
}

void SceneImpl::Remove( SceneGraphNode* node )
{
    SceneGraphNode* n = node;
    
    while (n != nullptr)
    {
        if (n->Parent() && n->Parent() == &m->root)
        {
            node->Parent()->children.remove( node );
            
            if (node->GetType() == SceneGraphNode::Type::Light)
            {
                LightImpl* li = static_cast< LightImpl* >( node );
                
                for (std::size_t i = 0; i < m->lights.size(); ++i)
                {
                    if (li == m->lights[ i ])
                    {
                        m->lights.erase( std::begin( m->lights ) + i );
                        return;
                    }
                }
            }
            else if (node->GetType() == SceneGraphNode::Type::Model)
            {
                ModelImpl* mi = static_cast< ModelImpl* >( node );
                
                for (std::size_t i = 0; i < m->models.size(); ++i)
                {
                    if (mi == m->models[ i ])
                    {
                        m->models.erase( std::begin( m->models ) + i );
                        return;
                    }
                }
            }
            else if (node->GetType() == SceneGraphNode::Type::Camera)
            {
                CameraImpl* ci = static_cast< CameraImpl* >( node );
                
                for (std::size_t i = 0; i < m->cameras.size(); ++i)
                {
                    if (ci == m->cameras[ i ])
                    {
                        m->cameras.erase( std::begin( m->cameras ) + i );
                        return;
                    }
                }
            }
        }
        
        n = n->Parent();
    }
}

Camera* SceneImpl::GetNonRTCamera()
{
    for (auto camera : m->cameras)
    {
        if (camera->GetRenderTarget() == nullptr)
        {
            return camera;
        }
    }
    
    return nullptr;
}

Material* SceneImpl::GetMaterial( const char* name )
{
    if (m->materials.find( name ) != m->materials.end())
    {
        return m->materials[ name ];
    }

    return nullptr;
}

Model* SceneImpl::GetModel( const char* name )
{
    for (std::size_t i = 0; i < m->models.size(); ++i)
    {
        if (m->models[ i ]->GetName() == std::string( name ))
        {
            return m->models[ i ];
        }
    }

    return nullptr;
}

void SceneImpl::Impl::ScreenPointToRay( int screenX, int screenY, Vec3& outRayOrigin, Vec3& outRayTarget ) const
{
    if (!activeCamera)
    {
        outRayOrigin = Vec3( 0, 0, 0 );
        outRayTarget = Vec3( 0, 0, 1 );
        return;
    }

    const float aspect = Screen::Height() / (float)Screen::Width();
    const float halfWidth = (float)(Screen::Width()) * 0.5f;
    const float halfHeight = (float)(Screen::Height()) * 0.5f;
    const float fov = activeCamera->FOV() * (MathUtil::pi / 180.0f);
    
    // Normalizes screen coordinates and scales them to the FOV.
    const float dx = std::tan( fov * 0.5f ) * (screenX / halfWidth - 1.0f) / aspect;
    const float dy = std::tan( fov * 0.5f ) * (screenY / halfHeight - 1.0f);
    
    Matrix44 invView;
    Matrix44::Invert( activeCamera->ViewMatrix(), invView );
    
    const float farp = activeCamera->FarClipPlane();
    
    outRayOrigin = activeCamera->GetPosition();
    outRayTarget = -Vec3( -dx * farp, dy * farp, farp );
    
    Matrix44::TransformPoint( outRayTarget, invView, &outRayTarget );
}

std::list< std::pair< Model*, std::list< std::string > > > SceneImpl::GetColliders( int screenX, int screenY, CollisionTest collisionTest, float maxDepth ) const
{
    if (maxDepth < 0)
    {
        Log::Print( "Scene::GetColliders: maxDepth must be > 0." );
        return std::list< std::pair< Model*, std::list< std::string > > >();
    }

    if (!m->activeCamera)
    {
        return std::list< std::pair< Model*, std::list< std::string > > >();
    }

    Vec3 rayOrigin, rayTarget;
    m->ScreenPointToRay( screenX, screenY, rayOrigin, rayTarget );
    
    std::list< std::pair< Model*, std::list< Mesh* > > > aabbIntersections;
    
    // Collects models and meshes that collide the ray.
    for (auto& o : m->models)
    {
        const Vec3 oMin = o->AABBMinWorld();
        const Vec3 oMax = o->AABBMaxWorld();
        
        const float modelDistance = MathUtil::IntersectRayAABB( rayOrigin, rayTarget, oMin, oMax );

        if (modelDistance < maxDepth)
        {
            auto intersection = std::make_pair( o, std::list< Mesh* >() );
            bool intersectedMeshes = false;
            
            for (auto& mesh : o->Meshes())
            {
                const Vec3 mMin = mesh.AABBMinWorld();
                const Vec3 mMax = mesh.AABBMaxWorld();
                
                const float meshDistance = MathUtil::IntersectRayAABB( rayOrigin, rayTarget, mMin, mMax );
                
                if (meshDistance > 0 && meshDistance < maxDepth)
                {
                    
                    if (collisionTest == CollisionTest::Triangles)
                    {
                        Matrix44 worldToObject;
                        Matrix44::Invert( o->modelMatrix, worldToObject );
                        
                        Vec3 rayOriginModelSpace;
                        Matrix44::TransformPoint( rayOrigin, worldToObject, &rayOriginModelSpace );
                        
                        Vec3 rayTargetModelSpace;
                        Matrix44::TransformPoint( rayTarget, worldToObject, &rayTargetModelSpace );
                        
                        const float meshTriangleDistance = MathUtil::IntersectRayTriangles( rayOriginModelSpace, rayTargetModelSpace,
                                                            mesh.GetVertexPositions(), mesh.Indices() );
                        if (meshTriangleDistance > 0)
                        {
                            intersection.second.push_back( &mesh );
                            intersectedMeshes = true;
                        }
                    }
                    else
                    {
                        intersection.second.push_back( &mesh );
                        intersectedMeshes = true;
                    }
                }
            }

            if (intersectedMeshes)
            {
                aabbIntersections.push_back( intersection );
            }
        }
    }
    
    return SortIntersectedModelsAndMeshes( aabbIntersections, m->activeCamera->GetPosition() );
    //return std::list< std::pair< Model*, std::list< std::string > > >();
}

std::list< std::pair< Model*, std::list< std::string > > > SceneImpl::GetColliders( const Vec3& aabbMin, const Vec3& aabbMax, CollisionTest collisionTest ) const
{
    if (!m->activeCamera)
    {
        return std::list< std::pair< Model*, std::list< std::string > > >();
    }

    std::list< std::pair< Model*, std::list< Mesh* > > > aabbIntersections;
    
    // Collects models and meshes that collide with the AABB
    for (auto& o : m->models)
    {
        Vec3 oMin = o->AABBMinWorld();
        Vec3 oMax = o->AABBMaxWorld();
        
        if (MathUtil::IntersectAABBandAABB( aabbMin, aabbMax, oMin, oMax ))
        {
            aabbIntersections.push_back( std::make_pair( o, std::list< Mesh* >() ) );
            
            for (auto& mesh : o->Meshes())
            {
                Vec3 mMin = mesh.AABBMinWorld();
                Vec3 mMax = mesh.AABBMaxWorld();
                
                if (MathUtil::IntersectAABBandAABB( aabbMin, aabbMax, mMin, mMax ))
                {
                    aabbIntersections.back().second.push_back( &mesh );
                }
            }
        }
    }

    if (collisionTest == CollisionTest::AABB)
    {
        return SortIntersectedModelsAndMeshes( aabbIntersections, m->activeCamera->GetPosition() );
    }
    else
    {
        Log::Assert( false, "Unhandled collision test!" );
    }
    
    return std::list< std::pair< Model*, std::list< std::string > > >();
}

bool SceneImpl::LoadFromFile( const char* path )
{
    std::ifstream ifs( path );
    if (!ifs)
    {
        Log::Print( "Scene::LoadFromFile: Could not open %s", path );
        return false;
    }

    std::stringstream stm;
    std::locale c_locale( "C" );
    stm.imbue( c_locale );

    enum class ReadingState
    {
        Model,
        Material,
        Light,
        Node,
        Camera
    };

    ReadingState readingState = ReadingState::Model;

    std::list< ModelImpl* > models;
    std::list< LightImpl* > lights;
    std::list< CameraImpl* > cameras;
    std::map< std::string, TextureImpl* > textures;
    std::string currentMaterial;
    SceneGraphNode* currentNode = nullptr;

    while (!ifs.eof())
    {
        std::string line;
        getline( ifs, line );
        stm.clear();
        stm.str( line );

        std::string token;
        stm >> token;

        if (token == "model")
        {
            readingState = ReadingState::Model;
            models.push_back( new ModelImpl() );
            std::string name;
            stm >> name;
            models.back()->SetName( name );
        }

        if (token == "node")
        {
            readingState = ReadingState::Node;
            SceneGraphNode* node = new SceneGraphNode();
            Root()->AddChild( node );
            
            std::string name;
            stm >> name;
            node->SetName( name );
            currentNode = node;
        }

        if (token == "camera")
        {
            readingState = ReadingState::Camera;
            cameras.push_back( new CameraImpl() );
            std::string name;
            stm >> name;
            cameras.back()->SetName( name );
        }
        
        if (token == "dirlight")
        {
            readingState = ReadingState::Light;
            lights.push_back( new LightImpl( Light::Type::Directional ) );
            std::string name;
            stm >> name;
            lights.back()->SetName( name );
        }

        if (token == "pointlight")
        {
            readingState = ReadingState::Light;
            lights.push_back( new LightImpl( Light::Type::Point ) );
            std::string name;
            stm >> name;
            lights.back()->SetName( name );
        }

        if (token == "spotlight")
        {
            readingState = ReadingState::Light;
            lights.push_back( new LightImpl( Light::Type::Spot ) );
            std::string name;
            stm >> name;
            lights.back()->SetName( name );
        }

        if (token == "texture")
        {
            std::string name, texPath;
            stm >> name;
            stm >> texPath;
            bool isNormalMap = texPath.find( "_ddn" ) != std::string::npos;
            isNormalMap |= texPath.find( "_n." ) != std::string::npos;
            textures[ name ] = new TextureImpl( texPath.c_str(), true, !isNormalMap, TextureImpl::Tiling::Repeat, TextureImpl::Filter::Linear );
        }

        if (token == "material")
        {
            stm >> currentMaterial;
            readingState = ReadingState::Material;
            m->materials[ currentMaterial ] = new MaterialImpl();
        }

        if (token == "transparent")
        {
            Log::Assert( readingState == ReadingState::Material, "transparent block without material." );
            m->materials[ currentMaterial ]->SetOpaque( false );
        }

        if (token == "param_texture")
        {
            Log::Assert( readingState == ReadingState::Material, "param_texture block without material." );
            std::string paramName, textureName;
            stm >> paramName >> textureName;
            
            if (textures.find( textureName ) == textures.end())
            {
                Log::Print( "Scene::LoadFromFile: Material %s tries to use texture that is not loaded: %s", currentMaterial.c_str(), textureName.c_str() );
            }

            m->materials[ currentMaterial ]->SetTexture( paramName, textures[ textureName ] );
        }

        if (token == "param_vec3")
        {
            Log::Assert( readingState == ReadingState::Material, "param_vec3 block without material." );
            std::string paramName;
            Vec3 vec;
            stm >> paramName >> vec.x >> vec.y >> vec.z;
            m->materials[ currentMaterial ]->SetVector( paramName, vec );
        }

        if (token == "param_vec4")
        {
            Log::Assert( readingState == ReadingState::Material, "param_vec4 block without material." );
            std::string paramName;
            Vec4 vec;
            stm >> paramName >> vec.x >> vec.y >> vec.z >> vec.w;
            m->materials[ currentMaterial ]->SetVector( paramName, vec );
        }

        if (token == "param_float")
        {
            Log::Assert( readingState == ReadingState::Material, "param_float block without material." );
            std::string paramName;
            float f;
            stm >> paramName >> f;
            m->materials[ currentMaterial ]->SetFloat( paramName, f );
        }

        if (token == "all_material")
        {
            Log::Assert( readingState == ReadingState::Model, "all_material block without model." );
            std::string material;
            stm >> material;
            models.back()->SetMaterialForMeshSubstring( m->materials[ material ], "" );
        }

        if (token == "mesh_material")
        {
            Log::Assert( readingState == ReadingState::Model, "mesh_material block without model." );

            std::string mesh;
            stm >> mesh;

            std::string material;
            stm >> material;
            models.back()->SetMaterialForMesh( m->materials[ material ], mesh );
        }

        if (token == "mesh_material_path")
        {
            Log::Assert( readingState == ReadingState::Model, "mesh_material block without model." );
            
            std::string mesh;
            stm >> mesh;
            
            std::string materialPath;
            stm >> materialPath;
            MaterialImpl* material = new MaterialImpl();
            material->LoadFromFile( materialPath.c_str() );
            m->materials[ material->Name() ] = material;
            models.back()->SetMaterialForMesh( m->materials[ material->Name() ], mesh );
        }

        if (token == "shaders")
        {
            Log::Assert( readingState == ReadingState::Material, "shaders block without material." );
            std::string vertex, fragment;
            stm >> vertex >> fragment;
            
            std::string defines;
            
            while (!stm.eof())
            {
                std::string define;
                stm >> define;
                
                if (defines.empty())
                {
                    defines = define;
                }
                else
                {
                    defines += " " + define;
                }
            }

            m->materials[ currentMaterial ]->SetShaders( vertex, fragment, "", defines );
        }

        if (token == "path")
        {
            std::string resourcePath;
            stm >> resourcePath;

            if (readingState == ReadingState::Model)
            {
                Error::Enum error = models.back()->LoadFromFile( resourcePath );

                if (error != Error::NO_ERRORS)
                {
                    Log::Print( "Could not load model %s", resourcePath.c_str() );
                }
            }
        }

        if (token == "position")
        {
            Vec3 position;
            stm >> position.x >> position.y >> position.z;

            if (readingState == ReadingState::Camera)
            {
                cameras.back()->SetPosition( position );
            }
            if (readingState == ReadingState::Model)
            {
                models.back()->SetPosition( position );
            }
            else if (readingState == ReadingState::Light)
            {
                lights.back()->SetPosition( position );
            }
            else if (readingState == ReadingState::Node)
            {
                Log::Assert( currentNode != nullptr, "Reading node, but node is null!" );
                currentNode->SetPosition( position );
            }
        }

        if (token == "axisangle")
        {
            Vec3 axis;
            stm >> axis.x >> axis.y >> axis.z;
            float angle;
            stm >> angle;

            Quaternion rotation;
            rotation.FromAxisAngle( axis, angle );

            if (readingState == ReadingState::Model)
            {
                models.back()->SetRotation( rotation );
            }
            else if (readingState == ReadingState::Light)
            {
                lights.back()->SetRotation( rotation );
            }
            else if (readingState == ReadingState::Node)
            {
                Log::Assert( currentNode != nullptr, "Reading node, but node is null!" );
                currentNode->SetRotation( rotation );
            }
        }

        if (token == "projection")
        {
            float fov, aspect, near, far;
            stm >> fov >> aspect >> near >> far;
            
            if (readingState == ReadingState::Camera)
            {
                cameras.back()->SetProjection( fov, aspect, near, far );
            }
        }

        if (token == "color")
        {
            Vec3 color;
            stm >> color.x >> color.y >> color.z;

            if (readingState == ReadingState::Light)
            {
                lights.back()->SetColor( color );
            }
        }

        if (token == "scale")
        {
            Log::Assert( readingState == ReadingState::Model || readingState == ReadingState::Node, "scale block without model or node." );

            float scale = 1;
            stm >> scale;
            
            if (readingState == ReadingState::Model)
            {
                models.back()->SetScale( scale );
            }
            else if (readingState == ReadingState::Node)
            {
                currentNode->SetScale( scale );
            }
        }

        if (token == "radius")
        {
            float radius = 1;
            stm >> radius;
            
            lights.back()->SetRadius( radius );
        }

        if (token == "direction")
        {
            Vec3 direction;
            stm >> direction.x >> direction.y >> direction.z;

            if (readingState == ReadingState::Light)
            {
                lights.back()->SetDirection( direction );
            }
        }
    }

    for (auto& model : models)
    {
        Add( model );
    }

    for (auto& light : lights)
    {
        Add( light );
    }

    for (auto& camera : cameras)
    {
        Add( camera );
    }

    m->modelsLoadedFromFile = models;
    return true;
}

bool SceneImpl::SaveToFile( const char* path )
{
    std::ofstream ofs( path );
    if (!ofs)
    {
        Log::Print( "Scene::SaveToFile: Could not open %s", path );
        return false;
    }

    std::locale c_locale( "C" );
    ofs.imbue( c_locale );

    for (const auto& node : m->emptyNodes)
    {
        Vec3 axis;
        float angleRad;
        node->GetRotation().GetAxisAngle( axis, angleRad );
        
        ofs << "node " << node->GetName() << std::endl;
        ofs << "position " << node->GetPosition().x << " " << node->GetPosition().y << " " << node->GetPosition().z << std::endl;
        ofs << "axisangle " << axis.x << " " << axis.y << " " << axis.z << " " << angleRad << std::endl;
        ofs << "scale " << node->GetScale() << std::endl << std::endl;
    }
    
    for (const auto& model : m->models)
    {
        Vec3 axis;
        float angleRad;
        model->GetRotation().GetAxisAngle( axis, angleRad );

        ofs << "model " << model->GetName() << std::endl;
        ofs << "path " << model->GetPath() << std::endl;
        ofs << "position " << model->GetPosition().x << " " << model->GetPosition().y << " " << model->GetPosition().z << std::endl;
        ofs << "axisangle " << axis.x << " " << axis.y << " " << axis.z << " " << angleRad << std::endl;
        ofs << "scale " << model->GetScale() << std::endl;
        
        for (auto mesh : model->Meshes())
        {
            if (mesh.Material() && !std::string( mesh.Material()->Path() ).empty())
            {
                ofs << "mesh_material_path " << mesh.Name() << " " << mesh.Material()->Path() << std::endl;
            }
        }
        
        ofs << std::endl;
    }

    for (const auto& light : m->lights)
    {
        if (light->GetLightType() == Light::Type::Point)
        {
            ofs << "pointlight " << light->GetName() << std::endl;
            ofs << "position " << light->GetPosition().x << " " << light->GetPosition().y << " " << light->GetPosition().z << std::endl;
        }
        else if (light->GetLightType() == Light::Type::Directional)
        {
            ofs << "dirlight " << light->GetName() << std::endl;
            ofs << "direction " << light->Direction().x << " " << light->Direction().y << " " << light->Direction().z << std::endl;
        }
        else if (light->GetLightType() == Light::Type::Spot)
        {
            ofs << "spotlight " << light->GetName() << std::endl;
            ofs << "direction " << light->Direction().x << " " << light->Direction().y << " " << light->Direction().z << std::endl;
            ofs << "position " << light->GetPosition().x << " " << light->GetPosition().y << " " << light->GetPosition().z << std::endl;
        }

        ofs << "color " << light->GetColor().x << " " << light->GetColor().y << " " << light->GetColor().z << std::endl;
        
        Vec3 axis;
        float angleRad;
        light->GetRotation().GetAxisAngle( axis, angleRad );
        ofs << "axisangle " << axis.x << " " << axis.y << " " << axis.z << " " << angleRad << std::endl;
        ofs << std::endl;
    }

    for (const auto& camera : m->cameras)
    {
        ofs << "camera " << camera->GetName() << std::endl;
        ofs << "position " << camera->GetPosition().x << " " << camera->GetPosition().y << " " << camera->GetPosition().z << std::endl;
        ofs << "projection " << camera->FOV() << " " << camera->Aspect() << " " << camera->NearClipPlane() << " " << camera->FarClipPlane() << std::endl;
        ofs << std::endl;
    }

    return true;
}

SceneGraphNode& SceneImpl::RootNode()
{
    return m->root;
}

void SceneImpl::SetActiveCamera( Camera* camera )
{
    if (!camera)
    {
        Log::Print( "SetActiveCamera: Tried to set a null camera." );
        return;
    }

    m->activeCamera = (CameraImpl*)camera;
}

void SceneImpl::Impl::AddVisibleLights( std::list< Mesh* >& outOpaqueMeshes, std::list< Mesh* >& outTransparentMeshes )
{
    for (auto meshList : { outOpaqueMeshes, outTransparentMeshes })
    {
        for (auto mesh : meshList)
        {
            mesh->ClearLights();

            for (const auto& li : lights)
            {
                bool pointLightVisible = true;

                // D3D11 and GL43 cull point lights on compute shader so no need to cull here.
#if AETHER3D_GL32 || AETHER3D_GLES3
                if (li->GetLightType() == Light::Type::Point)
                {
                    pointLightVisible = MathUtil::IntersectSphereAndAABB( li->GetPosition(), li->Radius(),
                                                                          mesh->AABBMinWorld(), mesh->AABBMaxWorld() );
                }
#endif
                if (pointLightVisible)
                {
                    mesh->AddLight( li );
                }
            }
        }
    }
}
    
void SceneImpl::Impl::AddVisibleMeshes( SceneGraphNode* node, std::list< Mesh* >& outOpaqueMeshes, std::list< Mesh* >& outTransparentMeshes )
{
    Log::Assert( node != nullptr, "Node is null." );
    Log::Assert( activeCamera != nullptr, "Scene has no active camera." );

    if (node->GetType() == SceneGraphNode::Type::Model && (node->GetLayerMask() & activeCamera->GetLayerMask()))
    {
        ModelImpl* modelImpl = (ModelImpl*) node;

        if (activeCamera->GetFrustum().BoxInFrustum( modelImpl->AABBMinWorld(), modelImpl->AABBMaxWorld() ))
        {
            // If the model only has one mesh, skip mesh's AABB test because it's the same as model's.
            if (modelImpl->Meshes().size() == 1)
            {
                Mesh& mesh = modelImpl->Meshes()[ 0 ];

                mesh.SetModelMatrix( modelImpl->ModelMatrix() );

                // FIXME: Should be able to set only once when model is created, but 
                //        does not work for some reason for cloned models, so have to set here.
                mesh.SetModel( modelImpl );
                
                (mesh.Material()->Opaque() ? outOpaqueMeshes : outTransparentMeshes).push_front( &mesh );
            }
            else
            {
                for (auto& mesh : modelImpl->Meshes())
                {
                    if (!mesh.IsMorphTarget() && activeCamera->GetFrustum().BoxInFrustum( mesh.AABBMinWorld(), mesh.AABBMaxWorld() ))
                    {
                        mesh.SetModelMatrix( modelImpl->ModelMatrix() );

                        // FIXME: Should be able to set only once when model is created, but 
                        //        does not work for some reason for cloned models, so have to set here.
                        mesh.SetModel( modelImpl );

                        (mesh.Material()->Opaque() ? outOpaqueMeshes : outTransparentMeshes).push_front( &mesh );
                    }
                }
            }
        }
    }

    for (auto& child : node->Children())
    {
        AddVisibleMeshes( child, outOpaqueMeshes, outTransparentMeshes );
    }
}

void SceneImpl::Cull( std::list< Mesh* >& outOpaqueMeshes, std::list< Mesh* >& outTransparentMeshes )
{
    m->AddVisibleMeshes( &m->root, outOpaqueMeshes, outTransparentMeshes );
    m->AddVisibleLights( outOpaqueMeshes, outTransparentMeshes );
}

void SceneImpl::OnNodeAdded( SceneGraphNode* node )
{
    m->TryAddNodeToList( node, node );
    GenerateAABB();
}

void SceneImpl::SetSkybox( const Texture* cubeMap )
{
    if (!cubeMap)
    {
        Log::Print( "SetSkybox: cubeMap is null." );
        return;
    }
    
    if (!((const TextureImpl*)cubeMap)->IsCubemap())
    {
        Log::Print( "SetSkybox: cubeMap is not a cube map." );
        return;
    }

    m->skybox = (const TextureImpl*)cubeMap;
}

SceneGraphNode* SceneImpl::Root()
{
    return &m->root;
}

void SceneImpl::GetCameras( SceneGraphNode* node, bool renderTextureCamera, std::list< CameraImpl* >& outCameras )
{
    if (node->GetType() == SceneGraphNode::Type::Camera)
    {
        if ( (renderTextureCamera && ((CameraImpl*)node)->GetRenderTarget()) ||
             (!renderTextureCamera && !((CameraImpl*)node)->GetRenderTarget()))
            
        outCameras.push_front( (CameraImpl*)node );
    }
    
    for (auto& child : node->Children())
    {
        GetCameras( child, renderTextureCamera, outCameras );
    }
}

void SceneImpl::GenerateAABB()
{
    const float maxValue = 99999999.0f;
    m->aabbMin = Vec3(  maxValue,  maxValue,  maxValue );
    m->aabbMax = Vec3( -maxValue, -maxValue, -maxValue );
    
    for (auto& o : m->models)
    {
        const Vec3 oAABBmin = o->AABBMinWorld();
        const Vec3 oAABBmax = o->AABBMaxWorld();
        
        if (oAABBmin.x < m->aabbMin.x)
        {
            m->aabbMin.x = oAABBmin.x;
        }
        
        if (oAABBmin.y < m->aabbMin.y)
        {
            m->aabbMin.y = oAABBmin.y;
        }
        
        if (oAABBmin.z < m->aabbMin.z)
        {
            m->aabbMin.z = oAABBmin.z;
        }
        
        if (oAABBmax.x > m->aabbMax.x)
        {
            m->aabbMax.x = oAABBmax.x;
        }
        
        if (oAABBmax.y > m->aabbMax.y)
        {
            m->aabbMax.y = oAABBmax.y;
        }
        
        if (oAABBmax.z > m->aabbMax.z)
        {
            m->aabbMax.z = oAABBmax.z;
        }
    }
}

void SceneImpl::Impl::TryAddNodeToList( SceneGraphNode* nodeRoot, SceneGraphNode* node )
{
    if (node->GetType() == SceneGraphNode::Type::Model)
    {
        bool found = false;
        
        for (ModelImpl* mi : models)
        {
            if (mi == (ModelImpl*)node)
            {
                found = true;
            }
        }
        
        if (!found)
        {
            models.push_back( (ModelImpl*) node);
        }
    }
    else if (node->GetType() == SceneGraphNode::Type::Camera)
    {
        bool found = false;
        
        for (CameraImpl* ci : cameras)
        {
            if (ci == (CameraImpl*)node)
            {
                found = true;
            }
        }
        
        if (!found)
        {
            cameras.push_back( (CameraImpl*) node);
        }
    }
    else if (node->GetType() == SceneGraphNode::Type::Light)
    {
        bool found = false;
        
        for (LightImpl* li : lights)
        {
            if (li == (LightImpl*)node)
            {
                found = true;
            }
        }
        
        if (!found)
        {
            LightImpl* li = static_cast< LightImpl* >( node );
            lights.push_back( li );
            
            if (li->GetLightType() == Light::Type::Point)
            {
                li->AddPointLightToTileCuller();
            }
            else if (li->GetLightType() == Light::Type::Spot)
            {
                li->AddSpotLightToTileCuller();
            }
        }
    }
    else if (node->GetType() == SceneGraphNode::Type::None)
    {
        bool found = false;
        
        for (SceneGraphNode* e : emptyNodes)
        {
            if (e == node)
            {
                found = true;
            }
        }
        
        if (!found)
        {
            emptyNodes.push_back( node );
        }
    }
    
    for (auto child : node->children)
    {
        TryAddNodeToList( nodeRoot, child );
    }
}

std::vector< LightImpl* >& SceneImpl::GetLights()
{
    return m->lights;
}

const TextureImpl* SceneImpl::Skybox() const
{
    return m->skybox;
}

void SceneImpl::SolveShadowFrustum( const LightImpl* light, const CameraImpl* eyeCamera, CameraImpl& shadowCamera )
{
    if (light->GetLightType() == Light::Type::Directional)
    {
        const Vec3 viewFrustumCentroid = eyeCamera->GetFrustum().Centroid();
        
        // Start at the centroid, and move back in the opposite direction of the light
        // by an amount equal to the camera's farClip. This is the temporary working position for the light.
        const Vec3 shadowCameraPosition = viewFrustumCentroid - light->Direction() * eyeCamera->GetFrustum().FarClipPlane();

        shadowCamera.LookAt( shadowCameraPosition, viewFrustumCentroid, Vec3( 0, 1, 0 ) );
        
        const Matrix44& shadowCameraView = shadowCamera.ViewMatrix();

        // Transforms view camera frustum points to shadow camera space.
        std::vector< Vec3 > viewFrustumLS( 8 );
            
        const Frustum& viewFrustum = eyeCamera->GetFrustum();
            
        Matrix44::TransformPoint( viewFrustum.NearTopLeft(), shadowCameraView, &viewFrustumLS[ 0 ] );
        Matrix44::TransformPoint( viewFrustum.NearTopRight(), shadowCameraView, &viewFrustumLS[ 1 ] );
        Matrix44::TransformPoint( viewFrustum.NearBottomLeft(), shadowCameraView, &viewFrustumLS[ 2 ] );
        Matrix44::TransformPoint( viewFrustum.NearBottomRight(), shadowCameraView, &viewFrustumLS[ 3 ] );
            
        Matrix44::TransformPoint( viewFrustum.FarTopLeft(), shadowCameraView, &viewFrustumLS[ 4 ] );
        Matrix44::TransformPoint( viewFrustum.FarTopRight(), shadowCameraView, &viewFrustumLS[ 5 ] );
        Matrix44::TransformPoint( viewFrustum.FarBottomLeft(), shadowCameraView, &viewFrustumLS[ 6 ] );
        Matrix44::TransformPoint( viewFrustum.FarBottomRight(), shadowCameraView, &viewFrustumLS[ 7 ] );
            
        // Gets light-space view frustum extremes.
        Vec3 viewMinLS, viewMaxLS;
        MathUtil::GetMinMax( viewFrustumLS, viewMinLS, viewMaxLS );
            
        // Transforms scene's AABB to light-space.
        Vec3 sceneAABBminLS, sceneAABBmaxLS;
        const Vec3 min = m->aabbMin;
        const Vec3 max = m->aabbMax;
            
        const Vec3 sceneCorners[] =
        {
            Vec3( min.x, min.y, min.z ),
            Vec3( max.x, min.y, min.z ),
            Vec3( min.x, max.y, min.z ),
            Vec3( min.x, min.y, max.z ),
            Vec3( max.x, max.y, min.z ),
            Vec3( min.x, max.y, max.z ),
            Vec3( max.x, max.y, max.z ),
            Vec3( max.x, min.y, max.z )
        };
            
        std::vector< Vec3 > sceneAABBLS( 8 );
        
        for (int i = 0; i < 8; ++i)
        {
            Matrix44::TransformPoint( sceneCorners[ i ], shadowCameraView, &sceneAABBLS[ i ] );
        }
            
        MathUtil::GetMinMax( sceneAABBLS, sceneAABBminLS, sceneAABBmaxLS );
            
        // Use world volume for near plane.
        viewMaxLS.z = sceneAABBmaxLS.z > viewMaxLS.z ? sceneAABBmaxLS.z : viewMaxLS.z;
        
        // Shimmering reduction. Doesn't seem to work.
        // Source: https://msdn.microsoft.com/en-us/library/windows/desktop/ee416324(v=vs.85).aspx
        // Possible solution: http://www.gamedev.net/topic/497259-stable-cascaded-shadow-maps/
        const float texelSize = 1.0f / (float)shadowCamera.GetRenderTarget()->Width();
        
        viewMinLS /= texelSize;
        viewMinLS = Vec3( MathUtil::Floor(viewMinLS.x), MathUtil::Floor(viewMinLS.y), MathUtil::Floor(viewMinLS.z) );
        viewMinLS *= texelSize;

        viewMaxLS /= texelSize;
        viewMaxLS = Vec3( MathUtil::Floor(viewMaxLS.x), MathUtil::Floor(viewMaxLS.y), MathUtil::Floor(viewMaxLS.z) );
        viewMaxLS *= texelSize;
        
        shadowCamera.SetProjection( viewMinLS.x, viewMaxLS.x, viewMinLS.y, viewMaxLS.y, -viewMaxLS.z, -viewMinLS.z );
    }
    else if (light->GetLightType() == Light::Type::Spot)
    {
        shadowCamera.LookAt( light->worldPosition, light->worldPosition + light->Direction() * 200, Vec3( 0, 1, 0 ) );
        shadowCamera.SetProjection( 45, 1, 1, 200 );
    }
    else
    {
        Log::Assert( false, "SolveShadowFrustum not implemented for point lights." );
    }
}

void SceneImpl::TransformLights()
{
    for (auto& light : m->lights)
    {
        if (light->IsDirty())
        {
            light->UpdateTransform();
        }
    }
}

void SceneImpl::UpdateAABBs()
{
    for (auto& model : m->models)
    {
        if (model->IsDirty())
        {
            model->SolveWorldAABB();
        }
    }
    
    // TODO: Generate only if something is dirty.
    GenerateAABB();
}
