#ifndef SCENE_IMPL_H
#define SCENE_IMPL_H

#ifndef INCLUDE_LIST_H
#    include <list>
#    define INCLUDE_LIST_H
#endif
#ifndef INCLUDE_MEMORY_H
#    include <memory>
#    define INCLUDE_MEMORY_H
#endif
#ifndef INCLUDE_VECTOR_H
#    include <vector>
#    define INCLUDE_VECTOR_H
#endif
#ifndef SCENE_H
#    include "Scene.hpp"
#endif

class CameraImpl;
class Camera;
class Light;
class LightImpl;
class Material;
class Mesh;
class Model;
class Texture;
class TextureImpl;
class SceneGraphNode;

class SceneImpl : public Scene
{
  public:
    SceneImpl();
    virtual ~SceneImpl();
    void Add( Camera* camera ) override;
    void Add( Light* light ) override;
    void Add( Model* model ) override;
    void Remove( SceneGraphNode* node ) override;
    Camera* GetNonRTCamera() override;
    Material* GetMaterial( const char* name ) override;
    Model* GetModel( const char* name ) override;
    std::list< std::pair< Model*, std::list< std::string > > > GetColliders( int screenX, int screenY, CollisionTest collisionTest, float maxDepth ) const override;
    std::list< std::pair< Model*, std::list< std::string > > > GetColliders( const Vec3& aabbMin, const Vec3& aabbMax, CollisionTest collisionTest ) const override;
    bool LoadFromFile( const char* path ) override;
    bool SaveToFile( const char* path ) override;
    virtual SceneGraphNode& RootNode() override;
    void SetSkybox( const Texture* cubeMap ) override;
    void OnNodeAdded( SceneGraphNode* node ) override;

    SceneGraphNode* Root();
    
    /**
       Finds all the cameras from the scene graph.
     
       \param node Node, should start from the root.
       \param renderTextureCamera If true, returns render-to-texture cameras.
       \return If renderTextureCamera is true, returns render texture Cameras, else returns normal cameras.
     */
    void GetCameras( SceneGraphNode* node, bool renderTextureCamera, std::list< CameraImpl* >& outCameras );

    /** \param outLights Returns lights. */
    std::vector< LightImpl* >& GetLights();
    
    // \return skybox.
    const TextureImpl* Skybox() const;

    void SetActiveCamera( Camera* camera );

    /// \return Active camera.
    CameraImpl* ActiveCamera() const;

    /**
       Returns visible opaque and transparent meshes and matrices.  Requires active camera.
     
       \param outOpaqueMeshes List of visible opaque meshes.
       \param outTransparentMeshes List of visible transparent meshes.
     */
    void Cull( std::list< Mesh* >& outOpaqueMeshes, std::list< Mesh* >& outTransparentMeshes );

    /** Generates an Axis-Aligned Bounding Box from objects' bounding boxes. */
    void GenerateAABB();

    /** Solves shadow frustum. Shadow camera should be active when calling this method. */
    void SolveShadowFrustum( const LightImpl* light, const CameraImpl* eyeCamera, CameraImpl& shadowCamera );

    /** Applies scene graph transfrom hierarchy to lights. */
    void TransformLights();

    /** Updates world-space AABBs for models that have dirty parents. */
    void UpdateAABBs();

private:
    struct Impl;
    std::unique_ptr< Impl > m;
};

#endif
