#ifndef SHADER_H
#define SHADER_H

#ifndef INCLUDE_MEMORY_H
#    include <memory>
#    define INCLUDE_MEMORY_H
#endif
#ifndef INCLUDE_STRING_H
#    include <string>
#    define INCLUDE_STRING_H
#endif
#ifndef ERRORS_H
#    include "Errors.hpp"
#endif

class TextureImpl;
struct Vec3;
struct Vec4;
struct Matrix44;

/** Shader used by Material. */
class Shader
{
public:
    /// Input Channel.
    enum Channel
    {
        POSITION    = 0,
        TEXCOORD    = 1,
        NORMAL      = 2,
        TANGENT     = 3,
        BONE_INDEX  = 4,
        BONE_WEIGHT = 5,
        MORPH_POS1  = 6,
        MORPH_POS2  = 7,
        COLOR       = 8,
        MODEL_MATRIX_INSTANCE = 9
    };

    static Error::Enum Reload( const std::string& vertexShaderPath,
                               const std::string& fragmentShaderPath,
                               const std::string& geometryShaderPath,
                               const std::string& defines );
    Shader();
    Shader( const Shader& other );
    virtual ~Shader();
    Shader& operator=( const Shader& other );
    static void Deinit();
    bool Equals( const Shader& other ) const;
    unsigned GLSLProgram() const;
    void Load( const std::string& vertexShaderPath,
                      const std::string& fragmentShaderPath,
                      const std::string& geometryShaderPath,
                      const std::string& defines, Error::Enum& outError );
    void SetFloat( const std::string& name, float value );
    void SetInt( const std::string& name, int value );
    void SetUInt( const std::string& name, unsigned value );
    void SetMatrix( const std::string& name, const Matrix44& matrix );
    void SetMatrices( const std::string& name, const Matrix44* matrices, int count );
    /**
       Sets a texture.  Its uv scale and offset should be present in the shader and is set automatically.
       If the texture name is "textureMap", its scale and offset are in 4-dimensional vector "textureMap_ST".
     
       \param name Name of the texture in the shader source.
       \param texture Texture.
       \param textureUnit Texture unit.  Not used in D3D11 as it queries the unit from source. Massive hack (D3D11): Set to 666 to unbind a texture.
    */
    void SetTexture( const std::string& name, const TextureImpl* texture, int textureUnit );
    void SetVector( const std::string& name, const Vec3& value );
    void SetVector( const std::string& name, const Vec4& value );
    void SetVectors( const std::string& name, const Vec3* values, int count );
    void SetVectors( const std::string& name, const Vec4* values, int count );
    /// Currently only used in D3D11 renderer. Uploads the variables.
    void SetConstantBuffers();
    void Use();
    /**
       Unbinds a texture.

       \param name Name of the texture in the shader code.
    */
    void UnbindTexture( const char* name );
    /** Unbinds textures.  Needed because render textures cannot be both rendered into and read from at the same time. */
    void UnbindTextures();

    /// \return Path to vertex shader source.
    const std::string& VertexPath() const;
    /// \return Path to fragment shader source.
    const std::string& FragmentPath() const;
    /// \return Path to geometry shader source.
    const std::string& GeometryPath() const;

    static unsigned ActiveGLProgram();
    static void Validate();

private:
    struct Impl;
    std::unique_ptr< Impl > m;
};
#endif
