#include "TextureImpl.hpp"
#include <fstream>
#include <map>
#include <vector>
#define STB_IMAGE_IMPLEMENTATION
#undef _CRT_SECURE_NO_WARNINGS
#include "stb_image.c"
#include "FileWatcher.hpp"
#include "Log.hpp"

#ifdef AETHER3D_WINDOWS
#    define AETHER3D_API __declspec(dllexport)
#else
#    define AETHER3D_API
#endif

// TODO constructor delegation for DRY.

#if AETHER3D_IOS
unsigned char* LoadImageCG( const std::string& outFileName, int& outWidth, int& outHeight, bool& outHasAlpha );
#endif

std::string PathWithCorrectSeparators( const std::string& path );

const char* GetFullPath( const char* fileName );

extern "C" AETHER3D_API Texture* CreateTexture( const char* path, bool useMipmaps, bool sRGB, Texture::Tiling tiling, Texture::Filter filter )
{
    return new TextureImpl( path, useMipmaps, sRGB, tiling, filter );
}

extern "C" AETHER3D_API Texture* CreateTextureFromAtlas( const char* atlasPath,
                                                         const char* atlasMetadataPath,
                                                         const char* fileName )
{
    return new TextureImpl( atlasPath, atlasMetadataPath, fileName, true );
}

extern "C" AETHER3D_API Texture* CreateRenderTexture( int width, int height )
{
    if (width <= 0)
    {
        Log::Print( "MakeRenderTexture: Invalid width %d, defaulting to 256.", width );
        width = 256;
    }

    if (height <= 0)
    {
        Log::Print( "MakeRenderTexture: Invalid height %d, defaulting to 256.", height );
        height = 256;
    }

    TextureImpl* texture = new TextureImpl();
    texture->CreateRenderTexture2D( width, height, TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Linear, TextureImpl::Tiling::Repeat, "app-created rt" );
    return static_cast< Texture* >( texture );
}

extern "C" AETHER3D_API Texture* CreateRenderTextureCube( int size )
{
    TextureImpl* texture = new TextureImpl();
    texture->CreateRenderTextureCube( size );
    return static_cast< Texture* >( texture );
}

extern "C" AETHER3D_API Texture* CreateTextureCube( const char* left,  const char* right,
                                                    const char* top,   const char* bottom,
                                                    const char* front, const char* back )
{
    return new TextureImpl( left, right, top, bottom, front, back );
}

namespace Global
{
    std::map< std::string, TextureImpl > pathToCachedTexture;
}

/// \todo Create StringUtil.hpp when more string methods are needed.
void Tokenize(const std::string& str,
              std::vector< std::string >& tokens,
              const std::string& delimiters = " ")
{
    // Skip delimiters at beginning.
    std::string::size_type lastPos = str.find_first_not_of( delimiters, 0 );
    // Find first "non-delimiter".
    std::string::size_type pos = str.find_first_of( delimiters, lastPos );

    while (std::string::npos != pos || std::string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back( str.substr( lastPos, pos - lastPos ) );
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of( delimiters, pos );
        // Find next "non-delimiter"
        pos = str.find_first_of( delimiters, lastPos );
    }
}

/**
 Loads texture data.

 \param path Texture file name.
 \param outWidth Width in pixels is returned here.
 \param outHeight Height in pixels is returned here.
 \param outComponents # of color components is returned here.
 \return Image data. Must be released by caller by calling FreeTextureData().
 */
unsigned char* LoadTextureData( const std::string& path, int& outWidth, int& outHeight, int& outComponents )
{
#if AETHER3D_IOS
    bool hasAlpha;
    unsigned char* data = LoadImageCG( path, outWidth, outHeight, hasAlpha );
    outComponents = hasAlpha ? 4 : 3;
    return data;
#else
    unsigned char* data = stbi_load( path.c_str(), &outWidth, &outHeight, &outComponents, 4 );

    if (!data)
    {
        const std::string reason( stbi_failure_reason() );
        Log::Print( "%s failed to load. stb_image's reason: %s", path.c_str(), reason.c_str() );
    }

    return data;
#endif
}

void FreeTextureData( unsigned char* data )
{
    stbi_image_free( data );
}

void TextureImpl::Reload( const std::string& path )
{
    if (Global::pathToCachedTexture.find( path ) == Global::pathToCachedTexture.end())
    {
        Log::Assert( false, "Tried to reload a texture that wasn't loaded." );
    }

    const bool cachedTextureHasMipmaps = Global::pathToCachedTexture[ path ].HasMipmaps();
    const bool cachedIssRGB = Global::pathToCachedTexture[ path ].IsSRGB();
    Global::pathToCachedTexture[ path ].Load( path, cachedTextureHasMipmaps, cachedIssRGB );
}

TextureImpl::TextureImpl()
  : scaleOffset( 1, 1, 0, 0 )
  , opaque( true )
  , sRGB( true )
  , hasMipmaps( false )
  , cubeMap( false )
  {
    sampler.tiling = Tiling::Repeat;
    sampler.filter = Filter::Linear;
  }

TextureImpl::TextureImpl( const char* aPath, bool useMipmaps, bool is_sRGB, Tiling tiling, Filter filter )
  : scaleOffset( 1, 1, 0, 0 )
  , opaque( true )
  , sRGB( is_sRGB )
  , hasMipmaps( useMipmaps )
  , cubeMap( false )
{
    sampler.filter = filter;
    sampler.tiling = tiling;

    const std::string fPath = PathWithCorrectSeparators( aPath );

    if (Global::pathToCachedTexture.find( GetFullPath( fPath.c_str() ) ) != Global::pathToCachedTexture.end())
    {
        *this = Global::pathToCachedTexture[ GetFullPath( fPath.c_str() ) ];
        return;
    }

    const bool loadedSuccessfully = Load( GetFullPath( fPath.c_str() ), useMipmaps, sRGB );

    if (loadedSuccessfully)
    {
        FileWatcher::Instance().AddTexture( GetFullPath( fPath.c_str() ) );
    }

    Global::pathToCachedTexture[ GetFullPath( fPath.c_str() ) ] = *this;
}

TextureImpl::TextureImpl( const std::string& atlasPath,
                          const std::string& atlasMetadataPath,
                          const std::string& fileName,
                          bool is_sRGB )
  : scaleOffset( 1, 1, 0, 0 )
  , opaque( true )
  , sRGB( is_sRGB )
  , hasMipmaps( false )
  , cubeMap( false )
{
    sampler.tiling = Tiling::Clamp;
    sampler.filter = Filter::Linear;

    if (Global::pathToCachedTexture.find( atlasPath ) != Global::pathToCachedTexture.end())
    {
        *this = Global::pathToCachedTexture[ atlasPath ];
    }
    else
    {
        const bool loadedSuccessfully = Load( atlasPath, hasMipmaps, sRGB );

        if (loadedSuccessfully)
        {
            FileWatcher::Instance().AddTexture( atlasPath.c_str() );
        }

        Global::pathToCachedTexture[ atlasPath ] = *this;
    }

    std::ifstream metaStream( atlasMetadataPath.c_str() );

    if (atlasMetadataPath.find( ".xml" ) != std::string::npos || atlasMetadataPath.find( ".XML" ) != std::string::npos )
    {
        // Assumes that .xml extension means CEGUI/Ogre XML format.
        std::string line;
        while (std::getline( metaStream, line ))
        {
            if (line.find( "<Image Name" ) != std::string::npos)
            {
                std::vector< std::string > tokens;
                Tokenize( line, tokens, "\"" );
                bool found = false;

                for (std::size_t t = 0; t < tokens.size(); ++t)
                {
                    if (tokens[ t ].find( "Name" ) != std::string::npos)
                    {
                        if (tokens[ t + 1 ] == fileName)
                        {
                            found = true;
                        }
                    }

                    if (found)
                    {
                        if (tokens[ t ].find( "XPos" ) != std::string::npos)
                        {
                            scaleOffset.z = std::stoi( tokens[ t + 1 ] ) / (float)width;
                        }
                        else if (tokens[ t ].find( "YPos" ) != std::string::npos)
                        {
                            scaleOffset.w = std::stoi( tokens[ t + 1 ] ) / (float)height;
                        }
                        else if (tokens[ t ].find( "Width" ) != std::string::npos)
                        {
                            const int w = std::stoi( tokens[ t + 1 ] );
                            scaleOffset.x = 1.0f / ((float)width / (float)w);
                            width = w;
                        }
                        else if (tokens[ t ].find( "Height" ) != std::string::npos)
                        {
                            const int h = std::stoi( tokens[ t + 1 ] );
                            scaleOffset.y = 1.0f / ((float)height / (float)h);
                            height = h;
                        }
                    }
                }

                if (found)
                {
                    return;
                }
            }
        }

        Log::Print( "Didn't find texture %s in atlas %s.", fileName.c_str(), atlasPath.c_str() );
#if !AETHER3D_D3D11
        id = 0;
#endif
        width = 64;
        height = 64;
        opaque = true;
        scaleOffset = Vec4( 1, 1, 0, 0 );
    }
}

TextureImpl::TextureImpl( const std::string& left,  const std::string& right,
                          const std::string& top,   const std::string& bottom,
                          const std::string& front, const std::string& back )
: scaleOffset( 1, 1, 0, 0 )
, opaque( true )
, sRGB( true )
, hasMipmaps( false )
, cubeMap( true )
{
    sampler.tiling = Tiling::Clamp;
    sampler.filter = Filter::Linear;

    const std::string cachePath = left + right + top + bottom + front + back;

    if (Global::pathToCachedTexture.find( cachePath ) != Global::pathToCachedTexture.end())
    {
        *this = Global::pathToCachedTexture[ cachePath ];
        return;
    }

    Load( left, right, top, bottom, front, back, true, true );

    Global::pathToCachedTexture[ cachePath ] = *this;
}
