#include "TextureImpl.hpp"
#include <cstdlib>
#include <algorithm>
#include <vector>
#include <map>
#include <fstream>
#if defined AETHER3D_GLES3 && !defined AETHER3D_IOS
#    include <GLES3/gl3.h>
#elif defined AETHER3D_GLES3 && defined AETHER3D_IOS
#    include <OpenGLES/ES3/glext.h>
#else
#    include "GL/glxw.h"
#    include "DDSLoader.hpp"
#endif
#include "Log.hpp"
#include "MathUtil.hpp"
#include "RendererImpl.hpp"
#include "Statistics.hpp"

unsigned char* LoadTextureData( const std::string& path, int& outWidth, int& outHeight, int& outComponents );
void FreeTextureData( unsigned char* data );
#if AETHER3D_IOS
bool loadPVR( const std::string& aFileName, unsigned& outId, bool& outOpaque, int& outWidth, int& outHeight );
#endif
bool HasExtension( const std::string& glExtension );
void ErrorCheck( const std::string& info );
void ErrorCheckFBO();

namespace Global
{
    extern std::map< std::string, TextureImpl > pathToCachedTexture;
    extern GLuint systemFBO;
}

static void SolveGLFormat( TextureImpl::Format format, GLenum& outType, GLenum& outInternalFormat, GLenum& outExternalFormat )
{
    if (format == TextureImpl::Format::UnsignedByte)
    {
        outType = GL_UNSIGNED_BYTE;
        outInternalFormat = GL_RGBA8;
        outExternalFormat = GL_RGBA;
    }
    else if (format == TextureImpl::Format::HalfFloat)
    {
        outType = GL_FLOAT;
        outInternalFormat = GL_RGBA16F;
        outExternalFormat = GL_RGBA;
    }
    else if (format == TextureImpl::Format::HalfFloatRG)
    {
        outType = GL_FLOAT;
#if AETHER3D_GLES3
        Log::Print( "CreateRenderTexture: Two-channel float format not available on GLES3 renderer, using 4-channel instead." );
        outExternalFormat = GL_RGBA;
#else
        outInternalFormat = GL_RG16F;
#endif
        outExternalFormat = GL_RGBA;
    }
    else if (format == TextureImpl::Format::Float)
    {
        outType = GL_FLOAT;
        outExternalFormat = GL_RGBA;
#if AETHER3D_GLES3
        if (!HasExtension( "GL_EXT_color_buffer_float" ))
        {
            Log::Print( "CreateRenderTexture: Tried to create a 32-bit float texture but extension GL_EXT_color_buffer_float is not supported! Falling back to 16-bit float." );
            outInternalFormat = GL_RGBA16F;
        }
        else
#endif
        {
            outInternalFormat = GL_RGBA32F;
        }
    }
    else
    {
        Log::Assert( false, "Unhandled texture format." );
        outType = GL_UNSIGNED_BYTE;
        outInternalFormat = GL_RGBA8;
        outExternalFormat = GL_RGBA;
    }
}

TextureImpl::~TextureImpl()
{
}

#if !AETHER3D_GLES3
/**
  Sets active texture's max anisotropy.

  \param filter Filter. Should be Anisotropic8 or Anisotropic16.
*/
static void SetAnisotropy( TextureImpl::Filter filter )
{
    float anisotropy = 1;

    if (filter == Texture::Filter::Anisotropic8)
    {
        anisotropy = 8;
    }
    else if (filter == Texture::Filter::Anisotropic16)
    {
        anisotropy = 16;
    }

    if (HasExtension( "GL_EXT_texture_filter_anisotropic" ) && anisotropy > 1)
    {
        ErrorCheck( "TextureImpl Set anisotropy begin" );

        glTexParameterf( GL_TEXTURE_2D, 0x84FE/*GL_TEXTURE_MAX_ANISOTROPY_EXT*/, anisotropy );
        ErrorCheck( "TextureImpl Set anisotropy end" );
    }
    else if (anisotropy > 1)
    {
        static bool seenThis = false;

        if (!seenThis)
        {
            Log::Print( "Asked for anisotropic texture filtering, but the driver doesn't support it." );
            seenThis = true;
        }
    }
}
#endif

void TextureImpl::CreateRenderTexture2D( int aWidth, int aHeight, Format format, Filter filter, Tiling tiling, const std::string& debugName )
{
    (void)debugName;
    ErrorCheck( "CreateRenderTexture2D begin" );

    width = aWidth;
    height = aHeight;
    sampler.filter = filter;
    sampler.tiling = tiling;

    glDeleteTextures( 1, &id );
    glGenTextures( 1, &id );
#if !AETHER3D_GLES3
    if (HasExtension( "GL_KHR_debug" ))
    {
        // FIXME: Bug in Mesa? glObjectLabel causes a GL_INVALID_VALUE even though the id and debugName are correct.
        // Tested on Ubuntu 20.04, Mesa 20.0.8, Radeon R9 Nano.
        //int maxLen = 0;
        //glGetIntegerv( GL_MAX_LABEL_LENGTH, &maxLen );
        //Log::Print("max len: %d, name len: %d, name: %s\n", maxLen, debugName.size(), debugName.c_str());
        //glObjectLabel( GL_TEXTURE, id, -1, debugName.c_str() );
    }
#endif

    const int msaaSamples = RendererImpl::Instance().MSAASamples();
#if !AETHER3D_GLES3
    const GLenum target = msaaSamples == 1 ? GL_TEXTURE_2D : GL_TEXTURE_2D_MULTISAMPLE;
#else
    const GLenum target = GL_TEXTURE_2D;
#endif
    glBindTexture( target, id );

    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, tiling == TextureImpl::Tiling::Repeat ? GL_REPEAT : GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, tiling == TextureImpl::Tiling::Repeat ? GL_REPEAT : GL_CLAMP_TO_EDGE );
    
    if (msaaSamples == 1)
    {
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter == TextureImpl::Filter::Point ? GL_NEAREST : GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter == TextureImpl::Filter::Point ? GL_NEAREST : GL_LINEAR );
    }

    GLenum type, internalFormat, externalFormat;
    SolveGLFormat( format, type, internalFormat, externalFormat );

    if (msaaSamples == 1)
    {
        glTexImage2D( GL_TEXTURE_2D, 0, internalFormat, width, height, 0, externalFormat, type, nullptr );
    }
    else
    {
#if !AETHER3D_GLES3
        glTexImage2DMultisample( GL_TEXTURE_2D_MULTISAMPLE, msaaSamples, internalFormat, width, height, true );
#endif
    }

    glBindTexture( target, 0 );

    // Creates a renderbuffer object to store depth info.
    glDeleteRenderbuffers( 1, &rboId );
    glGenRenderbuffers( 1, &rboId );

    glBindRenderbuffer( GL_RENDERBUFFER, rboId );

    if (msaaSamples == 1)
    {
        glRenderbufferStorage( GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, width, height );
    }
    else
    {
        glRenderbufferStorageMultisample( GL_RENDERBUFFER, msaaSamples, GL_DEPTH_COMPONENT24, width, height );
    }

    glBindRenderbuffer( GL_RENDERBUFFER, 0 );

    // Creates the Frame Buffer Object.
    glDeleteFramebuffers( 1, &fboId );
    glGenFramebuffers( 1, &fboId );
    glBindFramebuffer( GL_FRAMEBUFFER, fboId );
    
    ErrorCheck( "CreateRenderTexture2D middle" );

    glFramebufferTexture2D( GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, target, id, 0 );    
    glFramebufferRenderbuffer( GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, rboId );
    RendererImpl::Instance().ClearScreen();

    ErrorCheckFBO();
    ErrorCheck( "CreateRenderTexture2D end" );

    glBindFramebuffer( GL_FRAMEBUFFER, Global::systemFBO );
}

void TextureImpl::CreateRenderTextureCube( int size )
{
    ErrorCheck( "CreateRenderTextureCube begin" );

    width = height = size;
    cubeMap = true;

    glGenTextures( 1, &id );
    glBindTexture( GL_TEXTURE_CUBE_MAP, id );

    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );

    // Creates a renderbuffer object to store depth info.
    GLuint bufferId;
    glGenRenderbuffers( 1, &bufferId );

    glBindRenderbuffer( GL_RENDERBUFFER, bufferId );
    glRenderbufferStorage( GL_RENDERBUFFER, GL_DEPTH_COMPONENT16, width, height );
    glBindRenderbuffer( GL_RENDERBUFFER, 0 );

    glGenFramebuffers( 1, &fboId );
    glBindFramebuffer( GL_FRAMEBUFFER, fboId );

    const GLenum externalFormat = GL_RGBA;
    const GLenum internalFormat = GL_RGBA8;
    const GLenum dataType = GL_UNSIGNED_BYTE;
    
    for (int i = 0; i < 6; ++i)
    {
        glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, internalFormat,
                      size, size, 0, externalFormat,
                      dataType, nullptr );
    }

    glFramebufferRenderbuffer( GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, bufferId );
    glBindTexture( GL_TEXTURE_CUBE_MAP, 0 );

    ErrorCheckFBO();
    ErrorCheck( "CreateRenderTextureCube end." );
}

void TextureImpl::Bind( int targetUnit ) const
{
    ErrorCheck( "TextureImpl Bind begin" );
    Log::Assert( targetUnit < sMaxTextureUnit, "Texture unit too high!" );
    Log::Assert( id != 0, "Texture not created yet." );

    static GLuint cachedId[ sMaxTextureUnit ] = { 0 };
    static GLenum cachedType[ sMaxTextureUnit ] = { 0 };

#if !AETHER3D_GLES3
    const int msaaSamples = RendererImpl::Instance().MSAASamples();
    const GLenum target2d = (msaaSamples == 1 || fboId == 0) ? GL_TEXTURE_2D : GL_TEXTURE_2D_MULTISAMPLE;
#else
    const GLenum target2d = GL_TEXTURE_2D;
#endif
    const GLenum type = cubeMap ? GL_TEXTURE_CUBE_MAP : target2d;

    if (type != cachedType[ targetUnit ] || id != cachedId[ targetUnit ])
    {
        cachedId[ targetUnit ]   = id;
        cachedType[ targetUnit ] = type;

        glActiveTexture( GL_TEXTURE0 + targetUnit );
        // Unbinds previous texture to prevent using a wrong texture type which causes an OpenGL error
        // in some cases if the shader uses both types.
        glBindTexture( type == target2d ? GL_TEXTURE_CUBE_MAP : target2d, 0 );
        glBindTexture( type, id );
        Statistics::IncTextureSwitches();
    }

    ErrorCheck( "TextureImpl Bind end" );
}

void TextureImpl::CreateNoise( int size )
{
    ErrorCheck( "CreateNoise begin" );
    
    width = height = size;
    sampler.tiling = Tiling::Repeat;

    glGenTextures( 1, &id );
    
    glBindTexture( GL_TEXTURE_2D, id );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );

    std::vector< Vec3 > noise( size * size );
    
	for (int i = 0; i < size * size; ++i)
    {
		noise[ i ] = Vec3(
                           MathUtil::Random( -1.0f, 1.0f ),
                           MathUtil::Random( -1.0f, 1.0f ),
                           0
                           ).Normalized();
    }
    
#if !AETHER3D_GLES3
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA32F, width, height, 0, GL_RGB, GL_FLOAT, &noise[ 0 ] );
#else
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, &noise[ 0 ] );
#endif
    ErrorCheck( "CreateNoise end" );
}

bool TextureImpl::Load( const std::string& aPath, bool useMipmaps, bool is_sRGB )
{
    ErrorCheck( "TextureImpl Load begin" );

    //path = aPath;
    
#if AETHER3D_IOS
    if (aPath.find( ".pvr" )   != std::string::npos ||
        aPath.find( ".PVR" )   != std::string::npos ||
        aPath.find( ".pvrtc" ) != std::string::npos ||
        aPath.find( ".PVRTC" ) != std::string::npos)
    {
        std::ifstream ifs( aPath.c_str() );
        if (!ifs)
        {
            return false;
        }

        return loadPVR( aPath, id, opaque, width, height );
    }
#endif
    if (aPath.find( ".dds" ) != std::string::npos || aPath.find( ".DDS" ) != std::string::npos)
    {
#if !AETHER3D_GLES3
        if (id == 0)
        {
            glGenTextures( 1, &id );
            
            if (HasExtension( "GL_KHR_debug" ))
            {
                glObjectLabel( GL_TEXTURE, id, -1, aPath.c_str() );
            }
        }

        glActiveTexture( GL_TEXTURE0 );
        glBindTexture( GL_TEXTURE_2D, id );

        const GLenum magFilter = sampler.filter == Texture::Filter::Point ? GL_NEAREST : GL_LINEAR;
        const GLenum minFilter = sampler.filter == Texture::Filter::Point ? GL_NEAREST : (useMipmaps ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);

        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter );

        // \todo Generate mipmaps if the .dds doesn't already have them and useMipmaps is true.
        
        const bool success = DDSLoader::Load( aPath.c_str(), 0, width, height, opaque );

        if (!success)
        {
            Log::Print( "MakeTexture: Could not load %s", aPath.c_str() );
        }

        return success;
#else
        const bool defaultLoaded = Load( RendererImpl::Instance().BuiltinAssetPath() + "textures/checker.png", true, true );

        if (!defaultLoaded)
        {
            Log::Assert( false, "Could not load default texture." );
        }
#endif
    }

    // Checks for uncompressed formats in texture's file name.
    const std::string extensions[] =
    {
        ".png", ".PNG", ".jpg", ".JPG", ".tga", ".TGA",
        ".bmp", ".BMP", ".gif", ".GIF"
    };

    const bool extensionFound = std::any_of( std::begin( extensions ), std::end( extensions ),
                                             [&]( const std::string& extension ) { return aPath.find( extension ) != std::string::npos; } );
    if (extensionFound)
    {
        int comp = 0;
        unsigned char* data = LoadTextureData( aPath, width, height, comp );
        
        if (!data || comp == 1)
        {
            if (comp == 1)
            {
                Log::Print( "MakeTexture: Texture %s is not 32 or 24 bits.", aPath.c_str() );
            }
            else
            {
                Log::Print( "MakeTexture: Could not load %s", aPath.c_str() );
            }
            
            if (aPath == RendererImpl::Instance().BuiltinAssetPath() + "textures/checker.png")
            {
                Log::Assert( false, "Trying to load default texture unsuccessfully." );
                return false;
            }
            
            const bool ok = Load( RendererImpl::Instance().BuiltinAssetPath() + "textures/checker.png", true, true );
            
            if (!ok)
            {
                Log::Assert( false, "Could not load default texture." );
            }
            
            return false;
        }

        // The texture wasn't cached, so create it.
        opaque = (comp == 3 || comp == 1);
        
        GLenum internalFormat = GL_SRGB8_ALPHA8;
        const GLenum externalFormat = GL_RGBA;
        
        if (!is_sRGB)
        {
            internalFormat = externalFormat;
        }
        
        if (id == 0)
        {
            glGenTextures( 1, &id );
        }

        glActiveTexture( GL_TEXTURE0 );
        glBindTexture( GL_TEXTURE_2D, id );
        
        const GLenum magFilter = sampler.filter == Texture::Filter::Point ? GL_NEAREST : GL_LINEAR;
        const GLenum minFilter = sampler.filter == Texture::Filter::Point ? GL_NEAREST : (useMipmaps ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
        
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, sampler.tiling == TextureImpl::Tiling::Repeat ? GL_REPEAT : GL_CLAMP_TO_EDGE );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, sampler.tiling == TextureImpl::Tiling::Repeat ? GL_REPEAT : GL_CLAMP_TO_EDGE );
        
#if !AETHER3D_GLES3
        if (HasExtension( "GL_KHR_debug" ))
        {
            glObjectLabel( GL_TEXTURE, id, -1, aPath.c_str() );
        }

        SetAnisotropy( sampler.filter );
#endif
        glTexImage2D( GL_TEXTURE_2D, 0, internalFormat, width, height, 0, externalFormat, GL_UNSIGNED_BYTE, data );

        if (useMipmaps)
        {
            glGenerateMipmap( GL_TEXTURE_2D );
        }
        
        FreeTextureData( data );
        data = nullptr;
    }

    if (!extensionFound)
    {
        Log::Print( "MakeTexture: Unsupported extension in file: %s", aPath.c_str() );
    }
    
    ErrorCheck( "TextureImpl Load end" );
    return extensionFound;
}

bool TextureImpl::Load( const std::string& left,  const std::string& right,
                        const std::string& top,   const std::string& bottom,
                        const std::string& front, const std::string& back,
                        bool useMipmaps, bool usesRGB )
{
    ErrorCheck( "TextureImpl Load begin" );
    
    if (id == 0)
    {
        glGenTextures( 1, &id );
    }

    glActiveTexture( GL_TEXTURE0 );
    glBindTexture( GL_TEXTURE_CUBE_MAP, id );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, useMipmaps ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );

    const std::string fileNames[] = { right, left, bottom, top, front, back };
    const std::string extensions[] =
    {
        ".png", ".PNG", ".jpg", ".JPG", ".tga", ".TGA",
        ".bmp", ".BMP", ".gif", ".GIF"
    };

    for (int i = 0; i < 6; ++i)
    {
        bool loadedSuccessfully = true;

#if !AETHER3D_GLES3
        if (fileNames[ i ].find( ".dds" ) != std::string::npos || fileNames[ i ].find( ".DDS" ) != std::string::npos)
        {
            const bool ok = DDSLoader::Load( fileNames[ i ].c_str(), i + 1, width, height, opaque );

            if (!ok)
            {
                Log::Print( "MakeTexture: Could not load %s", fileNames[ i ].c_str() );
            }

            continue;
        }
#endif

        for (auto& e : extensions)
        {
            if (fileNames[ i ].find( e ) == std::string::npos)
            {
                continue;
            }

            int comp = 0;
            unsigned char* data = LoadTextureData( fileNames[ i ], width, height, comp );

            if (!data || comp == 1)
            {
                if (comp == 1)
                {
                    Log::Print( "MakeTexture: Texture %s is not 32 or 24 bits.", fileNames[ i ].c_str() );
                }
                else
                {
                    Log::Print( "MakeTextureCube: Could not load %s", fileNames[ i ].c_str() );
                }

                loadedSuccessfully = false;

                comp = 3;
                width = 64;
                height = 64;
                data = new unsigned char[ width * height * 4 ];
                std::fill_n( data, width * height * 4, 1 );
            }

            opaque = (comp == 3 || comp == 1);

            GLenum internalFormat = GL_SRGB8_ALPHA8;
            const GLenum externalFormat = GL_RGBA;

            if (!usesRGB)
            {
                internalFormat = externalFormat;
            }

            glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, internalFormat,
                         width, height, 0, externalFormat,
                         GL_UNSIGNED_BYTE, data );
            
            if (loadedSuccessfully)
            {
                FreeTextureData( data );
            }
            else
            {
                delete[] data;
            }

            data = nullptr;
        }
    }

    if (useMipmaps)
    {
        glGenerateMipmap( GL_TEXTURE_CUBE_MAP );
    }

    ErrorCheck( "TextureImpl Load end" );
    return true;
}
