#import <QuartzCore/QuartzCore.h>
#include <OpenGLES/ES3/glext.h>
#include <string>

const char* GetFullPath( const char* fileName )
{
    NSBundle *b = [NSBundle mainBundle];
    NSString *dir = [b resourcePath];
    NSString* fName = [NSString stringWithUTF8String: fileName];
    dir = [dir stringByAppendingString:fName];
    return [dir fileSystemRepresentation];
}

namespace PVRType
{
    enum Enum
    {
        kPVRTextureFlagTypePVRTC_2 = 24,
        kPVRTextureFlagTypePVRTC_4
    };
}

typedef struct
{
    uint32_t headerLength;
    uint32_t height;
    uint32_t width;
    uint32_t numMipmaps;
    uint32_t flags;
    uint32_t dataLength;
    uint32_t bpp;
    uint32_t bitmaskRed;
    uint32_t bitmaskGreen;
    uint32_t bitmaskBlue;
    uint32_t bitmaskAlpha;
    uint32_t pvrTag;
    uint32_t numSurfs;
} PVRTexHeader;

bool loadPVR( const std::string& aFileName, unsigned& outId, bool& outOpaque, int& outWidth, int& outHeight )
{
    // PVRTC textures generated by texturetool using command line:
    // texturetool -m -e PVRTC -f PVR -o checker.pvr checker.png
    NSString* pvrtcNSString = [NSString stringWithUTF8String: aFileName.c_str()];
    NSData* fileData = [NSData dataWithContentsOfFile:pvrtcNSString];
    PVRTexHeader* header = (PVRTexHeader*) [fileData bytes];
    uint32_t pvrTag = CFSwapInt32LittleToHost(header->pvrTag);
    const char gPVRTexIdentifier[] = "PVR!";
    
    if (gPVRTexIdentifier[0] != ((pvrTag >>  0) & 0xff) ||
        gPVRTexIdentifier[1] != ((pvrTag >>  8) & 0xff) ||
        gPVRTexIdentifier[2] != ((pvrTag >> 16) & 0xff) ||
        gPVRTexIdentifier[3] != ((pvrTag >> 24) & 0xff))
    {
        NSLog( @"PVRTC header is corrupted: %@\n", pvrtcNSString );
        return false;
    }
    
    glGenTextures( 1, &outId );

    glBindTexture( GL_TEXTURE_2D, outId );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    
    uint32_t flags = CFSwapInt32LittleToHost( header->flags );
    const uint32_t PVR_TEXTURE_FLAG_TYPE_MASK = 0xFF;
    uint32_t formatFlags = flags & PVR_TEXTURE_FLAG_TYPE_MASK;
    
    if (formatFlags == PVRType::kPVRTextureFlagTypePVRTC_4 ||
        formatFlags == PVRType::kPVRTextureFlagTypePVRTC_2)
    {
        uint8_t* bytes = (uint8_t*)([fileData bytes]) + sizeof( PVRTexHeader );
        NSMutableArray* imageData = [[NSMutableArray alloc] initWithCapacity:10];
        [imageData removeAllObjects];
        
        outOpaque = !CFSwapInt32LittleToHost( header->bitmaskAlpha );
        
        uint32_t dataLength = CFSwapInt32LittleToHost( header->dataLength );
        uint32_t dataOffset = 0;
        outWidth = CFSwapInt32LittleToHost( header->width );
        outHeight = CFSwapInt32LittleToHost( header->height );

        while (dataOffset < dataLength)
        {
            uint32_t blockSize;
            uint32_t widthBlocks = 0;
            uint32_t heightBlocks = 0;
            uint32_t bpp = 0;
            
            if (formatFlags == PVRType::kPVRTextureFlagTypePVRTC_4)
            {
                blockSize = 4 * 4; // Pixel by pixel block size for 4bpp
                widthBlocks = outWidth / 4;
                heightBlocks = outHeight / 4;
                bpp = 4;
            }
            else
            {
                blockSize = 8 * 4; // Pixel by pixel block size for 2bpp
                widthBlocks = outWidth / 8;
                heightBlocks = outHeight / 4;
                bpp = 2;
            }
            
            // Clamp to minimum number of blocks
            if (widthBlocks < 2)
            {
                widthBlocks = 2;
            }
            
            if (heightBlocks < 2)
            {
                heightBlocks = 2;
            }
            
            uint32_t dataSize = widthBlocks * heightBlocks * ((blockSize  * bpp) / 8);
            
            [imageData addObject:[NSData dataWithBytes:bytes+dataOffset length:dataSize]];
            
            dataOffset += dataSize;
            
            outWidth = MAX(outWidth >> 1, 1);
            outHeight = MAX(outHeight >> 1, 1);
        }
        
        // Stores mipmaps.
        outWidth = header->width;
        outHeight = header->height;
        
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, [imageData count] > 1 ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR );
        
        for (int i = 0; i < [imageData count]; ++i)
        {
            NSData* data = [imageData objectAtIndex:i];
            glCompressedTexImage2D( GL_TEXTURE_2D, i, !outOpaque ? GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG : GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG,
                                   outWidth, outHeight, 0, [data length], [data bytes] );
            outWidth = MAX( outWidth >> 1, 1 );
            outHeight = MAX( outHeight >> 1, 1 );
        }
        [imageData removeAllObjects];
        
        GLenum errorCode;
        while ( (errorCode = glGetError()) != GL_NO_ERROR )
        {
            NSLog( @"Error loading PVRTC texture %@\n", pvrtcNSString );
        }
        return true;
    }
    
    return false;
}

/**
 Loads an image file.
 
 \param aFileName File name.  Supported extensions are:
 \param aWidth Width in pixels is stored here.
 \param aHeight Height in pixels is stored here.
 \param aHasAlpha Presence of alpha channel is stored here.
 \param aData Pixel data that can be fed to glTexImage.  Caller must free the memory.
 */
unsigned char* LoadImageCG( const std::string& aFileName, int& aWidth, int& aHeight, bool& aHasAlpha )
{
    CGImageRef spriteImage;
    CGContextRef spriteContext;
    GLubyte *spriteData;
    
    // Creates a Core Graphics image from an image file.
    NSString *string = [NSString stringWithUTF8String: aFileName.c_str()];
    spriteImage = [UIImage imageWithContentsOfFile:string].CGImage;
    aWidth  = (int)CGImageGetWidth ( spriteImage );
    aHeight = (int)CGImageGetHeight( spriteImage );
    
    if (spriteImage)
    {
        aHasAlpha = CGImageGetAlphaInfo( spriteImage ) == kCGImageAlphaPremultipliedLast;
        
        // Allocate memory needed for the bitmap context.
        spriteData = (GLubyte *) calloc( aWidth * aHeight * 4, sizeof(GLubyte) );
        
        // Uses the bitmap creation function provided by the CG framework.
        spriteContext = CGBitmapContextCreate( spriteData, aWidth, aHeight, 8,
                                              aWidth * 4, CGImageGetColorSpace( spriteImage ),
                                              kCGImageAlphaPremultipliedLast);
        
        // Draws the sprite image to the context.
        CGContextDrawImage( spriteContext, CGRectMake( 0.0, 0.0,
                                                      (CGFloat)aWidth, (CGFloat)aHeight), spriteImage );
        CGContextRelease( spriteContext );
        
        return spriteData;
    }
    return NULL;
}
