#import "ViewController.h"
#include "CameraImpl.hpp"
#include "LightImpl.hpp"
#include "MaterialImpl.hpp"
#include "ModelImpl.hpp"
#include "RendererImpl.hpp"
#include "SceneImpl.hpp"
#include "TextureImpl.hpp"

const char* GetFullPath( const char* fileName );
GLKView *globalView = NULL;

void BindDrawable()
{
    [globalView bindDrawable];
}

@interface ViewController () {
    CameraImpl camera;
    ModelImpl model;
    SceneImpl scene;
    TextureImpl* texture;
    TextureImpl* normal;
    MaterialImpl skinMaterial;
    MaterialImpl cookTorranceMaterial;
    LightImpl* light;
    float angle;
}
@property (strong, nonatomic) EAGLContext *context;

- (void)setupGL;
- (void)tearDownGL;

@end

@implementation ViewController

- (void)viewDidLoad
{
    [super viewDidLoad];
    
    self.context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES3];

    if (!self.context)
    {
        NSLog(@"Failed to create ES context");
    }
    
    GLKView *view = (GLKView *)self.view;
    view.context = self.context;
    view.drawableDepthFormat = GLKViewDrawableDepthFormat24;
    
    globalView = view;

    [self setupGL];
}

- (void)dealloc
{    
    [self tearDownGL];
    
    if ([EAGLContext currentContext] == self.context) {
        [EAGLContext setCurrentContext:nil];
    }
}

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];

    if ([self isViewLoaded] && ([[self view] window] == nil)) {
        self.view = nil;
        
        [self tearDownGL];
        
        if ([EAGLContext currentContext] == self.context) {
            [EAGLContext setCurrentContext:nil];
        }
        self.context = nil;
    }
}

- (void)setupGL
{
    [EAGLContext setCurrentContext:self.context];

    glEnable(GL_DEPTH_TEST);

    RendererImpl::Instance().SetBuiltinAssetPath( "/Assets/builtin/" );
    RendererImpl::Instance().OpenWindow( self.view.bounds.size.width*2, self.view.bounds.size.height*2, true, 1, true );
    RendererImpl::Instance().SetClearColor( 1, 0, 0 );
    //RendererImpl::Instance().SetBloom( true, 0.85f, 2 );

    SceneGraphNode::RegisterScene( &scene );

    texture = new TextureImpl( "/Assets/textures/checker.pvr", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear );
    normal = new TextureImpl( "/table_n.png", false, false, Texture::Tiling::Repeat, Texture::Filter::Linear );

    skinMaterial.SetShaders( "unlit", "unlit", "", "SKIN" );
    skinMaterial.SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    skinMaterial.SetTexture( "textureMap", texture );

    cookTorranceMaterial.SetShaders( "cook_torrance", "cook_torrance", "", "" );
    cookTorranceMaterial.SetTexture( "diffuseMap", texture );
    cookTorranceMaterial.SetTexture( "normalMap", normal );
    cookTorranceMaterial.SetVector( "materialSpecular", Vec3( 0.2f, 0.2f, 0.2f ) );
    cookTorranceMaterial.SetFloat( "roughness", 0.1f );

    Error::Enum error = model.LoadFromFile( "/skinnedMesh.ae3d" );
    assert( error == Error::NO_ERRORS );
    //model.SetMaterialForMesh( &skinMaterial, "body" );
    model.SetMaterialForMesh( &cookTorranceMaterial, "body" );

    scene.Add( &model );

    light = new LightImpl( Light::Type::Directional );
    light->SetDirection( Vec3( 1, 0, 0 ) );
    light->SetCastShadow( false, 512 );
    scene.Add( light );

    camera.SetProjection( 55, (float)self.view.bounds.size.width / (float)self.view.bounds.size.height, 1, 200 );
    //camera->SetProjection( -10, 10, -10, 10, 1, 200 );
    camera.LookAt( Vec3( 0, 0, -20 ), Vec3( 0, 0, 2000 ), Vec3( 0, 1, 0 ) );
    scene.Add( &camera );
    model.ApplyAnimation( 0, 0 );
    angle = 0;
}

- (void)tearDownGL
{
    [EAGLContext setCurrentContext:self.context];
}

#pragma mark - GLKView and GLKViewController delegate methods

- (void)update
{
    ++angle;
    Quaternion rotation;
    Vec3 axis = Vec3( 1, 1, 1 ).Normalized();
    rotation.FromAxisAngle( axis, angle );
    model.SetRotation( rotation );
}

- (void)glkView:(GLKView *)view drawInRect:(CGRect)rect
{
    RendererImpl& renderer = RendererImpl::Instance();
    
    renderer.ClearScreen();
    
    renderer.Draw( &scene );
    renderer.Draw( "hello world", "aether3d_font", 1, 10, 280, Vec4( 1, 1, 1, 1 ) );
    renderer.DrawStatistics();
    
    renderer.FinishFrame();
    renderer.SwapBuffers();
}

@end
