#include <cassert>
#include "../common.hpp"

const int gWidth = 1280;
const int gHeight = 720;

int main()
{
    Aether3D_API dll;
    Error::Enum error = dll.LoadDLL( "../Renderers/libaether3d_gl32" );

    if (error == Error::DLL_NOT_FOUND)
    {
        assert( !"Could not find engine DLL!" );
        return (int)error;
    }
    else if (error == Error::FACTORY_NOT_FOUND)
    {
        assert( !"No entry point in DLL!" );
        return (int)error;
    }
    else if (error == Error::DLL_VERSION_MISMATCH)
    {
        assert( !"DLL version does not match header version!" );
        return (int)error;
    }

    Renderer* renderer = dll.MakeRenderer( "../Assets/builtin/" );
    assert( renderer );

    error = renderer->OpenWindow( gWidth, gHeight, false, 1, true );
    //error = renderer->OpenWindow( 0, 0, true, 4 );

    if (error != Error::NO_ERRORS)
    {
        assert( !"Could not create rendering context!" );
        renderer->Deinit();
        return (int)error;
    }

    std::unique_ptr< KeyboardInput > keyboardInput( dll.MakeKeyboardInput() );
    std::unique_ptr< MouseInput > mouseInput( dll.MakeMouseInput() );
    mouseInput->SetRelative( true );

    renderer->SetClearColor( 0.6f, 0.6f, 1 );
    renderer->SetWindowTitle( "Aether3D Example: RenderTexture" );

    std::unique_ptr< Texture > texture( dll.MakeTexture( "../Assets/builtin/textures/checker.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > texture2( dll.MakeTexture( "../Assets/textures/skybox/skybox_left.jpg", false, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > renderTexture( dll.MakeRenderTexture( 1024, 1024 ) );
    std::unique_ptr< Texture > renderTextureCube( dll.MakeRenderTextureCube( 1024 ) );
    std::unique_ptr< Texture > skybox( dll.MakeTexture( "../Assets/textures/skybox/skybox_left.jpg",
                                                        "../Assets/textures/skybox/skybox_right.jpg",
                                                        "../Assets/textures/skybox/skybox_top.jpg",
                                                        "../Assets/textures/skybox/skybox_bottom.jpg",
                                                        "../Assets/textures/skybox/skybox_front.jpg",
                                                        "../Assets/textures/skybox/skybox_back.jpg") );

    std::unique_ptr< Model > skinnedModel( dll.MakeModel( "Assets/Models/skinnedMesh.ae3d", error ) );
    std::unique_ptr< Model > model2( dll.MakeModel( "../Tests/assets/models/textured_cube.ae3d", error ) );
    std::unique_ptr< Model > cube( dll.MakeModel( "../Tests/assets/models/textured_cube.ae3d", error ) );
    skinnedModel->SetPosition( Vec3( 0, 6, 0 ) );
    model2->SetPosition( Vec3( 0, 0, -6 ) );
    model2->SetScale( 2 );
    cube->SetPosition( Vec3( 10, 0, -6 ) );
    cube->SetScale( 4 );
    
    std::unique_ptr< Material > material( dll.MakeMaterial() );
    material->SetShaders( "unlit", "unlit", "", "SKIN" );
    material->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    material->SetTexture( "textureMap", texture.get() );

    std::unique_ptr< Material > rtMaterial( dll.MakeMaterial() );
    rtMaterial->SetShaders( "unlit", "unlit", "", "" );
    rtMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    rtMaterial->SetTexture( "textureMap", renderTexture.get() );

    std::unique_ptr< Material > cubeMaterial( dll.MakeMaterial() );
    cubeMaterial->SetShaders( "envmap", "envmap", "", "" );
    cubeMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    cubeMaterial->SetTexture( "textureMap", renderTextureCube.get() );

    skinnedModel->SetMaterialForMesh( material.get(), "body" );
    model2->SetMaterialForMesh( rtMaterial.get(), "Cube" );
    cube->SetMaterialForMesh( cubeMaterial.get(), "Cube" );

    std::unique_ptr< Scene > scene( dll.MakeScene() );
    scene->Add( skinnedModel.get() );
    scene->Add( model2.get() );
    scene->Add( cube.get() );
    scene->SetSkybox( skybox.get() );

    std::unique_ptr< Camera > camera( dll.MakeCamera() );
    camera->SetProjection( 55, (float)renderer->Width() / (float)renderer->Height(), 1, 200 );
    //camera->SetProjection( -10, 10, -10, 10, 1, 200 );
    camera->LookAt( Vec3( 0, 0, -20 ), Vec3( 2000, 2000, 0 ), Vec3( 0, 1, 0 ) );
    scene->Add( camera.get() );

    std::unique_ptr< Camera > cubeCamera( dll.MakeCamera() );
    cubeCamera->SetProjection( 45, 1, 1, 200 );
    cubeCamera->SetPosition( Vec3( 0, 0, -6 ) );
    cubeCamera->SetRenderTarget( renderTextureCube.get() );
    scene->Add( cubeCamera.get() );

    std::unique_ptr< Camera > rtCamera( dll.MakeCamera() );
    rtCamera->SetProjection( 55, 1, 1, 200 );
    rtCamera->LookAt( Vec3( 0, 0, -20 ), Vec3( 0, 0, 2000 ), Vec3( 0, 1, 0 ) );
    rtCamera->SetRenderTarget( renderTexture.get() );
    scene->Add( rtCamera.get() );

    //double gpuTime = 0;

    /*std::unique_ptr< Light > sunLight( dll.MakeLight() );
    Vec3 lightDir( 0.5f, 0.5f, 0.0f );
    lightDir.Normalize();
    sunLight->SetDirection( lightDir );
    sunLight->SetCastShadow( true, 512 );
    scene->Add( sunLight.get() );
    */
    while (!keyboardInput->IsDown( ae3d::keyboardinput::KEY_ESC ))
    {
        //renderer->BeginQuery();

        mouseInput->SetRelative(renderer->HasFocus());

        HandleInput( keyboardInput.get(), camera.get() );
        HandleInput( mouseInput.get(), camera.get() );

        renderer->SetClearColor( 0.2f, 0.2f, 0.5f );
        renderer->Draw( scene.get() );

        //std::string gpuText = "GPU: ";
        //gpuText += std::to_string( gpuTime ) + " ms";
        //renderer->Draw( gpuText, "aether3d_font", 1, 5, 50, Vec4( 1, 1, 1, 1 ) );

        renderer->DrawStatistics();
        renderer->FinishFrame();

        renderer->SwapBuffers();
        //gpuTime = renderer->EndQuery();
    }

    renderer->Deinit();
    return 0;
}

