#include <algorithm>
#include <iostream>
#include <vector>
#include "Errors.hpp"
#include "MathUtil.hpp"
#include "Vec3.hpp"
#include "Quaternion.hpp"

int main()
{
    if (MathUtil::AlmostEquals( 0.1f, 0.2f ) || !MathUtil::AlmostEquals( 0.5f, 0.5f ))
    {
        std::cerr << "MathUtil: AlmostEquals failed!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    std::vector< Vec3 > points;
    Vec3 pointsMin, pointsMax;

    MathUtil::GetMinMax( points, pointsMin, pointsMax );

    points.push_back( Vec3( 0, 0, 0) );
    points.push_back( Vec3( 1, 1, 1) );

    MathUtil::GetMinMax( points, pointsMin, pointsMax );

    if (pointsMin != points[ 0 ] || pointsMax != points[ 1 ])
    {
        std::cerr << "MathUtil: GetMinMax failed!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    std::vector< Vec3 > corners;

    MathUtil::GetCorners( pointsMin, pointsMax, corners );

    if (corners.size() != 8 || std::find( std::begin( corners ), std::end( corners ), pointsMin ) == std::end( corners ))
    {
        std::cerr << "MathUtil: GetCorners failed!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    float f = 0.5f;

    if (MathUtil::Ceil( f ) != 1.0f)
    {
        std::cerr << "MathUtil: Ceil failed!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    if (MathUtil::Lerp( 0, 1, 0 ) != 0 || MathUtil::Lerp( 0, 1, 1 ) != 1)
    {
        std::cerr << "MathUtil: Lerp failed!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    if (!MathUtil::IsNaN( 0.0f / 0.0f ) || MathUtil::IsNaN( 1 ))
    {
        std::cerr << "MathUtil: IsNaN failed!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    if (!MathUtil::IsPowerOfTwo( 256 ) || MathUtil::IsPowerOfTwo( 255 ))
    {
        std::cerr << "MathUtil: IsPowerOfTwo failed!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    if (MathUtil::Min2< int >( -1, 1 ) != -1 || MathUtil::Min2< int >( 2, 3 ) != 2 ||
        MathUtil::Min2< int >( 0, 0 ) != 0)
    {
        std::cerr << "MathUtil: Min2 failed!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    if (MathUtil::Max2< int >( -1, 1 ) != 1 || MathUtil::Max2< int >( 2, 3 ) != 3 ||
        MathUtil::Max2< int >( 0, 0 ) != 0)
    {
        std::cerr << "MathUtil: Max2 failed!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    const Vec3 aabbMin( -5, -5, -5 );
    const Vec3 aabbMax(  5,  5,  5 );
    Vec3 sphereOrigin( 0, 0, 0 );
    const float sphereRadius = 2;

    if (!MathUtil::IntersectSphereAndAABB( sphereOrigin, sphereRadius, aabbMin, aabbMax ))
    {
        std::cerr << "MathUtil: IntersectSphereAndAABB failed when sphere is inside!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    sphereOrigin = Vec3( 8, 8, 8 );

    if (MathUtil::IntersectSphereAndAABB( sphereOrigin, sphereRadius, aabbMin, aabbMax ))
    {
        std::cerr << "MathUtil: IntersectSphereAndAABB failed when sphere is outside!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    const Vec3 aabb2Min( 10, 10, 10 );
    const Vec3 aabb2Max( 15, 15, 15 );

    if (MathUtil::IntersectAABBandAABB( aabbMin, aabbMax, aabb2Min, aabb2Max ))
    {
        std::cerr << "MathUtil: IntersectAABBandAABB failed when there is no intersection!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    if (MathUtil::IntersectAABBandAABB( aabbMin, aabbMax, aabb2Min, aabb2Max ) !=
        MathUtil::IntersectAABBandAABB( aabb2Min, aabb2Max, aabbMin, aabbMax ))
    {
        std::cerr << "MathUtil: IntersectAABBandAABB failed when switching order!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }

    if (!MathUtil::IntersectAABBandAABB( Vec3( 9, 9, 9 ), Vec3( 12, 12, 12 ), Vec3( 10, 10, 10 ), Vec3( 14, 14, 14 ) ))
    {
        std::cerr << "MathUtil: IntersectAABBandAABB failed when there is an intersection!" << std::endl;
        return (int)Error::CONTRACT_VIOLATION;
    }
}

