#include <cassert>
#include <iostream>
#include <vector>
#include "../../../Samples/common.hpp"

const int gWidth = 1280;
const int gHeight = 720;

#if AETHER3D_WINDOWS
#include <Windows.h>
INT WINAPI WinMain( HINSTANCE, HINSTANCE, LPSTR, INT )
#else
int main()
#endif
{
    Aether3D_API dll;
    Error::Enum error = dll.LoadDLL( "../../../Renderers/libaether3d_gl32.dylib" );

    if (error == Error::DLL_NOT_FOUND)
    {
        std::cerr << "Could not find engine DLL!" << std::endl;
        return (int)error;
    }
    else if (error == Error::FACTORY_NOT_FOUND)
    {
        std::cerr << "No entry point in DLL!" << std::endl;
        return (int)error;
    }
    else if (error == Error::DLL_VERSION_MISMATCH)
    {
        std::cerr << "DLL version does not match header version!" << std::endl;
        return (int)error;
    }

    Renderer* renderer = dll.MakeRenderer( "../../../Assets/builtin/" );
    assert( renderer );

    error = renderer->OpenWindow( gWidth, gHeight, false /* fullscreen */, 1, true );
    //error = renderer->OpenWindow( 0, 0, true, 1 );

    if (error != Error::NO_ERRORS)
    {
        std::cerr << "Could not create rendering context!" << std::endl;
        renderer->Deinit();
        return (int)error;
    }

    std::unique_ptr< KeyboardInput > keyboardInput( dll.MakeKeyboardInput() );
    std::unique_ptr< MouseInput > mouseInput( dll.MakeMouseInput() );
    mouseInput->SetRelative( true );
    
    renderer->SetClearColor( 0.6f, 0.6f, 1 );
    renderer->SetWindowTitle( "Aether3D Example: Model Loading" );
    renderer->AddShaderSearchPath( "../../../Assets/builtin/shaders/glsl150/" );

    std::unique_ptr< Texture > white( dll.MakeTexture( "../../../Assets/builtin/textures/white.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > trans50( dll.MakeTexture( "../../assets/textures/trans50.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > fontTexture( dll.MakeTexture( "../../../Assets/builtin/textures/font.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );

    std::unique_ptr< Material > material( dll.MakeMaterial() );
    material->SetShaders( "unlit", "unlit", "", "" );
    material->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    material->SetTexture( "textureMap", fontTexture.get() );
    material->SetOpaque( false );
    
    std::unique_ptr< Material > blueMaterial( dll.MakeMaterial() );
    blueMaterial->SetShaders( "unlit", "unlit", "", "" );
    blueMaterial->SetVector( "tint", Vec4( 0, 0, 1, 1 ) );
    blueMaterial->SetTexture( "textureMap", trans50.get() );
    blueMaterial->SetOpaque( false );

    std::unique_ptr< Material > redMaterial( dll.MakeMaterial() );
    redMaterial->SetShaders( "unlit", "unlit", "", "" );
    redMaterial->SetVector( "tint", Vec4( 1, 0, 0, 1 ) );
    redMaterial->SetTexture( "textureMap", trans50.get() );
    redMaterial->SetOpaque( true );
    
    std::unique_ptr< Scene > scene( dll.MakeScene() );

    auto model1 = dll.MakeModel( "../../../Samples/Assets/Models/sphere.ae3d", error );
    model1->SetPosition( Vec3( 0, -5, 0 ) );
    model1->SetMaterialForMeshSubstring( blueMaterial.get(), "" );
    scene->Add( model1 );

    auto model2 = dll.MakeModel( "../../../Samples/Assets/Models/sphere.ae3d", error );
    model2->SetPosition( Vec3( 4, -5, 0 ) );
    model2->SetMaterialForMeshSubstring( redMaterial.get(), "" );
    scene->Add( model2 );

    auto fontBall = dll.MakeModel( "../../../Samples/Assets/Models/sphere.ae3d", error );
    fontBall->SetPosition( Vec3( 0, -5, 4 ) );
    fontBall->SetMaterialForMeshSubstring( material.get(), "" );
    scene->Add( fontBall );

    /*std::unique_ptr< Texture > asphaltTexture( dll.MakeTexture( "../../../Samples/Assets/Textures/asphalt.jpg", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Material > asphaltMaterial( dll.MakeMaterial() );
    asphaltMaterial->SetShaders( "unlit", "unlit", "", "" );
    asphaltMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    asphaltMaterial->SetVector( "textureMap_ST", Vec4( 6, 6, 0, 0 ) );
    asphaltMaterial->SetTexture( "textureMap", asphaltTexture.get() );
    */
    std::unique_ptr< Texture > gridTexture( dll.MakeTexture( "../../../Samples/Assets/Textures/metal_grid.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );
    std::unique_ptr< Texture > gridTextureN( dll.MakeTexture( "../../../Samples/Assets/Textures/metal_grid_n2.png", true, false, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );
    std::unique_ptr< Texture > gridTextureS( dll.MakeTexture( "../../../Samples/Assets/Textures/metal_grid_s.png", true, false, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );

    std::unique_ptr< Material > asphaltMaterial( dll.MakeMaterial() );
    asphaltMaterial->SetShaders( "cook_torrance", "cook_torrance", "", "SPECULAR_MAP" );
    asphaltMaterial->SetVector( "diffuseMap_ST", Vec4( 10, 10, 0, 0 ) );
    asphaltMaterial->SetTexture( "diffuseMap", gridTexture.get() );
    asphaltMaterial->SetTexture( "normalMap", gridTextureN.get() );
    asphaltMaterial->SetTexture( "specularMap", gridTextureS.get() );
    asphaltMaterial->SetVector( "materialSpecular", Vec3( 0.2f, 0.2f, 0.2f ) );
    asphaltMaterial->SetFloat( "roughness", 0.3f );

    std::unique_ptr< Texture > skybox( dll.MakeTexture( "../../../Assets/textures/skybox/skybox_left.jpg",
                                                       "../../../Assets/textures/skybox/skybox_right.jpg",
                                                       "../../../Assets/textures/skybox/skybox_top.jpg",
                                                       "../../../Assets/textures/skybox/skybox_bottom.jpg",
                                                       "../../../Assets/textures/skybox/skybox_front.jpg",
                                                       "../../../Assets/textures/skybox/skybox_back.jpg") );

    std::unique_ptr< Model > floor( dll.MakeModel( "../../../Samples/Assets/Models/plane.ae3d", error ) );
    floor->SetScale( 10 );
    floor->SetMaterialForMesh( asphaltMaterial.get(), "Cube" );
    scene->Add( floor.get() );
    //scene->SetSkybox( skybox.get() );
    
    std::unique_ptr< Camera > camera( dll.MakeCamera() );
    camera->SetProjection( 55, (float)renderer->Width() / (float)renderer->Height(), 1, 200 );
    //camera->SetProjection( -10, 10, -10, 10, 1, 200 );
    camera->LookAt( Vec3( 0, 0, -20 ), Vec3( 0, 0, 2000 ), Vec3( 0, 1, 0 ) );
    scene->Add( camera.get() );
    
    std::unique_ptr< Light > light( dll.MakeLight( Light::Type::Directional ) );
    light->SetDirection( Vec3( 0, -1, 1 ).Normalized() );
    light->SetCastShadow( true, 1024 );
    scene->Add( light.get() );

    renderer->SetClearColor( 0.2f, 0.2f, 0.5f );
    
    //double gpuTime = 0;

    while (!keyboardInput->IsDown( ae3d::keyboardinput::KEY_ESC ))
    {
        // Timer Query broken in OS X since upgrading to Mavericks.
        //renderer->BeginQuery();
        
        HandleInput( keyboardInput.get(), camera.get() );
        HandleInput( mouseInput.get(), camera.get() );

        renderer->Draw( scene.get() );
        
        //std::string gpuText = "GPU: ";
        //gpuText += std::to_string( gpuTime ) + " ms";
        //renderer->Draw( gpuText, "aether3d_font", 1, 5, 50, 1, 1, 1, 1 );

        renderer->DrawStatistics();
        renderer->FinishFrame();

        renderer->SwapBuffers();
        //gpuTime = renderer->EndQuery();
    }

    renderer->Deinit();
    return 0;
}

