#pragma once
#ifndef COMMON_H
#define COMMON_H

#include <cassert>
#include <fstream>
#include <iostream>
#include <map>
#include <string>
#include <vector>
#include "Matrix.hpp"
#include "Vec3.hpp"

static const char* gAe3dVersion = "a9";

struct TexCoord
{
    TexCoord() : u( 0 ), v( 0 ) {}
    TexCoord( float aU, float aV ) : u( aU ), v( aV ) {}
    TexCoord operator-( const TexCoord& t ) const
    {
        return TexCoord( u - t.u, v - t.v );
    }

    TexCoord operator+( const TexCoord& t ) const
    {
        return TexCoord( u + t.u, v + t.v );
    }

    float u, v;
};

// There is one influence per vertex, and it has info about
// which joints and weights are influencing it.
struct Influence
{
    std::vector< int > jointIndices; // Joints that affect this vertex.
    std::vector< int > weightIndices; // Weights that affect this vertex.
};

// These influences are sent to GPU and are capped to 4.
struct ProcessedInfluence
{
    ProcessedInfluence()
    {
        jointIndices[ 0 ] = jointIndices[ 1 ] = jointIndices[ 2 ] = jointIndices[ 3 ] = 0;
        weights[ 1 ] = weights[ 2 ] = weights[ 3 ] = 0;
        weights[ 0 ] = 1;
    }

    float jointIndices[ 4 ]; // Some GPUs don't like integer indices.
    float weights[ 4 ];
};

/**
 Used to store intermediate vertex data that comes from
 the exported model.  Later it's converted to vertex array
 friendly format by removing duplicates and interleaving data.
 */
struct Face
{
    Face()
    {
        vInd[0] = vInd[1] = vInd[2] = 0;
        uvInd[0] = uvInd[1] = uvInd[2] = 0;
        vnInd[0] = vnInd[1] = vnInd[2] = 0;
        colInd[0] = colInd[1] = colInd[2] = 0;
    }

    unsigned short vInd[3];
    unsigned short uvInd[3];
    unsigned short vnInd[3];
    unsigned short colInd[3];
};

// Defines a face used in a vertex array.
// a, b and c are indices to Mesh::interleavedVertices.
struct VertexInd
{
    unsigned short a, b, c;
};

struct Joint
{
    std::vector< Joint > children;
    std::string name;
    Matrix44 transform;
    Matrix44 invBindMatrix;
};

struct Animation
{
    enum Interpolation
    {
        Linear = 0,
        Bezier
    };

    Animation() : inputIndex( 0 ), joint( 0 ) {}

    std::vector< float > input; // Time in seconds.
    std::vector< float > output; // Angles, positions or 4x4 matrixes.
    std::vector< Interpolation > interpolation;
    std::vector< float > inTangents;
    std::vector< float > outTangents;

    std::string name;
    int inputIndex;
    int joint; // Index to Mesh.jointArray.
    std::string operation; // rotateX.ANGLE, translate, transform etc.
};

struct Vertex
{
    Vec3 position;
    TexCoord texCoord;
    Vec3 normal;
    Vec4 tangent;
    Vec4 color;
};

struct Mesh
{
    Mesh()
    : name( "unnamed" )
    , morphTarget( false )
    {
    }

    void BuildVertexInfluences();
    void Interleave();
    void SolveAABB();
    void SolveFaceNormals();
    void SolveFaceTangents();
    void SolveVertexNormals();
    void SolveVertexTangents();
    bool AlmostEquals( const Vec3& v1, const Vec3& v2 ) const;
    bool AlmostEquals( const Vec4& v1, const Vec4& v2 ) const;
    bool AlmostEquals( const TexCoord& t1, const TexCoord& t2 ) const;

    std::string name;
    std::vector< Vec3 >     vertex;
    std::vector< Vec4 >     tangents;
    std::vector< Vec3 >     vnormal;
    std::vector< Vec4 >     colors;
    std::vector< Vec3 >     fnormal;
    std::vector< TexCoord > tcoord;
    std::vector< Face >     face;

    // Separate element arrays combined to one.
    // These are written to the .ae3d file.
    std::vector< Vertex >    interleavedVertices;
    std::vector< VertexInd > indices;

    // Used to calculate tangent-space handedness.
    std::vector< Vec3 > bitangents;  // For faces.
    std::vector< Vec3 > vbitangents; // For vertices.

    Joint skeletonRoot;
    std::vector< Influence > influences; // One item per vertex. Points to un-interleaved vertices.
    std::vector< Influence > interleavedInfluences; // One item per vertex. Points to interleaved vertices.
    std::vector< float > weights; // skin weights.

    // Used to determine which joint influences what.
    std::vector< std::string > jointNames;
    std::vector< std::string > jointIdRefs;

    // Joints from skeleton organized so that influence indices point right.
    std::vector< Joint > joints;

    // Transform of the bind shape before skinning.
    Matrix44 bindShapeMatrix;

    std::map< std::string, Animation > animations;

    // Influences sent as vertex attribute.
    std::vector< ProcessedInfluence > processedInfluences;

    // If mesh has morph targets, it's treated as a base mesh.
    std::vector< std::string > morphTargets;
    std::vector< float > morphWeights;
    bool morphTarget;

    Vec3 aabbMax;
    Vec3 aabbMin;
};

std::vector< Mesh > gMeshes;

void Mesh::SolveAABB()
{
    const float maxValue = 99999999.0f;

    aabbMin = Vec3(  maxValue,  maxValue,  maxValue );
    aabbMax = Vec3( -maxValue, -maxValue, -maxValue );

    for (size_t v = 0; v < vertex.size(); ++v)
    {
        const Vec3& pos = vertex[ v ];

        if (pos.x > maxValue || pos.y > maxValue || pos.z > maxValue)
        {
            std::cerr << "AABB's max value is too low in file " << __FILE__ << " near line " << __LINE__ << std::endl;
            std::cerr << "Vertex " << v << ": " << pos.x << ", " << pos.y << ", " << pos.z << std::endl;
        }
        
        if (pos.x < -maxValue || pos.y < -maxValue || pos.z < -maxValue)
        {
            std::cerr << "AABB's min value is too high in file " << __FILE__ << " near line " << __LINE__ << std::endl;
            std::cerr << "Vertex " << v << ": " << pos.x << ", " << pos.y << ", " << pos.z << std::endl;
        }

        aabbMin = Vec3::Min2( aabbMin, pos );
        aabbMax = Vec3::Max2( aabbMax, pos );
    }
}

void Mesh::SolveFaceNormals()
{
    fnormal.resize( indices.size() );

    for (std::size_t faceInd = 0; faceInd < indices.size(); ++faceInd)
    {
        const unsigned short& faceA = indices[ faceInd ].a;
        const unsigned short& faceB = indices[ faceInd ].b;
        const unsigned short& faceC = indices[ faceInd ].c;

        Vec3 va = interleavedVertices[ faceA ].position;
        Vec3 vb = interleavedVertices[ faceB ].position;
        Vec3 vc = interleavedVertices[ faceC ].position;

        vb -= va;
        vc -= va;

        vb = Vec3::Cross( vb, vc ).Normalized();

        fnormal[ faceInd ] = vb;
    }
}

/**
 Generates tangents for faces.

 Tangent vector (u-axis) will point in the direction of increasing U-coordinate.
 Bitangent vector (v-axis) will point in the direction of increasing V-coordinate.
 */
void Mesh::SolveFaceTangents()
{
    assert( !indices.empty() );

    tangents.resize  ( indices.size() );
    bitangents.resize( indices.size() );

    bool degenerateFound = false;

    // Algorithm source:
    // http://www.opengl-tutorial.org/intermediate-tutorials/tutorial-13-normal-mapping/#header-3
    for (std::size_t f = 0; f < indices.size(); ++f)
    {
        const Vec3& p1 = interleavedVertices[ indices[ f ].a ].position;
        const Vec3& p2 = interleavedVertices[ indices[ f ].b ].position;
        const Vec3& p3 = interleavedVertices[ indices[ f ].c ].position;

        const TexCoord& uv1 = interleavedVertices[ indices[ f ].a ].texCoord;
        const TexCoord& uv2 = interleavedVertices[ indices[ f ].b ].texCoord;
        const TexCoord& uv3 = interleavedVertices[ indices[ f ].c ].texCoord;

        const TexCoord deltaUV1 = uv2 - uv1;
        const TexCoord deltaUV2 = uv3 - uv1;

        const float area = deltaUV1.u * deltaUV2.v - deltaUV1.v * deltaUV2.u;

        Vec3 tangent;
        Vec3 bitangent;

        if (std::fabs( area ) > 0.00001f)
        {
            const Vec3 deltaP1 = p2 - p1;
            const Vec3 deltaP2 = p3 - p1;

            tangent   = (deltaP1 * deltaUV2.v - deltaP2 * deltaUV1.v) / area;
            bitangent = (deltaP2 * deltaUV1.u - deltaP1 * deltaUV2.u) / area;
        }
        else
        {
            degenerateFound = true;
        }

        const Vec4 tangent4( tangent.x, tangent.y, tangent.z, 0.0f );
        tangents[ f ]   = tangent4;
        bitangents[ f ] = bitangent;
    }

    if (degenerateFound)
    {
        std::cout << std::endl << "Warning: Degenerate UV map in mesh " << name << ".  Author needs to separate texture points." << std::endl;
    }
}

void Mesh::SolveVertexNormals()
{
    // For every vertex, check how many faces touches
    // it and calculate the average of those face normals.
    for (std::size_t vertInd = 0; vertInd < vertex.size(); ++vertInd)
    {
        Vec3 normal;

        for (size_t faceInd = 0; faceInd < face.size(); ++faceInd)
        {
            face[ faceInd ].vnInd[ 0 ] = face[ faceInd ].vInd[ 0 ];
            face[ faceInd ].vnInd[ 1 ] = face[ faceInd ].vInd[ 1 ];
            face[ faceInd ].vnInd[ 2 ] = face[ faceInd ].vInd[ 2 ];

            unsigned short& faceA = face[ faceInd ].vInd[ 0 ];
            unsigned short& faceB = face[ faceInd ].vInd[ 1 ];
            unsigned short& faceC = face[ faceInd ].vInd[ 2 ];

            Vec3 va = vertex[ faceA ];
            Vec3 vb = vertex[ faceB ];
            Vec3 vc = vertex[ faceC ];

            vb -= va;
            vc -= va;

            vb = Vec3::Cross( vb, vc ).Normalized();

            if (face[ faceInd ].vInd[ 0 ] == vertInd ||
                face[ faceInd ].vInd[ 1 ] == vertInd ||
                face[ faceInd ].vInd[ 2 ] == vertInd)
            {
                normal = normal + vb;
            }
        }
        
        normal = normal.Normalized();
        vnormal.push_back( normal );
    }
}

void Mesh::SolveVertexTangents()
{
    assert( !tangents.empty() );

    for (std::size_t v = 0; v < interleavedVertices.size(); ++v)
    {
        interleavedVertices[ v ].tangent = Vec4( 0.0f, 0.0f, 0.0f, 0.0f );
    }

    vbitangents.resize( interleavedVertices.size() );

    // http://www.terathon.com/code/tangent.html
    // "To find the tangent vectors for a single vertex, we average the tangents for all triangles sharing that vertex"
    for (std::size_t vertInd = 0; vertInd < interleavedVertices.size(); ++vertInd)
    {
        Vec4 tangent( 0.0f, 0.0f, 0.0f, 0.0f );
        Vec3 bitangent;

        for (std::size_t faceInd = 0; faceInd < indices.size(); ++faceInd)
        {
            if (indices[ faceInd ].a == vertInd ||
                indices[ faceInd ].b == vertInd ||
                indices[ faceInd ].c == vertInd)
            {
                tangent   += tangents  [ faceInd ];
                bitangent += bitangents[ faceInd ];
            }
        }

        interleavedVertices[ vertInd ].tangent = tangent;
        vbitangents[ vertInd ] = bitangent;
    }

    for (std::size_t v = 0; v < interleavedVertices.size(); ++v)
    {
        Vec4& tangent = interleavedVertices[ v ].tangent;
        const Vec3& normal = interleavedVertices[ v ].normal;
        const Vec4 normal4( normal.x, normal.y, normal.z, 0.0 );

        // Gram-Schmidt orthonormalization.
        tangent -= normal4 * normal4.Dot( tangent );
        tangent.Normalize();

        // Handedness. TBN must form a right-handed coordinate system,
        // i.e. cross(n,t) must have the same orientation as b.
        const Vec3 cp = Vec3::Cross( normal, Vec3( tangent.x, tangent.y, tangent.z ) );
        tangent.w = Vec3::Dot( cp, vbitangents[v] ) >= 0.0f ? 1.0f : -1.0f;
    }
}

// Creates an interleaved vertex array.
void Mesh::Interleave()
{
    VertexInd newFace;

    for (size_t f = 0; f < face.size(); ++f)
    {
        // vertind 0
        Vec3 tvertex     = vertex [ face[ f ].vInd [ 0 ] ];
        Vec3 tnormal     = vnormal[ face[ f ].vnInd[ 0 ] ];
        TexCoord ttcoord = tcoord.empty() ? TexCoord() : tcoord[ face[ f ].uvInd[ 0 ] ];
        Vec4 tcolor = colors.empty() ? Vec4( 0, 0, 0, 1 ) : colors[ face[ f ].colInd[ 0 ] ];

        // Searches face f's vertex a from the vertex list.
        // If it's not found, it's added.
        bool found = false;
        for (std::size_t i = 0; i < indices.size(); ++i)
        {
            if (AlmostEquals( interleavedVertices[ indices[ i ].a ].position, tvertex ) &&
                AlmostEquals( interleavedVertices[ indices[ i ].a ].normal,   tnormal ) &&
                AlmostEquals( interleavedVertices[ indices[ i ].a ].texCoord, ttcoord ) &&
                AlmostEquals( interleavedVertices[ indices[ i ].a ].color, tcolor ))
            {
                found = true;
                newFace.a = indices[ i ].a;
                break;
            }
        }

        if (!found)
        {
            Vertex newVertex;
            newVertex.position = tvertex;
            newVertex.normal   = tnormal;
            newVertex.texCoord = ttcoord;
            newVertex.color    = tcolor;

            interleavedVertices.push_back( newVertex );

            newFace.a = (unsigned short)(interleavedVertices.size() - 1);

            // Maps the un-interleaved array's weight index to interleaved.
            if (weights.size() > 0)
            {
                interleavedInfluences.push_back( influences[ face[ f ].vInd[ 0 ] ] );
            }
        }

        // vertind 1
        tvertex = vertex [ face[ f ].vInd [ 1 ] ];
        tnormal = vnormal[ face[ f ].vnInd[ 1 ] ];
        ttcoord = tcoord.empty() ? TexCoord() : tcoord [ face[ f ].uvInd[ 1 ] ];
        tcolor = colors.empty() ? Vec4( 0, 0, 0, 1 ) : colors[ face[ f ].colInd[ 1 ] ];
        
        found = false;
        for (std::size_t i = 0; i < indices.size(); ++i)
        {
            if (AlmostEquals( interleavedVertices[ indices[ i ].b ].position, tvertex ) &&
                AlmostEquals( interleavedVertices[ indices[ i ].b ].normal,   tnormal ) &&
                AlmostEquals( interleavedVertices[ indices[ i ].b ].texCoord, ttcoord ) &&
                AlmostEquals( interleavedVertices[ indices[ i ].b ].color, tcolor ))
                
            {
                found = true;

                newFace.b = indices[ i ].b;
                break;
            }
        }

        if (!found)
        {
            Vertex newVertex;
            newVertex.position = tvertex;
            newVertex.normal   = tnormal;
            newVertex.texCoord = ttcoord;
            newVertex.color    = tcolor;

            interleavedVertices.push_back( newVertex );

            newFace.b = (unsigned short)(interleavedVertices.size() - 1);

            // Maps the un-interleaved array's weight index to interleaved.
            if (weights.size() > 0)
            {
                interleavedInfluences.push_back( influences[ face[ f ].vInd[ 1 ] ] );
            }
        }

        // vertind 2
        tvertex = vertex [ face[ f ].vInd [ 2 ] ];
        tnormal = vnormal[ face[ f ].vnInd[ 2 ] ];
        ttcoord = tcoord.empty() ? TexCoord() :  tcoord [ face[ f ].uvInd[ 2 ] ];
        tcolor = colors.empty() ? Vec4( 0, 0, 0, 1 ) : colors[ face[ f ].colInd[ 2 ] ];

        found = false;
        for (std::size_t i = 0; i < indices.size(); ++i)
        {
            if (AlmostEquals( interleavedVertices[ indices[ i ].c ].position, tvertex ) &&
                AlmostEquals( interleavedVertices[ indices[ i ].c ].normal,   tnormal ) &&
                AlmostEquals( interleavedVertices[ indices[ i ].c ].texCoord, ttcoord ) &&
                AlmostEquals( interleavedVertices[ indices[ i ].c ].color, tcolor ))
            {
                found = true;

                newFace.c = indices[ i ].c;
                break;
            }
        }

        if (!found)
        {
            Vertex newVertex;
            newVertex.position = tvertex;
            newVertex.normal   = tnormal;
            newVertex.texCoord = ttcoord;
            newVertex.color    = tcolor;

            interleavedVertices.push_back( newVertex );

            newFace.c = (unsigned short)(interleavedVertices.size() - 1);

            // Maps the un-interleaved array's weight index to interleaved.
            if (weights.size() > 0)
            {
                interleavedInfluences.push_back( influences[ face[ f ].vInd[ 2 ] ] );
            }
        }

        indices.push_back( newFace );
    }

    if (interleavedVertices.size() > 65536)
    {
        std::cerr << "Mesh " << name << " has more than 65536 vertices!" << std::endl;
        exit( 1 );
    }
}

bool Mesh::AlmostEquals( const Vec3& v1, const Vec3& v2 ) const
{
    if (std::fabs( v1.x - v2.x ) > 0.0001f) { return false; }
    if (std::fabs( v1.y - v2.y ) > 0.0001f) { return false; }
    if (std::fabs( v1.z - v2.z ) > 0.0001f) { return false; }
    return true;
}

bool Mesh::AlmostEquals( const Vec4& v1, const Vec4& v2 ) const
{
    if (std::fabs( v1.x - v2.x ) > 0.0001f) { return false; }
    if (std::fabs( v1.y - v2.y ) > 0.0001f) { return false; }
    if (std::fabs( v1.z - v2.z ) > 0.0001f) { return false; }
    if (std::fabs( v1.w - v2.w ) > 0.0001f) { return false; }
    return true;
}

bool Mesh::AlmostEquals( const TexCoord& t1, const TexCoord& t2 ) const
{
    if (std::fabs( t1.u - t2.u ) > 0.0001f) { return false; }
    if (std::fabs( t1.v - t2.v ) > 0.0001f) { return false; }
    return true;
}

// Builds per-vertex influences that are ready to be sent to GPU.
void Mesh::BuildVertexInfluences()
{
    assert( !interleavedVertices.empty() );

    processedInfluences.resize( interleavedVertices.size() );

    for (std::size_t i = 0; i < processedInfluences.size(); ++i)
    {
        for (std::size_t ji = 0; ji < interleavedInfluences[ i ].jointIndices.size(); ++ji)
        {
            processedInfluences[ i ].jointIndices[ ji ] = (float)interleavedInfluences[ i ].jointIndices[ ji ];
            processedInfluences[ i ].weights[ ji ] = weights[ interleavedInfluences[ i ].weightIndices[ ji ] ];
        }
    }
}

/**
 Caps weight count to 4 and uses at most 4 most significant weights.
 
 \param weights Weights to be tested.
 \param outInfluence Influence to be modified.
 */
void Take4mostSignificantWeights( const std::vector< float >& weights, Influence& outInfluence )
{
    assert( !weights.empty() );
    
    std::vector< float > iWeights( outInfluence.weightIndices.size() );
    
    for (std::size_t i = 0; i < outInfluence.weightIndices.size(); ++i)
    {
        iWeights[ i ] = weights[ outInfluence.weightIndices[ i ] ];
    }
    
    std::sort( iWeights.begin(), iWeights.end() );
    std::reverse( iWeights.begin(), iWeights.end() );
    
    std::vector< int > newWeightIndices;
    std::vector< int > newJointIndices;
    
    for (std::size_t w = 0; w < (std::size_t)std::min( 4, (int)weights.size() ); ++w)
    {
        // Finds correct index for this weight.
        for (std::size_t i = 0; i < outInfluence.weightIndices.size(); ++i)
        {
            if (weights[ outInfluence.weightIndices[ i ] ] == iWeights[ w ])
            {
                newWeightIndices.push_back( outInfluence.weightIndices[ i ] );
                newJointIndices.push_back( outInfluence.jointIndices[ i ] );
                break;
            }
        }
    }
    
    outInfluence.weightIndices = newWeightIndices;
    outInfluence.jointIndices = newJointIndices;
}

/**
 bytes  data
 (2)    magic number, e.g. "a7"
 (4*6)  Object's AABB min, AABB max.
 (2)    # of meshes
 (4*6)      Mesh's AABB min, AABB max.
 (2)        mesh name length in bytes.
 (*)        mesh name (1 character = 1 byte)
 (2)        # of vertices.
 (*)        Vertex data array of type Vertex.
 (2)        # of faces
 (*)        faces
 (*)        face normals
 (1)        1 if the mesh is a morph target, 0 otherwise.
 (1)        # of morph target names
 (*)        (1) name length, name (# names times)
 (1)        0 if the mesh has no skeleton, 1 otherwise.
 If there is a skeleton, it starts here.
 (2)    # of joints.
 (*)    joint name lengths and names.  Length is 2 bytes.
 (4 * 4 * processedInfluences.size()) joint indices
 (4 * 4 * processedInfluences.size()) weights
 After this line comes joints.
 (2)    index of joint name
 (16*4) inverse bind matrix
 (16*4) transformation matrix
 (2)    # of children
 (*)    indices to children joints, 2 bytes each.
 (2)    # of animations
 (2)    animation name length
 (*)    animation name
 (2)    # of inputs
 (*)    inputs, 4 bytes each.
 (2)    # of outputs
 (*)    outputs, 4 bytes each.
 (2)    # of interpolations
 (*)    interpolations, 1 byte each.
 (2)    # of in tangents
 (*)    in tangents
 (2)    # of out tangents
 (*)    out tangents
 (2)    animation joint index.
 (2)    length of operation name
 (*)    operation name.
 (1)    Terminator magic: 100
 */

/**
 Writes a .ae3d model to a file.

 \param aOutFile File name to save the model into.
 \todo split into parts like in DLL.
 */
void WriteAe3d( const std::string& aOutFile )
{
    static_assert( sizeof( Vertex    ) == 64, "" );
    static_assert( sizeof( Vec3      ) == 12, "" );
    static_assert( sizeof( VertexInd ) ==  6, "" );

    std::ofstream ofs( aOutFile.c_str(), std::ios::binary );

    if (!ofs.is_open())
    {
        std::cerr << "Couldn't open file for writing!" << std::endl;
        exit( 1 );
    }

    for (size_t m = 0; m < gMeshes.size(); ++m)
    {
        gMeshes[ m ].SolveAABB();

        if (gMeshes[ m ].vnormal.empty())
        {
            gMeshes[ m ].SolveVertexNormals();
        }

        gMeshes[ m ].Interleave();

        if (gMeshes[ m ].skeletonRoot.children.size() > 0)
        {
            gMeshes[ m ].BuildVertexInfluences();
        }

        gMeshes[ m ].SolveFaceNormals();
        gMeshes[ m ].SolveFaceTangents();
        gMeshes[ m ].SolveVertexTangents();
    }

    // Calculates model's AABB by finding extreme values from meshes' AABBs.
    const float maxValue = 99999999.0f;
    Vec3 aabbMin(  maxValue,  maxValue,  maxValue );
    Vec3 aabbMax( -maxValue, -maxValue, -maxValue );

    for (size_t m = 0; m < gMeshes.size(); ++m)
    {
        aabbMin = Vec3::Min2( aabbMin, gMeshes[ m ].aabbMin );
        aabbMax = Vec3::Max2( aabbMax, gMeshes[ m ].aabbMax );
    }

    // The file starts with identification bytes.
    ofs.write( gAe3dVersion, 2 );

    ofs.write( reinterpret_cast< char* >( &aabbMin.x ), 3 * 4 );
    ofs.write( reinterpret_cast< char* >( &aabbMax.x ), 3 * 4 );

    const size_t meshes = gMeshes.size();
    // # of meshes.
    ofs.write( reinterpret_cast< const char* >( &meshes ), 2 );

    for (size_t m = 0; m < meshes; ++m)
    {
        assert( gMeshes[ m ].fnormal.size() == gMeshes[ m ].indices.size() );

        // AABB.
        ofs.write( reinterpret_cast< char* >( &gMeshes[ m ].aabbMin.x ), 3 * 4 );
        ofs.write( reinterpret_cast< char* >( &gMeshes[ m ].aabbMax.x ), 3 * 4 );

        // Writes name's length.
        const unsigned short nameLength = (unsigned short)gMeshes[ m ].name.length();
        ofs.write( reinterpret_cast< char* >((char*)&nameLength ), 2 );
        // Writes name.
        ofs.write(reinterpret_cast<char*>((char*)gMeshes[m].name.data()),
                  nameLength);
        
        // Writes # of vertices.
        const unsigned short nVertices = (unsigned short)gMeshes[ m ].interleavedVertices.size();
        ofs.write( reinterpret_cast< char* >((char*)&nVertices), 2 );

        // Writes vertex data.
        ofs.write( (char*)&gMeshes[m].interleavedVertices[ 0 ], gMeshes[m].interleavedVertices.size() * sizeof( Vertex ) );

        // Writes # of indices.
        const unsigned short nIndices = (unsigned short)gMeshes[m].indices.size();
        ofs.write( reinterpret_cast< char* >((char*)&nIndices), 2 );

        // Writes indices.
        ofs.write( (char*)&gMeshes[m].indices[ 0 ], gMeshes[m].indices.size() * sizeof( VertexInd ) );

        // Writes face normals.
        ofs.write( (char*)&gMeshes[m].fnormal[ 0 ].x, gMeshes[m].fnormal.size() * sizeof( Vec3 ) );

        unsigned char byte = gMeshes[m].morphTarget ? 1 : 0;
        ofs.write( (char*)&byte, 1 );

        // Morph target count.
        assert( gMeshes[m].morphTargets.size() < 256 );
        byte = (unsigned char)gMeshes[m].morphTargets.size();
        ofs.write( (char*)&byte, 1 );

        for (std::size_t targetIndex = 0; targetIndex < gMeshes[m].morphTargets.size(); ++targetIndex)
        {
            // Morph target name length.
            byte = (unsigned char)gMeshes[m].morphTargets[ targetIndex ].length();
            ofs.write( (char*)&byte, 1 );

            // Morph target name.
            ofs.write( (char*)gMeshes[m].morphTargets[ targetIndex ].data(), byte );
        }

        byte = 0;

        if (!gMeshes[ m ].skeletonRoot.children.empty())
        {
            // Skeleton exists, so write 1.
            byte = 1;
            ofs.write( (char*)&byte, 1 );

            // Writes # of joint names.
            const short jointNameArraySize = (short)gMeshes[ m ].jointNames.size();
            ofs.write( (char*)&jointNameArraySize, 2 );

            // Writes joint names.
            for (std::size_t j = 0; j < gMeshes[ m ].jointNames.size(); ++j)
            {
                // Writes length of joint name.
                const short len = (short)gMeshes[ m ].jointNames[ j ].length();
                ofs.write( (char*)&len, 2 );

                // Writes joint name.
                ofs.write( (char*)gMeshes[ m ].jointNames[ j ].data(), len );
            }

            for (std::size_t i = 0; i < gMeshes[ m ].processedInfluences.size(); ++i)
            {
                ofs.write( (char*)&gMeshes[ m ].processedInfluences[ i ].jointIndices[ 0 ], 4 * sizeof( float ) );
                ofs.write( (char*)&gMeshes[ m ].processedInfluences[ i ].weights[ 0 ], 4 * sizeof( float ) );
            }

            // Writes joint hierarchy.
            for (std::size_t j = 0; j < gMeshes[ m ].jointNames.size(); ++j)
            {
                std::size_t t = 0;
                // Finds the joint by name.
                for (t = 0; t < gMeshes[ m ].joints.size(); ++t)
                {
                    if (gMeshes[ m ].joints[ t ].name == gMeshes[ m ].jointNames[ j ])
                    {
                        break;
                    }
                }

                // Writes inverse bind matrix for this joint.
                for (std::size_t i = 0; i < 16; ++i)
                {
                    ofs.write( (char*)&gMeshes[ m ].joints[t].invBindMatrix.m[ i ], 4 );
                }

                // Writes transformation matrix for this joint.
                for (std::size_t i = 0; i < 16; ++i)
                {
                    ofs.write( (char*)&gMeshes[ m ].joints[t].transform.m[ i ], 4 );
                }

                // Writes # of children.
                const std::size_t nChildren = gMeshes[ m ].joints[t].children.size();
                ofs.write( (char*)&nChildren, 2 );

                // Writes indices to children joints.
                for (std::size_t c = 0; c < nChildren; ++c)
                {
                    std::string& childName = gMeshes[ m ].joints[ t ].children[ c ].name;
                    // Finds the child joint by name.
                    std::size_t jc = 0;
                    
                    for (jc = 0; jc < gMeshes[ m ].joints.size(); ++jc)
                    {
                        if (gMeshes[ m ].joints[ jc ].name == childName)
                        {
                            break;
                        }
                    }
                    
                    ofs.write( (char*)&jc, 2 );
                }
            }
        }
        else // Skeleton not present.
        {
            ofs.write( (char*)&byte, 1 );
        }

        // Writes # of animations.
        const short nAnimations = (short)gMeshes[ m ].animations.size();
        ofs.write( (char*) &nAnimations, 2 );

        // Writes animations.
        for (auto a = gMeshes[ m ].animations.begin(); a != gMeshes[ m ].animations.end(); ++a)
        {
            // Writes animation name length.
            const short len = (short)a->second.name.length();
            ofs.write( (char*) &len, 2 );

            // Writes animation name.
            ofs.write( (char*)a->second.name.data(), len );

            // Writes # of inputs.
            const short nInputs = (short)a->second.input.size();
            ofs.write( (char*) &nInputs, 2 );

            // Writes inputs.
            for (std::size_t i = 0; i < a->second.input.size(); ++i)
            {
                ofs.write( (char*)&a->second.input[ i ], 4 );
            }

            // Writes # of outputs.
            const short nOutputs = (short)a->second.output.size();
            ofs.write( (char*) &nOutputs, 2 );

            // Writes outputs.
            for (std::size_t i = 0; i < a->second.output.size(); ++i)
            {
                ofs.write( (char*) &a->second.output[ i ], 4 );
            }

            // Writes # of interpolations.
            const short nInterpolations = (short)a->second.interpolation.size();
            ofs.write( (char*) &nInterpolations, 2 );

            // Writes interpolations.
            for (std::size_t i = 0; i < a->second.interpolation.size(); ++i)
            {
                ofs.write( (char*) &a->second.interpolation[ i ], 1 );
            }

            // Writes # of in tangents.
            const short nInTangents = (short)a->second.inTangents.size();
            ofs.write( (char*) &nInTangents, 2 );

            // Writes in tangents.
            for (std::size_t i = 0; i < a->second.inTangents.size(); ++i)
            {
                ofs.write( (char*) &a->second.inTangents[ i ], 4 );
            }

            // Writes # of out tangents.
            const short nOutTangents = (short)a->second.outTangents.size();
            ofs.write( (char*) &nOutTangents, 2 );

            // Writes out tangents.
            for (std::size_t i = 0; i < a->second.outTangents.size(); ++i)
            {
                ofs.write( (char*) &a->second.outTangents[ i ], 4 );
            }

            // Writes joint index.
            ofs.write( (char*)&a->second.joint, 2 );

            // Writes operation length.
            const short opLen = (short)a->second.operation.length();
            ofs.write( (char*)&opLen, 2 );

            // Writes operation string.
            ofs.write( (char*)a->second.operation.data(), opLen );
        }
    }

    // Terminator.
    unsigned char byte = 100;
    ofs.write( (char*)&byte, 1 );

    std::cout << "Wrote " << aOutFile << std::endl;
}

#endif
